/* KageHinata.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "KageHinata/KageHinata.c"

#include <windows.h>
#include <commctrl.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <windowsx.h>
#include <HTMLHELP.H>
#include <crtdbg.h>

#include "resource.h"

#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"

#pragma comment(lib, "Comctl32.lib")
#pragma comment(lib, "shlwapi.lib")
#pragma comment(lib, "version.lib")
#pragma comment(lib, "HTMLHELP.LIB")

/*
#pragma comment(lib, "util.lib")
#pragma comment(lib, "gpgez.lib")
#pragma comment(lib, "gpgew.lib")
*/

ENCFILETYPE g_encfiletype_pub[] = {
  { FILETYPE_GPGENCPUBLIC, "gpg encrypt(*.gpg)", IDS_FILETYPE_GPGENCPUBLIC, NULL, "gpg", TRUE, FALSE, -1 },
  { FILETYPE_GPGENCPUBLIC_ASCII, "gpg encrypt ascii armored(*.txt)", IDS_FILETYPE_GPGENCPUBLIC_ASCII, NULL, "txt", TRUE, TRUE, -1 },
  { FILETYPE_GPGENCSYMMETRIC, "gpg encrypt symmetric(*.gpg)", IDS_FILETYPE_GPGENCSYMMETRIC, NULL, "gpg", FALSE, FALSE, 8 },
  { FILETYPE_GPGENCSYMMETRIC_ASCII, "gpg encrypt symmetric ascii armored(*.txt)", IDS_FILETYPE_GPGENCSYMMETRIC_ASCII, NULL, "txt", FALSE, TRUE, 8 },
  { FILETYPE_ZIPENC, "zip(*.zip)", IDS_FILETYPE_ZIPENC, NULL, "zip", FALSE, FALSE, 0 },
/*
  { FILETYPE_ZIPENCAES, "zip aes encrypt(*.zip)", IDS_FILETYPE_ZIPENCAES, NULL, "zip", FALSE, FALSE, 1 },
  { FILETYPE_7ZENC, "7z(*.7z)", IDS_FILETYPE_7ZENC, NULL, "7z", FALSE, FALSE, 0 },
  { FILETYPE_ZIPPLAIN, "zip(*.zip)", IDS_FILETYPE_ZIPPLAIN, NULL, "zip", FALSE, FALSE, -1 },
  { FILETYPE_7ZPLAIN, "7z(*.7z)", IDS_FILETYPE_7ZPLAIN, NULL, "7z", FALSE, FALSE, -1 },
  { FILETYPE_TARGZ, "tar gzip(*.tgz)", IDS_FILETYPE_TARGZ, NULL, "tgz", FALSE, FALSE, -1 },
  { FILETYPE_TARBZ2, "tar bzip2(*.tar.bz2)", IDS_FILETYPE_TARBZ2, NULL, "tar.bz2", FALSE, FALSE, -1 },
  { FILETYPE_TAR, "tar(*.tar)", IDS_FILETYPE_TAR, NULL, "tar", FALSE, FALSE, -1 },
*/
  { FILETYPE_SFX, "self extract(*.exe)", IDS_FILETYPE_SFX, NULL, "exe", FALSE, FALSE, 0 },
  { FILETYPE_NONE, "", 0, NULL, "" , FALSE, FALSE, -1},
};

ENCFILETYPE g_encfiletype_sym[] = {
  { FILETYPE_GPGENCSYMMETRIC, "gpg encrypt symmetric(*.gpg)", IDS_FILETYPE_GPGENCSYMMETRIC, NULL, "gpg", FALSE, FALSE, 8 },
  { FILETYPE_GPGENCSYMMETRIC_ASCII, "gpg encrypt symmetric ascii armored(*.txt)", IDS_FILETYPE_GPGENCSYMMETRIC_ASCII, NULL, "txt", FALSE, TRUE, 8 },
  { FILETYPE_ZIPENC, "zip(*.zip)", IDS_FILETYPE_ZIPENC, NULL, "zip", FALSE, FALSE, 0 },
  { FILETYPE_SFX, "self extract(*.exe)", IDS_FILETYPE_SFX, NULL, "exe", FALSE, FALSE, 0 },
  { FILETYPE_NONE, "", 0, NULL, "" , FALSE, FALSE, -1},
};

ENCFILETYPE g_encfiletype_gpg_pub[] = {
  { FILETYPE_GPGENCPUBLIC, "gpg encrypt(*.gpg)", IDS_FILETYPE_GPGENCPUBLIC, NULL, "gpg", TRUE, FALSE, -1 },
  { FILETYPE_GPGENCPUBLIC_ASCII, "gpg encrypt ascii armored(*.txt)", IDS_FILETYPE_GPGENCPUBLIC_ASCII, NULL, "txt", TRUE, TRUE, -1 },
  { FILETYPE_GPGENCSYMMETRIC, "gpg encrypt symmetric(*.gpg)", IDS_FILETYPE_GPGENCSYMMETRIC, NULL, "gpg", FALSE, FALSE, 8 },
  { FILETYPE_GPGENCSYMMETRIC_ASCII, "gpg encrypt symmetric ascii armored(*.txt)", IDS_FILETYPE_GPGENCSYMMETRIC_ASCII, NULL, "txt", FALSE, TRUE, 8 },
  { FILETYPE_NONE, "", 0, NULL, "" , FALSE, FALSE, -1},
};

ENCFILETYPE g_encfiletype_gpg_sym[] = {
  { FILETYPE_GPGENCSYMMETRIC, "gpg encrypt symmetric(*.gpg)", IDS_FILETYPE_GPGENCSYMMETRIC, NULL, "gpg", FALSE, FALSE, 8 },
  { FILETYPE_GPGENCSYMMETRIC_ASCII, "gpg encrypt symmetric ascii armored(*.txt)", IDS_FILETYPE_GPGENCSYMMETRIC_ASCII, NULL, "txt", FALSE, TRUE, 8 },
  { FILETYPE_NONE, "", 0, NULL, "" , FALSE, FALSE, -1},
};

DECFILETYPE g_decfiletype[] = {
  { EXTENSION_GPG, "gpg" },
  { EXTENSION_ZIP, "zip" },
  { EXTENSION_7Z, "7z" },
  { EXTENSION_LZH, "lzh" },
  { EXTENSION_CAB, "cab" },
  { EXTENSION_TGZ, "tgz" },
  { EXTENSION_TARGZ, "tar.gz" },
  { EXTENSION_TBZ2, "tbz2" },
  { EXTENSION_TARBZ2, "tar.bz2" },
  { EXTENSION_TAR, "tar" },
  { EXTENSION_GZ, "gz" },
  { EXTENSION_BZ2, "bz2" },
  { EXTENSION_ERROR, ""}
};

ENCFILETYPE g_encfiletype_key[] = {
  { FILETYPE_KEYBIN, "gpg key(*.gpg)", IDS_FILETYPE_KEY_BIN, NULL, "gpg", FALSE, FALSE, -1 },
  { FILETYPE_KEYASC, "gpg key ascii armored(*.txt)", IDS_FILETYPE_KEY_ASCII, NULL, "txt", FALSE, TRUE, -1 },
  { FILETYPE_NONE, "", 0, NULL, "" , FALSE, FALSE, -1},
};

typedef struct janus_variables {
  PGPGEZBASE pgpgezbase;
} JANUS_VAR, *PJANUS_VAR;

HINSTANCE g_hInstance;
HANDLE g_hBitmap_enc1, g_hBitmap_enc2;
HANDLE g_hBitmap_dec1, g_hBitmap_dec2;
HWND g_hStatic1, g_hStatic2;
WNDPROC defProcWndForStatic1;
WNDPROC defProcWndForStatic2;
DWORD cookieHelp = 0;
INT DebugLevel = 1;
static CHAR szExeFileName[MAX_PATH + 1];

#define MAX_FILES 65530
#define APPNAME "KageHinata"
#define KEYNAME_GPGEXE "gpgexe"
#define KEYNAME_CONFDIR "homedir"
#define KEYNAME_TAREXE "tarexe"
#define KEYNAME_7ZEXE "7zexe"
#define KEYNAME_NKFEXE "nkfexe"
#define KEYNAME_IFDEBUG "debug"
#define APPDATA "kzworks\\KageHinata"
#define HELP_FILENAME "KageHinata.chm"
#define INI_FILENAME "KageHinata.ini"
#define CONFDIR "conf"
#define GPGEXE "bin\\gpg.exe"
#define TAREXE "bin\\bsdtar.exe"
#define SEVENZIPEXE "bin\\7za.exe"
#define NKFEXE "bin\\nkfwin.exe"
#define LARGEICON_ENC "bitmap_enc"
#define LARGEICON_DEC "bitmap_dec"
#define LARGEICON_WIDTH 128
#define LARGEICON_HEIGHT 128
#define MAX_BUFF 1024

#define WM_CREATEPROCESS   (WM_USER + 3)
#define WM_PROCESSCMDLINE  (WM_USER + 4)
#define WM_PROCESSCMDLINE2  (WM_USER + 5)

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

#define DO_DEBUG(hr, fn) if (FAILED((hr) = (fn))) { DEBUGERROR(0, (hr)); goto finish_func;}


CRITICAL_SECTION cs;

HRESULT ShowAboutDialog(HWND hwnd);
HRESULT KageHinata_CanExportKageKoboshi(VOID);
HRESULT KageHinata_ExportKageKoboshi(HWND hwnd, PCSTR pHomeDir, PCSTR pFpr);

/*------------------------------------------------------------*/
/**
 * \fn static HRESULT zgpencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PTARGET_FILELIST ptfl);
 *
 *
 */
static HRESULT
zgpencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PTARGET_FILELIST ptfl) {
  HRESULT hr = S_OK;
  if (gpgezw_getMyFpr(pgpgezbase)) {
    hr = gpgezw_zgpencrypt(pgpgezbase, hwnd, ptfl, NULL, NULL, NULL, FALSE, TRUE, g_encfiletype_pub);
  } else {
    hr = gpgezw_zgpencrypt(pgpgezbase, hwnd, ptfl, NULL, NULL, NULL, FALSE, TRUE, g_encfiletype_sym);
  }
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn static HRESULT zgpdecrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, PCSTR pDir);
 *
 *
 */
static HRESULT
zgpdecrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, PCSTR pDir) {
  HRESULT hr = S_OK;

  hr = gpgezw_zgpdecrypt(pgpgezbase, hwnd, pszFileName, pDir, NULL, NULL, FALSE, TRUE, g_decfiletype);

  return hr;
}

/*------------------------------------------------------------*/
static HRESULT
FlashPicture(HWND hwnd, HANDLE hImage1, HANDLE hImage2, HANDLE hImage3) {
  HRESULT hr = S_OK;
  int i = 0;
  for (i = 0; i < 2; ++i) {
    SendMessage(hwnd, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hImage2);
    Sleep(50);
    SendMessage(hwnd, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hImage1);
    Sleep(50);
  }
  SendMessage(hwnd, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hImage3);
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn static HRESULT SelectFilesAndEncrypt(HWND hwnd);
 *
 * j[ŁuÍvIꍇ̏
 * TARGET_FILELIST\̂Ɋi[t@CIAÍ
 * tH_̑IłȂH
 *
 * hwnd ͊{IɃCEChE
 */
static HRESULT
SelectFilesAndEncrypt(HWND hwnd) {
  HRESULT hr = S_OK;
  PTARGET_FILELIST ptfl = NULL;
  OPENFILENAME ofn;
  CHAR files[MAX_FILES + 1];
  HANDLE hFind;
  WIN32_FIND_DATA wfd;
  PSTR p;
  PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);

  FlashPicture(g_hStatic1, g_hBitmap_enc1, g_hBitmap_enc2, g_hBitmap_enc1);
  files[0]= '\0';
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.hInstance = g_hInstance;
  ofn.lpstrFilter = "All Files(*.*)\0*.*\0\0";
  ofn.lpstrFile = files;
  ofn.nMaxFile = MAX_FILES;
  ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_ALLOWMULTISELECT | OFN_EXPLORER;
  if (GetOpenFileName(&ofn)) {
    if ((hFind = FindFirstFile(files, &wfd)) == INVALID_HANDLE_VALUE) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    FindClose(hFind);

    /* Iꂽt@CtH_łꍇ̓t@CIĂ */
    /* FIX ME [gtH_łꍇ̏ */
    if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
      HRESULT hr = S_OK;

      if (FAILED(hr = target_filelist_create(&ptfl, files))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }

      p = files + util_strlen(files, -1) + 1;
      while (*p) {
        if (FAILED(hr = target_filelist_add_node(ptfl, p))) {
          DEBUGERROR(0, hr);
          goto finish_func;
        }
        p += util_strlen(p, -1);
        p++;
      }
    } else {
      PSTR buff = NULL;
      DWORD len = 0;
      /* Iꂽt@Cʏt@C̏ꍇAPt@Cɑ΂鏈 */
      *(PathFindFileName(files) - 1) = '\0';
      len = util_strlen(files, -1);

      if (FAILED(hr = target_filelist_create(&ptfl, files))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
      if (FAILED(hr = target_filelist_add_node(ptfl, files + len + 1))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
    hr = zgpencrypt(pvar->pgpgezbase, hwnd, ptfl);
finish_func:
    if (ptfl) target_filelist_release(&ptfl);
  } else {
    /* G[ */
  }
  SendMessage(g_hStatic1, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)g_hBitmap_enc1);
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn static DWORD SelectFileAndDecrypt(HWND hwnd);
 *
 * j[ŁuvIꍇ̏
 * t@CIŁA
 *
 */
static HRESULT
SelectFileAndDecrypt(HWND hwnd) {
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR filename[MAX_PATH + 1];
  PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);

  FlashPicture(g_hStatic2, g_hBitmap_dec1, g_hBitmap_dec2, g_hBitmap_dec1);
  filename[0]= '\0';
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.hInstance = g_hInstance;
  ofn.lpstrFilter = "gpg Files(*.gpg)\0*.gpg\0All Files(*.*)\0*.*\0\0";
  ofn.lpstrFile = filename;
  ofn.nMaxFile = MAX_PATH;
  ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_EXPLORER;
  if (GetOpenFileName(&ofn)) {
    zgpdecrypt(pvar->pgpgezbase, hwnd, filename, NULL);
  } else {
    /* G[ */
  }
  SendMessage(g_hStatic2, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)g_hBitmap_dec1);
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn static DWORD SelectFileAndDecryptOnly(HWND hwnd);
 *
 * j[Łû݁vIꍇ̏
 * t@CIŁA
 *
 */
static HRESULT
SelectFileAndDecryptOnly(HWND hwnd) {
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR filename[MAX_PATH + 1];
  PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);

  FlashPicture(g_hStatic2, g_hBitmap_dec1, g_hBitmap_dec2, g_hBitmap_dec1);
  filename[0]= '\0';
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.hInstance = g_hInstance;
  ofn.lpstrFilter = "gpg Files(*.gpg)\0*.gpg\0All Files(*.*)\0*.*\0\0";
  ofn.lpstrFile = filename;
  ofn.nMaxFile = MAX_PATH;
  ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_EXPLORER;
  if (GetOpenFileName(&ofn)) {
    gpgezw_zgpdecryptonly(pvar->pgpgezbase, hwnd, filename, NULL, NULL, NULL, TRUE);
  } else {
    /* G[ */
  }
  SendMessage(g_hStatic2, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)g_hBitmap_dec1);
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn LRESULT CALLBACK procWndStatic1(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
 *
 * ̃X^eBbNRg[̃R[obN֐
 * t@C̃hbvꍇɁAhbvꂽt@CXgAbvÍ
 *
 */
LRESULT CALLBACK
procWndStatic1(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  if (uMsg == WM_DROPFILES) {
    PTARGET_FILELIST ptfl = NULL;
    HRESULT hr = S_OK;
    HWND hwndMain = GetParent(hwnd);
    PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwndMain, GWL_USERDATA);

    SetForegroundWindow(hwndMain);
    FlashPicture(g_hStatic1, g_hBitmap_enc1, g_hBitmap_enc2, g_hBitmap_enc1);

    hr = gpgezw_listupFilesFromDrop((HDROP)wParam, &ptfl);
    DragFinish((HDROP)wParam);
    if (SUCCEEDED(hr) && (target_filelist_get_number_of_files(ptfl) > 0)) {
      hr = zgpencrypt(pvar->pgpgezbase, hwndMain, ptfl);
    }
    target_filelist_release(&ptfl);
    SendMessage(g_hStatic1, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)g_hBitmap_enc1);
    ResetDebugError();
    return 0L;
  } else {
    return CallWindowProc(defProcWndForStatic1, hwnd, uMsg, wParam, lParam);
  }
  return 0L;
}

/*------------------------------------------------------------*/
/**
 * \fn LRESULT CALLBACK@procWndStatic2(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
 *
 * ẼX^eBbNRg[̃R[obN֐
 * t@C̃hbvꍇɁAhbvꂽt@C𕜍
 *
 */
LRESULT CALLBACK
procWndStatic2(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  if (uMsg == WM_DROPFILES) {
    HRESULT hr = S_OK;
    CHAR buff[MAX_PATH + 1];
    HANDLE hFind;
    WIN32_FIND_DATA wfd;
    HWND hwndMain = GetParent(hwnd);
    PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwndMain, GWL_USERDATA);

    SetForegroundWindow(hwndMain);
    FlashPicture(g_hStatic2, g_hBitmap_dec1, g_hBitmap_dec2, g_hBitmap_dec1);
    /* hbvꂽt@CłtH_ł肵ꍇ̓G[ */
    if (DragQueryFile((HDROP)wParam, -1, NULL, 0) != 1) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    DragQueryFile((HDROP)wParam, 0, buff, MAX_PATH);
    hFind = FindFirstFile(buff, &wfd);
    if (hFind == INVALID_HANDLE_VALUE) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    FindClose(hFind);
    if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    /* G[ȂΕ */
    hr = zgpdecrypt(pvar->pgpgezbase, hwndMain, buff, NULL);
  finish_func:
    SendMessage(g_hStatic2, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)g_hBitmap_dec1);
    ResetDebugError();
    return 0L;
  } else {
    return CallWindowProc(defProcWndForStatic2, hwnd, uMsg, wParam, lParam);
  }
  return 0L;
}

/*------------------------------------------------------------*/
static HRESULT
LoadOptions(PJANUS_VAR pvar) {
  HRESULT hr = S_OK;
  CHAR szAppPF[MAX_PATH + 1];
  CHAR szIniFileName[MAX_PATH + 1];
  CHAR szPath[MAX_PATH + 1];
  CHAR szGpgExe[MAX_PATH + 1] = "";
  CHAR szConfDir[MAX_PATH + 1] = "";
  CHAR szTarExe[MAX_PATH + 1] = "";
  CHAR sz7ZipExe[MAX_PATH + 1] = "";
  CHAR szNkfExe[MAX_PATH + 1] = "";
  HANDLE hFind = INVALID_HANDLE_VALUE;
  WIN32_FIND_DATA wfd;

  GetModuleFileName(NULL, szAppPF, MAX_PATH);
  *(PathFindFileName(szAppPF)) = '\0';
  DO_DEBUG(hr, util_strcpy(szIniFileName, MAX_PATH, szAppPF));
  DO_DEBUG(hr, util_strcat(szIniFileName, MAX_PATH, INI_FILENAME));

  if (!GetPrivateProfileString(APPNAME, KEYNAME_GPGEXE, "", szPath, MAX_PATH, szIniFileName)) {
    DO_DEBUG(hr, util_strcpy(szGpgExe, MAX_PATH, szAppPF));
    DO_DEBUG(hr, util_strcat(szGpgExe, MAX_PATH, GPGEXE));
  } else {
    if (!ExpandEnvironmentStrings(szPath, szGpgExe, MAX_PATH)) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  if (!GetPrivateProfileString(APPNAME, KEYNAME_CONFDIR, "", szPath, MAX_PATH, szIniFileName)) {
    DO_DEBUG(hr, util_strcpy(szConfDir, MAX_PATH, szAppPF));
    DO_DEBUG(hr, util_strcat(szConfDir, MAX_PATH, CONFDIR));
  } else {
    if (!ExpandEnvironmentStrings(szPath, szConfDir, MAX_PATH)) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  if (!GetPrivateProfileString(APPNAME, KEYNAME_TAREXE, "", szPath, MAX_PATH, szIniFileName)) {
    DO_DEBUG(hr, util_strcpy(szTarExe, MAX_PATH, szAppPF));
    DO_DEBUG(hr, util_strcat(szTarExe, MAX_PATH, TAREXE));
  } else {
    if (!ExpandEnvironmentStrings(szPath, szTarExe, MAX_PATH)) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  if (!GetPrivateProfileString(APPNAME, KEYNAME_7ZEXE, "", szPath, MAX_PATH, szIniFileName)) {
    DO_DEBUG(hr, util_strcpy(sz7ZipExe, MAX_PATH, szAppPF));
    DO_DEBUG(hr, util_strcat(sz7ZipExe, MAX_PATH, SEVENZIPEXE));
  } else {
    if (!ExpandEnvironmentStrings(szPath, sz7ZipExe, MAX_PATH)) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  if (!GetPrivateProfileString(APPNAME, KEYNAME_NKFEXE, "", szPath, MAX_PATH, szIniFileName)) {
    DO_DEBUG(hr, util_strcpy(szNkfExe, MAX_PATH, szAppPF));
    DO_DEBUG(hr, util_strcat(szNkfExe, MAX_PATH, NKFEXE));
  } else {
    if (!ExpandEnvironmentStrings(szPath, szNkfExe, MAX_PATH)) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  hFind = FindFirstFile(szGpgExe, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    hr = HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  FindClose(hFind);

  hFind = FindFirstFile(szTarExe, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    hr = HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  FindClose(hFind);

  hFind = FindFirstFile(sz7ZipExe, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    hr = HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  FindClose(hFind);

  hFind = FindFirstFile(szNkfExe, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    hr = HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  FindClose(hFind);

  /* folderȂ΂ */
  if (FAILED(hr = util_createfolder(szConfDir))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  if (FAILED(hr = gpgezw_initialize(&(pvar->pgpgezbase), szGpgExe, szConfDir, NULL))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (FAILED(hr = gpgezw_initialize_tar(pvar->pgpgezbase, szTarExe))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (FAILED(hr = gpgezw_initialize_7Zip(pvar->pgpgezbase, sz7ZipExe))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (FAILED(hr = gpgezw_initialize_nkf(pvar->pgpgezbase, szNkfExe))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn static HRESULT processCmdLine(HWND hwnd, PCSTR lpCmdLine, BOOL fClose)
 *
 * R}hC̏
 *
 * hwnd ͊{IɃCEChE
 */
static HRESULT
processCmdLine(HWND hwnd, PCSTR pCmdLine, BOOL fClose) {
  HRESULT hr = S_OK;
  PTARGET_FILELIST ptfl = NULL;
  INT i;
  PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);

  /* R}hCXCb`L̊mF */
  if (__argc == 0) goto finish_func;

  if (!util_strcmp(__argv[1], "-e")) {
    /* XCb`'e'̏ꍇ */
    if (__argc < 4) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    /* tfl\̂Ƀf[^i[ */
    if (FAILED(hr = target_filelist_create(&ptfl, __argv[2]))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    /* t@CXgɊi[ */
    for (i = 3; i < __argc; ++i) {
      if (FAILED(hr = target_filelist_add_node(ptfl, __argv[i]))) {
        DEBUGERROR(0, hr);
        target_filelist_release(&ptfl);
        ptfl = NULL;
        goto finish_func;
      }
    }
    if (FAILED(hr)) {
      DEBUGERROR(0, hr);
      target_filelist_release(&ptfl);
      ptfl = NULL;
      goto finish_func;
    }
    zgpencrypt(pvar->pgpgezbase, hwnd, ptfl);

    target_filelist_release(&ptfl);

    if (fClose) {
      PostQuitMessage(0);
    }
  } else if (!util_strcmp(__argv[1], "-d")) {
    if (__argc != 3) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    zgpdecrypt(pvar->pgpgezbase, hwnd, __argv[2], NULL);
    if (fClose) {
      PostQuitMessage(0);
    }
  } else if (!util_strcmp(__argv[1], "-dh")) {
    if (__argc != 4) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    zgpdecrypt(pvar->pgpgezbase, hwnd, __argv[3], __argv[2]);
    if (fClose) {
      PostQuitMessage(0);
    }
  }
finish_func:
  if (ptfl) target_filelist_release(&ptfl);
  return hr;
}

/*------------------------------------------------------------*/
/*
 * rbg}bvɁA]rbg}bv𐶐
 *
 */
static HRESULT
generate_rev_bitmap(HBITMAP hsrc, HBITMAP *phdst1) {
  HRESULT hr = S_OK;
  HBITMAP hprev = NULL;
  HDC memdc1 = NULL, memdc2 = NULL;

  memdc1 = CreateCompatibleDC(NULL);
  if (!memdc1) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    goto finish_func;
  }
  SelectObject(memdc1, hsrc);

  *phdst1 = CreateCompatibleBitmap(memdc1, LARGEICON_WIDTH, LARGEICON_HEIGHT);
  if (!(*phdst1)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    goto finish_func;
  }

  memdc2 = CreateCompatibleDC(memdc1);
  if (!memdc2) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    goto finish_func;
  }
  hprev = SelectObject(memdc2, (*phdst1));

  if (!BitBlt(memdc2, 0, 0, LARGEICON_WIDTH, LARGEICON_HEIGHT, memdc1, 0, 0, NOTSRCCOPY)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    SelectObject(memdc2, hprev);
    goto finish_func;
  }

  DeleteDC(memdc2);
  memdc2 = NULL;
  DeleteDC(memdc1);
  memdc1 = NULL;

finish_func:
  if (memdc2) DeleteDC(memdc2);
  if (memdc1) DeleteDC(memdc1);
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn static BOOL Cls_OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct);
 *
 * EChE̍쐬
 *
 */
static BOOL
Cls_OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct) {
  HRESULT hr = S_OK;
  UINT borderwidth = 0;
  PJANUS_VAR pvar = NULL;
  RECT rc0, rc0c, rc1, rc2;

  pvar = (PJANUS_VAR)util_malloc(sizeof(JANUS_VAR));
  if (!pvar) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    return FALSE;
  }
  ZeroMemory(pvar, sizeof(JANUS_VAR));
  pvar->pgpgezbase = NULL;
  SetWindowLong(hwnd, GWL_USERDATA, (LONG)pvar);

  if (FAILED(hr = LoadOptions(pvar))) {
    return FALSE;
  }

  if (gpgezw_getMyFpr(pvar->pgpgezbase)) {
  } else {
    HMENU hMenu = NULL;
    
    if (hMenu = GetMenu(hwnd)) {
      EnableMenuItem(hMenu, IDM_EDIT_KEY, MF_BYCOMMAND | MF_DISABLED | MF_GRAYED);
      DrawMenuBar(hwnd);
    }
  }

  /* ẼX^eBbNRg[쐬 */
  g_hStatic1 = CreateWindowEx(WS_EX_CLIENTEDGE | WS_EX_ACCEPTFILES, "STATIC", "", WS_CHILD | WS_BORDER | WS_VISIBLE | SS_NOTIFY | SS_BITMAP, 0, 0, LARGEICON_WIDTH, LARGEICON_HEIGHT, hwnd, NULL, g_hInstance, NULL);
  GetWindowRect(g_hStatic1, &rc0);
  GetClientRect(g_hStatic1, &rc0c);
  borderwidth = (rc0.right - rc0.left - (rc0c.right - rc0c.left)) / 2;
  MoveWindow(g_hStatic1,  0, 0, LARGEICON_WIDTH + borderwidth * 2, LARGEICON_HEIGHT + borderwidth * 2, TRUE);
  GetWindowRect(g_hStatic1, &rc1);
  g_hStatic2 = CreateWindowEx(WS_EX_CLIENTEDGE | WS_EX_ACCEPTFILES, "STATIC", "", WS_CHILD | WS_BORDER | WS_VISIBLE | SS_NOTIFY | SS_BITMAP, rc1.right - rc1.left, 0, LARGEICON_HEIGHT + borderwidth * 2, LARGEICON_HEIGHT + borderwidth * 2, hwnd, NULL, g_hInstance, NULL);
  GetWindowRect(g_hStatic2, &rc2);

  /* Bitmapǂݍ */
  {
    CHAR szAppPF[MAX_PATH + 1];
    CHAR szIniFileName[MAX_PATH + 1];
    CHAR szPath[MAX_PATH + 1];
    CHAR szFileName[MAX_PATH + 1];

    GetModuleFileName(NULL, szAppPF, MAX_PATH);
    *(PathFindFileName(szAppPF)) = '\0';
    if (FAILED(hr = util_strcpy(szIniFileName, MAX_PATH, szAppPF))) {
      DEBUGERROR(hr, 0);
      return FALSE;
    }
    if (FAILED(hr = util_strcat(szIniFileName, MAX_PATH, INI_FILENAME))) {
      DEBUGERROR(hr, 0);
      return FALSE;
    }

    /* encpBitmap */
    g_hBitmap_enc1 = NULL;
    if (GetPrivateProfileString(APPNAME, LARGEICON_ENC, "", szPath, MAX_PATH, szIniFileName)) {
      if (ExpandEnvironmentStrings(szPath, szFileName, MAX_PATH)) {
        g_hBitmap_enc1 = LoadImage(NULL, szFileName, IMAGE_BITMAP, LARGEICON_WIDTH, LARGEICON_HEIGHT, LR_LOADFROMFILE);
      }
    }
    /* Bitmapt@Cꍇ͊̉ʂ\ */
    if (!g_hBitmap_enc1) {
      g_hBitmap_enc1 = LoadImage(g_hInstance, MAKEINTRESOURCE(IDB_BITMAP_ENC1F), IMAGE_BITMAP, LARGEICON_WIDTH, LARGEICON_HEIGHT, LR_DEFAULTCOLOR);
      if (!g_hBitmap_enc1) {
        hr = HRESULT_FROM_WIN32(GetLastError());
        DEBUGERROR(hr, 0);
        return FALSE;
      }
    }
    /* ]rbg}bv */
    if (FAILED(hr = generate_rev_bitmap((HBITMAP)g_hBitmap_enc1, &((HBITMAP)g_hBitmap_enc2)))) {
      DEBUGERROR(hr, 0);
      return FALSE;
    }
    

    /* decpBitmap */
    g_hBitmap_dec1 = NULL;
    if (GetPrivateProfileString(APPNAME, LARGEICON_DEC, "", szPath, MAX_PATH, szIniFileName)) {
      if (ExpandEnvironmentStrings(szPath, szFileName, MAX_PATH)) {
        g_hBitmap_dec1 = LoadImage(NULL, szFileName, IMAGE_BITMAP, LARGEICON_WIDTH, LARGEICON_HEIGHT, LR_LOADFROMFILE);
      }
    }
    /* Bitmapt@Cꍇ͊̉ʂ\ */
    if (!g_hBitmap_dec1) {
      g_hBitmap_dec1 = LoadImage(g_hInstance, MAKEINTRESOURCE(IDB_BITMAP_DEC1F), IMAGE_BITMAP, LARGEICON_WIDTH, LARGEICON_HEIGHT, LR_DEFAULTCOLOR);
      if (!g_hBitmap_dec1) {
        hr = HRESULT_FROM_WIN32(GetLastError());
        DEBUGERROR(hr, 0);
        return FALSE;
      }
    }
    /* ]rbg}bv */
    if (FAILED(hr = generate_rev_bitmap((HBITMAP)g_hBitmap_dec1, &((HBITMAP)g_hBitmap_dec2)))) {
      DEBUGERROR(hr, 0);
      return FALSE;
    }

  }
    
  /* X^eBbNRg[ɃC[W\ */
  SendMessage(g_hStatic1, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)g_hBitmap_enc1);
  defProcWndForStatic1 = (WNDPROC) GetWindowLong(g_hStatic1, GWL_WNDPROC);
  SetWindowLong(g_hStatic1, GWL_WNDPROC, (LONG)procWndStatic1);

  /* EX^eBbNRg[ɃC[W\ */
  SendMessage(g_hStatic2, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)g_hBitmap_dec1);
  defProcWndForStatic2 = (WNDPROC) GetWindowLong(g_hStatic2, GWL_WNDPROC);
  SetWindowLong(g_hStatic2, GWL_WNDPROC, (LONG)procWndStatic2);

  GetWindowRect(hwnd, &rc0);
  GetClientRect(hwnd, &rc0c);
  borderwidth = (rc0.right - rc0.left - (rc0c.right - rc0c.left)) / 2;
  MoveWindow(hwnd, rc0.left, rc0.top, rc1.right - rc1.left + rc2.right - rc2.left + borderwidth * 2, rc1.bottom - rc1.top + (rc0.bottom - rc0.top) - (rc0c.bottom - rc0c.top), TRUE);

  /* wvVXe */
  HtmlHelp(NULL, NULL, HH_INITIALIZE, (DWORD) &cookieHelp);

  /* e@t邩ǂ`FbNAȂ΃GNX|[gj[gpsɂ */
  if (FAILED(KageHinata_CanExportKageKoboshi())) {
    HMENU hMenu = GetMenu(hwnd);
    EnableMenuItem(hMenu, IDM_EXPORTKAGEKOBOSHI, MF_BYCOMMAND | MF_GRAYED);
    DrawMenuBar(hwnd);
  }

  return TRUE;
}

/*------------------------------------------------------------*/
static void
Cls_OnDestroy(HWND hwnd) {
  PJANUS_VAR pvar = NULL;

  pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);
  if (pvar) {
    gpgezw_release(&(pvar->pgpgezbase));
    util_free(pvar);
  }
  HtmlHelp(NULL, NULL, HH_UNINITIALIZE, cookieHelp);
  RemoveProp(hwnd, szExeFileName);
  PostQuitMessage(0);
  return;
}

/*------------------------------------------------------------*/
static void
Cls_OnClose(HWND hwnd) {
  DestroyWindow(hwnd);
}

/*------------------------------------------------------------*/
static void
Cls_OnPaint(HWND hwnd) {

  return;
}

/*------------------------------------------------------------*/
void
Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {

  if (hwndCtl == g_hStatic1) {
    if (codeNotify == STN_DBLCLK) {
      SelectFilesAndEncrypt(hwnd);
    }
  } else if (hwndCtl == g_hStatic2) {
    if (codeNotify == STN_DBLCLK) {
      SelectFileAndDecrypt(hwnd);
    }
  } else {
    switch(id) {
    case IDM_ENCRYPT:
      {
        SelectFilesAndEncrypt(hwnd);
        ResetDebugError();
      }
      break;
    case IDM_DECRYPT:
      {
        SelectFileAndDecrypt(hwnd);
        ResetDebugError();
      }
      break;
    case IDM_DECRYPTONLY:
      {
        SelectFileAndDecryptOnly(hwnd);
        ResetDebugError();
      }
      break;
    case IDM_FORYOUREYESONLY:
      {
        PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);
        FlashPicture(g_hStatic1, g_hBitmap_enc1, g_hBitmap_enc2, g_hBitmap_enc1);
        if (gpgezw_getMyFpr(pvar->pgpgezbase)) {
          gpgezw_encrypt_fyeo(pvar->pgpgezbase, hwnd, NULL, NULL, NULL, FALSE, TRUE, g_encfiletype_gpg_pub);
        } else {
          gpgezw_encrypt_fyeo(pvar->pgpgezbase, hwnd, NULL, NULL, NULL, FALSE, TRUE, g_encfiletype_gpg_sym);
        }
        ResetDebugError();
      }
      break;
    case IDM_SIGN:
      {
  /*      SelectFileAndSign(hwnd); */
      }
      break;
    case IDM_VERIFY:
      {
  /*      SelectFileAndVerify(hwnd); */
      }
      break;
    case IDM_QUIT:
      {
        SendMessage(hwnd, WM_CLOSE, 0, 0L);
      }
      break;
    case IDM_EDIT_MYKEY:
      {
        PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);

        if (!gpgezw_getMyFpr(pvar->pgpgezbase)) {
          CHAR m_buff[MAX_PATH + 1];

          if (LoadString(g_hInstance, IDS_CONFIRM_GENERATEKEY, m_buff, MAX_PATH) == 0) {
            util_strcpy(m_buff, MAX_PATH, "Create a GPG Key Pair");
          }
          if (MessageBox(hwnd, m_buff, "Confirmation", MB_YESNO | MB_ICONQUESTION) == IDNO) {
            break;
          }
          if (FAILED(gpgezw_prepareKeypair(pvar->pgpgezbase, hwnd))) {
            break;
          }
          if (!(gpgezw_getMyFpr(pvar->pgpgezbase))) {
            /* TODO: Show Error Message */
            break;
          }
          {
            HMENU hMenu = NULL;
            
            if (hMenu = GetMenu(hwnd)) {
              EnableMenuItem(hMenu, IDM_EDIT_KEY, MF_BYCOMMAND | MF_ENABLED);
              DrawMenuBar(hwnd);
            }
          }
        }
        if (gpgezw_fNeedPassphrase(pvar->pgpgezbase)) {
          if (!gpgezw_getPassphrseForPublicKeyEncryption(pvar->pgpgezbase, hwnd, 2)) {
            MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
            break;
          }
        }
        gpgezw_editMyGpgKey(pvar->pgpgezbase, hwnd, NULL, NULL, FALSE, g_encfiletype_key);
        ResetDebugError();
      }
      break;
    case IDM_EDIT_KEY:
      {
        PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);

        if (gpgezw_fNeedPassphrase(pvar->pgpgezbase)) {
          if (!gpgezw_getPassphrseForPublicKeyEncryption(pvar->pgpgezbase, hwnd, 2)) {
            MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
            break;
          }
        }
        gpgezw_editGpgKey(pvar->pgpgezbase, hwnd, NULL, NULL, FALSE, g_encfiletype_key);
        ResetDebugError();
      }
      break;
    case IDM_EXPORTKAGEKOBOSHI:
      {
        PJANUS_VAR pvar = (PJANUS_VAR)GetWindowLong(hwnd, GWL_USERDATA);
        KageHinata_ExportKageKoboshi(hwnd, gpgezw_getConfDir(pvar->pgpgezbase), gpgezw_getMyFpr(pvar->pgpgezbase));
        ResetDebugError();
      }
    case IDM_SHOW_HELP:
      {
        CHAR szHelpFilename[MAX_PATH + 1] = "";

        GetModuleFileName(NULL, szHelpFilename, MAX_PATH);
        *(PathFindFileName(szHelpFilename)) = '\0';
        if (util_strlen(szHelpFilename, -1) + util_strlen(HELP_FILENAME, -1) > MAX_PATH) break;
        util_strcat(szHelpFilename, MAX_PATH, HELP_FILENAME);
        HtmlHelp(hwnd, szHelpFilename, HH_DISPLAY_TOC, 0);
      }
      break;
    case IDM_ABOUT:
      {
        ShowAboutDialog(hwnd);
        ResetDebugError();
      }
      break;
    }
  }
}

/*------------------------------------------------------------*/
LRESULT CALLBACK
procWnd(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_MSG(hwnd, WM_CREATE, Cls_OnCreate);
    HANDLE_MSG(hwnd, WM_DESTROY, Cls_OnDestroy);
    HANDLE_MSG(hwnd, WM_COMMAND, Cls_OnCommand);
    HANDLE_MSG(hwnd, WM_CLOSE, Cls_OnClose);
  case WM_PROCESSCMDLINE:
    processCmdLine(hwnd, (PCSTR)lParam, TRUE);
    break;
  case WM_PROCESSCMDLINE2:
    processCmdLine(hwnd, (PCSTR)lParam, FALSE);
    break;
  default:
    return (DefWindowProc(hwnd, uMsg, wParam, lParam));
  }
  return 0L;
}

/*------------------------------------------------------------*/
int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR lpCmdLine, int nCmdShow)
{
  MSG msg;
  WNDCLASSEX wc;
  BOOL bRet;
  static CHAR szMainWindowClassName[] = APPNAME;
  HANDLE hMutex; /* pXt[YǗꗂȂ悤ɂ邽߁AdNh~ */

  g_hInstance = hInstance;

  /* dN̖h~ */
  {
    PSTR p = szExeFileName;
    if (!GetModuleFileName(NULL, szExeFileName, MAX_PATH)) {
      return 0;
    }
    while (*p) {
      if (IsDBCSLeadByte(*p)) {
        ++p;
        if (!*p) {
          return 0;
        }
      } else {
        if (*p == '\\') *p = '/';
      }
      ++p;
    }
    hMutex = CreateMutex(NULL, TRUE, szExeFileName);
    if (!hMutex) return 0;
    if (GetLastError() == ERROR_ALREADY_EXISTS) {
      /* ̃EChẼANeBx[g */
      HWND hwnd = GetWindow(GetDesktopWindow(), GW_CHILD);
      while (hwnd) {
        if (GetProp(hwnd, szExeFileName)) {
          if (IsIconic(hwnd)) {
            ShowWindow(hwnd, SW_RESTORE);
          }
          SetForegroundWindow(hwnd);
          if (util_strlen(lpCmdLine, -1) > 0) PostMessage(hwnd, WM_PROCESSCMDLINE2, 0, (LPARAM)lpCmdLine);
          return 0;
        }
        hwnd = GetWindow(hwnd, GW_HWNDNEXT);
      }
      return 0;
    }
  }

  /* fobO[h̐ݒ */
  {
    CHAR szAppPF[MAX_PATH + 1] = "";
    CHAR szIniFileName[MAX_PATH + 1] = "";
    GetModuleFileName(NULL, szAppPF, MAX_PATH);
    *(PathFindFileName(szAppPF)) = '\0';
    util_strcpy(szIniFileName, MAX_PATH, szAppPF);
    util_strcat(szIniFileName, MAX_PATH, INI_FILENAME);
    if (GetPrivateProfileInt(APPNAME, KEYNAME_IFDEBUG, 0, szIniFileName)) {
      CHAR szFileName[MAX_PATH + 1];
      util_strcpy(szFileName, MAX_PATH, szAppPF);
      util_strcat(szFileName, MAX_PATH, "debug.log");
      InitDebugError(szFileName);

      util_strcpy(szFileName, MAX_PATH, szAppPF);
      util_strcat(szFileName, MAX_PATH, "gpgstatus.log");
      gpgez_setstatuslog(szFileName);
    } else {
      InitDebugError("");
      gpgez_setstatuslog("");
    }
  }
  InitCommonControls();
  CoInitialize(NULL);
  InitializeCriticalSection(&cs);

  /* ϐAPPDATAmF */
  {
    CHAR szPath[MAX_PATH + 1] = "";
    DWORD dwLen = 0;

    dwLen = GetEnvironmentVariable("APPDATA", szPath, MAX_PATH);
    if (dwLen >= MAX_PATH) {
      return 0;
    } else if (dwLen == 0) {
      HANDLE hFind;
      WIN32_FIND_DATA wfd;

      if (!SHGetSpecialFolderPath(NULL, szPath, CSIDL_APPDATA, FALSE)) {
        DEBUGERROR(0, HRESULT_FROM_WIN32(GetLastError()));
        return 0;
      }
      /* User AppData Folder Ȃ΃G[ */
      hFind = FindFirstFile(szPath, &wfd);
      if (hFind == INVALID_HANDLE_VALUE) {
        DEBUGERROR(0, HRESULT_FROM_WIN32(ERROR_INVALID_DATA));
        return 0;
      }
      FindClose(hFind);

      if (!SetEnvironmentVariable("APPDATA", szPath)) {
        return 0;
      }
    }
  }
    
  /* ϐAPPBINݒ */
  {
    CHAR szPath[MAX_PATH + 1] = "";

    if (!GetModuleFileName(NULL, szPath, MAX_PATH)) {
      return 0;
    }
    *(PathFindFileName(szPath) - 1) = '\0';
    if (!SetEnvironmentVariable("APPBIN", szPath)) {
      return 0;
    }
  }

  /* 񃊃\[Xǂݍ */
  {
    int i;
    CHAR m_buff[MAX_PATH + 1] = "";

    for (i = 0; g_encfiletype_pub[i].filetype != FILETYPE_NONE; ++i) {
      m_buff[0] = '\0';
      if (LoadString(g_hInstance, g_encfiletype_pub[i].dwStr, m_buff, MAX_PATH)) {
        g_encfiletype_pub[i].filedesc1 = util_strdup(m_buff);
      }
    }
    for (i = 0; g_encfiletype_sym[i].filetype != FILETYPE_NONE; ++i) {
      m_buff[0] = '\0';
      if (LoadString(g_hInstance, g_encfiletype_sym[i].dwStr, m_buff, MAX_PATH)) {
        g_encfiletype_sym[i].filedesc1 = util_strdup(m_buff);
      }
    }
    for (i = 0; g_encfiletype_gpg_pub[i].filetype != FILETYPE_NONE; ++i) {
      m_buff[0] = '\0';
      if (LoadString(g_hInstance, g_encfiletype_gpg_pub[i].dwStr, m_buff, MAX_PATH)) {
        g_encfiletype_gpg_pub[i].filedesc1 = util_strdup(m_buff);
      }
    }
    for (i = 0; g_encfiletype_gpg_sym[i].filetype != FILETYPE_NONE; ++i) {
      m_buff[0] = '\0';
      if (LoadString(g_hInstance, g_encfiletype_gpg_sym[i].dwStr, m_buff, MAX_PATH)) {
        g_encfiletype_gpg_sym[i].filedesc1 = util_strdup(m_buff);
      }
    }
  }

  ZeroMemory(&wc, sizeof(WNDCLASSEX));
  wc.cbSize = sizeof(WNDCLASSEX);
  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = procWnd;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0; /* DLGWINDOWEXTRA */;
  wc.hInstance = hInstance;
  wc.hIcon = LoadImage(g_hInstance, MAKEINTRESOURCE(IDI_ICON2), IMAGE_ICON, 32, 32, LR_CREATEDIBSECTION);
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = (HBRUSH) COLOR_WINDOW;
  wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU1);
  wc.lpszClassName = szMainWindowClassName;
  wc.hIconSm = LoadImage(g_hInstance, MAKEINTRESOURCE(IDI_ICON2), IMAGE_ICON, 16, 16, LR_CREATEDIBSECTION);
  RegisterClassEx(&wc);

  msg.hwnd = CreateWindow(szMainWindowClassName, APPNAME, WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, hInstance, NULL);
  if (!SetProp(msg.hwnd, szExeFileName, (HANDLE)1)) {
    return 0;
  }

  ShowWindow(msg.hwnd, nCmdShow);
  UpdateWindow(msg.hwnd);

  if (util_strlen(lpCmdLine, -1) > 0) PostMessage(msg.hwnd, WM_PROCESSCMDLINE, 0, (LPARAM)lpCmdLine);

  while ((bRet = GetMessage(&msg, NULL, 0, 0)) != 0) {
    if (bRet == -1) {
      MessageBox(msg.hwnd, "GetMessage Error", "Error", MB_OK);
      break;
    } else {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  }

  /* 񃊃\[X */
  {
    int i;
    for (i = 0; g_encfiletype_pub[i].filetype != FILETYPE_NONE; ++i) {
      if (g_encfiletype_pub[i].filedesc1) {
        util_free(g_encfiletype_pub[i].filedesc1);
      }
    }
    for (i = 0; g_encfiletype_sym[i].filetype != FILETYPE_NONE; ++i) {
      if (g_encfiletype_sym[i].filedesc1) {
        util_free(g_encfiletype_sym[i].filedesc1);
      }
    }
    for (i = 0; g_encfiletype_gpg_pub[i].filetype != FILETYPE_NONE; ++i) {
      if (g_encfiletype_gpg_pub[i].filedesc1) {
        util_free(g_encfiletype_gpg_pub[i].filedesc1);
      }
    }
    for (i = 0; g_encfiletype_gpg_sym[i].filetype != FILETYPE_NONE; ++i) {
      if (g_encfiletype_gpg_sym[i].filedesc1) {
        util_free(g_encfiletype_gpg_sym[i].filedesc1);
      }
    }
  }

  DeleteCriticalSection(&cs);
  CoUninitialize();
  return (int)msg.wParam;
}

/*------------------------------------------------------------*/
