/**
 * about.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "KageHinata/about.c"

#include <windows.h>
#include <windowsx.h>

#include "../include/util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

/*------------------------------------------------------------*/
static BOOL
inputbox_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  DWORD dwSize = 0, dwLang = 0, dwErr = ERROR_SUCCESS, dwHandle = 0;
  CHAR szFilename[MAX_PATH + 1];
  LPVOID pBuff = NULL;
  LPSTR buff = NULL;
  CHAR szProductName[MAX_PATH + 1] = "", szProductVersion[MAX_PATH + 1] = "", szLegalCopyright[MAX_PATH + 1] = "";
  UINT uLen = 0, uLabelLen = 0;
  CHAR szLabel[MAX_PATH * 3 + 1] = "";
  
  if (GetModuleFileName(NULL, szFilename, MAX_PATH) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if ((dwSize = GetFileVersionInfoSize(szFilename, NULL)) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  pBuff = (PBYTE)util_malloc(dwSize);
  if (!pBuff) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ZeroMemory(pBuff, dwSize);
  if (GetFileVersionInfo(szFilename, dwHandle, dwSize, pBuff) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (VerQueryValue(pBuff, "\\StringFileInfo\\041104b0\\ProductName", (LPVOID *)&buff, &uLen) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (uLen > 0) {
    ZeroMemory(szProductName, MAX_PATH);
    util_strncpy(szProductName, MAX_PATH, buff, uLen);
  }
  if (VerQueryValue(pBuff, "\\StringFileInfo\\041104B0\\ProductVersion", (LPVOID *)&buff, &uLen) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (uLen > 0) {
    ZeroMemory(szProductVersion, MAX_PATH);
    util_strncpy(szProductVersion, MAX_PATH, buff, uLen);
  }
  if (VerQueryValue(pBuff, "\\StringFileInfo\\041104B0\\LegalCopyright", (LPVOID *)&buff, &uLen) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (uLen > 0) {
    ZeroMemory(szLegalCopyright, MAX_PATH);
    util_strncpy(szLegalCopyright, MAX_PATH, buff, uLen);
  }
#ifndef PORTABLE
  wsprintf(szLabel, "%s Version %s", szProductName, szProductVersion);
#else
  wsprintf(szLabel, "%s Portable Version %s", szProductName, szProductVersion);
#endif
  Static_SetText(GetDlgItem(hwnd, IDC_STATIC_PRODUCTNAME), szLabel);
  Static_SetText(GetDlgItem(hwnd, IDC_STATIC_LEGALCOPYRIGHT), szLegalCopyright);
finish_func:
  if (pBuff) util_free(pBuff);
  if (dwErr != ERROR_SUCCESS) EndDialog(hwnd, (int)dwErr);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
inputbox_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)ERROR_SUCCESS);
  return;
}

/*------------------------------------------------------------*/
static VOID
inputbox_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
BOOL CALLBACK
procAboutDialog(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, inputbox_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, inputbox_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, inputbox_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
HRESULT
ShowAboutDialog(HWND hwnd) {
  INT iRet;
  iRet = DialogBox(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_VERSIONINFO), hwnd, procAboutDialog);
  return HRESULT_FROM_WIN32((DWORD)iRet);
}

/*------------------------------------------------------------*/
