/* export_KageKoboshi.c
 * Copyright (C) 2006-2008 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "KageHinata/export_KageKoboshi.c"

#include <windows.h>
#include <commctrl.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <crtdbg.h>

#include "resource.h"

#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"

#pragma comment(lib, "Comctl32.lib")
#pragma comment(lib, "shlwapi.lib")

#define KAGEKOBOSHI_DIR "KageKoboshi"
#define KAGEKOBOSHI_EXE "KageKoboshi.exe"
#define KAGEKOBOSHI_CONFDIR "conf"
#define KAGEKOBOSHI_PUBRING "conf\\pubring.gpg"
#define SEVENZIP_EXE "bin\\7za.exe"
#define GPG_EXE "bin\\gpg.exe"

HRESULT
KageHinata_CanExportKageKoboshi(VOID) {
  CHAR szPath[MAX_PATH + 1] = "";
  CHAR szKageHinataDir[MAX_PATH + 1] = "";
  HANDLE hFind;
  WIN32_FIND_DATA wfd;
  HRESULT hr = S_OK;

  GetModuleFileName(NULL, szKageHinataDir, MAX_PATH);
  *(PathFindFileName(szKageHinataDir)) = '\0';

  if (FAILED(hr = util_strcpy(szPath, MAX_PATH, szKageHinataDir))) {
    return E_FAIL;
  }
  if (FAILED(hr = util_strcat(szPath, MAX_PATH, KAGEKOBOSHI_DIR "\\" KAGEKOBOSHI_EXE))) {
    return E_FAIL;
  }

  hFind = FindFirstFile(szPath, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    return E_FAIL;
  }
  FindClose(hFind);

  if (FAILED(hr = util_strcpy(szPath, MAX_PATH, szKageHinataDir))) {
    return E_FAIL;
  }
  if (FAILED(hr = util_strcat(szPath, MAX_PATH, SEVENZIP_EXE))) {
    return E_FAIL;
  }

  hFind = FindFirstFile(szPath, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    return E_FAIL;
  }
  FindClose(hFind);

  if (FAILED(hr = util_strcpy(szPath, MAX_PATH, szKageHinataDir))) {
    return E_FAIL;
  }
  if (FAILED(hr = util_strcat(szPath, MAX_PATH, GPG_EXE))) {
    return E_FAIL;
  }
  hFind = FindFirstFile(szPath, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    return E_FAIL;
  }
  FindClose(hFind);

  if (FAILED(hr = util_strcpy(szPath, MAX_PATH, szKageHinataDir))) {
    return E_FAIL;
  }
  if (FAILED(hr = util_strcat(szPath, MAX_PATH, KAGEKOBOSHI_DIR "\\" KAGEKOBOSHI_CONFDIR))) {
    return E_FAIL;
  }
  hFind = FindFirstFile(szPath, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    if (!CreateDirectory(szPath, NULL)) {
      return E_FAIL;
    }
  } else {
    if (!(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
      return E_FAIL;
    }
    FindClose(hFind);
  }

  return S_OK;
}

HRESULT
KageHinata_ExportKageKoboshi(HWND hwnd, PCSTR pHomeDir, PCSTR pFpr) {
  CHAR szKageKoboshiZip[MAX_PATH + 1] = "KageKoboshi.zip";
  CHAR szKageHinataDir[MAX_PATH + 1] = "";
  OPENFILENAME ofn;
  HANDLE hFind;
  WIN32_FIND_DATA wfd;
  HRESULT hr = S_OK;

  GetModuleFileName(NULL, szKageHinataDir, MAX_PATH);
  *(PathFindFileName(szKageHinataDir)) = '\0';

  if (!pFpr) {
    if (IDYES != MessageBox(hwnd, "You can use symmetric encryption only.\nProceed anyway ?", "Confirmation", MB_YESNO)) {
      return E_FAIL;
    }
  }

  /* GNX|[gt@Cw */
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrTitle = "Export KageKoboshi";
  ofn.lpstrFilter = "zip(*.zip)\0*.zip\0All files(*.*)\0*.*\0\0";
  ofn.lpstrDefExt = "zip";
  ofn.lpstrFile = szKageKoboshiZip;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT;
  if (!GetSaveFileName(&ofn)) {
    DWORD dwErr = CommDlgExtendedError();
    /* G[`FbN */
    goto finish_func;
  }

  /* KAGEKOBOSHIZIP폜 */
  hFind = FindFirstFile(szKageKoboshiZip, &wfd);
  if (hFind != INVALID_HANDLE_VALUE) {
    if (!DeleteFile(szKageKoboshiZip)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    FindClose(hFind);
  }

  {
    /* KageKoboshicd */
    /* ..\bin\\7za.exe a -tzip %KAGEKOBOSHI_ZIP% KageKoboshi.exe KageKoboshi.chm KageKoboshi.ini Readme.txt */
    CHAR szCommandLine[MAX_PATH * 3 + 1] = "";
    CHAR szCurrentDirectory[MAX_PATH + 1] = "";
    STARTUPINFO si;
    PROCESS_INFORMATION pi;

    if (FAILED(hr = util_strcpy(szCurrentDirectory, MAX_PATH, szKageHinataDir))) {
      return E_FAIL;
    }

    if (FAILED(hr = util_strcat(szCurrentDirectory, MAX_PATH, KAGEKOBOSHI_DIR))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    szCommandLine[0] = '"'; szCommandLine[1] = '\0';
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 3, szKageHinataDir))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 3, SEVENZIP_EXE "\" a -tzip \""))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 3, szKageKoboshiZip))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 3, "\" KageKoboshi.exe KageKoboshi.chm KageKoboshi.ini Readme.txt"))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    ZeroMemory(&si, sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESHOWWINDOW;
    si.wShowWindow = SW_HIDE;
    
    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
    
    if (!CreateProcess(NULL, szCommandLine, NULL, NULL, FALSE, 0, NULL, szCurrentDirectory, &si, &pi)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    CloseHandle(pi.hThread);
    WaitForSingleObject(pi.hProcess, INFINITE);
    CloseHandle(pi.hProcess);
  }

  {
    /* KageHinatacd */
    /* ..\bin\\7za.exe a -tzip %KAGEKOBOSHI_ZIP% gpgezw.dll bin\archive1.dll bin\bsdtar.exe bin\bzip2.dll bin\gpg.exe bin\iconv.dll bin\nkfwin.exe bin\zlib1.dll */
    CHAR szCommandLine[MAX_PATH * 3 + 1] = "";
    CHAR szCurrentDirectory[MAX_PATH + 1] = "";
    STARTUPINFO si;
    PROCESS_INFORMATION pi;

    if (FAILED(hr = util_strcpy(szCurrentDirectory, MAX_PATH, szKageHinataDir))) {
      return E_FAIL;
    }

    szCommandLine[0] = '"'; szCommandLine[1] = '\0';
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 3, szKageHinataDir))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 3, SEVENZIP_EXE "\" a -tzip \""))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 3, szKageKoboshiZip))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 3, "\" gpgezw.dll bin\\archive1.dll bin\\bsdtar.exe bin\\bzip2.dll bin\\gpg.exe bin\\iconv.dll bin\\nkfwin.exe bin\\zlib1.dll"))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    ZeroMemory(&si, sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESHOWWINDOW;
    si.wShowWindow = SW_HIDE;
    
    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
    
    if (!CreateProcess(NULL, szCommandLine, NULL, NULL, FALSE, 0, NULL, szCurrentDirectory, &si, &pi)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    CloseHandle(pi.hThread);
    WaitForSingleObject(pi.hProcess, INFINITE);
    CloseHandle(pi.hProcess);
  }

  /* %EXEGPG% --homedir=%GPGHOMEDIR% --output PUBRING --export %MYFPR% */
  if (pFpr) {
    CHAR szCommandLine[MAX_PATH * 4 + 1] = "";
    CHAR szPath[MAX_PATH + 1] = "";
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    DWORD dwExitCode = 0;

    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));

    szCommandLine[0] = '"'; szCommandLine[1] = '\0';
    if (FAILED(util_strcat(szCommandLine, MAX_PATH * 4, szKageHinataDir))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, GPG_EXE))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, "\" --batch --no-tty --yes --homedir=\""))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, pHomeDir))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, "\" --output=\""))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, szKageHinataDir))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, KAGEKOBOSHI_DIR "\\" KAGEKOBOSHI_PUBRING))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, "\" --export "))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, pFpr))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    ZeroMemory(&si, sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESHOWWINDOW;
    si.wShowWindow = SW_HIDE;
    
    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
    
    if (!CreateProcess(NULL, szCommandLine, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    CloseHandle(pi.hThread);
    WaitForSingleObject(pi.hProcess, INFINITE);
    CloseHandle(pi.hProcess);

    /* %EXEGPG% --homedir=%GPGHOMEDIR% --export %MYFPR% | %EXE7ZA% a -tzip %KAGEKOBOSHI_ZIP% -siconf\pubring.gpg */
    szCommandLine[0] = '"'; szCommandLine[1] = '\0';
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, szKageHinataDir))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, SEVENZIP_EXE))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, "\" a -tzip \""))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, szKageKoboshiZip))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szCommandLine, MAX_PATH * 4, "\" \"" KAGEKOBOSHI_PUBRING "\" "))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    ZeroMemory(&si, sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESHOWWINDOW;
    si.wShowWindow = SW_HIDE;

    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));

    if (FAILED(util_strcpy(szPath, MAX_PATH, szKageHinataDir))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szPath, MAX_PATH, KAGEKOBOSHI_DIR))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (!CreateProcess(NULL, szCommandLine, NULL, NULL, FALSE, 0, NULL, szPath, &si, &pi)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    CloseHandle(pi.hThread);
    WaitForSingleObject(pi.hProcess, INFINITE);
    CloseHandle(pi.hProcess);
  }

  MessageBox(hwnd, "KageKoboshi Exported", "Information", MB_OK);

finish_func:
  if (FAILED(hr)) {
    MessageBox(hwnd, "Failed to export KageKoboshi", "Error", MB_OK);
  }
  return hr;
}


