/* KageHinataOp.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of JanusDGX.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "KageHinataOp/KageHinataOp.c"

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <strsafe.h>
#include <crtdbg.h>

#include "resource.h"

#pragma comment(lib, "Comctl32.lib")
#pragma comment(lib, "shlwapi.lib")
#pragma comment(lib, "version.lib")

#define OPTAPPNAME "KageHinataOp"
#define APPNAME "KageHinata"
#define INIFILE "KageHinata.ini"

#define OPT_DEBUG "debug"

typedef struct struct_option_vars {
  CHAR pIniFileName[MAX_PATH + 1];
  BOOL fDebug;
} OPTIONVARS, *POPTIONVARS;

HWND hDebug;

HINSTANCE g_hInstance;

/*------------------------------------------------------------*/
static PSTR
util_itoa(INT i, PSTR buff, INT max) {
  int i2 = 0;
  PSTR p = buff;
  INT cnt = 0;

  SetLastError(ERROR_SUCCESS);

  /* i ̌𐔂 */
  if (i == 0) cnt = 1;
  else {
    for (i2 = i; i2; i2 /= 10) ++cnt;
    if (max < cnt - 1) {
      *buff = '\0';
       SetLastError(ERROR_ARITHMETIC_OVERFLOW);
      goto finish_func;
    }
  }

  i2 = i;
  buff[cnt] = '\0';
  while (cnt > 0) {
    --cnt;
    buff[cnt] = '0' + i2 % 10;
    i2 /= 10;
  }
finish_func:
  return buff;
}
/*------------------------------------------------------------*/
static BOOL
optionpage1_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  POPTIONVARS pvar = (POPTIONVARS)lParam;
  BOOL fRet = FALSE;
  
  hDebug = GetDlgItem(hwnd, IDC_CHECK_DEBUG);

  SetWindowLong(hwnd, GWL_USERDATA, lParam);

  if (pvar->fDebug) Button_SetCheck(hDebug, TRUE);

  return fRet;
}

/*------------------------------------------------------------*/
static VOID
optionpage1_Cls_OnClose(HWND hwnd) {
  PostQuitMessage(0);
  return;
}

/*------------------------------------------------------------*/
static VOID
optionpage1_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      POPTIONVARS pvar = (POPTIONVARS)GetWindowLong(hwnd, GWL_USERDATA);
      CHAR buff[11] = "";

      pvar->fDebug = Button_GetCheck(hDebug);
      WritePrivateProfileString(APPNAME, OPT_DEBUG, pvar->fDebug ? "-1" : "0", pvar->pIniFileName);
      PostQuitMessage(0);
    }
    break;
  case IDCANCEL:
    {
      PostQuitMessage(0);
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlg(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_MSG(hDlg, WM_INITDIALOG, optionpage1_Cls_OnInitDialog);
    HANDLE_MSG(hDlg, WM_CLOSE, optionpage1_Cls_OnClose);
    HANDLE_MSG(hDlg, WM_COMMAND, optionpage1_Cls_OnCommand);
  }
  return DefWindowProc(hDlg, uMsg, wParam, lParam);
}

/*------------------------------------------------------------*/
int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR lpCmdLine, int nCmdShow)
{
  MSG msg;
  WNDCLASSEX wc;
  BOOL bRet;
  static CHAR szMainWindowClassName[] = OPTAPPNAME;
  HANDLE hMutex;
  HANDLE hMutexOrig;
  HWND hDlg = NULL;
  OPTIONVARS ovs;

  g_hInstance = hInstance;

  InitCommonControls();

  /* dN̖h~ */
  {
    CHAR szExeFileName[MAX_PATH + 1];
    PSTR p = szExeFileName;

    if (!GetModuleFileName(NULL, szExeFileName, MAX_PATH)) {
      return 0;
    }
    while (*p) {
      if (IsDBCSLeadByte(*p)) {
        ++p;
        if (!*p) {
          return 0;
        }
      } else {
        if (*p == '\\') *p = '/';
      }
      ++p;
    }
    hMutex = CreateMutex(NULL, TRUE, szExeFileName);
    if (!hMutex) return 0;
    if (GetLastError() == ERROR_ALREADY_EXISTS) {
      return 0;
    }
  }

  /* KageHinataN͋NȂ */
  {
    CHAR szExeFileName[MAX_PATH + 1];
    PSTR p = szExeFileName;

    if (!GetModuleFileName(NULL, szExeFileName, MAX_PATH)) {
      return 0;
    }
    *(PathFindFileName(szExeFileName)) = '\0';
    if (FAILED(StringCbCat(szExeFileName, MAX_PATH, APPNAME))) {
      return 0;
    }
    if (FAILED(StringCbCat(szExeFileName, MAX_PATH, ".exe"))) {
      return 0;
    }
    while (*p) {
      if (IsDBCSLeadByte(*p)) {
        ++p;
        if (!*p) {
          return 0;
        }
      } else {
        if (*p == '\\') *p = '/';
      }
      ++p;
    }
    hMutexOrig = CreateMutex(NULL, TRUE, szExeFileName);
    if (!hMutexOrig) return 0;
    if (GetLastError() == ERROR_ALREADY_EXISTS) {
      MessageBox(NULL, "Please close KageHinata.", "ERROR", MB_OK);
      return 0;
    }
  }

  /* pIniFileName 擾 */
  {
    CHAR szFileName[MAX_PATH + 1];

    if (!GetModuleFileName(NULL, szFileName, MAX_PATH)) {
      return 0;
    }
    *(PathFindFileName(szFileName)) = '\0';
    if (FAILED(StringCbCat(szFileName, MAX_PATH, INIFILE))) {
      return 0;
    }
    if (FAILED(StringCbCopy(ovs.pIniFileName, MAX_PATH, szFileName))) {
      return 0;
    }
    ovs.fDebug = GetPrivateProfileInt(APPNAME, OPT_DEBUG, 0, ovs.pIniFileName);
  }

  ZeroMemory(&wc, sizeof(WNDCLASSEX));
  wc.cbSize = sizeof(WNDCLASSEX);
  wc.style = 0; //CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = procDlg;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = DLGWINDOWEXTRA; /* DLGWINDOWEXTRA */;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon(NULL,IDI_APPLICATION);
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = (HBRUSH) COLOR_WINDOW;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = szMainWindowClassName;
  wc.hIconSm = NULL;
  if (!RegisterClassEx(&wc)) {
    DWORD dwErr = 0;
    dwErr = GetLastError();
    return 0;
  }

  hDlg = CreateDialogParam(g_hInstance, MAKEINTRESOURCE(IDD_OPTIONPAGE1), NULL, procDlg, (LPARAM)&ovs);
  if (!hDlg) {
    DWORD dwErr = 0;
    dwErr = GetLastError();
    MessageBox(NULL, "Failed to create main window.", "ERROR", MB_OK);
    return 0;
  }

  ShowWindow(hDlg, nCmdShow);
  UpdateWindow(hDlg);

  while ((bRet = GetMessage(&msg, NULL, 0, 0)) != 0) {
    if (bRet == -1) {
      MessageBox(msg.hwnd, "GetMessage Error", "Error", MB_OK);
      break;
    } else {
      if (!IsDialogMessage(hDlg, &msg)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
    }
  }

  return (int)msg.wParam;
}

/*------------------------------------------------------------*/
