/**
 * gpgez.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgez/gpgez.c"

#include <windows.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <strsafe.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "gpgez_util.h"

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024


extern HINSTANCE g_hInstance;

#define GPGSTATUSLINE_GET_LINE "[GNUPG:] GET_LINE"
#define GPGSTATUSLINE_GET_HIDDEN "[GNUPG:] GET_HIDDEN"
#define GPGSTATUSLINE_GET_BOOL "[GNUPG:] GET_BOOL"

#define NUMBUFF_LEN 10

#define DO_DEBUG(hr, fn) if (FAILED((hr) = (fn))) { DEBUGERROR(0, (hr)); goto finish_func;}
#define DO_DEBUG_RETC(hr, fn, c) if ((fn)) { DWORD dw = (c); (hr) = HRESULT_FROM_WIN32(dw); DEBUGERROR(0, (hr)); goto finish_func;}
#define STRINGLIST_ADDOPTION(hr, pctx, psl, s, f) if (psl) {\
  PSTRINGLIST_NODE pnode = NULL;\
  for (pnode = stringlist_get_head((psl)); pnode; pnode = stringlist_get_next(pnode)) {\
    DO_DEBUG((hr), cmdez_add_option((pctx), s, stringlist_get_string(pnode), (f)));\
  }\
}

/*------------------------------------------------------------*/
typedef struct struct_param_gpgstatus {
  HANDLE hStatus;
  PSTRINGLIST plstatus;
} PARAM_GPGSTATUS, *PPARAM_GPGSTATUS;

/*------------------------------------------------------------*/
typedef struct struct_buffhandle {
  PHANDLE h;
  PCSTR buff;
  DWORD len;
} BUFFHANDLE, *PBUFFHANDLE;

/*------------------------------------------------------------*/
static DWORD WINAPI
procWriteBuff(LPVOID lpv) {
  DWORD dwErr = ERROR_SUCCESS;
  PBUFFHANDLE pbh = (PBUFFHANDLE)lpv;
  DWORD dwWritten;

  if (!WriteFile(pbh->h, pbh->buff, pbh->len, &dwWritten, NULL)) {
    dwErr = GetLastError();
  }
  CloseHandle(pbh->h);
  return dwErr;
}

/*------------------------------------------------------------*/
/**
 * \fn DllExport VOID unescapeTarMessage(LPSTR pLine);
 *
 * bsdtarGXP[vĂ܂ɖ߂֐iɓ{܂j
 * ̕KZȂ邽߁A㏑ŏĂ_ɒ
 *
 */
DllExport VOID
unescapeTarMessage(PSTR pLine)
{
  LPSTR ps, pd;
  for (ps = pLine, pd = pLine; *ps; ps++) {
    if (*ps != '\\') {
      *pd = *ps;
    } else {
      ps++;
      if (*ps == '\\') {
        *pd = '\\';
      } else if (*ps == 'a') {
        *pd = '\a';
      } else if (*ps == 'b') {
        *pd = '\b';
      } else if (*ps == 'f') {
        *pd = '\f';
      } else if (*ps == 'n') {
        *pd = '\n';
      } else if (*ps == 't') {
        *pd = '\t';
      } else if ((*ps >= '0') && (*ps <= '9')) {
        *pd = (*ps - '0') * 64 + (*(ps + 1) - '0') * 8 + (*(ps + 2) - '0');
        ps += 2;
      }
    }
    pd++;
  }
  *pd = '\0';

  for (ps = pLine; *ps; ps++) {
    if (!util_strncmp(ps, ": Can't set permissions: Function not implemented", 49)) {
      *ps = '\0';
      break;
    }
  }
  while ((ps > pLine + 2) && (*(ps - 1) == '\n') && (*(ps - 2) == '\n')) {
    ps--;
    *ps = '\0';
  }
}

/*------------------------------------------------------------*/
DllExport VOID
unescapePercent(PCSTR ps, PSTR pd, DWORD dwLen)
{
  while (*ps) {
    if (*ps != '%') {
      *pd = *ps;
    } else {
      ++ps;
      if ((*ps >= '0') && (*ps <= '9')) {
        *pd = *ps - '0';
      } else if ((*ps >= 'a') && (*ps <= 'f')) {
        *pd = *ps - 'a' + 10;
      } else if ((*ps >= 'A') && (*ps <= 'F')) {
        *pd = *ps - 'A' + 10;
      } else {
        *pd = '\0';
        break;
      }
      ++ps;
      *pd <<= 4;
      if ((*ps >= '0') && (*ps <= '9')) {
        *pd |= *ps - '0';
      } else if ((*ps >= 'a') && (*ps <= 'f')) {
        *pd |= *ps - 'a' + 10;
      } else if ((*ps >= 'A') && (*ps <= 'F')) {
        *pd |= *ps - 'A' + 10;
      } else {
        *pd = '\0';
        break;
      }
    }
    ++pd;
    ++ps;
    --dwLen;
    if (dwLen < 0) break;
  }
  *pd = '\0';
}

/*------------------------------------------------------------*/
/*
* ߂l
* 0: 
* 1: OZ
* 2: sȕ܂܂Ă
*/
DllExport DWORD
ValidateNameForGpg(PCSTR szName)
{
  DWORD dwRet = 0;
  LPCSTR pSrc;

  if (util_strlen(szName, -1) < 5) {
    dwRet = 1;
    goto finish_func;
  }
  if ((*szName >= '0') && (*szName <= '9')) {
    dwRet = 2;
    goto finish_func;
  }
  for (pSrc = szName; *pSrc; pSrc++) {
    if ((*pSrc == '<') || (*pSrc == '>') || (*pSrc == '(') || (*pSrc == ')')) {
      dwRet = 2;
      goto finish_func;
    }
  }
finish_func:
  return dwRet;
}

/*------------------------------------------------------------*/
/*
* DWORD ValidateEmailForGpg(PCSTR szName)
* ߂l
* 0: 
* 1: `s
* 2: sȕ܂܂Ă
*/
DllExport DWORD
ValidateEmailForGpg(PCSTR szEmail)
{
  DWORD dwRet = 0;
  BOOL fAt = FALSE;
  PCSTR p;

  if ((*szEmail == '@') ||
      (*(szEmail + strlen(szEmail) - 1) == '@') ||
      (*(szEmail + strlen(szEmail) - 1) == '.')) {
    dwRet = 1;
    goto finish_func;
  }
  for (p = szEmail; *p; ++p) {
    if ((*p == '.') && (*(p + 1) == '.')) {
      dwRet = 1;
      goto finish_func;
    }
  }

  for (p = szEmail; *p; ++p) {
    if (*p == '@') {
      if (fAt) {
        dwRet = 1;
        goto finish_func;
      } else {
        fAt = TRUE;
      }
    }
  }
  if (!fAt) {
    dwRet = 1;
    goto finish_func;
  }

  fAt = FALSE;
  for (p = szEmail; *p; ++p) {
    if (*p & 0x80) {
      dwRet = 2;
      goto finish_func;
    }
    if (*p == '@') {
      fAt = TRUE;
    } else {
      if ((!fAt) &&
          (!(((*p >= '0') && (*p <= '9')) || 
             ((*p >= 'a') && (*p <= 'z')) ||
             ((*p >= 'A') && (*p <= 'Z')))) && /* isalnum */
          (*p != '+') &&
          (*p != '-') &&
          (*p != '_') &&
          (*p != '.')) {
        dwRet = 1;
        goto finish_func;
      } else if ((fAt)  &&
                 (!(((*p >= '0') && (*p <= '9')) || 
                   ((*p >= 'a') && (*p <= 'z')) ||
                   ((*p >= 'A') && (*p <= 'Z')))) && /* isalnum */
                 (*p != '+') &&
                 (*p != '-') &&
                 (*p != '_') &&
                 (*p != '.')) {
        dwRet = 1;
        goto finish_func;
      }
    }
  }
finish_func:
  return dwRet;
}

/*------------------------------------------------------------*/
DllExport DWORD
ValidateCommentForGpg(PCSTR szComment)
{
  DWORD dwRet = 0;
  PCSTR p;

  for (p = szComment; *p; ++p) {
    if ((*p == '(') || (*p == ')')) {
      dwRet = 2;
      goto finish_func;
    }
  }
finish_func:
  return dwRet;
}

/*------------------------------------------------------------*/
/*
 * [AhX̌`ƂāAȉ3
 * 1. user1@example.com `
 * 2. name<user1@example.com> `
 * 3. name(comment)<user1@example.com> `
 */
/*------------------------------------------------------------*/
DllExport HRESULT
GetNameFromGpgId(PCSTR pGpgId, PSTR pName, DWORD dwMax)
{
  HRESULT hr = S_OK;
  PCSTR p = pGpgId, ps = NULL;
  PCSTR pe = NULL; /* '<'邢'('̈ʒu */

  pName[0] = '\0';

  /* '<''('邩mF */
  for (p = pGpgId; *p; ++p) {
    if ((*p == '(') || (*p == '<')) {
      pe = p;
      break;
    }
  }

  if (pe) {
    ps = pGpgId;
    p = pe;
    while (*ps == ' ') ++ps;
    if (ps == p) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
    while (*(p - 1) == ' ') --p;
    if (FAILED(hr = StringCbCopyN(pName, dwMax, ps, p - ps))) {
      goto finish_func;
    }
    pName[p - ps] = '\0';
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
GetEmailFromGpgId(PCSTR pGpgId, PSTR pEmail, DWORD dwMax)
{
  HRESULT hr = S_OK;
  PCSTR p = pGpgId, ps = NULL;

  pEmail[0] = '\0';

  /* '<'邩mF */
  for (p = pGpgId; *p; ++p) {
    if (*p == '<') {
      ps = p;
      break;
    }
  }

  if (ps) {
    ++ps;
    /* 擪̋󔒂ǂݔ΂ */
    while (*ps == ' ') ++ps;
    p = ps;
    while (*p) {
      if (*p == '>') break;
      ++p;
    }
    /* '>'Ȃ΃G[ */
    if (!*p) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
    /* ̗]vȋ󔒂 */
    while (*(p - 1) == ' ') --p;
    if (FAILED(hr = StringCbCopyN(pEmail, dwMax - 1, ps, p - ps))) {
      goto finish_func;
    }
    pEmail[p - ps] = '\0';
  } else { /* 1܂4` */
    ps = pGpgId;
    while (*ps == ' ') ++ps;
    if (!*ps) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
    p = ps;
    while (*p) ++p;
    while (*(p - 1) == ' ') --p;
    if (FAILED(hr = StringCbCopyN(pEmail, dwMax - 1, ps, p - ps))) {
      goto finish_func;
    }
    pEmail[p - ps] = '\0';
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
GetCommentFromGpgId(PCSTR pGpgId, PSTR pComment, DWORD dwMax)
{
  HRESULT hr = S_OK;
  PCSTR p = pGpgId, ps = NULL;

  pComment[0] = '\0';

  /* '('邩mF */
  for (p = pGpgId; *p; ++p) {
    if (*p == '(') {
      ps = p;
      break;
    }
  }

  if (ps) {
    ++ps;
    /* 擪̋󔒂ǂݔ΂ */
    while (*ps == ' ') ++ps;
    p = ps;
    while (*p) {
      if (*p == ')') break;
      ++p;
    }
    /* ')'Ȃ΃G[ */
    if (!*p) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
    /* ̗]vȋ󔒂 */
    while (*(p - 1) == ' ') --p;
    if (FAILED(hr = StringCbCopyN(pComment, dwMax - 1, ps, p - ps))) {
      goto finish_func;
    }
    pComment[p - ps] = '\0';
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_ReadFileLine(PREADFILEBUFF pfile, PSTR buff, DWORD len, PCSTR pLine)
{
  HRESULT hr = S_OK;

  while (TRUE) {
    if (FAILED(hr = ReadFileLine(pfile, buff, len))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (!util_strncmp(buff, pLine, util_strlen(pLine, -1))) {
      hr = S_OK;
      goto finish_func;
    }
    if ((fEofFileLine(pfile)) ||
        (!util_strncmp(buff, GPGSTATUSLINE_GET_LINE, util_strlen(GPGSTATUSLINE_GET_LINE, -1))) ||
        (!util_strncmp(buff, GPGSTATUSLINE_GET_HIDDEN, util_strlen(GPGSTATUSLINE_GET_HIDDEN, -1))) ||
        (!util_strncmp(buff, GPGSTATUSLINE_GET_BOOL, util_strlen(GPGSTATUSLINE_GET_BOOL, -1)))) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_open_file_inherit_read(PHANDLE ph, PCSTR pFileName)
{
  SECURITY_ATTRIBUTES sa;
  DWORD dwErr = ERROR_SUCCESS;
  HRESULT hr = S_OK;

  ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = TRUE;

  *ph = CreateFile(pFileName, GENERIC_READ, FILE_SHARE_READ, &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (*ph == INVALID_HANDLE_VALUE) {
    dwErr = GetLastError();
    hr = HRESULT_FROM_WIN32(dwErr);
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_open_file_inherit_write(PHANDLE ph, PCSTR pFileName)
{
  SECURITY_ATTRIBUTES sa;
  DWORD dwErr = ERROR_SUCCESS;
  HRESULT hr = S_OK;

  ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = TRUE;

  *ph = CreateFile(pFileName, GENERIC_WRITE, 0, &sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  if (*ph == INVALID_HANDLE_VALUE) {
    dwErr = GetLastError();
    hr = HRESULT_FROM_WIN32(dwErr);
  }
  return hr;
}

/*------------------------------------------------------------*/
HRESULT
gpgez_stringlist_addoption(PCMDEZCONTEXT pctx, PSTRINGLIST psl, PCSTR pstr, BOOL fDq)
{
  HRESULT hr = S_OK;
  PSTRINGLIST_NODE pnode = NULL;

  if (psl) {
    for (pnode = stringlist_get_head((psl)); pnode; pnode = stringlist_get_next(pnode)) {
      if (FAILED(hr = cmdez_add_option(pctx, pstr, stringlist_get_string(pnode), fDq))) goto finish_func;
    }
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
HRESULT
gpgez_stringlist_addoperand(PCMDEZCONTEXT pctx, PSTRINGLIST psl, BOOL fDq)
{
  HRESULT hr = S_OK;
  PSTRINGLIST_NODE pnode = NULL;

  if (psl) {
    for (pnode = stringlist_get_head((psl)); pnode; pnode = stringlist_get_next(pnode)) {
      if (FAILED(hr = cmdez_add_operand(pctx, stringlist_get_string(pnode), fDq))) goto finish_func;
    }
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_loadmyfpr(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTR *ppFpr)
{
  HRESULT hr = S_OK;
  PCMDEZCONTEXT pgctx = NULL;
  UINT iSec = 0; /* --list-secret-key ̌sec:so񐔂JEg */
  PREADFILEBUFF pfilebuff = NULL;
  HANDLE hStdOutP = INVALID_HANDLE_VALUE, hStdOutC = INVALID_HANDLE_VALUE;
  CHAR buff[MAX_LINE + 1];

  *ppFpr = NULL;
  /* gpg --list-secret-keys ŌւmF */
  DO_DEBUG(hr, cmdez_initialize(&pgctx, pGpgExe, NULL));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--list-secret-keys", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--with-colons", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--fixed-list-mode", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--with-fingerprint", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--with-fingerprint", NULL, FALSE));
  DO_DEBUG(hr, util_create_read_pipe(&hStdOutP, &hStdOutC));
  DO_DEBUG(hr, cmdez_set_handle(pgctx, HANDLE_TYPE_STDOUT, hStdOutC));
  DO_DEBUG(hr, cmdez_execute(pgctx));
  CloseHandleClean(&hStdOutC);

  DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, hStdOutP));
  while (!fEofFileLine(pfilebuff)) {
    DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
    if (!util_strncmp("sec:", buff, 4)) {
      ++iSec;
    }
    if ((iSec == 1) && (!(*ppFpr)) && (!util_strncmp("fpr:", buff, 4))) {
      CHAR fpr[MAX_LINE + 1];
      DO_DEBUG(hr, sv_get(buff, ':', 9, fpr, MAX_LINE));
      *ppFpr = util_strdup(fpr);
      DO_DEBUG_RETC(hr, (!*ppFpr), ERROR_OUTOFMEMORY);
    }
  }
  DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));
  DO_DEBUG(hr, ReleaseReadFileBuff(&pfilebuff));
  CloseHandleClean(&hStdOutP);
  cmdez_finish(pgctx);
  cmdez_release(&pgctx);
  pgctx = NULL;

finish_func:
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  if (hStdOutP != INVALID_HANDLE_VALUE) CloseHandle(hStdOutP);
  if (hStdOutC != INVALID_HANDLE_VALUE) CloseHandle(hStdOutC);
  if (pgctx) cmdez_release(&pgctx);
  return hr;
}

/*------------------------------------------------------------*/
struct {
  PCSTR pstatus;
  INT value;
} gpgstatus_status_value[] = {
  { "ALREADY_SIGNED", GPGSTATUS_ALREADY_SIGNED },
  { "ATTRIBUTE", GPGSTATUS_ATTRIBUTE },
  { "BACKUP_KEY_CREATED", GPGSTATUS_BACKUP_KEY_CREATED },
  { "BAD_PASSPHRASE", GPGSTATUS_BAD_PASSPHRASE },
  { "BADARMOR", GPGSTATUS_BADARMOR },
  { "BADSIG", GPGSTATUS_BADSIG },
  { "BEGIN_DECRYPTION", GPGSTATUS_BEGIN_DECRYPTION },
  { "BEGIN_ENCRYPTION", GPGSTATUS_BEGIN_ENCRYPTION },
  { "BEGIN_STREAM", GPGSTATUS_BEGIN_STREAM },
  { "CARDCTRL", GPGSTATUS_CARDCTRL },
  { "DECRYPTION_FAILED", GPGSTATUS_DECRYPTION_FAILED },
  { "DECRYPTION_OKAY", GPGSTATUS_DECRYPTION_OKAY },
  { "DELETE_PROBLEM", GPGSTATUS_DELETE_PROBLEM },
  { "ENC_TO", GPGSTATUS_ENC_TO },
  { "END_DECRYPTION", GPGSTATUS_END_DECRYPTION },
  { "END_ENCRYPTION", GPGSTATUS_END_ENCRYPTION },
  { "END_STREAM", GPGSTATUS_END_STREAM },
  { "ERROR", GPGSTATUS_ERROR },
  { "ERRSIG", GPGSTATUS_ERRSIG },
  { "EXPKEYSIG", GPGSTATUS_EXPKEYSIG },
  { "EXPSIG", GPGSTATUS_EXPSIG },
  { "FILE_DONE", GPGSTATUS_FILE_DONE },
  { "FILE_START", GPGSTATUS_FILE_START },
  { "GET_BOOL", GPGSTATUS_GET_BOOL },
  { "GET_HIDDEN", GPGSTATUS_GET_HIDDEN },
  { "GET_LINE", GPGSTATUS_GET_LINE },
  { "GOOD_PASSPHRASE", GPGSTATUS_GOOD_PASSPHRASE },
  { "GOODSIG", GPGSTATUS_GOODSIG },
  { "GOT_IT", GPGSTATUS_GOT_IT },
  { "IMPORT_CHECK", GPGSTATUS_IMPORT_CHECK },
  { "IMPORT_OK", GPGSTATUS_IMPORT_OK },
  { "IMPORT_PROBLEM", GPGSTATUS_IMPORT_PROBLEM },
  { "IMPORT_RES", GPGSTATUS_IMPORT_RES },
  { "IMPORTED", GPGSTATUS_IMPORTED },
  { "INV_RECP", GPGSTATUS_INV_RECP },
  { "KEY_CREATED", GPGSTATUS_KEY_CREATED },
  { "KEY_NOT_CREATED", GPGSTATUS_KEY_NOT_CREATED },
  { "KEYEXPIRED", GPGSTATUS_KEYEXPIRED },
  { "KEYREVOKED", GPGSTATUS_KEYREVOKED },
  { "MISSING_PASSPHRASE", GPGSTATUS_MISSING_PASSPHRASE },
  { "NEED_PASSPHRASE", GPGSTATUS_NEED_PASSPHRASE },
  { "NEED_PASSPHRASE_PIN", GPGSTATUS_NEED_PASSPHRASE_PIN },
  { "NEED_PASSPHRASE_SYM", GPGSTATUS_NEED_PASSPHRASE_SYM },
  { "NEWSIG", GPGSTATUS_NEWSIG },
  { "NO_PUBKEY", GPGSTATUS_NO_PUBKEY },
  { "NO_RECP", GPGSTATUS_NO_RECP },
  { "NO_SECKEY", GPGSTATUS_NO_SECKEY },
  { "NODATA", GPGSTATUS_NODATA },
  { "NOTATION_DATA", GPGSTATUS_NOTATION_DATA },
  { "NOTATION_NAME", GPGSTATUS_NOTATION_NAME },
  { "PLAINTEXT", GPGSTATUS_PLAINTEXT },
  { "PLAINTEXT_LENGTH", GPGSTATUS_PLAINTEXT_LENGTH },
  { "POLICY_URL", GPGSTATUS_POLICY_URL },
  { "PROGRESS", GPGSTATUS_PROGRESS },
  { "REVKEYSIG", GPGSTATUS_REVKEYSIG },
  { "RSA_OR_IDEA", GPGSTATUS_RSA_OR_IDEA },
  { "SC_OP_FAILURE", GPGSTATUS_SC_OP_FAILURE },
  { "SC_OP_SUCCESS", GPGSTATUS_SC_OP_SUCCESS },
  { "SESSION_KEY", GPGSTATUS_SESSION_KEY },
  { "SHM_GET", GPGSTATUS_SHM_GET },
  { "SHM_GET_BOOL", GPGSTATUS_SHM_GET_BOOL },
  { "SHM_GET_HIDDEN", GPGSTATUS_SHM_GET_HIDDEN },
  { "SHM_INFO", GPGSTATUS_SHM_INFO },
  { "SIG_CREATED", GPGSTATUS_SIG_CREATED },
  { "SIG_ID", GPGSTATUS_SIG_ID },
  { "SIG_SUBPACKET", GPGSTATUS_SIG_SUBPACKET },
  { "SIGEXPIRED", GPGSTATUS_SIGEXPIRED },
  { "TRUNCATED", GPGSTATUS_TRUNCATED },
  { "TRUST_FULLY", GPGSTATUS_TRUST_FULLY },
  { "TRUST_MARGINAL", GPGSTATUS_TRUST_MARGINAL },
  { "TRUST_NEVER", GPGSTATUS_TRUST_NEVER },
  { "TRUST_ULTIMATE", GPGSTATUS_TRUST_ULTIMATE },
  { "TRUST_UNDEFINED", GPGSTATUS_TRUST_UNDEFINED },
  { "UNEXPECTED", GPGSTATUS_UNEXPECTED },
  { "USERID_HINT", GPGSTATUS_USERID_HINT },
  { "VALIDSIG", GPGSTATUS_VALIDSIG },
  { "", GPGSTATUS_ZZZZZZZZ }
};
  

DllExport INT
gpgez_interpret_status(PCSTR pStatusStr)
{
  CHAR buff[MAX_PATH + 1] = "";
  INT iRet = GPGSTATUS_ZZZZZZZZ;
  INT i = 0;

  if (!sv_cmp(pStatusStr, ' ', 0, "[GNUPG:]")) goto finish_func;
  if (FAILED(sv_get(pStatusStr, ' ', 1, buff, MAX_PATH))) goto finish_func;
  while (gpgstatus_status_value[i].value != GPGSTATUS_ZZZZZZZZ) {
    if (!util_strcmp(buff, gpgstatus_status_value[i].pstatus)) {
      iRet = gpgstatus_status_value[i].value;
    }
    ++i;
  }
finish_func:
  return iRet;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_ispublicencrypted(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFileName, PBOOL pfPublic) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;

  *pfPublic = FALSE;

  /* eXgp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFileName));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt_listonly(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, NULL, hIn, INVALID_HANDLE_VALUE, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);

  /* gpgI҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* ʌ */
  {
    PSTRINGLIST pslstatus = gpgez_gpg_get_statuslist(pgpg);
    PSTRINGLIST_NODE pnode;

    for (pnode = stringlist_get_head(pslstatus); pnode; pnode = stringlist_get_next(pnode)) {
      PCSTR p = stringlist_get_string(pnode);
      if (!util_strncmpl("[GNUPG:] ENC_TO", p)) {
        *pfPublic = TRUE;
        break;
      }
    }
  }

finish_func:
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hIn);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_ispublicencrypted_buff(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, PBOOL pfPublic) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hInP = INVALID_HANDLE_VALUE, hInC = INVALID_HANDLE_VALUE;
  BUFFHANDLE bh;
  HANDLE hThread1 = NULL;
  DWORD dwtid;

  *pfPublic = FALSE;

  /* eXgp\̍쐬 */
  DO_DEBUG(hr, util_create_write_pipe(&hInP, &hInC));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt_listonly(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, NULL, hInC, INVALID_HANDLE_VALUE, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hInC);

  /* obt@oXbhJn */
  bh.h = hInP;
  bh.buff = pbuff;
  bh.len = dwLen;
  hThread1 = CreateThread(NULL, 0, procWriteBuff, (LPVOID)&bh, 0, &dwtid);
  if (!hThread1) {
    DWORD dwErr = GetLastError();
    hr = HRESULT_FROM_WIN32(dwErr);
    goto finish_func;
  }
  WaitForSingleObject(hThread1, INFINITE);
  CloseHandle(hThread1);
  hThread1 = NULL;
  hInP = INVALID_HANDLE_VALUE;

  /* gpgI҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* ʌ */
  {
    PSTRINGLIST pslstatus = gpgez_gpg_get_statuslist(pgpg);
    PSTRINGLIST_NODE pnode;

    for (pnode = stringlist_get_head(pslstatus); pnode; pnode = stringlist_get_next(pnode)) {
      PCSTR p = stringlist_get_string(pnode);
      if (!util_strncmpl("[GNUPG:] ENC_TO", p)) {
        *pfPublic = TRUE;
        break;
      }
    }
  }

finish_func:
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hInC);
  CloseHandleClean(&hInP);
  return hr;
}

/*------------------------------------------------------------*/
/* TODO:
 * ܂At@C擾ƏmFATCYmF̂߁Agpg𑖂点A
 * Kvɉāut@CȂA*.taȑꍇ̂݁vtarœWJ`FbN悤ɂق
 * ̂ł͂Ȃ낤
*/
DllExport HRESULT
gpgez_isdecryptverified(PCSTR pGpgExe, PCSTR pTarExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFileName, PSTRINGLIST pslVerifiedName, PSTRINGLIST pslVerifiedFpr, PSTR pXFilename, DWORD dwLen, PDWORD pdwSize, PBOOL pfTar) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  PGPGEZ_CMD ptar = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;
  HANDLE hGpgOutP = INVALID_HANDLE_VALUE, hGpgOutC = INVALID_HANDLE_VALUE;
  HANDLE hTarInP = INVALID_HANDLE_VALUE, hTarInC = INVALID_HANDLE_VALUE;
  BOOL fDecrypted = FALSE;
  DWORD dwSize = 0;
  
  if (pfTar) {
    *pfTar = FALSE;
  }
  if (pdwSize) {
    *pdwSize = 0;
  }

  /* ̃vZXԃnh쐬 */
  DO_DEBUG(hr, util_create_read_pipe(&hGpgOutP, &hGpgOutC));

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFileName));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, FALSE, pPassphrase, hIn, hGpgOutC, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hGpgOutC);

  /* tareXgp\̍쐬 */
  DO_DEBUG(hr, util_create_write_pipe(&hTarInP, &hTarInC));
  DO_DEBUG(hr, gpgez_testtar_create(pTarExe, hTarInC, INVALID_HANDLE_VALUE, &ptar));
  CloseHandleClean(&hTarInC);

  {
    CHAR buff[MAX_LINE + 1];
    DWORD dwRead, dwWritten;
    while(TRUE) {
      if (!ReadFile(hGpgOutP, buff, MAX_LINE, &dwRead, NULL)) {
        DWORD dwErr = GetLastError();
        if (dwErr != ERROR_BROKEN_PIPE) {
          hr = HRESULT_FROM_WIN32(dwErr);
          goto finish_func;
        }
        CloseHandleClean(&hGpgOutP);
        break;
      }
      dwSize += dwRead;
      if (hTarInP != INVALID_HANDLE_VALUE) {
        if (!WriteFile(hTarInP, buff, dwRead, &dwWritten, NULL)) {
          CloseHandleClean(&hTarInP);
        }
      }
    }
  }
  CloseHandleClean(&hTarInP);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(ptar));

  /* ؂ł (VALIDSIG邩ǂ) */
  /* t@C͖ߍ܂Ă邩 */
  {
    PSTRINGLIST pslstatus = gpgez_gpg_get_statuslist(pgpg);
    PSTRINGLIST_NODE pnode;

    for (pnode = stringlist_get_head(pslstatus); pnode; pnode = stringlist_get_next(pnode)) {
      PCSTR p = stringlist_get_string(pnode);
      switch(gpgez_interpret_status(p)) {
      case GPGSTATUS_DECRYPTION_OKAY:
        fDecrypted = TRUE;
        break;
      case GPGSTATUS_BAD_PASSPHRASE:
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_PASSWORD);
        goto finish_func;
        break;
      case GPGSTATUS_VALIDSIG:
        if (pslVerifiedFpr) {
          PCSTR p1 = sv_pos(p, ' ', 2);
          PCSTR p2 = sv_pos(p1, ' ', 1);

          if (*p2) --p2; /* 񖖂łȂ󔒁{1̈ʒu܂Ń|C^ił̂1߂ */
          DO_DEBUG(hr, stringlist_add_node_len(pslVerifiedFpr, p1, p2 - p1));
        }
        break;
      case GPGSTATUS_GOODSIG:
        if (pslVerifiedName) {
          CHAR szIdUtf8[MAX_PATH * 6 + 1];
          PCSTR p1 = sv_pos(p, ' ', 3);
          PSTR p2 = NULL;

          unescapePercent(p1, szIdUtf8, MAX_PATH * 6);
          DO_DEBUG(hr, util_utf8tosjis(szIdUtf8, &p2));
          if (FAILED(hr = stringlist_add_node(pslVerifiedName, p2))) {
            DEBUGERROR(0, hr);
            util_free(p2);
            goto finish_func;
          }
          util_free(p2);
        }
        break;
      case GPGSTATUS_PLAINTEXT:
        {
          PCSTR p1 = sv_pos(p, ' ', 4);
          CHAR buff1[MAX_PATH * 3 + 1];
          PSTR p2 = NULL;

          unescapePercent(p1, buff1, MAX_PATH * 3 * 1);
          if (FAILED(hr = util_utf8tosjis(buff1, &p2))) {
            DEBUGERROR(0, hr);
            pXFilename[0] = '\0';
            goto finish_func;
          }
          if (FAILED(hr = util_strcpy(pXFilename, dwLen, p2))) {
            util_free(p2);
            goto finish_func;
          }
          util_free(p2);
        }
        break;
      }
    }
  }
  DO_DEBUG_RETC(hr, (!fDecrypted), ERROR_DECRYPTION_FAILED);
  if (gpgez_cmd_get_exitcode(ptar) == 0) {
    if (pfTar) {
      *pfTar = TRUE;
    }
  }
  if (pdwSize) *pdwSize = dwSize;

finish_func:
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (ptar) gpgez_cmd_release(&ptar);
  CloseHandleClean(&hIn);
  CloseHandleClean(&hGpgOutP);
  CloseHandleClean(&hGpgOutC);
  CloseHandleClean(&hTarInP);
  CloseHandleClean(&hTarInC);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_is7zencrypted(PCSTR p7ZipExe, PCSTR pFileName, PBOOL pfEncrypted) {
  HRESULT hr = S_OK;
  PCMDEZCONTEXT pzctx = NULL;

  /* 7zip s */
  DO_DEBUG(hr, cmdez_initialize(&pzctx, p7ZipExe, NULL));
  DO_DEBUG(hr, cmdez_add_option2(pzctx, "t -y ", pFileName, " ",TRUE));

  /* 7zip s */
  DO_DEBUG(hr, cmdez_execute(pzctx));

  /* 7zip I҂ */
  DO_DEBUG(hr, cmdez_finish(pzctx));

  *pfEncrypted = FALSE;
  if (cmdez_get_exitcode(pzctx) == 0) {
    *pfEncrypted = FALSE;
  } else if (cmdez_get_exitcode(pzctx) == 2) {
    *pfEncrypted = TRUE;
  } else {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
  }
finish_func:
  if (pzctx) cmdez_release(&pzctx);

  return hr;
}

/*------------------------------------------------------------*/
/*
 *  gpg R}hQ
 *
 */
/*------------------------------------------------------------*/
#if 0
DllExport HRESULT
gpgez_genkey(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
             PCSTR pName, PCSTR pEmail, PCSTR pComment, PCSTR pPassphrase, PSTR *ppFpr)
{
  HRESULT hr = S_OK;
  PGPGEZ_GENKEY pgpg = NULL;
  PCMDEZCONTEXT pgctx = NULL;

  if (FAILED(hr = gpgez_genkey_create(pGpgExe, pGpgHome, pslSecRing, pslPubRing, INVALID_HANDLE_VALUE, &pgpg))) goto finish_func;
  /* 񏑂 */
  if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Key-Type", "DSA"))) goto finish_func;
  if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Key-Length", "1024"))) goto finish_func;
  if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Subkey-Type:", "ELG-E"))) goto finish_func;
  if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Subkey-Length:", "2048"))) goto finish_func;
  if ((pName) && (*(pName))) {
    if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Name-Real:", pName))) goto finish_func;
  }
  if ((pEmail) && (*(pEmail))) {
    if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Name-Email:", pEmail))) goto finish_func;
  }
  if ((pComment) && (*(pComment))) {
    if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Name-Comment:", pComment))) goto finish_func;
  }
  if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Expire-Date:", "0"))) goto finish_func;
  if ((pPassphrase) && (*(pPassphrase))) {
    if (FAILED(hr = gpgez_genkey_write_param(pgpg, "Passphrase:", pPassphrase))) goto finish_func;
  }
  if (FAILED(hr = gpgez_genkey_write_cmd(pgpg, "%commit"))) goto finish_func;
  if (FAILED(hr = gpgez_genkey_close_write(pgpg))) goto finish_func;
  /* gpg I҂ */
  if (FAILED(hr = gpgez_genkey_finish(pgpg))) goto finish_func;

  /* qvZX̖߂l`FbN */
  if (gpgez_genkey_get_exitcode(pgpg) != 0) {
  }

finish_func:
  if (pgpg) gpgez_genkey_release(&pgpg);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_encrypt(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
              PCSTR pFile, PCSTR pZfile, PSTRINGLIST pRecipients, PCSTR pPassphraseSym, BOOL fSign, BOOL fAscii)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  PCMDEZCONTEXT pgctx = NULL;
  PCSTR pFileName = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE;

  /* Íp\̍쐬 */
  if (FAILED(hr = gpgez_open_file_inherit_read(&hIn, pFile))) goto finish_func;
  if (FAILED(hr = gpgez_open_file_inherit_write(&hOut, pZfile))) goto finish_func;
  pFileName = PathFindFileName(pFile);
  if (FAILED(hr = gpgez_gpg_create_encrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, pRecipients, pPassphraseSym,
    fSign, fAscii, pFileName, hIn, hOut, INVALID_HANDLE_VALUE, &pgpg))) goto finish_func;
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);

  /* gpg I҂ */
  if (FAILED(hr = gpgez_gpg_finish(pgpg))) goto finish_func;

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_encryptfrombuff(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                      PCSTR pBuff, DWORD dwLen, PCSTR pZfile,
                      PSTRINGLIST pRecipients, PCSTR pPassphraseSym, BOOL fSign, BOOL fAscii) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hInP = INVALID_HANDLE_VALUE, hInC = INVALID_HANDLE_VALUE;
  HANDLE hOut = INVALID_HANDLE_VALUE;
  DWORD dwWritten;

  /* Íp\̍쐬 */
  DO_DEBUG(hr, util_create_write_pipe(&hInP, &hInC));
  if (FAILED(hr = gpgez_open_file_inherit_write(&hOut, pZfile))) goto finish_func;
  if (FAILED(hr = gpgez_gpg_create_encrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, pRecipients, pPassphraseSym,
    fSign, fAscii, "_CONSOLE", hInC, hOut, INVALID_HANDLE_VALUE, &pgpg))) goto finish_func;
  CloseHandleClean(&hInC);
  CloseHandleClean(&hOut);

  /* obt@o */
  WriteFile(hInP, pBuff, dwLen, &dwWritten, NULL);
  CloseHandleClean(&hInP);

  /* gpg I҂ */
  if (FAILED(hr = gpgez_gpg_finish(pgpg))) goto finish_func;

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hInC);
  CloseHandleClean(&hInP);
  CloseHandleClean(&hOut);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_encrypt_buff2buff(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                        PCSTR pbuff, DWORD dwLen, PSTR *ppbuff, PDWORD pdwLen,
                        PSTRINGLIST pRecipients, PCSTR pPassphraseSym, BOOL fSign, BOOL fAscii) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hInP = INVALID_HANDLE_VALUE, hInC = INVALID_HANDLE_VALUE;
  HANDLE hOutP = INVALID_HANDLE_VALUE, hOutC = INVALID_HANDLE_VALUE;
  PBINBUFF pbinbuff = NULL;
  BUFFHANDLE bh;
  HANDLE hThread1 = NULL;
  DWORD dwtid;

  /* gpg s */
  *pdwLen = 0;
  *ppbuff = NULL;
  DO_DEBUG(hr, binbuff_create(&pbinbuff));

  /* Íp\̍쐬 */
  DO_DEBUG(hr, util_create_write_pipe(&hInP, &hInC));
  DO_DEBUG(hr, util_create_read_pipe(&hOutP, &hOutC));
  DO_DEBUG(hr, gpgez_gpg_create_encrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, pRecipients, pPassphraseSym,
    fSign, fAscii, "_CONSOLE", hInC, hOutC, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hInC);
  CloseHandleClean(&hOutC);

  /* obt@oXbhJn */
  bh.h = hInP;
  bh.buff = pbuff;
  bh.len = dwLen;
  hThread1 = CreateThread(NULL, 0, procWriteBuff, (LPVOID)&bh, 0, &dwtid);
  if (!hThread1) {
    DWORD dwErr = GetLastError();
    hr = HRESULT_FROM_WIN32(dwErr);
    goto finish_func;
  }
  CloseHandle(hThread1);
  hThread1 = NULL;
  hInP = INVALID_HANDLE_VALUE;

  /* qvZX̕Wo͓ǂݎ */
  {
    CHAR buff1[MAX_LINE + 1];
    DWORD dwRead;

    while (ReadFile(hOutP, buff1, MAX_LINE, &dwRead, NULL)) {
      DO_DEBUG(hr, binbuff_append(pbinbuff, buff1, dwRead));
    }
    CloseHandleClean(&hOutP);
  }

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  }
  *pdwLen = binbuff_len(pbinbuff);
  *ppbuff = binbuff_detach(pbinbuff);

finish_func:
  /* I */
  if (hThread1) CloseHandle(hThread1);
  if (pbinbuff) binbuff_detach(pbinbuff);
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hInC);
  CloseHandleClean(&hInP);
  CloseHandleClean(&hOutC);
  CloseHandleClean(&hOutP);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_tarencrypt(PCSTR pTarExe, PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                 PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile,
                 PSTRINGLIST pRecipients, PCSTR pPassphraseSym, BOOL fSign, BOOL fAscii)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  PGPGEZ_CMD ptar = NULL;
  HANDLE hOut = INVALID_HANDLE_VALUE;
  HANDLE hTarOutP = INVALID_HANDLE_VALUE, hTarOutC = INVALID_HANDLE_VALUE;
  CHAR szInterFileName[MAX_PATH + 1] = "";

  /* ̃vZXԃnh쐬 */
  DO_DEBUG(hr,util_create_read_pipe(&hTarOutP, &hTarOutC));

  /* Tarp\̍쐬 */
  DO_DEBUG(hr, gpgez_tar_create(pTarExe, TAR_COMPRESS_NONE, pBaseDirectory,
    pFiles, hTarOutC, INVALID_HANDLE_VALUE, &ptar));
  CloseHandle(hTarOutC);
  hTarOutC = INVALID_HANDLE_VALUE;

  /* nhtւihTarOutPpł悤ɂj */
  {
    HANDLE h = hTarOutP;
    if (!DuplicateHandle(GetCurrentProcess(), h, GetCurrentProcess(), &hTarOutP, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    CloseHandle(h);
  }

  /* ԃt@C쐬 */
  {
    DO_DEBUG(hr, util_strcpy(szInterFileName, MAX_PATH, stringlist_get_string(stringlist_get_head(pFiles))));
    {
      PSTR p = util_pathfindextension(szInterFileName);
      if ((p - szInterFileName) < MAX_PATH - 4) {
        util_strcpy(p, MAX_PATH, ".tar");
      }
    }
  }

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, pZfile));
  DO_DEBUG(hr, gpgez_gpg_create_encrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, pRecipients, pPassphraseSym,
    fSign, fAscii, szInterFileName, hTarOutP, hOut, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandle(hOut);
  hOut = INVALID_HANDLE_VALUE;

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(ptar));

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(ptar) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  } else {
    if (gpgez_gpg_get_exitcode(pgpg) != 0) {
      hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
      goto finish_func;
    }
  }

finish_func:
  /* I */
  if (ptar) gpgez_cmd_release(&ptar);
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hOut);
  CloseHandleClean(&hTarOutP);
  CloseHandleClean(&hTarOutC);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_zipencrypt(PCSTR p7ZipExe, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PCSTR pPassphraseSym) {
  HRESULT hr = S_OK;
  PGPGEZ_CMD psevenzip = NULL;

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_sevenzip_create(p7ZipExe, SEVENZIP_ZIPENC, pBaseDirectory,
    pFiles, pPassphraseSym, pZfile, INVALID_HANDLE_VALUE, &psevenzip));

  /* 7zI҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(psevenzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(psevenzip) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (psevenzip) gpgez_cmd_release(&psevenzip);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_7zencrypt(PCSTR p7ZipExe, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZFile, PCSTR pPassphraseSym, INT iMode) {
  HRESULT hr = S_OK;
  PGPGEZ_CMD psevenzip = NULL;

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_sevenzip_create(p7ZipExe, iMode, pBaseDirectory,
    pFiles, pPassphraseSym, pZFile, INVALID_HANDLE_VALUE, &psevenzip));

  /* 7zI҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(psevenzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(psevenzip) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (psevenzip) gpgez_cmd_release(&psevenzip);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_tar(PCSTR pTarExe, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, INT iMode) {
  HRESULT hr = S_OK;
  PGPGEZ_CMD ptar = NULL;
  HANDLE hOut = INVALID_HANDLE_VALUE;

  /* Tarp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, pZfile));
  DO_DEBUG(hr, gpgez_tar_create(pTarExe, iMode, pBaseDirectory,
    pFiles, hOut, INVALID_HANDLE_VALUE, &ptar));
  CloseHandleClean(&hOut);

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(ptar));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(ptar) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (ptar) gpgez_cmd_release(&ptar);
  CloseHandleClean(&hOut);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_encryptsfx(PCSTR p7ZipExe, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZFile, PCSTR pPassphraseSym) {
  HRESULT hr = S_OK;
  PGPGEZ_CMD psevenzip = NULL;

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_sevenzip_create(p7ZipExe, SEVENZIP_SFX, pBaseDirectory,
    pFiles, pPassphraseSym, pZFile, INVALID_HANDLE_VALUE, &psevenzip));

  /* 7zI҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(psevenzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(psevenzip) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (psevenzip) gpgez_cmd_release(&psevenzip);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_decrypt(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFile, PCSTR pXFile) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE;

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFile));
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, pXFile));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, TRUE, pPassphrase, hIn, hOut, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_decrypt2buff(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFile, PSTR *ppbuff, PDWORD pdwLen) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;
  HANDLE hOutP = INVALID_HANDLE_VALUE, hOutC = INVALID_HANDLE_VALUE;
  PBINBUFF pbinbuff = NULL;

  /* o̓obt@ */
  *pdwLen = 0;
  *ppbuff = NULL;
  DO_DEBUG(hr, binbuff_create(&pbinbuff));

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFile));
  DO_DEBUG(hr, util_create_read_pipe(&hOutP, &hOutC));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, TRUE, pPassphrase, hIn, hOutC, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOutC);

  /* qvZX̕Wo͓ǂݎ */
  {
    CHAR buff1[MAX_LINE + 1];
    DWORD dwRead;

    while (ReadFile(hOutP, buff1, MAX_LINE, &dwRead, NULL)) {
      DO_DEBUG(hr, binbuff_append(pbinbuff, buff1, dwRead));
    }
    CloseHandleClean(&hOutP);
  }

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

  *pdwLen = binbuff_len(pbinbuff);
  *ppbuff = binbuff_detach(pbinbuff);

finish_func:
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (pbinbuff) binbuff_release(&pbinbuff);
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOutP);
  CloseHandleClean(&hOutC);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_decrypt_buff2buff(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pbuff, DWORD dwLen, PSTR *ppbuff, PDWORD pdwLen, PSTRINGLIST pslVerified) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hInP = INVALID_HANDLE_VALUE, hInC = INVALID_HANDLE_VALUE;
  HANDLE hOutP = INVALID_HANDLE_VALUE, hOutC = INVALID_HANDLE_VALUE;
  HANDLE hPassW = INVALID_HANDLE_VALUE, hPassR = INVALID_HANDLE_VALUE;
  PBINBUFF pbinbuff = NULL;
  BUFFHANDLE bh;
  HANDLE hThread1 = NULL;
  DWORD dwtid;
  BOOL fDecrypted = FALSE;

  /* o̓obt@ */
  *pdwLen = 0;
  *ppbuff = NULL;
  DO_DEBUG(hr, binbuff_create(&pbinbuff));

  /* p\̍쐬 */
  DO_DEBUG(hr, util_create_write_pipe(&hInP, &hInC));
  DO_DEBUG(hr, util_create_read_pipe(&hOutP, &hOutC));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, TRUE, pPassphrase, hInC, hOutC, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hInC);
  CloseHandleClean(&hOutC);

  /* obt@oXbhJn */
  bh.h = hInP;
  bh.buff = pbuff;
  bh.len = dwLen;
  hThread1 = CreateThread(NULL, 0, procWriteBuff, (LPVOID)&bh, 0, &dwtid);
  if (!hThread1) {
    DWORD dwErr = GetLastError();
    hr = HRESULT_FROM_WIN32(dwErr);
    goto finish_func;
  }
  CloseHandle(hThread1);
  hThread1 = NULL;
  hInP = INVALID_HANDLE_VALUE;

  /* qvZX̕Wo͓ǂݎ */
  {
    CHAR buff1[MAX_LINE + 1];
    DWORD dwRead;

    while (ReadFile(hOutP, buff1, MAX_LINE, &dwRead, NULL)) {
      DO_DEBUG(hr, binbuff_append(pbinbuff, buff1, dwRead));
    }
    CloseHandleClean(&hOutP);
  }

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* ؂ł (VALIDSIG邩ǂ) */
  {
    PSTRINGLIST pslstatus = gpgez_gpg_get_statuslist(pgpg);
    PSTRINGLIST_NODE pnode;

    for (pnode = stringlist_get_head(pslstatus); pnode; pnode = stringlist_get_next(pnode)) {
      PCSTR p = stringlist_get_string(pnode);
      if (!util_strncmpl("[GNUPG:] DECRYPTION_OKAY", p)) {
        fDecrypted = TRUE;
      } else if (!util_strncmpl("[GNUPG:] BAD_PASSPHRASE ", p)) {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_PASSWORD);
        goto finish_func;
      } else if (!util_strncmpl("[GNUPG:] VALIDSIG ", p)) {
/*        *pfVerified = TRUE; */
      } else if ((pslVerified) && (!util_strncmpl("[GNUPG:] GOODSIG ", p))) {
        PCSTR p1 = p + 17;
        int i = 0;
        while (*p1) {
          if (*p1 == ' ') {
            ++i;
            if (i == 1) {
              CHAR szIdUtf8[MAX_PATH * 6 + 1];
              PSTR pstr1;
              unescapePercent(p1 + 1, szIdUtf8, MAX_PATH * 6);
              DO_DEBUG(hr, util_utf8tosjis(szIdUtf8, &pstr1));
              DO_DEBUG(hr, stringlist_add_node(pslVerified, pstr1));
              break;
            }
          }
          ++p1;
        }
      }
    }
  }
  if (!fDecrypted) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

  *pdwLen = binbuff_len(pbinbuff);
  *ppbuff = binbuff_detach(pbinbuff);

finish_func:
  if (hThread1) CloseHandle(hThread1);
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (pbinbuff) binbuff_release(&pbinbuff);
  CloseHandleClean(&hInP);
  CloseHandleClean(&hInC);
  CloseHandleClean(&hOutP);
  CloseHandleClean(&hOutC);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_decrypttar(PCSTR pGpgExe, PCSTR pTarExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFileName, PCSTR pBaseDirectory, BOOL fOverWrite)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  PGPGEZ_CMD puntar = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;
  HANDLE hGpgOutP = INVALID_HANDLE_VALUE, hGpgOutC = INVALID_HANDLE_VALUE;

  /* ̃vZXԃnh쐬 */
  DO_DEBUG(hr, util_create_read_pipe(&hGpgOutP, &hGpgOutC));

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFileName));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, TRUE, pPassphrase, hIn, hGpgOutC, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hGpgOutC);

  /* nhtւihGpgOutPpł悤ɂj */
  {
    HANDLE h = hGpgOutP;
    if (!DuplicateHandle(GetCurrentProcess(), h, GetCurrentProcess(), &hGpgOutP, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    CloseHandle(h);
  }

  /* tarWJp\̍쐬 */
  DO_DEBUG(hr, gpgez_untar_create(pTarExe, pBaseDirectory, fOverWrite,
    &hGpgOutP, INVALID_HANDLE_VALUE, &puntar));
  CloseHandleClean(&hGpgOutP);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(puntar));

  /* qvZX̖߂l`FbN */
  /* NOTICE: bsdtar̖untar͕KFail̂Ń`FbNȂ */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (puntar) gpgez_cmd_release(&puntar);
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hIn);
  CloseHandleClean(&hGpgOutP);
  CloseHandleClean(&hGpgOutC);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_7zdecrypt(PCSTR p7ZipExe, PCSTR pPassphrase, PCSTR pFileName, PCSTR pBaseDirectory, BOOL fOverWrite)
{
  HRESULT hr = S_OK;
  PGPGEZ_CMD punsevenzip = NULL;

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_unsevenzip_create(p7ZipExe, pFileName, pBaseDirectory,
    fOverWrite, pPassphrase, INVALID_HANDLE_VALUE, &punsevenzip));

  /* 7zip I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(punsevenzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(punsevenzip) == 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (punsevenzip) gpgez_cmd_release(&punsevenzip);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_untar(PCSTR pTarExe, PCSTR pFileName, PCSTR pBaseDirectory, BOOL fOverWrite) {
  HRESULT hr = S_OK;
  PGPGEZ_CMD puntar = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;

  /* tarWJp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFileName));
  DO_DEBUG(hr, gpgez_untar_create(pTarExe, pBaseDirectory, fOverWrite,
    &hIn, INVALID_HANDLE_VALUE, &puntar));
  CloseHandleClean(&hIn);

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(puntar));

  /* NOTICE: bsdtar̖untar͕KFail̂Ń`FbNȂ */

finish_func:
  /* I */
  if (puntar) gpgez_cmd_release(&puntar);
  if (hIn != INVALID_HANDLE_VALUE) CloseHandle(hIn);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_7zgunzip(PCSTR p7ZipExe, PCSTR pFileNameIn, PCSTR pFileNameOut)
{
  HRESULT hr = S_OK;
  HANDLE hOut = INVALID_HANDLE_VALUE;
  PGPGEZ_CMD punsevenzipunzip = NULL;

  /* WJp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, pFileNameOut));
  DO_DEBUG(hr, gpgez_unsevenzipunzip_create(p7ZipExe, pFileNameIn,
    hOut, INVALID_HANDLE_VALUE, &punsevenzipunzip));

  /* 7zip I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(punsevenzipunzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(punsevenzipunzip) == 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (punsevenzipunzip) gpgez_cmd_release(&punsevenzipunzip);
  CloseHandleClean(&hOut);
  return hr;
}
#endif

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_changeuserid(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFpr, PCSTR pName, PCSTR pEmail, PCSTR pComment, PCSTR pPassphrase) {
  HRESULT hr = S_OK;
  PCMDEZCONTEXT pgctx = NULL;
  HANDLE hCommandW = INVALID_HANDLE_VALUE, hCommandR = INVALID_HANDLE_VALUE;
  HANDLE hStatusW = INVALID_HANDLE_VALUE, hStatusR = INVALID_HANDLE_VALUE;
  CHAR buff[MAX_LINE + 1];
  PREADFILEBUFF pfilebuff = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&pgctx, pGpgExe, NULL));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--homedir", pGpgHome, TRUE));
/*
  if (pslPubRing) {
    PSTRINGLIST_NODE pnode = NULL;
    for (pnode = stringlist_get_head(pslPubRing); pnode; pnode = stringlist_get_next(pnode)) {
      if (FAILED(hr = cmdez_add_option(pgctx, "--keyring", stringlist_get_string(pnode), TRUE))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
  }
  if (pslSecRing) {
    PSTRINGLIST_NODE pnode = NULL;
    for (pnode = stringlist_get_head(pslSecRing); pnode; pnode = stringlist_get_next(pnode)) {
      if (FAILED(hr = cmdez_add_option(pgctx, "--secret-keyring", stringlist_get_string(pnode), TRUE))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
  }
  if (fNoDefault) {
    if (FAILED(hr = cmdez_add_option(pgctx, "--no-default-keyring", NULL, FALSE))) {
      goto finish_func;
    }
  }
*/
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--no-random-seed-file", NULL, FALSE));
  DO_DEBUG(hr, util_create_write_pipe(&hCommandW, &hCommandR));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--command-fd", util_itoa((INT)hCommandR, buff, 10), FALSE));
  DO_DEBUG(hr, util_create_read_pipe(&hStatusR, &hStatusW));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--status-fd", util_itoa((INT)hStatusW, buff, 10), FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--edit-key", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_operand(pgctx, pFpr, TRUE));
  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgctx));
  CloseHandleClean(&hCommandR);
  CloseHandleClean(&hStatusW);

  DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, hStatusR));
  /* adduid R}hs */
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keyedit.prompt"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "adduid"));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keygen.name"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, pName));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keygen.email"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, pEmail));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keygen.comment"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, pComment));
  if (pPassphrase) {
    DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_HIDDEN passphrase.enter"));
    DO_DEBUG(hr, WriteFileLine1(hCommandW, pPassphrase));
  }
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keyedit.prompt"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "1"));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keyedit.prompt"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "deluid"));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_BOOL keyedit.remove.uid.okay"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "y"));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keyedit.prompt"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "1"));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keyedit.prompt"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "trust"));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE edit_ownertrust.value"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "5"));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_BOOL edit_ownertrust.set_ultimate.okay"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "y"));
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE keyedit.prompt"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "save"));

  while (!fEofFileLine(pfilebuff)) {
    DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
  }
  DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));

  CloseHandleClean(&hStatusR);
  CloseHandleClean(&hCommandW);

  /* gpg I҂ */
  DO_DEBUG(hr, cmdez_finish(pgctx));
  /* qvZX̖߂l`FbN */
  if (cmdez_get_exitcode(pgctx) != 0) {
  } else {
  }

finish_func:
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  if (hStatusW != INVALID_HANDLE_VALUE) CloseHandle(hStatusW);
  if (hStatusR != INVALID_HANDLE_VALUE) CloseHandle(hStatusR);
  if (hCommandW != INVALID_HANDLE_VALUE) CloseHandle(hCommandW);
  if (hCommandR != INVALID_HANDLE_VALUE) CloseHandle(hCommandR);
  if (pgctx) cmdez_release(&pgctx);

  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_changepassphrase(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFpr, PCSTR pOldPass, PCSTR pNewPass) {
  HRESULT hr = S_OK;
  PCMDEZCONTEXT pgctx = NULL;
  HANDLE hStatusW = INVALID_HANDLE_VALUE, hStatusR = INVALID_HANDLE_VALUE;
  HANDLE hCommandR = INVALID_HANDLE_VALUE, hCommandW = INVALID_HANDLE_VALUE;
  CHAR buff[MAX_LINE + 1];
  PREADFILEBUFF pfilebuff = NULL;

  DO_DEBUG(hr, util_create_read_pipe(&hStatusR, &hStatusW));
  DO_DEBUG(hr, util_create_write_pipe(&hCommandW, &hCommandR));
  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&pgctx, pGpgExe, NULL));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--homedir", pGpgHome, TRUE));
  STRINGLIST_ADDOPTION(hr, pgctx, pslPubRing, "--keyring", TRUE);
  STRINGLIST_ADDOPTION(hr, pgctx, pslSecRing, "--secret-keyring", TRUE);
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--no-random-seed-file --always-trust", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--status-fd", util_itoa((INT)hStatusW, buff, 10), FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--command-fd", util_itoa((INT)hCommandR, buff, 10), FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--edit-key", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, pFpr, NULL, FALSE));

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgctx));
  CloseHandleClean(&hStatusW);
  CloseHandleClean(&hCommandR);
    
  DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, hStatusR));

  /* passwd R}hs */
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "passwd"));
  if (pOldPass && (*pOldPass)) {
    /* keyɃpXt[YĂꍇ */
    DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_HIDDEN"));
    DO_DEBUG(hr, WriteFileLine1(hCommandW, pOldPass));
  }
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_HIDDEN"));
  if (pNewPass && (*pNewPass)) {
    /* VpXt[Yɒꍇ */
    DO_DEBUG(hr, WriteFileLine1(hCommandW, pNewPass));
  } else {
    /* pXt[YOꍇ */
    DO_DEBUG(hr, WriteFileLine1(hCommandW, ""));
    DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_BOOL"));
    DO_DEBUG(hr, WriteFileLine1(hCommandW, "y"));
  }
  DO_DEBUG(hr, gpgez_ReadFileLine(pfilebuff, buff, MAX_LINE, "[GNUPG:] GET_LINE"));
  DO_DEBUG(hr, WriteFileLine1(hCommandW, "save"));
  while (!fEofFileLine(pfilebuff)) {
    DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
  }
  DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));

  CloseHandleClean(&hStatusR);
  CloseHandleClean(&hCommandW);

  /* gpg I҂ */
  DO_DEBUG(hr, cmdez_finish(pgctx));
  /* qvZX̖߂l`FbN */
  if (cmdez_get_exitcode(pgctx) != 0) {
  } else {
  }

finish_func:
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  if (hStatusW != INVALID_HANDLE_VALUE) CloseHandle(hStatusW);
  if (hStatusR != INVALID_HANDLE_VALUE) CloseHandle(hStatusR);
  if (hCommandW != INVALID_HANDLE_VALUE) CloseHandle(hCommandW);
  if (hCommandR != INVALID_HANDLE_VALUE) CloseHandle(hCommandR);
  if (pgctx) cmdez_release(&pgctx);

  return hr;
}

#if 0
/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_import(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefault, PCSTR pFileName) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;

  /* C|[gp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFileName));
  DO_DEBUG(hr, gpgez_gpg_create_import(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, hIn, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }

finish_func:
  /* I */
  CloseHandleClean(&hIn);
  if (pgpg) gpgez_gpg_release(&pgpg);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_export(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFpr, PCSTR pFileName, BOOL fAscii) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hOut = INVALID_HANDLE_VALUE;

  /* GNX|[gp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, pFileName));
  DO_DEBUG(hr, gpgez_gpg_create_export(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, pFpr, fAscii, hOut, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hOut);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }

finish_func:
  /* I */
  CloseHandleClean(&hOut);
  if (pgpg) gpgez_gpg_release(&pgpg);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_delete(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefault, PCSTR pFpr) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;

  /* 폜\̍쐬 */
  DO_DEBUG(hr, gpgez_gpg_create_deletekey(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, pFpr, INVALID_HANDLE_VALUE, &pgpg));

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_detach_sign(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFile, PCSTR pSfile, PCSTR pPassphraseSym, BOOL fAscii) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE;

  /* p\̍쐬 */
  if (pFile) {
    DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFile));
  }
  if (pSfile) {
    DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, pSfile));
  }
  DO_DEBUG(hr, gpgez_gpg_create_sign(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, pPassphraseSym, fAscii, TRUE, hIn, hOut, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_verify(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFile, PCSTR pSfile, PCSTR pPassphraseSym) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;

  /* ؗp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFile));
  DO_DEBUG(hr, gpgez_gpg_create_verify(pGpgExe, pGpgHome, pslSecRing, pslPubRing,
    FALSE, pPassphraseSym, pSfile, hIn, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hIn);
  return hr;
}
#endif

/*------------------------------------------------------------*/
enum {
  BLOCKTYPE_NONE = 0,
  BLOCKTYPE_SEC,
  BLOCKTYPE_PUB,
  BLOCKTYPE_OTHER
};

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_listkeys(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefaultKeyRing, BOOL fSec, PSTRINGLIST pKeys)
{
  HRESULT hr = S_OK;
  PCMDEZCONTEXT pgctx = NULL;
  PREADFILEBUFF pfilebuff = NULL;
  HANDLE hP = INVALID_HANDLE_VALUE, hC = INVALID_HANDLE_VALUE;
  INT blocktype;

  DO_DEBUG_RETC(hr, (!pKeys), ERROR_INVALID_FUNCTION);

  {
    CHAR buff[MAX_LINE + 1] = "";
    CHAR key[MAX_LINE + 1] = "";

    DO_DEBUG(hr, cmdez_initialize(&pgctx, pGpgExe, NULL));
    DO_DEBUG(hr, cmdez_add_option(pgctx, "--homedir", pGpgHome, TRUE));
    if (fNoDefaultKeyRing) {
      DO_DEBUG(hr, cmdez_add_option(pgctx, "--no-default-keyring", NULL, FALSE));
    }
    if (fSec) {
      DO_DEBUG(hr, cmdez_add_option(pgctx, "--list-secret-keys", NULL, FALSE));
    } else {
      DO_DEBUG(hr, cmdez_add_option(pgctx, "--list-keys", NULL, FALSE));
    }
    DO_DEBUG(hr, cmdez_add_option(pgctx, "--with-colons", NULL, FALSE));
    DO_DEBUG(hr, cmdez_add_option(pgctx, "--fixed-list-mode", NULL, FALSE));
    DO_DEBUG(hr, cmdez_add_option(pgctx, "--with-fingerprint", NULL, FALSE));
    DO_DEBUG(hr, cmdez_add_option(pgctx, "--with-fingerprint", NULL, FALSE));
    STRINGLIST_ADDOPTION(hr, pgctx, pslPubRing, "--keyring", TRUE);
    STRINGLIST_ADDOPTION(hr, pgctx, pslSecRing, "--secret-keyring", TRUE);
    DO_DEBUG(hr, util_create_read_pipe(&hP, &hC));
    DO_DEBUG(hr, cmdez_set_handle(pgctx, HANDLE_TYPE_STDOUT, hC));
    DO_DEBUG(hr, cmdez_execute(pgctx));
    CloseHandleClean(&hC);

    DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, hP));
    while (!fEofFileLine(pfilebuff)) {
      DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
      if (!util_strncmp(buff, "sec:", 4)) {
        if (key[0]) {
          DO_DEBUG(hr, stringlist_add_node(pKeys, key));
        }
        blocktype = BLOCKTYPE_SEC;
        key[0] = '\0';
      } else if (!util_strncmp(buff, "pub:", 4)) {
        if (key[0]) {
          DO_DEBUG(hr, stringlist_add_node(pKeys, key));
        }
        blocktype = BLOCKTYPE_PUB;
        key[0] = '\0';
      } else if (!util_strncmp(buff, "fpr:", 4)) {
        if ((blocktype == BLOCKTYPE_SEC) || (blocktype == BLOCKTYPE_PUB)) {
          DO_DEBUG(hr, sv_get(buff, ':', 9, key, MAX_LINE));
          DO_DEBUG(hr, util_strcat(key, MAX_LINE, ":"));
        }
      } else if (!util_strncmp(buff, "uid:", 4)) {
        if ((blocktype == BLOCKTYPE_SEC) || (blocktype == BLOCKTYPE_PUB)) {
          WCHAR szTemp16[MAX_LINE * 2 + 1];
          CHAR szUid[MAX_LINE + 1];
          PCSTR pc;
          INT len = 0;

          DO_DEBUG_RETC(hr, (!key[0]), ERROR_INVALID_DATA);
          pc = util_strchr(key, ':');
          DO_DEBUG_RETC(hr, (!pc), ERROR_INVALID_DATA);
          if (!*(pc + 1)) {
            DO_DEBUG(hr, sv_get(buff, ':', 9, szUid, MAX_LINE));
            if (MultiByteToWideChar(CP_UTF8, 0, szUid, -1, NULL, 0) * 2 < MAX_LINE) {
              if (!MultiByteToWideChar(CP_UTF8, 0, szUid, -1, szTemp16, MAX_LINE)) {
                DO_DEBUG_RETC(hr, TRUE, GetLastError());
              }
              if (WideCharToMultiByte(CP_ACP, 0, szTemp16, -1, NULL, 0, NULL, NULL) < MAX_LINE) {
                if (!WideCharToMultiByte(CP_ACP, 0, szTemp16, -1, szUid, MAX_LINE, NULL, NULL)) {
                  DO_DEBUG_RETC(hr, TRUE, GetLastError());
                }
              } else {
                szUid[0] = '\0';
              }
            } else {
              szUid[0] = '\0';
            }
            DO_DEBUG(hr, util_strcat(key, MAX_LINE, szUid));
          }
        }
      } else {
        blocktype = BLOCKTYPE_OTHER;
      }
    }
    if (key[0]) {
      DO_DEBUG(hr, stringlist_add_node(pKeys, key));
    }
    DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));
    CloseHandleClean(&hP);
    cmdez_finish(pgctx);
    cmdez_release(&pgctx);
    pgctx = NULL;
  }
finish_func:
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  if (hP != INVALID_HANDLE_VALUE) CloseHandle(hP);
  if (hC != INVALID_HANDLE_VALUE) CloseHandle(hC);
  if (pgctx) cmdez_release(&pgctx);

  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_armor(PCSTR pGpgExe, PCSTR pGpgHome, PCSTR pFileNameIn, PCSTR pFileNameOut, BOOL fEnArmor) {
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE;

  /* Ascii Armorp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pFileNameIn));
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, pFileNameOut));
  DO_DEBUG(hr, gpgez_gpg_create_armor(pGpgExe, pGpgHome, fEnArmor, hIn, hOut, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);
  return hr;
}
/*------------------------------------------------------------*/

