#define THISFILENAME "gpgez/gpgez_genkey.c"

#include <windows.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "gpgez_util.h"

#define NUMBUFF_LEN 10

/*------------------------------------------------------------*/
typedef struct struct_gpgez_genkey {
  PCMDEZCONTEXT pgctx;
  HANDLE hStdInW;
} GPGEZ_GENKEY;

/*------------------------------------------------------------*/
static DWORD
WriteFileLine2(HANDLE hFile, LPCSTR s1, LPCSTR s2) {
  HRESULT hr = S_OK;
  DWORD dwWritten;
  
  if (!WriteFile(hFile, s1, util_strlen(s1, -1), &dwWritten, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WriteFile(hFile, " ", util_strlen(" ", -1), &dwWritten, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WriteFile(hFile, s2, util_strlen(s2, -1), &dwWritten, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WriteFile(hFile, "\r\n", util_strlen("\r\n", -1), &dwWritten, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_genkey_create(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                    HANDLE hStatusW, PGPGEZ_GENKEY *ppgpgez_genkey)
{
  HRESULT hr = S_OK;
  PGPGEZ_GENKEY pgpg = NULL;
  HANDLE hStdInR = INVALID_HANDLE_VALUE;
  
  pgpg = (PGPGEZ_GENKEY)util_malloc(sizeof(GPGEZ_GENKEY));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->hStdInW = INVALID_HANDLE_VALUE;

  /* gpg s */
  if (FAILED(hr = cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL))) goto finish_func;
  if (FAILED(hr = cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE))) goto finish_func;
/*
  STRINGLIST_ADDOPTION(hr, pgctx, pslPubRing, "--keyring", TRUE);
  STRINGLIST_ADDOPTION(hr, pgctx, pslSecRing, "--secret-keyring", TRUE);
  DO_DEBUG_IF(hr, fNoDefault, cmdez_add_option(pgctx, "--no-default-keyring", NULL, FALSE));
*/
  if (FAILED(hr = cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE))) goto finish_func;
  if (hStatusW != INVALID_HANDLE_VALUE) {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    if (FAILED(hr = cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)hStatusW, nmb, NUMBUFF_LEN), FALSE))) goto finish_func;
  }
  if (FAILED(hr = cmdez_add_option(pgpg->pgctx, "--gen-key", NULL, FALSE))) goto finish_func;
  if (FAILED(hr = util_create_write_pipe(&(pgpg->hStdInW), &hStdInR))) goto finish_func;
  if (FAILED(hr = cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDIN, hStdInR))) goto finish_func;
  /* gpg s */
  if (FAILED(hr = cmdez_execute(pgpg->pgctx))) goto finish_func;
  CloseHandle(hStdInR);
  hStdInR = INVALID_HANDLE_VALUE;
finish_func:
  if (hStdInR != INVALID_HANDLE_VALUE) CloseHandle(hStdInR);
  if (FAILED(hr)) {
    gpgez_genkey_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_genkey = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_genkey_release(PGPGEZ_GENKEY *ppgpgez_genkey)
{
  PGPGEZ_GENKEY pgpg = *ppgpgez_genkey;
  HRESULT hr = S_OK;
  
  if (pgpg) {
    if (pgpg->pgctx) {
      cmdez_release(&(pgpg->pgctx));
    }
    util_free(pgpg);
    pgpg = NULL;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HANDLE
gpgez_genkey_get_process_handle(PGPGEZ_GENKEY pgpgez_genkey)
{
  return cmdez_get_process(pgpgez_genkey->pgctx);
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_genkey_write_param(PGPGEZ_GENKEY pgpgez_genkey, PCSTR pkey, PCSTR pval)
{
  return WriteFileLine2(pgpgez_genkey->hStdInW, pkey, pval);
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_genkey_write_cmd(PGPGEZ_GENKEY pgpgez_genkey, PCSTR pcmd)
{
  return WriteFileLine1(pgpgez_genkey->hStdInW, pcmd);
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_genkey_close_write(PGPGEZ_GENKEY pgpgez_genkey)
{
  if (pgpgez_genkey->hStdInW != INVALID_HANDLE_VALUE) {
    CloseHandle(pgpgez_genkey->hStdInW);
    pgpgez_genkey->hStdInW = INVALID_HANDLE_VALUE;
  }
  return S_OK;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_genkey_finish(PGPGEZ_GENKEY pgpgez_genkey)
{
  return cmdez_finish(pgpgez_genkey->pgctx);
}

/*------------------------------------------------------------*/
DllExport INT
gpgez_genkey_get_exitcode(PGPGEZ_GENKEY pgpgez_genkey)
{
  return cmdez_get_exitcode(pgpgez_genkey->pgctx);
}

/*------------------------------------------------------------*/
