#define THISFILENAME "gpgez/gpgez_gpg.c"

#include <windows.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "gpgez_util.h"

#define MAX_LINE 1024
#define NUMBUFF_LEN 10

#define DO_DEBUG(hr, fn) if (FAILED((hr) = (fn))) { DEBUGERROR(0, (hr)); goto finish_func;}

/*------------------------------------------------------------*/
typedef struct struct_gpgez_gpg {
  PCMDEZCONTEXT pgctx;
  PGPGEZ_PARAM_GPGSTATUS pstatus;
  HANDLE hThreadStatus;
} GPGEZ_GPG, *PGPGEZ_GPG;

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_encrypt(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                     BOOL fNoDefaultKeyRing, PSTRINGLIST pslRecipients, PCSTR pPassphrase,
                     BOOL fSign, BOOL fAscii, PCSTR pFileName,
                     HANDLE hIn, HANDLE hOut, HANDLE hErr, PGPGEZ_GPG *ppgpgez_encrypt)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hPassW = INVALID_HANDLE_VALUE, hPassR = INVALID_HANDLE_VALUE;
  DWORD dwThreadId;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslRecipients, "--recipient", TRUE));
  if (fSign) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--sign", NULL, FALSE));
  }
  if (fAscii) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "-a", NULL, FALSE));
  }
  if (pFileName) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--set-filename", pFileName, TRUE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  if (pPassphrase) {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, util_create_write_pipe(&hPassW, &hPassR));
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--passphrase-fd", util_itoa((INT)(hPassR), nmb, NUMBUFF_LEN), FALSE));
  }
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  if (pslRecipients && stringlist_get_len(pslRecipients) > 0) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--encrypt", NULL, FALSE));
  } else {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--symmetric", NULL, FALSE));
  }
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }
  if (hIn != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDIN, hIn));
  }
  if (hOut != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDOUT, hOut));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));

  /* pXt[Yo */
  if (pPassphrase) {
    DWORD dwWritten;

    CloseHandle(hPassR);
    hPassR = INVALID_HANDLE_VALUE;
    WriteFile(hPassW, pPassphrase, util_strlen(pPassphrase, -1), &dwWritten, NULL);
    CloseHandle(hPassW);
    hPassW = INVALID_HANDLE_VALUE;
  }
  gpgez_gpgstatus_close_handle_write(pgpg->pstatus);
  pgpg->hThreadStatus = CreateThread(NULL, 0, procObserveGpgStatus, (LPVOID)(pgpg->pstatus), 0, &dwThreadId);

finish_func:
  if (hPassW != INVALID_HANDLE_VALUE) CloseHandle(hPassW);
  if (hPassR != INVALID_HANDLE_VALUE) CloseHandle(hPassR);
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_encrypt = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_decrypt(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                     BOOL fNoDefaultKeyRing, BOOL fSkipVerify, PCSTR pPassphrase,
                     HANDLE hIn, HANDLE hOut, HANDLE hErr, PGPGEZ_GPG *ppgpgez_decrypt)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hPassW = INVALID_HANDLE_VALUE, hPassR = INVALID_HANDLE_VALUE;
  DWORD dwThreadId;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  if (fSkipVerify) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--skip-verify", NULL, FALSE));
  }
  if (pPassphrase) {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, util_create_write_pipe(&hPassW, &hPassR));
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--passphrase-fd", util_itoa((INT)(hPassR), nmb, NUMBUFF_LEN), FALSE));
  }
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--decrypt", NULL, FALSE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }
  if (hIn != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDIN, hIn));
  }
  if (hOut != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDOUT, hOut));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));

  /* pXt[Yo */
  if (pPassphrase) {
    DWORD dwWritten;

    CloseHandle(hPassR);
    hPassR = INVALID_HANDLE_VALUE;
    WriteFile(hPassW, pPassphrase, util_strlen(pPassphrase, -1), &dwWritten, NULL);
    CloseHandle(hPassW);
    hPassW = INVALID_HANDLE_VALUE;
  }
  gpgez_gpgstatus_close_handle_write(pgpg->pstatus);
  pgpg->hThreadStatus = CreateThread(NULL, 0, procObserveGpgStatus, (LPVOID)(pgpg->pstatus), 0, &dwThreadId);

finish_func:
  if (hPassW != INVALID_HANDLE_VALUE) CloseHandle(hPassW);
  if (hPassR != INVALID_HANDLE_VALUE) CloseHandle(hPassR);
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_decrypt = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_decrypt_listonly(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                                  BOOL fNoDefaultKeyRing, PCSTR pPassphrase, 
                                  HANDLE hIn, HANDLE hOut, HANDLE hErr, PGPGEZ_GPG *ppgpgez_decrypt)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hPassW = INVALID_HANDLE_VALUE, hPassR = INVALID_HANDLE_VALUE;
  DWORD dwThreadId;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  if (pPassphrase) {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, util_create_write_pipe(&hPassW, &hPassR));
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--passphrase-fd", util_itoa((INT)(hPassR), nmb, NUMBUFF_LEN), FALSE));
  }
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--list-only --decrypt", NULL, FALSE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }
  if (hIn != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDIN, hIn));
  }
  if (hOut != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDOUT, hOut));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));

  /* pXt[Yo */
  if (pPassphrase) {
    DWORD dwWritten;

    CloseHandle(hPassR);
    hPassR = INVALID_HANDLE_VALUE;
    WriteFile(hPassW, pPassphrase, util_strlen(pPassphrase, -1), &dwWritten, NULL);
    CloseHandle(hPassW);
    hPassW = INVALID_HANDLE_VALUE;
  }
  gpgez_gpgstatus_close_handle_write(pgpg->pstatus);
  pgpg->hThreadStatus = CreateThread(NULL, 0, procObserveGpgStatus, (LPVOID)(pgpg->pstatus), 0, &dwThreadId);

finish_func:
  if (hPassW != INVALID_HANDLE_VALUE) CloseHandle(hPassW);
  if (hPassR != INVALID_HANDLE_VALUE) CloseHandle(hPassR);
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_decrypt = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_import(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                    BOOL fNoDefaultKeyRing, HANDLE hIn,
                    HANDLE hErr, PGPGEZ_GPG *ppgpgez_import)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  DWORD dwThreadId;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--import", NULL, FALSE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }
  if (hIn != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDIN, hIn));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));
  gpgez_gpgstatus_close_handle_write(pgpg->pstatus);
  pgpg->hThreadStatus = CreateThread(NULL, 0, procObserveGpgStatus, (LPVOID)(pgpg->pstatus), 0, &dwThreadId);

finish_func:
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_import = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_export(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                    BOOL fNoDefaultKeyRing, PCSTR pFpr, BOOL fAscii,
                    HANDLE hOut, HANDLE hErr, PGPGEZ_GPG *ppgpgez_export)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  DWORD dwThreadId;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  if (fAscii) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "-a", NULL, FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--export", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_operand(pgpg->pgctx, pFpr, TRUE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }
  if (hOut != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDOUT, hOut));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));
  gpgez_gpgstatus_close_handle_write(pgpg->pstatus);
  pgpg->hThreadStatus = CreateThread(NULL, 0, procObserveGpgStatus, (LPVOID)(pgpg->pstatus), 0, &dwThreadId);

finish_func:
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_export = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_deletekey(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                    BOOL fNoDefaultKeyRing, PCSTR pFpr, HANDLE hErr, PGPGEZ_GPG *ppgpgez_delete)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  DWORD dwThreadId;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--delete-key", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_operand(pgpg->pgctx, pFpr, TRUE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));
  gpgez_gpgstatus_close_handle_write(pgpg->pstatus);
  pgpg->hThreadStatus = CreateThread(NULL, 0, procObserveGpgStatus, (LPVOID)(pgpg->pstatus), 0, &dwThreadId);

finish_func:
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_delete = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_sign(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                     BOOL fNoDefaultKeyRing, PCSTR pPassphrase, BOOL fAscii, BOOL fDetach,
                     HANDLE hIn, HANDLE hOut, HANDLE hErr, PGPGEZ_GPG *ppgpgez_sign)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hPassW = INVALID_HANDLE_VALUE, hPassR = INVALID_HANDLE_VALUE;
  DWORD dwThreadId;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  if (fAscii) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "-a", NULL, FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  if (pPassphrase) {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, util_create_write_pipe(&hPassW, &hPassR));
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--passphrase-fd", util_itoa((INT)(hPassR), nmb, NUMBUFF_LEN), FALSE));
  }
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  if (fDetach) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--detach-sign", NULL, FALSE));
  } else {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--sign", NULL, FALSE));
  }
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }
  if (hIn != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDIN, hIn));
  }
  if (hOut != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDOUT, hOut));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));

  /* pXt[Yo */
  if (pPassphrase) {
    DWORD dwWritten;

    CloseHandle(hPassR);
    hPassR = INVALID_HANDLE_VALUE;
    WriteFile(hPassW, pPassphrase, util_strlen(pPassphrase, -1), &dwWritten, NULL);
    CloseHandle(hPassW);
    hPassW = INVALID_HANDLE_VALUE;
  }
  gpgez_gpgstatus_close_handle_write(pgpg->pstatus);
  pgpg->hThreadStatus = CreateThread(NULL, 0, procObserveGpgStatus, (LPVOID)(pgpg->pstatus), 0, &dwThreadId);

finish_func:
  if (hPassW != INVALID_HANDLE_VALUE) CloseHandle(hPassW);
  if (hPassR != INVALID_HANDLE_VALUE) CloseHandle(hPassR);
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_sign = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_verify(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                     BOOL fNoDefaultKeyRing, PCSTR pPassphrase, PCSTR pSignFileName,
                     HANDLE hIn, HANDLE hErr, PGPGEZ_GPG *ppgpgez_verify)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hPassW = INVALID_HANDLE_VALUE, hPassR = INVALID_HANDLE_VALUE;
  DWORD dwThreadId;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  if (pPassphrase) {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, util_create_write_pipe(&hPassW, &hPassR));
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--passphrase-fd", util_itoa((INT)(hPassR), nmb, NUMBUFF_LEN), FALSE));
  }
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--verify", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_operand(pgpg->pgctx, pSignFileName, TRUE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }
  if (hIn != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDIN, hIn));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));

  /* pXt[Yo */
  if (pPassphrase) {
    DWORD dwWritten;

    CloseHandle(hPassR);
    hPassR = INVALID_HANDLE_VALUE;
    WriteFile(hPassW, pPassphrase, util_strlen(pPassphrase, -1), &dwWritten, NULL);
    CloseHandle(hPassW);
    hPassW = INVALID_HANDLE_VALUE;
  }
  gpgez_gpgstatus_close_handle_write(pgpg->pstatus);
  pgpg->hThreadStatus = CreateThread(NULL, 0, procObserveGpgStatus, (LPVOID)(pgpg->pstatus), 0, &dwThreadId);

finish_func:
  if (hPassW != INVALID_HANDLE_VALUE) CloseHandle(hPassW);
  if (hPassR != INVALID_HANDLE_VALUE) CloseHandle(hPassR);
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_verify = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_create_armor(PCSTR pGpgExe, PCSTR pGpgHome, BOOL fEnArmor, HANDLE hIn, HANDLE hOut, HANDLE hErr, PGPGEZ_GPG *ppgpgez_armor)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  
  pgpg = (PGPGEZ_GPG)util_malloc(sizeof(GPGEZ_GPG));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->pstatus = NULL;
  pgpg->hThreadStatus = NULL;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, gpgez_gpgstatus_create(&(pgpg->pstatus)));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, GPGBASECMD, NULL, FALSE));
  {
    CHAR nmb[NUMBUFF_LEN + 1] = "";
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)gpgez_gpgstatus_get_handle_write(pgpg->pstatus), nmb, NUMBUFF_LEN), FALSE));
  }
  if (fEnArmor) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--enarmor", NULL, FALSE));
  } else {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--dearmor", NULL, FALSE));
  }
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDERR, hErr));
  }
  if (hIn != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDIN, hIn));
  }
  if (hOut != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pgpg->pgctx, HANDLE_TYPE_STDOUT, hOut));
  }

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));

finish_func:
  if (FAILED(hr)) {
    gpgez_gpg_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_armor = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_release(PGPGEZ_GPG *ppgpgez_gpg)
{
  PGPGEZ_GPG pgpg = *ppgpgez_gpg;
  HRESULT hr = S_OK;
  
  if (pgpg) {
    if (pgpg->pgctx) {
      cmdez_release(&(pgpg->pgctx));
    }
    if (pgpg->hThreadStatus) {
      CloseHandle(pgpg->hThreadStatus);
    }
    if (pgpg->pstatus) {
      gpgez_gpgstatus_release(&(pgpg->pstatus));
    }
    util_free(pgpg);
    pgpg = NULL;
  }
  *ppgpgez_gpg = NULL;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HANDLE
gpgez_gpg_get_process_handle(PGPGEZ_GPG pgpgez_gpg)
{
  return cmdez_get_process(pgpgez_gpg->pgctx);
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_finish(PGPGEZ_GPG pgpgez_gpg)
{
  HRESULT hr = S_OK;

  DO_DEBUG(hr, cmdez_finish(pgpgez_gpg->pgctx));
  if (pgpgez_gpg->hThreadStatus) {
    DWORD dwExitCode;

    if (WaitForSingleObject(pgpgez_gpg->hThreadStatus, INFINITE) == WAIT_FAILED) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      goto finish_func;
    }
    if (!GetExitCodeThread(pgpgez_gpg->hThreadStatus, &dwExitCode)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      goto finish_func;
    }
    if (dwExitCode != ERROR_SUCCESS) {
      hr = HRESULT_FROM_WIN32(dwExitCode);
      goto finish_func;
    }
    CloseHandle(pgpgez_gpg->hThreadStatus);
    pgpgez_gpg->hThreadStatus = NULL;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport INT
gpgez_gpg_get_exitcode(PGPGEZ_GPG pgpgez_gpg)
{
  return cmdez_get_exitcode(pgpgez_gpg->pgctx);
}

/*------------------------------------------------------------*/
DllExport PSTRINGLIST
gpgez_gpg_get_statuslist(PGPGEZ_GPG pgpgez_gpg)
{
  PSTRINGLIST plist = NULL;

  if (pgpgez_gpg->pstatus) {
    plist = gpgez_gpgstatus_get_status(pgpgez_gpg->pstatus);
  }
  return plist;
}
/*------------------------------------------------------------*/

/*------------------------------------------------------------*/
typedef struct struct_gpgez_gpg_editkey {
  PCMDEZCONTEXT pgctx;
  HANDLE hCommand;
  HANDLE hStatus;
  PREADFILEBUFF pfilebuff;
} GPGEZ_GPG_EDITKEY, *PGPGEZ_GPG_EDITKEY;

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_editkey_create(PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                         BOOL fNoDefaultKeyRing, PCSTR pFpr,
                         PGPGEZ_GPG_EDITKEY *ppgpg)
{
  HRESULT hr = S_OK;
  PGPGEZ_GPG_EDITKEY pgpg = NULL;
  HANDLE hStatusC = INVALID_HANDLE_VALUE;
  HANDLE hCommandC = INVALID_HANDLE_VALUE;
  CHAR nmb[NUMBUFF_LEN + 1];

  pgpg = (PGPGEZ_GPG_EDITKEY)util_malloc(sizeof(GPGEZ_GPG_EDITKEY));
  if (!pgpg) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->pgctx = NULL;
  pgpg->hCommand = INVALID_HANDLE_VALUE;
  pgpg->hStatus = INVALID_HANDLE_VALUE;

  /* gpg s */
  DO_DEBUG(hr, cmdez_initialize(&(pgpg->pgctx), pGpgExe, NULL));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--homedir", pGpgHome, TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslSecRing, "--secret-keyring", TRUE));
  DO_DEBUG(hr, gpgez_stringlist_addoption(pgpg->pgctx, pslPubRing, "--keyring", TRUE));
  if (fNoDefaultKeyRing) {
    DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-default-keyring", NULL, FALSE));
  }
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--no-random-seed-file", NULL, FALSE));
  DO_DEBUG(hr, util_create_write_pipe(&(pgpg->hCommand), &hCommandC));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--command-fd", util_itoa((INT)hCommandC, nmb, NUMBUFF_LEN), FALSE));
  DO_DEBUG(hr, util_create_read_pipe(&(pgpg->hStatus), &hStatusC));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--status-fd", util_itoa((INT)hStatusC, nmb, NUMBUFF_LEN), FALSE));
  DO_DEBUG(hr, cmdez_add_option(pgpg->pgctx, "--edit-key", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_operand(pgpg->pgctx, pFpr, TRUE));

  /* gpg s */
  DO_DEBUG(hr, cmdez_execute(pgpg->pgctx));
  CloseHandleClean(&hStatusC);
  CloseHandleClean(&hCommandC);
  DO_DEBUG(hr, CreateReadFileBuff(&(pgpg->pfilebuff), pgpg->hStatus));

finish_func:
  if (FAILED(hr)) {
    gpgez_gpg_editkey_release(&pgpg);
    if (pgpg->pfilebuff) ReleaseReadFileBuff(&(pgpg->pfilebuff));
    pgpg = NULL;
  }
  *ppgpg = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_editkey_release(PGPGEZ_GPG_EDITKEY *ppgpgez_gpg_editkey)
{
  PGPGEZ_GPG_EDITKEY pgpg = *ppgpgez_gpg_editkey;
  HRESULT hr = S_OK;

  if (pgpg) {
    if (pgpg->pfilebuff) {
      while (!fEofFileLine(pgpg->pfilebuff)) {
        CHAR buff[MAX_LINE + 1];

        if (FAILED(ReadFileLine(pgpg->pfilebuff, buff, MAX_LINE))) {
          break;
        }
      }
      DetatchReadFileLine(pgpg->pfilebuff);
      ReleaseReadFileBuff(&(pgpg->pfilebuff));
    }
    if (pgpg->pgctx) {
      cmdez_release(&(pgpg->pgctx));
    }
    if (pgpg->hCommand) {
      CloseHandle(pgpg->hCommand);
    }
    if (pgpg->hStatus) {
      CloseHandle(pgpg->hStatus);
    }
    util_free(pgpg);
    pgpg = NULL;
  }
  return hr;
}
  
/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_editkey_write_command(PGPGEZ_GPG_EDITKEY pgpg, PCSTR pstr)
{
  return WriteFileLine1(pgpg->hCommand, pstr);
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_editkey_wait_write_command(PGPGEZ_GPG_EDITKEY pgpg, INT iStatus, PCSTR pHint, PCSTR pstr)
{
  HRESULT hr = S_OK;
  CHAR buff[MAX_LINE + 1];

  while (TRUE) {
    INT iGpgStatus = 0;

    if (FAILED(hr = ReadFileLine(pgpg->pfilebuff, buff, MAX_LINE + 1))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    iGpgStatus = gpgez_interpret_status(buff);
    if (iGpgStatus == iStatus) {
      if ((!pHint) || (!util_strncmp(pHint, sv_pos(buff, ' ', 2), util_strlen(pHint, -1)))) {
        hr = S_OK;
        goto finish_func;
      }
    }
    if ((fEofFileLine(pgpg->pfilebuff)) ||
        (iGpgStatus == GPGSTATUS_GET_LINE) ||
        (iGpgStatus == GPGSTATUS_GET_HIDDEN) ||
        (iGpgStatus == GPGSTATUS_GET_BOOL)) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_gpg_editkey_finish(PGPGEZ_GPG_EDITKEY pgpg)
{
  HRESULT hr = S_OK;

  CloseHandleClean(&(pgpg->hStatus));
  CloseHandleClean(&(pgpg->hCommand));
  DO_DEBUG(hr, cmdez_finish(pgpg->pgctx));
finish_func:
  return hr;
}

/*------------------------------------------------------------*/

/*------------------------------------------------------------*/
typedef struct struct_gpgez_cmd {
  PCMDEZCONTEXT pctx;
} GPGEZ_CMD, *PGPGEZ_CMD;

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_tar_create(PCSTR pTarExe, INT iCompressMode, PCSTR pBaseDirectory, PSTRINGLIST pslFiles,
                 HANDLE hOut, HANDLE hErr, 
                 PGPGEZ_CMD *ppgpgez_tar)
{
  HRESULT hr = S_OK;
  PGPGEZ_CMD ptar = NULL;
  
  ptar = (PGPGEZ_CMD)util_malloc(sizeof(GPGEZ_CMD));
  if (!ptar) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  ptar->pctx = NULL;

  /* tar s */
  DO_DEBUG(hr, cmdez_initialize(&(ptar->pctx), pTarExe, pBaseDirectory));
  switch (iCompressMode) {
  case TAR_COMPRESS_NONE:
    DO_DEBUG(hr, cmdez_add_option(ptar->pctx, "-cvf -", NULL, FALSE));
    break;
  case TAR_COMPRESS_GZIP:
    DO_DEBUG(hr, cmdez_add_option(ptar->pctx, "-czvf -", NULL, FALSE));
    break;
  case TAR_COMPRESS_BZIP2:
    DO_DEBUG(hr, cmdez_add_option(ptar->pctx, "-cjvf -", NULL, FALSE));
    break;
  default:
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  DO_DEBUG(hr, gpgez_stringlist_addoperand(ptar->pctx, pslFiles, TRUE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(ptar->pctx, HANDLE_TYPE_STDERR, hErr));
  }
  DO_DEBUG(hr, cmdez_set_handle(ptar->pctx, HANDLE_TYPE_STDOUT, hOut));

  /* tar s */
  DO_DEBUG(hr, cmdez_execute(ptar->pctx));

finish_func:
  if (FAILED(hr)) {
    gpgez_cmd_release(&ptar);
    ptar = NULL;
  }
  *ppgpgez_tar = ptar;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_untar_create(PCSTR pTarExe, PCSTR pBaseDirectory, BOOL fOverWrite,
                 HANDLE hIn, HANDLE hErr,
                 PGPGEZ_CMD *ppgpgez_untar)
{
  HRESULT hr = S_OK;
  PGPGEZ_CMD puntar = NULL;
  
  puntar = (PGPGEZ_CMD)util_malloc(sizeof(GPGEZ_CMD));
  if (!puntar) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  puntar->pctx = NULL;

  /* tar s */
  DO_DEBUG(hr, cmdez_initialize(&(puntar->pctx), pTarExe, pBaseDirectory));
  if (fOverWrite) {
    DO_DEBUG(hr, cmdez_add_option(puntar->pctx, "-xvf -", NULL, FALSE));
  } else {
    DO_DEBUG(hr, cmdez_add_option(puntar->pctx, "-xkvf -", NULL, FALSE));
  }
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(puntar->pctx, HANDLE_TYPE_STDERR, hErr));
  }
  DO_DEBUG(hr, cmdez_set_handle(puntar->pctx, HANDLE_TYPE_STDIN, hIn));

  /* tar s */
  DO_DEBUG(hr, cmdez_execute(puntar->pctx));

finish_func:
  if (FAILED(hr)) {
    gpgez_cmd_release(&puntar);
    puntar = NULL;
  }
  *ppgpgez_untar = puntar;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_testtar_create(PCSTR pTarExe, HANDLE hIn, HANDLE hErr, PGPGEZ_CMD *ppgpgez_testtar)
{
  HRESULT hr = S_OK;
  PGPGEZ_CMD ptesttar = NULL;
  
  ptesttar = (PGPGEZ_CMD)util_malloc(sizeof(GPGEZ_CMD));
  if (!ptesttar) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  ptesttar->pctx = NULL;

  /* tar s */
  DO_DEBUG(hr, cmdez_initialize(&(ptesttar->pctx), pTarExe, NULL));
  DO_DEBUG(hr, cmdez_add_option(ptesttar->pctx, "-tf -", NULL, FALSE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(ptesttar->pctx, HANDLE_TYPE_STDERR, hErr));
  }
  DO_DEBUG(hr, cmdez_set_handle(ptesttar->pctx, HANDLE_TYPE_STDIN, hIn));

  /* tar s */
  DO_DEBUG(hr, cmdez_execute(ptesttar->pctx));

finish_func:
  if (FAILED(hr)) {
    gpgez_cmd_release(&ptesttar);
    ptesttar = NULL;
  }
  *ppgpgez_testtar = ptesttar;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_sevenzip_create(PCSTR pSevenZipExe, INT iCompressMode, PCSTR pBaseDirectory, PSTRINGLIST pslFiles,
                      PCSTR pPassphrase, PCSTR pZFileName, HANDLE hErr, 
                      PGPGEZ_CMD *ppgpgez_sevenzip)
{
  HRESULT hr = S_OK;
  PGPGEZ_CMD psevenzip = NULL;
  PSTRINGLIST_NODE pnode1;
  PCSTR petype = "-t7z";
  
  psevenzip = (PGPGEZ_CMD)util_malloc(sizeof(GPGEZ_CMD));
  if (!psevenzip) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  psevenzip->pctx = NULL;

  /* 7-Zip s */
  DO_DEBUG(hr, cmdez_initialize(&(psevenzip->pctx), pSevenZipExe, pBaseDirectory));
  DO_DEBUG(hr, cmdez_add_option(psevenzip->pctx, "a", NULL, FALSE));
  switch (iCompressMode) {
  case SEVENZIP_ZIPENC:
    petype = "-tzip";
    break;
  case SEVENZIP_ZIPENCAES:
    petype = "-tzip -mem=AES256";
    break;
  case SEVENZIP_7ZENC:
    petype = "-t7z";
    break;
  case SEVENZIP_ZIPPLAIN:
    petype = "-tzip";
    break;
  case SEVENZIP_7ZPLAIN:
    petype = "-t7z";
    break;
  case SEVENZIP_SFX:
    petype = "-sfx7z.sfx";
    break;
  default:
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  DO_DEBUG(hr, cmdez_add_option2(psevenzip->pctx, petype, pZFileName, " ", TRUE));

  DO_DEBUG(hr, cmdez_add_option(psevenzip->pctx, "-r", NULL, FALSE));

  if ((iCompressMode == SEVENZIP_ZIPENC) ||
      (iCompressMode == SEVENZIP_ZIPENCAES) ||
      (iCompressMode == SEVENZIP_7ZENC) ||
      (iCompressMode == SEVENZIP_SFX)) {
    if (pPassphrase && (*pPassphrase)) {
      DO_DEBUG(hr, cmdez_add_option2(psevenzip->pctx, "-p", pPassphrase, "", FALSE));
    }
  }

  for (pnode1 = stringlist_get_head(pslFiles); pnode1; pnode1 = stringlist_get_next(pnode1)) {
    DO_DEBUG(hr, cmdez_add_operand(psevenzip->pctx, stringlist_get_string(pnode1), TRUE));
  }

  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(psevenzip->pctx, HANDLE_TYPE_STDERR, hErr));
  }

  /* o̓t@Cɂ΍폜 */
  {
    HANDLE hFind = INVALID_HANDLE_VALUE;
    WIN32_FIND_DATA wfd;

    hFind = FindFirstFile(pZFileName, &wfd);
    if (hFind != INVALID_HANDLE_VALUE) {
      FindClose(hFind);
      if (!DeleteFile(pZFileName)) {
        DWORD dw = GetLastError();
        hr = HRESULT_FROM_WIN32(dw);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
  }

  /* 7-Zip s */
  DO_DEBUG(hr, cmdez_execute(psevenzip->pctx));

finish_func:
  if (FAILED(hr)) {
    gpgez_cmd_release(&psevenzip);
    psevenzip = NULL;
  }
  *ppgpgez_sevenzip = psevenzip;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_unsevenzip_create(PCSTR pSevenZipExe, PCSTR pFileName, PCSTR pBaseDirectory,
                        BOOL fOverWrite, PCSTR pPassphrase, HANDLE hErr, 
                        PGPGEZ_CMD *ppgpgez_unsevenzip)
{
  HRESULT hr = S_OK;
  PGPGEZ_CMD punsevenzip = NULL;
  
  punsevenzip = (PGPGEZ_CMD)util_malloc(sizeof(GPGEZ_CMD));
  if (!punsevenzip) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  punsevenzip->pctx = NULL;

  /* 7-Zip s */
  DO_DEBUG(hr, cmdez_initialize(&(punsevenzip->pctx), pSevenZipExe, pBaseDirectory));
  DO_DEBUG(hr, cmdez_add_option(punsevenzip->pctx, "x -y ", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_option2(punsevenzip->pctx, "-o", pBaseDirectory, "", TRUE));
  if (fOverWrite) {
    DO_DEBUG(hr, cmdez_add_option(punsevenzip->pctx, "-aoa", NULL, FALSE));
  } else {
    DO_DEBUG(hr, cmdez_add_option(punsevenzip->pctx, "-aos", NULL, FALSE));
  }
  if (pPassphrase) {
    DO_DEBUG(hr, cmdez_add_option2(punsevenzip->pctx, "-p", pPassphrase, "", FALSE));
  }
  DO_DEBUG(hr, cmdez_add_operand(punsevenzip->pctx, pFileName, TRUE));

  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(punsevenzip->pctx, HANDLE_TYPE_STDERR, hErr));
  }

  /* 7-Zip s */
  DO_DEBUG(hr, cmdez_execute(punsevenzip->pctx));

finish_func:
  if (FAILED(hr)) {
    gpgez_cmd_release(&punsevenzip);
    punsevenzip = NULL;
  }
  *ppgpgez_unsevenzip = punsevenzip;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_unsevenzipunzip_create(PCSTR pSevenZipExe, PCSTR pFileNameIn, HANDLE hOut,
                             HANDLE hErr, PGPGEZ_CMD *ppgpgez_unsevenzipunzip)
{
  HRESULT hr = S_OK;
  PGPGEZ_CMD punsevenzipunzip = NULL;
  
  punsevenzipunzip = (PGPGEZ_CMD)util_malloc(sizeof(GPGEZ_CMD));
  if (!punsevenzipunzip) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  punsevenzipunzip->pctx = NULL;

  /* 7-Zip s */
  DO_DEBUG(hr, cmdez_initialize(&(punsevenzipunzip->pctx), pSevenZipExe, NULL));
  DO_DEBUG(hr, cmdez_add_option(punsevenzipunzip->pctx, "x -y ", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_option(punsevenzipunzip->pctx, "-so", NULL, FALSE));
  DO_DEBUG(hr, cmdez_add_operand(punsevenzipunzip->pctx, pFileNameIn, TRUE));
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(punsevenzipunzip->pctx, HANDLE_TYPE_STDERR, hErr));
  }
  DO_DEBUG(hr, cmdez_set_handle(punsevenzipunzip->pctx, HANDLE_TYPE_STDOUT, hOut));

  /* 7-Zip s */
  DO_DEBUG(hr, cmdez_execute(punsevenzipunzip->pctx));

finish_func:
  if (FAILED(hr)) {
    gpgez_cmd_release(&punsevenzipunzip);
    punsevenzipunzip = NULL;
  }
  *ppgpgez_unsevenzipunzip = punsevenzipunzip;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_nkf_create(PCSTR pNkfExe, PCSTR pCodeSetIn, PCSTR pCodeSetOut,
                 HANDLE hIn, HANDLE hOut, HANDLE hErr, PGPGEZ_CMD *ppgpgez_nkf)
{
  HRESULT hr = S_OK;
  PGPGEZ_CMD pnkf = NULL;
  
  pnkf = (PGPGEZ_CMD)util_malloc(sizeof(GPGEZ_CMD));
  if (!pnkf) {
    hr =HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pnkf->pctx = NULL;

  /* nkf s */
  DO_DEBUG(hr, cmdez_initialize(&(pnkf->pctx), pNkfExe, NULL));
  DO_DEBUG(hr, cmdez_add_option(pnkf->pctx, "--ic", pCodeSetIn, FALSE));
  DO_DEBUG(hr, cmdez_add_option(pnkf->pctx, "--oc", pCodeSetOut, FALSE));
  DO_DEBUG(hr, cmdez_add_option(pnkf->pctx, "-Lw", NULL, FALSE));
  if (hIn != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pnkf->pctx, HANDLE_TYPE_STDIN, hIn));
  }
  if (hOut != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pnkf->pctx, HANDLE_TYPE_STDOUT, hOut));
  }
  if (hErr != INVALID_HANDLE_VALUE) {
    DO_DEBUG(hr, cmdez_set_handle(pnkf->pctx, HANDLE_TYPE_STDERR, hErr));
  }

  /* nkf s */
  DO_DEBUG(hr, cmdez_execute(pnkf->pctx));

finish_func:
  if (FAILED(hr)) {
    gpgez_cmd_release(&pnkf);
    pnkf = NULL;
  }
  *ppgpgez_nkf = pnkf;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_cmd_release(PGPGEZ_CMD *ppgpgez_cmd)
{
  PGPGEZ_CMD pcmd = *ppgpgez_cmd;
  HRESULT hr = S_OK;
  
  if (pcmd) {
    if (pcmd->pctx) {
      cmdez_release(&(pcmd->pctx));
    }
    util_free(pcmd);
    pcmd = NULL;
  }
  *ppgpgez_cmd = NULL;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HANDLE
gpgez_cmd_get_process_handle(PGPGEZ_CMD pgpgez_cmd)
{
  return cmdez_get_process(pgpgez_cmd->pctx);
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgez_cmd_finish(PGPGEZ_CMD pgpgez_cmd)
{
  HRESULT hr = S_OK;

  DO_DEBUG(hr, cmdez_finish(pgpgez_cmd->pctx));
finish_func: 
  return hr;
}

/*------------------------------------------------------------*/
DllExport INT
gpgez_cmd_get_exitcode(PGPGEZ_CMD pgpgez_cmd)
{
  return cmdez_get_exitcode(pgpgez_cmd->pctx);
}

/*------------------------------------------------------------*/
