#define THISFILENAME "gpgez/gpgez_status.c"

#include <windows.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "gpgez_util.h"

#define NUMBUFF_LEN 10
#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

CHAR gpgstatus_log[MAX_PATH + 1] = "";

/*------------------------------------------------------------*/
typedef struct struct_param_gpgstatus {
  HANDLE hStatusR;
  HANDLE hStatusW;
  PSTRINGLIST plstatus;
} GPGEZ_PARAM_GPGSTATUS;

/*------------------------------------------------------------*/
DllExport VOID
gpgez_setstatuslog(PCSTR pFileName) {
  util_strcpy(gpgstatus_log, MAX_PATH, pFileName);
}

/*------------------------------------------------------------*/
HRESULT
gpgez_gpgstatus_create(PGPGEZ_PARAM_GPGSTATUS *ppgpgez_param_gpgstatus)
{
  HRESULT hr = S_OK;
  PGPGEZ_PARAM_GPGSTATUS pgpg = NULL;

  pgpg = util_malloc(sizeof(GPGEZ_PARAM_GPGSTATUS));
  if (!pgpg) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pgpg->hStatusR = INVALID_HANDLE_VALUE;
  pgpg->hStatusW = INVALID_HANDLE_VALUE;

  if (FAILED(hr = util_create_read_pipe(&(pgpg->hStatusR), &(pgpg->hStatusW)))) goto finish_func;
  if (FAILED(hr = stringlist_create(&(pgpg->plstatus)))) goto finish_func;

finish_func:
  if (FAILED(hr)) {
    gpgez_gpgstatus_release(&pgpg);
    pgpg = NULL;
  }
  *ppgpgez_param_gpgstatus = pgpg;
  return hr;
}

/*------------------------------------------------------------*/
HRESULT
gpgez_gpgstatus_release(PGPGEZ_PARAM_GPGSTATUS *ppgpgez_param_gpgstatus)
{
  if (*ppgpgez_param_gpgstatus) {
    if ((*ppgpgez_param_gpgstatus)->hStatusR != INVALID_HANDLE_VALUE) {
      CloseHandle((*ppgpgez_param_gpgstatus)->hStatusR);
    }
    if ((*ppgpgez_param_gpgstatus)->hStatusW != INVALID_HANDLE_VALUE) {
      CloseHandle((*ppgpgez_param_gpgstatus)->hStatusW);
    }
    if ((*ppgpgez_param_gpgstatus)->plstatus) {
      stringlist_release(&((*ppgpgez_param_gpgstatus)->plstatus));
    }
    util_free(*ppgpgez_param_gpgstatus);
  }
  return S_OK;
}

/*------------------------------------------------------------*/
HANDLE
gpgez_gpgstatus_get_handle_write(PGPGEZ_PARAM_GPGSTATUS pgpg)
{
  return pgpg->hStatusW;
}

/*------------------------------------------------------------*/
HRESULT
gpgez_gpgstatus_close_handle_write(PGPGEZ_PARAM_GPGSTATUS pgpg)
{
  CloseHandle(pgpg->hStatusW);
  pgpg->hStatusW = INVALID_HANDLE_VALUE;
  return S_OK;
}

/*------------------------------------------------------------*/
PSTRINGLIST
gpgez_gpgstatus_get_status(PGPGEZ_PARAM_GPGSTATUS pgpg)
{
  return pgpg->plstatus;
}

/*------------------------------------------------------------*/
DWORD WINAPI
procObserveGpgStatus(LPVOID pparam)
{
  HRESULT hr = S_OK;
  PGPGEZ_PARAM_GPGSTATUS pgpg = (PGPGEZ_PARAM_GPGSTATUS)pparam;
  PREADFILEBUFF pfile = NULL;
  CHAR buff[MAX_LINE + 1];
  HANDLE hFileLog = INVALID_HANDLE_VALUE;

  if (FAILED(hr = CreateReadFileBuff(&pfile, pgpg->hStatusR))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (gpgstatus_log[0]) {
    hFileLog = CreateFile(gpgstatus_log, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  }
  while (!fEofFileLine(pfile)) {
    if (FAILED(hr = ReadFileLine(pfile, buff, MAX_LINE))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = stringlist_add_node(pgpg->plstatus, buff))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (hFileLog != INVALID_HANDLE_VALUE) {
      DWORD dwWritten;
      WriteFile(hFileLog, buff, util_strlen(buff, -1), &dwWritten, NULL);
      WriteFile(hFileLog, "\n", 1, &dwWritten, NULL);
    }
  }
  if (FAILED(hr = DetatchReadFileLine(pfile))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
finish_func:
  if (hFileLog != INVALID_HANDLE_VALUE) {
    CloseHandle(hFileLog);
  }
  CloseHandle(pgpg->hStatusR);
  pgpg->hStatusR = INVALID_HANDLE_VALUE;
  if (pfile) {
    ReleaseReadFileBuff(&pfile);
  }
  if (FAILED(hr)) {
    return HRESULT_CODE(hr);
  } else {
    return ERROR_SUCCESS;
  }
}

/*------------------------------------------------------------*/

