/**
 * changeUserid.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/changeUserid.c"

#include <windows.h>
#include <windowsx.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

typedef struct struct_param_chuid {
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pName;
  PCSTR pEmail;
  PCSTR pComment;
  PCSTR pPassphrase;
  PCSTR pUid;
  PCSTR pFpr;
} PARAM_CHUID, *PPARAM_CHUID;

/*------------------------------------------------------------*/
static HRESULT
chuid_dwChangeUserId(HWND hwnd)
{
  HRESULT hr = S_OK;
  PPARAM_CHUID pchuid;
  CHAR szName[MAX_PATH + 1];
  WCHAR szNameUtf16[MAX_PATH + 1];
  CHAR szNameUtf8[MAX_PATH * 6 + 1];
  CHAR szEmail[MAX_PATH + 1];
  CHAR szComment[MAX_PATH + 1];
  WCHAR szCommentUtf16[MAX_PATH + 1];
  CHAR szCommentUtf8[MAX_PATH * 6 + 1];
  HCURSOR hOldCursor;

  pchuid = (PPARAM_CHUID)GetWindowLong(hwnd, GWL_USERDATA);

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_NAME), szName, MAX_PATH);

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_EMAIL), szEmail, MAX_PATH);

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_COMMENT), szComment, MAX_PATH);

  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_NAME), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_EMAIL), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_COMMENT), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), "");

  hOldCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));

  /* szName ̏`FbN */
  if (ValidateNameForGpg(szName) != 0) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSERNAME, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Invalid User Name");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    /* Validation error ɂ */
    hr = S_OK;
    goto finish_func;
  }

  /* szEmail ̏`FbN */
  if (ValidateEmailForGpg(szEmail) != 0) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSEREMAIL, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Invalid User Email Address Format");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    /* Validation error ɂ */
    hr = S_OK;
    goto finish_func;
  }

  /* szComment ̏`FbN */
  if (ValidateCommentForGpg(szComment) != 0) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSERCOMMENT, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Invalid User Comment");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    /* Validation error ɂ */
    hr = S_OK;
    goto finish_func;
  }

  if (!MultiByteToWideChar(CP_ACP, 0, szName, -1, szNameUtf16, MAX_PATH)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WideCharToMultiByte(CP_UTF8, 0, szNameUtf16, -1, szNameUtf8, MAX_PATH * 6, NULL, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  pchuid->pName = szNameUtf8;
  pchuid->pEmail = szEmail;
  if (!MultiByteToWideChar(CP_ACP, 0, szComment, -1, szCommentUtf16, MAX_PATH)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WideCharToMultiByte(CP_UTF8, 0, szCommentUtf16, -1, szCommentUtf8, MAX_PATH * 6, NULL, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  pchuid->pComment = szCommentUtf8;
  hr = gpgez_changeuserid(pchuid->pGpgExe, pchuid->pGpgHome, NULL, NULL, pchuid->pFpr, szNameUtf8, szEmail, szCommentUtf8, pchuid->pPassphrase);
  /* gpgez_changeuserid()ŃG[\͊Ă̂ŁAł̓G[\͂Ȃ */

  SetCursor(hOldCursor);

finish_func:
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_NAME), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_EMAIL), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_COMMENT), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);
  return hr;
}

/*------------------------------------------------------------*/
static BOOL
genkey_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
  PPARAM_CHUID pchuid = (PPARAM_CHUID)lParam;

  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_NAME), MAX_PATH);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_EMAIL), MAX_PATH);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_COMMENT), MAX_PATH);
  Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), "");
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), FALSE);
  if (pchuid->pUid) {
    CHAR buff[MAX_PATH + 1] = "";

    if (SUCCEEDED(GetNameFromGpgId(pchuid->pUid, buff, MAX_PATH))) {
      if (buff[0]) {
        Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_NAME), buff);
      }
    }
    buff[0] = '\0';
    if (SUCCEEDED(GetEmailFromGpgId(pchuid->pUid, buff, MAX_PATH))) {
      if (buff[0]) {
        Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_EMAIL), buff);
      }
    }
    buff[0] = '\0';
    if (SUCCEEDED(GetCommentFromGpgId(pchuid->pUid, buff, MAX_PATH))) {
      if (buff[0]) {
        Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_COMMENT), buff);
      }
    }
  }
  SetWindowLong(hwnd, GWL_USERDATA, (LPARAM)pchuid);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
genkey_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
  DWORD dwErr;

  switch(id) {
  case IDOK:
    dwErr = chuid_dwChangeUserId(hwnd);
    if (dwErr != ERROR_SUCCESS) {
    } else {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDCANCEL:
    EndDialog(hwnd, (INT)ERROR_CANCELLED);
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgUserId(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, genkey_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, genkey_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_changeUserId(PGPGEZBASE pgpgezbase, HWND hwnd,
                    PCSTR pUid, PCSTR pFpr, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing)
{
  HRESULT hr = S_OK;
  PARAM_CHUID chuid;
  DWORD dwErr = ERROR_SUCCESS;

  chuid.pGpgExe = pgpgezbase->pGpgExe;
  chuid.pGpgHome = pgpgezbase->pConfDir;
  chuid.pslSecRing = pslSecRing;
  chuid.pslPubRing = pslPubRing;
  chuid.pUid = pUid;
  chuid.pFpr = pFpr;
  if (pgpgezbase->pPassphrase && (*(pgpgezbase->pPassphrase))) {
    chuid.pPassphrase = pgpgezbase->pPassphrase;
  } else {
    chuid.pPassphrase = NULL;
  }
  dwErr = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_USERID), hwnd, procDlgUserId, (LPARAM)&chuid);
  hr = HRESULT_FROM_WIN32(dwErr);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_changeGpgPassphrase(PGPGEZBASE pgpgezbase, HWND hwnd,
                           PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pSelfFpr)
{
  HRESULT hr = S_OK;
  CHAR m_buff[MAX_LINE + 1] = "";
  CHAR buff[MAX_LINE + 1] = "";

  if (LoadString(g_hInstance, IDS_MESSAGE_CHANGEPASSPHRASE, m_buff, MAX_LINE) == 0) {
    util_strcpy(m_buff, MAX_LINE, "Input New Passphrase");
  }

  if (FAILED(hr = gpgezw_InputBoxEx(hwnd, m_buff, "New Passphrase(more than 8 chars)", buff, 8, 260, INPUTBOX_WITHCANCEL | INPUTBOX_SECRET | INPUTBOX_CONFIRM))) {
    goto finish_func;
  }
  if (!(pgpgezbase->pPassphrase)) {
    if (!*buff) {
      goto finish_func;
    }
  } else {
    if (!util_strcmp(buff, pgpgezbase->pPassphrase)) {
      goto finish_func;
    }
  }
  if (FAILED(hr = gpgez_changepassphrase(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing, pSelfFpr, pgpgezbase->pPassphrase, buff))) {
    goto finish_func;
  }
  util_free(pgpgezbase->pPassphrase);
  if (buff[0]) {
    pgpgezbase->pPassphrase = util_strdup(buff);
    if (!(pgpgezbase->pPassphrase)) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      goto finish_func;
    }
  } else {
    pgpgezbase->pPassphrase = NULL;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
