/**
 * editKeyRing.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/editKeyRing.c"

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <crtdbg.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

typedef struct editkey_params {
  PGPGEZBASE pgpgezbase;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  BOOL fNoDefault;
  PENCFILETYPE pet;
} EDITKEY_PARAMS, *PEDITKEY_PARAMS;

/*------------------------------------------------------------*/
static DWORD
editkeyring_show(HWND hwnd)
{
  HRESULT hr = S_OK;
  PSTRINGLIST pslkey = NULL;
  PSTRINGLIST_NODE pnode;
  PEDITKEY_PARAMS pkey = NULL;
  HWND hwndList = NULL, hwndFpr = NULL, hwndUid = NULL;
  LV_ITEM item;
  INT iLine = 0;
  BOOL fUid = FALSE;

  pkey = (PEDITKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);

  if (!(hwndList = GetDlgItem(hwnd, IDC_LIST_KEYS))) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!ListView_DeleteAllItems(hwndList)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  if (FAILED(hr = stringlist_create(&pslkey))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (FAILED(hr = gpgez_listkeys(pkey->pgpgezbase->pGpgExe, pkey->pgpgezbase->pConfDir, pkey->pslSecRing, pkey->pslPubRing, FALSE, FALSE, pslkey))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  for (pnode = stringlist_get_head(pslkey); pnode; pnode = stringlist_get_next(pnode)) {
    /* ListView ɒǉ */
    CHAR buff[MAX_LINE + 1] = " ";
    PSTR p = NULL;

    if (FAILED(hr = util_strcpy(buff, MAX_LINE, stringlist_get_string(pnode)))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    p = buff + (util_strchr(buff, ':') - buff);
    if (p) *p = '\0';
    item.mask = LVIF_TEXT;
    if (p) {
      item.pszText = p + 1;
    } else {
      item.pszText = "";
    }
    item.iItem = iLine;
    item.iSubItem = 0;
    ListView_InsertItem(hwndList, &item);

    {
      CHAR buff2[MAX_LINE + 3] = "";

      if (FAILED(hr = decorate_fpr(buff2, MAX_LINE + 3, buff, gpgezw_getMyFpr(pkey->pgpgezbase)))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }

      item.pszText = buff2;
      item.iItem = iLine;
      item.iSubItem = 1;
      ListView_SetItem(hwndList, &item);
    }

    ++iLine;
  }

finish_func:
  if (pslkey) stringlist_release(&pslkey);
  return hr;
}

/*------------------------------------------------------------*/
static UINT CALLBACK
OFNHookProcSaveFile(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch(msg){
    case WM_NOTIFY:
      {
        LPOFNOTIFY pofn = (LPOFNOTIFY)lParam;
        if (pofn->hdr.code == CDN_TYPECHANGE) {
          if (pofn->lpOFN->nFilterIndex == 1) {
            SendMessage(hdlg, CDM_SETDEFEXT, 0, (LPARAM)"gpg");
          } else if (pofn->lpOFN->nFilterIndex == 2) {
            SendMessage(hdlg, CDM_SETDEFEXT, 0, (LPARAM)"txt");
          } else {
            SendMessage(hdlg, CDM_SETDEFEXT, 0, (LPARAM)"gpg");
          }
          return TRUE;
        }
      }
     return FALSE;
  }
  return FALSE;
}

/*------------------------------------------------------------*/
static HRESULT
editkeyring_import(HWND hwnd)
{
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR szFileName[MAX_PATH + 1] = "";
  PEDITKEY_PARAMS pkey = NULL;

  pkey = (PEDITKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
  _ASSERT(pkey);

  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = "gpg(*.gpg)\0*.gpg\0Ascii Armored(*.txt)\0*.txt\0\0";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.lpfnHook = OFNHookProcSaveFile;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_ENABLEHOOK;
  ofn.lpstrDefExt = "gpg";
  if (GetOpenFileName(&ofn)) {
    if (FAILED(hr = gpgezw_import(pkey->pgpgezbase, hwnd, pkey->pslSecRing, pkey->pslPubRing, pkey->fNoDefault, szFileName))) {
      DEBUGERROR(0, hr);
    }
    editkeyring_show(hwnd);
  } else {
    /* G[`FbN */
  }
  return hr;
}

/*------------------------------------------------------------*/
static HRESULT
editkeyring_export(HWND hwnd)
{
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  LVITEM item, item1;
  HWND hwndKeyList;
  INT iItem = -1;
  CHAR szFileName[MAX_PATH + 1] = "";
  CHAR szLine[MAX_LINE + 1] = "";
  CHAR szLine1[MAX_LINE + 1] = "";
  PEDITKEY_PARAMS pkey = NULL;
  PSTR pfilter = NULL;
  PENCFILETYPE penc1;

  pkey = (PEDITKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
  _ASSERT(pkey);

  penc1 = pkey->pet;

  item.mask = LVIF_TEXT;
  item.pszText = szLine;
  item.cchTextMax = MAX_LINE;
  item1.mask = LVIF_TEXT;
  item1.pszText = szLine1;
  item1.cchTextMax = MAX_LINE;
  hwndKeyList = GetDlgItem(hwnd, IDC_LIST_KEYS);
  iItem = ListView_GetNextItem(hwndKeyList, -1, LVNI_ALL | LVNI_SELECTED);
  if (iItem == -1) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_NOKEYSELECTED, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "No Key Selected");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK | MB_ICONEXCLAMATION);
    goto finish_func;
  }
  if (ListView_GetNextItem(hwndKeyList, iItem, LVNI_ALL | LVNI_SELECTED) != -1) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_MORETHANONEKEYSELECTEDFOREXPORT, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Please select one key to export");
    }
    MessageBox(hwnd, m_buff, "ERROR", MB_OK | MB_ICONEXCLAMATION);
    goto finish_func;
  }
  item.iItem = iItem;
  item.iSubItem = 0;
  ListView_GetItem(hwndKeyList, &item);
  item1.iItem = iItem;
  item1.iSubItem = 1;
  ListView_GetItem(hwndKeyList, &item1);

  if (FAILED(hr = GetEmailFromGpgId(item.pszText, szFileName, MAX_PATH))) goto finish_func;
/*
  szFileName[0] = '\0';
  if (util_strlen(item.pszText, -1) < MAX_PATH - 5) {
    if (FAILED(hr = util_strcpy(szFileName, MAX_PATH, item.pszText))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }
*/
  if (FAILED(hr = create_encrypt_filefilter(pkey->pet, &pfilter))) goto finish_func;

  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = pfilter;
  ofn.lpstrDefExt = "gpg";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.lCustData = (DWORD)(pkey->pet);
  ofn.lpfnHook = OFNHookProcSaveFile;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_ENABLEHOOK;
  if (GetSaveFileName(&ofn)) {
    CHAR buff2[MAX_LINE + 3] = "";

    {
      UINT i;

      for (i = 1; i < ofn.nFilterIndex; ++i) {
        ++penc1;
      }
    }

    if (FAILED(hr = undecorate_fpr(buff2, MAX_LINE + 3, item1.pszText))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = gpgezw_export(pkey->pgpgezbase, hwnd, pkey->pslSecRing, pkey->pslPubRing, buff2, ofn.lpstrFile, penc1->fascii))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  } else {
    /* G[`FbN */
  }
finish_func:
  if (pfilter) util_free(pfilter);
  return hr;
}

/*------------------------------------------------------------*/
static HRESULT
editkeyring_delete(HWND hwnd)
{
  HRESULT hr = S_OK;
  LVITEM item;
  HWND hwndKeyList;
  INT iItem = -1;
  PEDITKEY_PARAMS pkey = NULL;
  CHAR szLine[MAX_LINE + 1];

  pkey = (PEDITKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
  _ASSERT(pkey);

  /* IԂ̌𒲂ׂ */
  item.mask = LVIF_TEXT;
  item.pszText = szLine;
  item.cchTextMax = MAX_LINE;
  hwndKeyList = GetDlgItem(hwnd, IDC_LIST_KEYS);
  iItem = ListView_GetNextItem(hwndKeyList, -1, LVNI_ALL | LVNI_SELECTED);
  if (iItem == -1) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_NOKEYSELECTEDFORDELETE, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Please select a key to delete");
    }
    MessageBox(hwnd, m_buff, "ERROR", MB_OK);
    goto finish_func;
  }
  if (ListView_GetNextItem(hwndKeyList, iItem, LVNI_ALL | LVNI_SELECTED) != -1) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_MORETHANONEKEYSELECTEDFORDELETE, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Please select one key to delete");
    }
    MessageBox(hwnd, m_buff, "ERROR", MB_OK);
    goto finish_func;
  }
  item.iItem = iItem;
  item.iSubItem = 0;
  item.iSubItem = 1;
  ListView_GetItem(hwndKeyList, &item);
  {
    CHAR buff2[MAX_LINE + 3] = "";

    if (FAILED(hr = undecorate_fpr(buff2, MAX_LINE + 3, item.pszText))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = gpgezw_delete(pkey->pgpgezbase, hwnd, pkey->pslSecRing, pkey->pslPubRing, pkey->fNoDefault, buff2))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  editkeyring_show(hwnd);

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
static BOOL
editkeyring_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
  LV_COLUMN lvcol;

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)lParam);

  lvcol.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
  lvcol.fmt = LVCFMT_LEFT;
  lvcol.cx = 200;
  lvcol.pszText = "Name";
  lvcol.iSubItem = 0;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 0, &lvcol);

  lvcol.cx = 352;
  lvcol.pszText = "Fingerprint";
  lvcol.iSubItem = 1;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 1, &lvcol);

  editkeyring_show(hwnd);

  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
editkeyring_Cls_OnShowWindow(HWND hwnd, BOOL fShow, UINT status)
{
/*
  if (fShow) {
    editkeyring_show(hwnd);
  }
  */
  return;
}

/*------------------------------------------------------------*/
static VOID
editkeyring_Cls_OnClose(HWND hwnd)
{
  EndDialog(hwnd, (INT)ERROR_SUCCESS); /* ERROR_CANCELLEDɂKv͂Ȃ */
  return;
}

/*------------------------------------------------------------*/
static VOID
editkeyring_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
  switch(id) {
  case IDOK:
    {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDC_BUTTON_IMPORTKEY:
    editkeyring_import(hwnd);
    break;
  case IDC_BUTTON_EXPORTKEY:
    editkeyring_export(hwnd);
    break;
  case IDC_BUTTON_DELETEKEY:
    editkeyring_delete(hwnd);
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgEditKey(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, editkeyring_Cls_OnInitDialog);
/*    HANDLE_DLG_MSG(hDlg, WM_SHOWWINDOW, editkeyring_Cls_OnShowWindow); */
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, editkeyring_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, editkeyring_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_editGpgKey(PGPGEZBASE pgpgezbase, HWND hwnd,
                  PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                  BOOL fNoDefault, PENCFILETYPE pet)
{
  HRESULT hr = S_OK;
  EDITKEY_PARAMS keyparams;
  DWORD dwCode = ERROR_SUCCESS;

  keyparams.pgpgezbase = pgpgezbase;
  keyparams.pslSecRing = pslSecRing;
  keyparams.pslPubRing = pslPubRing;
  keyparams.fNoDefault = fNoDefault;
  keyparams.pet = pet;
  dwCode = (DWORD)DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_EDIT_KEYRING), hwnd, procDlgEditKey, (LPARAM)&keyparams);
  hr = HRESULT_FROM_WIN32(dwCode);
  return hr;
}

/*------------------------------------------------------------*/
