/**
 * editMyKeyRing.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/editMyKeyRing.c"

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <crtdbg.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

typedef struct editmykey_params {
  PGPGEZBASE pgpgezbase;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  BOOL fNoDefault;
  PSTR pMyUid;
  BOOL fMyUidChanged;
  PENCFILETYPE pet;
} EDITMYKEY_PARAMS, *PEDITMYKEY_PARAMS;

/*------------------------------------------------------------*/
static DWORD
editmykeyring_show(HWND hwnd)
{
  HRESULT hr = S_OK;
  PSTRINGLIST pslkey = NULL;
  PSTRINGLIST_NODE pnode;
  PEDITMYKEY_PARAMS pkey = NULL;
  HWND hwndFpr = NULL, hwndUid = NULL;
  DWORD dwLen = 0;

  pkey = (PEDITMYKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);

  if (!(hwndFpr = GetDlgItem(hwnd, IDC_EDIT_EDITFPR))) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!(hwndUid = GetDlgItem(hwnd, IDC_EDIT_EDITUID))) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  /* FingerPrint̕\ */
  {
    CHAR buff[MAX_LINE + 1] = "";
    PCSTR p1 = gpgezw_getMyFpr(pkey->pgpgezbase);
    PSTR p2 = buff;
    UINT iCnt = 0;

    while (*p1) {
      if (iCnt == 0) {
/*
      } else if (iCnt == 20) {
        *p2 = '\r';
        ++p2;
        *p2 = '\n';
        ++p2;
*/
      } else if (iCnt % 4 == 0) {
        *p2 = ' ';
        ++p2;
      }
      *p2 = *p1;
      ++p1; ++p2; ++iCnt;
    }
    *p2 = '\0';
    SendMessage(hwndFpr, WM_SETTEXT, 0, (LPARAM)buff);
  }

  dwLen = util_strlen(gpgezw_getMyFpr(pkey->pgpgezbase), -1);
  if (FAILED(hr = stringlist_create(&pslkey))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (FAILED(hr = gpgez_listkeys(pkey->pgpgezbase->pGpgExe, pkey->pgpgezbase->pConfDir, pkey->pslSecRing, pkey->pslPubRing, FALSE, FALSE, pslkey))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  for (pnode = stringlist_get_head(pslkey); pnode; pnode = stringlist_get_next(pnode)) {
    PCSTR p = stringlist_get_string(pnode);
    if ((!util_strncmp(p, gpgezw_getMyFpr(pkey->pgpgezbase), dwLen)) && (*(p + dwLen) == ':')) {
      if (pkey->pMyUid) util_free(pkey->pMyUid);
      if (!(pkey->pMyUid = util_strdup(p + dwLen + 1))) {
        hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
      break;
    }
  }
  stringlist_release(&pslkey);
  /* Uid ̕\ */
  SendMessage(hwndUid, WM_SETTEXT, 0, (LPARAM)(pkey->pMyUid));

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
static UINT CALLBACK
OFNHookProcSaveFile(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch(msg){
    case WM_NOTIFY:
      {
        LPOFNOTIFY pofn = (LPOFNOTIFY)lParam;
        if (pofn->hdr.code == CDN_TYPECHANGE) {
          if (pofn->lpOFN->nFilterIndex == 1) {
            SendMessage(hdlg, CDM_SETDEFEXT, 0, (LPARAM)"gpg");
          } else if (pofn->lpOFN->nFilterIndex == 2) {
            SendMessage(hdlg, CDM_SETDEFEXT, 0, (LPARAM)"txt");
          } else {
            SendMessage(hdlg, CDM_SETDEFEXT, 0, (LPARAM)"gpg");
          }
          return TRUE;
        }
      }
     return FALSE;
  }
  return FALSE;
}

/*------------------------------------------------------------*/
static HRESULT
editmykeyring_export(HWND hwnd)
{
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  INT iItem = -1;
  CHAR szFileName[MAX_PATH + 1] = "";
  CHAR szLine[MAX_LINE + 1] = "";
  CHAR szLine1[MAX_LINE + 1] = "";
  PEDITMYKEY_PARAMS pkey = NULL;
  PSTR pfilter = NULL;
  PENCFILETYPE penc1;

  pkey = (PEDITMYKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
  _ASSERT(pkey);

  penc1 = pkey->pet;

  GetEmailFromGpgId(pkey->pMyUid, szFileName, MAX_PATH);
/*
  if (FAILED(util_strcpy(szFileName, MAX_PATH, pkey->pMyUid))) {
    szFileName[0] = '\0';
  }
*/
  if (FAILED(hr = create_encrypt_filefilter(pkey->pet, &pfilter))) goto finish_func;

  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = pfilter;
  ofn.lpstrDefExt = "gpg";
  ofn.lpstrFile = szFileName;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.lCustData = (DWORD)(pkey->pet);
  ofn.lpfnHook = OFNHookProcSaveFile;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_ENABLEHOOK;
  ofn.lpstrDefExt = "gpg";
  if (GetSaveFileName(&ofn)) {
    {
      UINT i;

      for (i = 1; i < ofn.nFilterIndex; ++i) {
        ++penc1;
      }
    }
    if (FAILED(hr = gpgezw_export(pkey->pgpgezbase, hwnd, pkey->pslSecRing, pkey->pslPubRing, gpgezw_getMyFpr(pkey->pgpgezbase), ofn.lpstrFile, penc1->fascii))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  } else {
    /* G[`FbN */
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
static BOOL
editmykeyring_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)lParam);

  editmykeyring_show(hwnd);

  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
editmykeyring_Cls_OnShowWindow(HWND hwnd, BOOL fShow, UINT status)
{
/*
  if (fShow) {
    editmykeyring_show(hwnd);
  }
  */
  return;
}

/*------------------------------------------------------------*/
static VOID
editmykeyring_Cls_OnClose(HWND hwnd)
{
  EndDialog(hwnd, (INT)ERROR_SUCCESS);
  return;
}

/*------------------------------------------------------------*/
static VOID
editmykeyring_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
  switch(id) {
  case IDOK:
    {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDC_BUTTON_EXPORTKEY:
    editmykeyring_export(hwnd);
    break;
  case IDC_BUTTON_EDITUID:
    {
      HRESULT hr = S_OK;
      PEDITMYKEY_PARAMS pkey = (PEDITMYKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
      hr = gpgezw_changeUserId(pkey->pgpgezbase, hwnd, pkey->pMyUid, gpgezw_getMyFpr(pkey->pgpgezbase), NULL, NULL);
      if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
      } else {
        pkey->fMyUidChanged = TRUE;
        editmykeyring_show(hwnd);
      }
    }
    break;
  case IDC_BUTTON_EDITPASSPHRASE:
    {
      PEDITMYKEY_PARAMS pkey = (PEDITMYKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
      gpgezw_changeGpgPassphrase(pkey->pgpgezbase, hwnd, NULL, NULL, gpgezw_getMyFpr(pkey->pgpgezbase));
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgEditMyKey(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, editmykeyring_Cls_OnInitDialog);
/*    HANDLE_DLG_MSG(hDlg, WM_SHOWWINDOW, editmykeyring_Cls_OnShowWindow); */
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, editmykeyring_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, editmykeyring_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_editMyGpgKey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                    BOOL fNoDefault, PENCFILETYPE pet)
{
  HRESULT hr = S_OK;
  EDITMYKEY_PARAMS keyparams;
  DWORD dwCode = ERROR_SUCCESS;

  keyparams.pgpgezbase = pgpgezbase;
  keyparams.pslSecRing = pslSecRing;
  keyparams.pslPubRing = pslPubRing;
  keyparams.fNoDefault = fNoDefault;
  keyparams.pMyUid = NULL;
  keyparams.fMyUidChanged = FALSE;
  keyparams.pet = pet;

  dwCode = (DWORD)DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_EDIT_MYKEY), hwnd, procDlgEditMyKey, (LPARAM)&keyparams);
  hr = HRESULT_FROM_WIN32(dwCode);

  if (keyparams.pMyUid) util_free(keyparams.pMyUid);

  if (SUCCEEDED(hr)) {
    if (keyparams.fMyUidChanged) {
      hr = S_OK;
    } else {
      hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
    }
  }

  return hr;
}

/*------------------------------------------------------------*/
