/**
 * generateKey.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/generateKey.c"

#include <windows.h>
#include <windowsx.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

typedef struct struct_param_genkey {
  PGPGEZBASE pgpgezbase;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  LPCSTR pName;
  LPCSTR pEmail;
  LPCSTR pComment;
  LPCSTR pPassphrase;
  LPSTR pFpr;
  HWND hwnd;
} PARAM_GENKEY, *PPARAM_GENKEY;

/*------------------------------------------------------------*/
static HRESULT
genkey_exportPubKey(HWND hwnd, PCSTR pGpgExe, PCSTR pGpgHome, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                    LPSTR pFpr, LPSTR pName)
{
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR szFileTitle[MAX_PATH + 1] = "";
  CHAR szFileName[MAX_PATH + 1] = "";
  PPARAM_GENKEY pgen = (PPARAM_GENKEY)GetWindowLong(hwnd, GWL_USERDATA);

  if (FAILED(hr = util_strcpy(szFileTitle, MAX_PATH, pName))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (util_strlen(pName, -1) < MAX_PATH - 5) {
    if (FAILED(hr = util_strcat(szFileTitle, MAX_PATH, ".gpg"))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = "gpg(*.gpg)\0*.gpg\0Ascii Armored(*.gpg;*.asc;*.txt)\0*.gpg;*.asc;*.txt\0All files(*.*)\0*.*\0\0";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT;
  ofn.lpstrDefExt = "gpg";
  if (GetSaveFileName(&ofn)) {
    if (FAILED(hr = gpgezw_export(pgen->pgpgezbase, hwnd, pslSecRing, pslPubRing, pFpr, szFileName, ofn.nFilterIndex == 2 ? TRUE : FALSE))) {
      goto finish_func;
    }
  } else {
    /* G[`FbN */
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
static DWORD
genkey_dwGenerateGPGKeys(HWND hwnd)
{
  HRESULT hr = S_OK;
  PPARAM_GENKEY pgen;
  CHAR szName[MAX_PATH + 1];
  WCHAR szNameUtf16[MAX_PATH + 1];
  CHAR szNameUtf8[MAX_PATH * 6 + 1];
  CHAR szEmail[MAX_PATH + 1];
  CHAR szComment[MAX_PATH + 1];
  WCHAR szCommentUtf16[MAX_PATH + 1];
  CHAR szCommentUtf8[MAX_PATH * 6 + 1];
  HCURSOR hOldCursor;

  pgen = (PPARAM_GENKEY)GetWindowLong(hwnd, GWL_USERDATA);

  pgen->pFpr = NULL;

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_NAME), szName, MAX_PATH);

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_EMAIL), szEmail, MAX_PATH);

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_COMMENT), szComment, MAX_PATH);

  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_NAME), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_EMAIL), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_COMMENT), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), "");

  hOldCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));

  /* szName ̏`FbN */
  if (ValidateNameForGpg(szName) != 0) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSERNAME, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Invalid User Name");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    /* Validation error ɂ */
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }

  /* szEmail ̏`FbN */
  if (ValidateEmailForGpg(szEmail) != 0) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSEREMAIL, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Invalid User Email Address Format");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    /* Validation error ɂ */
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }

  /* szComment ̏`FbN */
  if (ValidateCommentForGpg(szComment) != 0) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSERCOMMENT, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Invalid User Comment");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    /* Validation error ɂ */
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }

  if (!MultiByteToWideChar(CP_ACP, 0, szName, -1, szNameUtf16, MAX_PATH)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WideCharToMultiByte(CP_UTF8, 0, szNameUtf16, -1, szNameUtf8, MAX_PATH * 6, NULL, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  pgen->pName = szNameUtf8;
  pgen->pEmail = szEmail;
  if (!MultiByteToWideChar(CP_ACP, 0, szComment, -1, szCommentUtf16, MAX_PATH)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WideCharToMultiByte(CP_UTF8, 0, szCommentUtf16, -1, szCommentUtf8, MAX_PATH * 6, NULL, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  pgen->pComment = szCommentUtf8;
  pgen->pPassphrase = NULL;

  hr = gpgezw_genkey(pgen->pgpgezbase, hwnd, NULL, NULL, pgen->pName, pgen->pEmail, pgen->pComment, pgen->pPassphrase, &(pgen->pFpr));
/*  hr = genkey_dwGenerateGPGKeys2(hwnd, pgen); */
  /* genkey_dwGenerateGPGKeys2()ŃG[\͊Ă̂ŁAł̓G[\͂Ȃ */

  SetCursor(hOldCursor);
#if 0
  if (SUCCEEDED(hr)) {
    if (pgen->pFpr) {
      int iRet;
      CHAR szLine[MAX_LINE + 1], szTitle[MAX_LINE + 1];
      LoadString(g_hInstance, IDS_PROMPT_KEYGENSUCCEEDED, szLine, MAX_LINE);
      LoadString(g_hInstance, IDS_TITLE_KEYGENSUCCEEDED, szTitle, MAX_LINE);
      iRet = MessageBox(pgen->hwnd, szLine, szTitle, MB_YESNO | MB_ICONQUESTION);
      if (iRet == IDYES) {
        if (FAILED(hr = genkey_exportPubKey(hwnd, pgen->pGpgExe, pgen->pGpgHome, pgen->pslSecRing, pgen->pslPubRing, pgen->pFpr, szName))) {
        /* genkey_exportPubKey()ŃG[\͊Ă̂ŁAł̓G[\͂Ȃ */
          goto finish_func;
        }
      }
    }
  }
#endif
finish_func:
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_NAME), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_EMAIL), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_COMMENT), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);
  return HRESULT_CODE(hr);
}

/*------------------------------------------------------------*/
static BOOL
genkey_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
  PPARAM_GENKEY pgen = (PPARAM_GENKEY)lParam;

  pgen->hwnd = hwnd;
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_NAME), MAX_PATH);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_EMAIL), MAX_PATH);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_COMMENT), MAX_PATH);
  Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), "");
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), FALSE);
  SetWindowLong(hwnd, GWL_USERDATA, (LPARAM)pgen);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
genkey_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
  DWORD dwErr;

  switch(id) {
  case IDOK:
    dwErr = genkey_dwGenerateGPGKeys(hwnd);
    if (dwErr != ERROR_SUCCESS) {
    } else {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDCANCEL:
    EndDialog(hwnd, (INT)ERROR_CANCELLED);
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgGenkey(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, genkey_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, genkey_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_generateGpgKey(PGPGEZBASE pgpgezbase, HWND hwnd,
                      PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing)
{
  HRESULT hr = S_OK;
  PARAM_GENKEY gen;
  DWORD dwErr = ERROR_SUCCESS;

  gen.pgpgezbase = pgpgezbase;
  gen.pslSecRing = pslSecRing;
  gen.pslPubRing = pslPubRing;
  gen.pFpr = NULL;
  dwErr = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_USERID), hwnd, procDlgGenkey, (LPARAM)&gen);
  if (dwErr == ERROR_SUCCESS) {
    pgpgezbase->pMyFpr = gen.pFpr;
    pgpgezbase->fPublic = TRUE;
    pgpgezbase->fNeedPassphrase = FALSE;
  }
  hr = HRESULT_FROM_WIN32(dwErr);
  return hr;
}
/*------------------------------------------------------------*/

