/**
 * gpgezw.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/gpgezw.c"

#include <windows.h>
#include <commctrl.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <windowsx.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

enum {
  WM_CREATEPROCESS_GENKEY = WM_USER + 11,
  WM_CREATEPROCESS_ENCRYPT,
  WM_CREATEPROCESS_TARENCRYPT,
  WM_CREATEPROCESS_ZIPENCRYPT,
  WM_CREATEPROCESS_7ZENCRYPT,
  WM_CREATEPROCESS_TAR,
  WM_CREATEPROCESS_ENCRYPTSFX,
  WM_CREATEPROCESS_DECRYPT,
  WM_CREATEPROCESS_DECRYPTTAR,
  WM_CREATEPROCESS_7ZDECRYPT,
  WM_CREATEPROCESS_UNTAR,
  WM_CREATEPROCESS_7ZGUNZIP,
  WM_CREATEPROCESS_IMPORT,
  WM_CREATEPROCESS_EXPORT,
  WM_CREATEPROCESS_DELETE,
  WM_CREATEPROCESS_DETACHSIGN,
  WM_CREATEPROCESS_VERIFY,
  WM_CREATEPROCESS_ARMOR
};

typedef struct struct_gpgezw_genkey {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pName;
  PCSTR pEMail;
  PCSTR pComment;
  PCSTR pPassphrase;
  PSTR *ppFpr;
} GPGEZW_GENKEY, *PGPGEZW_GENKEY;

typedef struct struct_gpgezw_encrypt {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pFile;
  PCSTR pZfile;
  PSTRINGLIST pRecipients;
  PCSTR pPassphraseSym;
  BOOL fSign;
  BOOL fAscii;
} GPGEZW_ENCRYPT, *PGPGEZW_ENCRYPT;

typedef struct struct_gpgezw_tarencrypt {
  HWND hwnd;
  PCSTR pTarExe;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pBaseDirectory;
  PSTRINGLIST pFiles;
  PCSTR pZfile;
  PSTRINGLIST pRecipients;
  PCSTR pPassphraseSym;
  BOOL fSign;
  BOOL fAscii;
} GPGEZW_TARENCRYPT, *PGPGEZW_TARENCRYPT;

typedef struct struct_gpgezw_zipencrypt {
  HWND hwnd;
  PCSTR p7ZipExe;
  PCSTR pBaseDirectory;
  PSTRINGLIST pFiles;
  PCSTR pZfile;
  PCSTR pPassphraseSym;
} GPGEZW_ZIPENCRYPT, *PGPGEZW_ZIPENCRYPT;

typedef struct struct_gpgezw_7zencrypt {
  HWND hwnd;
  PCSTR p7ZipExe;
  PCSTR pBaseDirectory;
  PSTRINGLIST pFiles;
  PCSTR pZfile;
  PCSTR pPassphraseSym;
  DWORD dwMode;
} GPGEZW_7ZENCRYPT, *PGPGEZW_7ZENCRYPT;

typedef struct struct_gpgezw_tar {
  HWND hwnd;
  PCSTR pTarExe;
  PCSTR pBaseDirectory;
  PSTRINGLIST pFiles;
  PCSTR pZfile;
  DWORD dwMode;
} GPGEZW_TAR, *PGPGEZW_TAR;

typedef struct struct_gpgezw_encryptsfx {
  HWND hwnd;
  PCSTR p7ZipExe;
  PCSTR pBaseDirectory;
  PSTRINGLIST pFiles;
  PCSTR pZfile;
  PCSTR pPassphraseSym;
} GPGEZW_ENCRYPTSFX, *PGPGEZW_ENCRYPTSFX;

typedef struct struct_gpgezw_decrypt {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pPassphrase;
  PCSTR pFile;
  PCSTR pXFile;
} GPGEZW_DECRYPT, *PGPGEZW_DECRYPT;

typedef struct struct_gpgezw_decrypttar {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pTarExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pPassphrase;
  PCSTR pFileName;
  PCSTR pBaseDirectory;
  BOOL fOverWrite;
} GPGEZW_DECRYPTTAR, *PGPGEZW_DECRYPTTAR;

typedef struct struct_gpgezw_7zdecrypt {
  HWND hwnd;
  PCSTR p7ZipExe;
  PCSTR pPassphrase;
  PCSTR pFileName;
  PCSTR pBaseDirectory;
  BOOL fOverWrite;
} GPGEZW_7ZDECRYPT, *PGPGEZW_7ZDECRYPT;

typedef struct struct_gpgezw_untar {
  HWND hwnd;
  PCSTR pTarExe;
  PCSTR pFileName;
  PCSTR pBaseDirectory;
  DWORD dwMode;
  BOOL fOverWrite;
} GPGEZW_UNTAR, *PGPGEZW_UNTAR;

typedef struct struct_gpgezw_7zgunzip {
  HWND hwnd;
  PCSTR p7ZipExe;
  PCSTR pFileNameIn;
  PCSTR pFileNameOut;
} GPGEZW_7ZGUNZIP, *PGPGEZW_7ZGUNZIP;

typedef struct struct_gpgezw_import {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  BOOL fNoDefault;
  PCSTR pFileName;
} GPGEZW_IMPORT, *PGPGEZW_IMPORT;

typedef struct struct_gpgezw_export {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pFpr;
  PCSTR pFileName;
  BOOL fAscii;
} GPGEZW_EXPORT, *PGPGEZW_EXPORT;

typedef struct struct_gpgezw_delete {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  BOOL fNoDefault;
  PCSTR pFpr;
} GPGEZW_DELETE, *PGPGEZW_DELETE;

typedef struct struct_gpgezw_detachsign {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pFile;
  PCSTR pSfile;
  PCSTR pPassphraseSym;
  BOOL fAscii;
} GPGEZW_DETACHSIGN, *PGPGEZW_DETACHSIGN;

typedef struct struct_gpgezw_verify {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pFile;
  PCSTR pSfile;
  PCSTR pPassphraseSym;
} GPGEZW_VERIFY, *PGPGEZW_VERIFY;

typedef struct struct_gpgezw_armor {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PCSTR pFileIn;
  PCSTR pFileOut;
  BOOL fEnArmor;
} GPGEZW_ARMOR, *PGPGEZW_ARMOR;

typedef struct struct_gpgezw_launch {
  UINT uMsg;
  PVOID pParam;
  PHANDLE phProcess;
  HRESULT hr;
} GPGEZW_LAUNCH, *PGPGEZW_LAUNCH;

typedef struct struct_gpgezw_gpgmsg {
  PSTRING pmsg;
  PHANDLE phGpgErr;
} GPGEZW_GPGMSG, *PGPGEZW_GPGMSG;

#define DO_DEBUG(hr, fn) if (FAILED((hr) = (fn))) { DEBUGERROR(0, (hr)); goto finish_func;}

/*------------------------------------------------------------*/
DllExport VOID
gpgezw_free(PVOID p) {
  util_free(p);
}
/*------------------------------------------------------------*/
static HRESULT
ShowProcessMessage(HWND hwnd, PCSTR pszMsg) {
  HRESULT hr = S_OK;
  HWND hwndmsg = GetDlgItem(hwnd, IDC_EDIT_MESSAGE);

  if (hwndmsg) {
    INT len = 0;
    len = SendMessage(hwndmsg, WM_GETTEXTLENGTH, 0, 0L);
    SendMessage(hwndmsg, EM_SETSEL, (WPARAM)len, (LPARAM)len);
    SendMessage(hwndmsg, EM_REPLACESEL, (WPARAM)FALSE, (LPARAM)pszMsg);
  }
  return hr;
}

/*------------------------------------------------------------*/
static HRESULT
ShowProcessMessageFromHandle(HWND hwnd, HANDLE hRead) {
  HRESULT hr = S_OK;
  PREADFILEBUFF pfilebuff = NULL;
  CHAR buff[MAX_LINE + 1] = "";

  DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, hRead));
  while (!fEofFileLine(pfilebuff)) {
    DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
    DO_DEBUG(hr, ShowProcessMessage(hwnd, buff));
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "\n"));
  }
  DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));
finish_func:
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  return hr;
}

/*------------------------------------------------------------*/
static HRESULT
ShowProcessMessageFromTarHandle(HWND hwnd, HANDLE hRead) {
  HRESULT hr = S_OK;
  PREADFILEBUFF pfilebuff = NULL;
  CHAR buff[MAX_LINE + 1] = "";

  DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, hRead));
  while (!fEofFileLine(pfilebuff)) {
    DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
    if ((buff[0] == 'a') && (buff[1] == ' ')) {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "adding "));
      unescapeTarMessage(buff + 2);
      DO_DEBUG(hr, ShowProcessMessage(hwnd, buff + 2));
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "\n"));
    } else if ((buff[0] == 'x') && (buff[1] == ' ')) {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "extracting "));
      unescapeTarMessage(buff + 2);
      DO_DEBUG(hr, ShowProcessMessage(hwnd, buff + 2));
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "\n"));
    } else {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, buff));
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "\n"));
    }
  }
  DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));
finish_func:
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  return hr;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
ReadProcessMessageFromGpgHandle(LPVOID lpParam) {
  PGPGEZW_GPGMSG pmsg = (PGPGEZW_GPGMSG)lpParam;
  PREADFILEBUFF pfilebuff = NULL;
  CHAR buff[MAX_LINE + 1] = "";
  HRESULT hr = S_OK;

  DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, *(pmsg->phGpgErr)));
  while (!fEofFileLine(pfilebuff)) {
    DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
    DO_DEBUG(hr, string_append(pmsg->pmsg, buff));
    DO_DEBUG(hr, string_append(pmsg->pmsg, "\n"));
  }
  DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));
finish_func:
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  return hr;
}

/*------------------------------------------------------------*/
static BOOL
gpgezw_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  HWND hwndmsg = GetDlgItem(hwnd, IDC_EDIT_MESSAGE);
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)lParam;

  Edit_Enable(hwndmsg, TRUE);
  Edit_SetText(hwndmsg, "");
  Edit_SetReadOnly(hwndmsg, FALSE);

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);
  SetWindowLong(hwnd, GWL_USERDATA, (LONG)pgl);
  PostMessage(hwnd, pgl->uMsg, 0, 0L);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
gpgezw_Cls_OnClose(HWND hwnd) {
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);

  if ((pgl->phProcess) && (*(pgl->phProcess) != INVALID_HANDLE_VALUE)) {
    TerminateProcess(*(pgl->phProcess), -1);
  }
  pgl->hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
  EndDialog(hwnd, ERROR_CANCELLED);
}

/*------------------------------------------------------------*/
static HRESULT
gpgezw_Cls_OnCreateProcess_Genkey_WriteKeyInfo(PGPGEZ_GENKEY pgpg, PCSTR pName, PCSTR pEmail, PCSTR pComment, PCSTR pPassphrase)
{
  HRESULT hr = S_OK;

  DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Key-Type:", "DSA"));
  DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Key-Length:", "1024"));
  DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Subkey-Type:", "ELG-E"));
  DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Subkey-Length:", "2048"));
  if ((pName) && (*(pName))) {
    DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Name-Real:", pName));
  }
  if ((pEmail) && (*(pEmail))) {
    DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Name-Email:", pEmail));
  }
  if ((pComment) && (*(pComment))) {
    DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Name-Comment:", pComment));
  }
  DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Expire-Date:", "0"));
  if ((pPassphrase) && (*(pPassphrase))) {
    DO_DEBUG(hr, gpgez_genkey_write_param(pgpg, "Passphrase:", pPassphrase));
  }
  DO_DEBUG(hr, gpgez_genkey_write_cmd(pgpg, "%commit"));

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Genkey(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_GENKEY pwgenkey = (PGPGEZW_GENKEY)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GENKEY pgpg = NULL;
  HANDLE hStatusW = INVALID_HANDLE_VALUE, hStatusR = INVALID_HANDLE_VALUE;
  PREADFILEBUFF pfilebuff = NULL;
  CHAR buff[MAX_LINE + 1];

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* p\̍쐬 */
  DO_DEBUG(hr, util_create_read_pipe(&hStatusR, &hStatusW));
  DO_DEBUG(hr, gpgez_genkey_create(pwgenkey->pGpgExe, pwgenkey->pGpgHome, pwgenkey->pslSecRing, pwgenkey->pslPubRing, hStatusW, &pgpg));
  CloseHandleClean(&hStatusW);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_genkey_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* 񏑂 */
  DO_DEBUG(hr, gpgezw_Cls_OnCreateProcess_Genkey_WriteKeyInfo(pgpg, pwgenkey->pName, pwgenkey->pEMail, pwgenkey->pComment, pwgenkey->pPassphrase));
  DO_DEBUG(hr, gpgez_genkey_close_write(pgpg));

  /* gpg̃Xe[^Xo͓ǂݎ */
  DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, hStatusR));
  while (!fEofFileLine(pfilebuff)) {
    INT iStatus = GPGSTATUS_ZZZZZZZZ;

    DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
    iStatus = gpgez_interpret_status(buff);

    switch(iStatus) {
    case GPGSTATUS_PROGRESS:
      {
        CHAR buff2[2] = "";

        DO_DEBUG(hr, sv_get(buff, ' ', 3, buff2, 2));
        buff2[1] = '\0';
        ShowProcessMessage(hwnd, buff2);
      }
      break;
    case GPGSTATUS_KEY_CREATED:
      {
        CHAR buff2[MAX_PATH + 1] = "";
        DWORD dwLen = 0;

        DO_DEBUG(hr, sv_get(buff, ' ', 3, buff2, MAX_PATH));
        *(pwgenkey->ppFpr) = util_strdup(buff2);
        if (!(*(pwgenkey->ppFpr))) {
          hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
          goto finish_func;
        }
        DO_DEBUG(hr, ShowProcessMessage(hwnd, "Key Created.\n"));
        DO_DEBUG(hr, ShowProcessMessage(hwnd, *(pwgenkey->ppFpr)));
        DO_DEBUG(hr, ShowProcessMessage(hwnd, "\n"));
      }
      break;
    case GPGSTATUS_KEY_NOT_CREATED:
      {
        DO_DEBUG(hr, ShowProcessMessage(hwnd, "Key Not Created.\n"));
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        goto finish_func;
      }
      break;
    default:
      /* do nothing */
      break;
    }
  }
  DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  CloseHandleClean(&hStatusR);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_genkey_finish(pgpg));

  /* qvZX̖߂l`FbN͂Ȃ */

finish_func:
  /* I */
  if (hStatusW != INVALID_HANDLE_VALUE) CloseHandle(hStatusW);
  if (hStatusR != INVALID_HANDLE_VALUE) CloseHandle(hStatusR);
  if (pgpg) gpgez_genkey_release(&pgpg);

  if (SUCCEEDED(hr)) {
    EndDialog(hwnd, ERROR_SUCCESS);
  }
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Encrypt(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_ENCRYPT pencrypt = (PGPGEZW_ENCRYPT)(pgl->pParam);
  HRESULT hr = S_OK;
  PCSTR pFileName = NULL;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE;
  HANDLE hErrR = INVALID_HANDLE_VALUE, hErrW = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pencrypt->pFile));
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, pencrypt->pZfile));
  DO_DEBUG(hr, util_create_read_pipe(&hErrR, &hErrW));
  pFileName = PathFindFileName(pencrypt->pFile);
  DO_DEBUG(hr, gpgez_gpg_create_encrypt(pencrypt->pGpgExe, pencrypt->pGpgHome, pencrypt->pslSecRing, pencrypt->pslPubRing,
    FALSE, pencrypt->pRecipients, pencrypt->pPassphraseSym,
    pencrypt->fSign, pencrypt->fAscii, pFileName, hIn, hOut, hErrW, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);
  CloseHandleClean(&hErrW);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromHandle(hwnd, hErrR));
  CloseHandleClean(&hErrR);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Completed\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Failed\n"));
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (hIn != INVALID_HANDLE_VALUE) CloseHandle(hIn);
  if (hOut != INVALID_HANDLE_VALUE) CloseHandle(hOut);
  if (hErrR != INVALID_HANDLE_VALUE) CloseHandle(hErrR);
  if (hErrW != INVALID_HANDLE_VALUE) CloseHandle(hErrW);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_TarEncrypt(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_TARENCRYPT ptarencrypt = (PGPGEZW_TARENCRYPT)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  PGPGEZ_CMD ptar = NULL;
  HANDLE hOut = INVALID_HANDLE_VALUE;
  HANDLE hErrP = INVALID_HANDLE_VALUE, hErrC = INVALID_HANDLE_VALUE;
  HANDLE hTarOutP = INVALID_HANDLE_VALUE, hTarOutC = INVALID_HANDLE_VALUE;
  GPGEZW_GPGMSG gpgmsg = {NULL, NULL};
  PSTRING pgpgmsg = NULL;
  HANDLE hGpgErrP = INVALID_HANDLE_VALUE, hGpgErrC = INVALID_HANDLE_VALUE;
  HANDLE hThread = NULL;
  DWORD dwThreadId;
  CHAR szInterFileName[MAX_PATH + 1] = "";

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* ̃vZXԃnh쐬 */
  DO_DEBUG(hr, util_create_read_pipe(&hTarOutP, &hTarOutC));

  /* Tarp\̍쐬 */
  DO_DEBUG(hr, util_create_read_pipe(&hErrP, &hErrC));
  DO_DEBUG(hr, gpgez_tar_create(ptarencrypt->pTarExe, TAR_COMPRESS_NONE, ptarencrypt->pBaseDirectory,
    ptarencrypt->pFiles, hTarOutC, hErrC, &ptar));
  CloseHandleClean(&hTarOutC);
  CloseHandleClean(&hErrC);

  /* nhtւihTarOutPpł悤ɂj */
  {
    HANDLE h = hTarOutP;

    if (!DuplicateHandle(GetCurrentProcess(), h, GetCurrentProcess(), &hTarOutP, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      goto finish_func;
    }
    CloseHandle(h);
  }

  /* ԃt@C쐬 */
  {
    if (FAILED(hr = util_strcpy(szInterFileName, MAX_PATH, stringlist_get_string(stringlist_get_head(ptarencrypt->pFiles))))) goto finish_func;
    {
      PSTR p = util_pathfindextension(szInterFileName);
      if ((p - szInterFileName) < MAX_PATH - 4) {
        util_strcpy(p, MAX_PATH, ".tar");
      }
    }
  }

  /* gpg̃bZ[W󂯎鏀 */
  DO_DEBUG(hr, util_create_read_pipe(&hGpgErrP, &hGpgErrC));
  DO_DEBUG(hr, string_create(&pgpgmsg));
  gpgmsg.pmsg = pgpgmsg;
  gpgmsg.phGpgErr = &hGpgErrP;

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, ptarencrypt->pZfile));
  DO_DEBUG(hr, gpgez_gpg_create_encrypt(ptarencrypt->pGpgExe, ptarencrypt->pGpgHome, ptarencrypt->pslSecRing, ptarencrypt->pslPubRing,
    FALSE, ptarencrypt->pRecipients, ptarencrypt->pPassphraseSym,
    ptarencrypt->fSign, ptarencrypt->fAscii, szInterFileName, hTarOutP, hOut, hGpgErrC, &pgpg));
  CloseHandleClean(&hTarOutP);
  CloseHandleClean(&hOut);
  CloseHandleClean(&hGpgErrC);
  hThread = CreateThread(NULL, 0, ReadProcessMessageFromGpgHandle, (LPVOID)&gpgmsg, 0, &dwThreadId);
  if (!hThread) {
    DWORD dwErr = GetLastError();
    hr = HRESULT_FROM_WIN32(dwErr);
    goto finish_func;
  }

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_cmd_get_process_handle(ptar);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromTarHandle(hwnd, hErrP));
  CloseHandleClean(&hErrP);

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(ptar));

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* gpgǂݏoXbhI҂ */
  WaitForSingleObject(hThread, INFINITE);
  CloseHandle(hThread);
  hThread = NULL;
  CloseHandleClean(&hGpgErrP);
  DO_DEBUG(hr, ShowProcessMessage(hwnd, string_c_str(pgpgmsg)));
  string_release(&pgpgmsg);

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(ptar) != 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Tar Archive Failed\n"));
  } else {
    if (gpgez_gpg_get_exitcode(pgpg) != 0) {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Failed\n"));
    } else {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Completed\n"));
    }
  }

finish_func:
  /* I */
  if (ptar) gpgez_cmd_release(&ptar);
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (pgpgmsg) string_release(&pgpgmsg);
  if (hOut != INVALID_HANDLE_VALUE) CloseHandle(hOut);
  if (hErrP != INVALID_HANDLE_VALUE) CloseHandle(hErrP);
  if (hErrC != INVALID_HANDLE_VALUE) CloseHandle(hErrC);
  if (hTarOutP != INVALID_HANDLE_VALUE) CloseHandle(hTarOutP);
  if (hTarOutC != INVALID_HANDLE_VALUE) CloseHandle(hTarOutC);
  if (hGpgErrP != INVALID_HANDLE_VALUE) CloseHandle(hGpgErrP);
  if (hGpgErrC != INVALID_HANDLE_VALUE) CloseHandle(hGpgErrC);
  if (hThread) CloseHandle(hThread);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_ZipEncrypt(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_ZIPENCRYPT pzipencrypt = (PGPGEZW_ZIPENCRYPT)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_CMD psevenzip = NULL;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_sevenzip_create(pzipencrypt->p7ZipExe, SEVENZIP_ZIPENC, pzipencrypt->pBaseDirectory,
    pzipencrypt->pFiles, pzipencrypt->pPassphraseSym, pzipencrypt->pZfile, INVALID_HANDLE_VALUE, &psevenzip));

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_cmd_get_process_handle(psevenzip);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* 7zI҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(psevenzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(psevenzip) != 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Failed\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Completed\n"));
  }

finish_func:
  /* I */
  if (psevenzip) gpgez_cmd_release(&psevenzip);

  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_7zEncrypt(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_7ZENCRYPT psevenzipencrypt = (PGPGEZW_7ZENCRYPT)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_CMD psevenzip = NULL;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_sevenzip_create(psevenzipencrypt->p7ZipExe, psevenzipencrypt->dwMode, psevenzipencrypt->pBaseDirectory,
    psevenzipencrypt->pFiles, psevenzipencrypt->pPassphraseSym, psevenzipencrypt->pZfile, INVALID_HANDLE_VALUE, &psevenzip));

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_cmd_get_process_handle(psevenzip);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* 7zI҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(psevenzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(psevenzip) != 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Failed\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Completed\n"));
  }

finish_func:
  /* I */
  if (psevenzip) gpgez_cmd_release(&psevenzip);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Tar(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_TAR p = (PGPGEZW_TAR)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_CMD ptar = NULL;
  HANDLE hOut = INVALID_HANDLE_VALUE;
  HANDLE hErrP = INVALID_HANDLE_VALUE, hErrC = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* Tarp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, p->pZfile));
  DO_DEBUG(hr, util_create_read_pipe(&hErrP, &hErrC));
  DO_DEBUG(hr, gpgez_tar_create(p->pTarExe, p->dwMode, p->pBaseDirectory,
    p->pFiles, hOut, hErrC, &ptar));
  CloseHandleClean(&hOut);
  CloseHandleClean(&hErrC);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_cmd_get_process_handle(ptar);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromTarHandle(hwnd, hErrP));
  CloseHandleClean(&hErrP);

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(ptar));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(ptar) != 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Tar Archive Failed\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Tar Archive Completed\n"));
  }

finish_func:
  /* I */
  if (ptar) gpgez_cmd_release(&ptar);
  if (hOut != INVALID_HANDLE_VALUE) CloseHandle(hOut);
  if (hErrP != INVALID_HANDLE_VALUE) CloseHandle(hErrP);
  if (hErrC != INVALID_HANDLE_VALUE) CloseHandle(hErrC);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_EncryptSfx(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_ENCRYPTSFX p = (PGPGEZW_ENCRYPTSFX)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_CMD psevenzip = NULL;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_sevenzip_create(p->p7ZipExe, SEVENZIP_SFX, p->pBaseDirectory,
    p->pFiles, p->pPassphraseSym, p->pZfile, INVALID_HANDLE_VALUE, &psevenzip));

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_cmd_get_process_handle(psevenzip);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* 7zI҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(psevenzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(psevenzip) != 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Failed\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Encryption Completed\n"));
  }

finish_func:
  /* I */
  if (psevenzip) gpgez_cmd_release(&psevenzip);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Decrypt(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_DECRYPT p = (PGPGEZW_DECRYPT)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;

  HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE;
  HANDLE hErrR = INVALID_HANDLE_VALUE, hErrW = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, p->pFile));
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, p->pXFile));
  DO_DEBUG(hr, util_create_read_pipe(&hErrR, &hErrW));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(p->pGpgExe, p->pGpgHome, p->pslSecRing, p->pslPubRing,
    FALSE, TRUE, p->pPassphrase, hIn, hOut, hErrW, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);
  CloseHandleClean(&hErrW);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromHandle(hwnd, hErrR));
  CloseHandleClean(&hErrR);

  CloseHandleClean(&hErrR);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Decryption Completed\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Decryption Failed\n"));
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (hIn != INVALID_HANDLE_VALUE) CloseHandle(hIn);
  if (hOut != INVALID_HANDLE_VALUE) CloseHandle(hOut);
  if (hErrR != INVALID_HANDLE_VALUE) CloseHandle(hErrR);
  if (hErrW != INVALID_HANDLE_VALUE) CloseHandle(hErrW);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_DecryptTar(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_DECRYPTTAR pdecrypttar = (PGPGEZW_DECRYPTTAR)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  PGPGEZ_CMD puntar = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;
  HANDLE hErrP = INVALID_HANDLE_VALUE, hErrC = INVALID_HANDLE_VALUE;
  HANDLE hGpgOutP = INVALID_HANDLE_VALUE, hGpgOutC = INVALID_HANDLE_VALUE;
  GPGEZW_GPGMSG gpgmsg = {NULL, NULL};
  PSTRING pgpgmsg = NULL;
  HANDLE hGpgErrP = INVALID_HANDLE_VALUE, hGpgErrC = INVALID_HANDLE_VALUE;
  HANDLE hThread = NULL;
  DWORD dwThreadId;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* ̃vZXԃnh쐬 */
  DO_DEBUG(hr, util_create_read_pipe(&hGpgOutP, &hGpgOutC));

  /* gpg̃bZ[W󂯎鏀 */
  DO_DEBUG(hr, util_create_read_pipe(&hGpgErrP, &hGpgErrC));
  DO_DEBUG(hr, string_create(&pgpgmsg));
  gpgmsg.pmsg = pgpgmsg;
  gpgmsg.phGpgErr = &hGpgErrP;

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pdecrypttar->pFileName));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(pdecrypttar->pGpgExe, pdecrypttar->pGpgHome, pdecrypttar->pslSecRing, pdecrypttar->pslPubRing,
    FALSE, TRUE, pdecrypttar->pPassphrase, hIn, hGpgOutC, hGpgErrC, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hGpgOutC);
  CloseHandleClean(&hGpgErrC);
  hThread = CreateThread(NULL, 0, ReadProcessMessageFromGpgHandle, (LPVOID)&gpgmsg, 0, &dwThreadId);
  if (!hThread) {
    DWORD dwErr = GetLastError();
    hr = HRESULT_FROM_WIN32(dwErr);
    goto finish_func;
  }

  /* nhtւihGpgOutPpł悤ɂj */
  {
    HANDLE h = hGpgOutP;
    if (!DuplicateHandle(GetCurrentProcess(), h, GetCurrentProcess(), &hGpgOutP, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      goto finish_func;
    }
    CloseHandle(h);
  }

  /* tarWJp\̍쐬 */
  DO_DEBUG(hr, util_create_read_pipe(&hErrP, &hErrC));
  DO_DEBUG(hr, gpgez_untar_create(pdecrypttar->pTarExe, pdecrypttar->pBaseDirectory, pdecrypttar->fOverWrite,
    hGpgOutP, hErrC, &puntar));
  CloseHandleClean(&hGpgOutP);
  CloseHandleClean(&hErrC);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromTarHandle(hwnd, hErrP));
  CloseHandleClean(&hErrP);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(puntar));

  /* gpgǂݏoXbhI҂ */
  WaitForSingleObject(hThread, INFINITE);
  CloseHandle(hThread);
  hThread = NULL;
  CloseHandleClean(&hGpgErrP);
  DO_DEBUG(hr, ShowProcessMessage(hwnd, string_c_str(pgpgmsg)));
  string_release(&pgpgmsg);

  /* qvZX̖߂l`FbN */
  /* NOTICE: bsdtar̖untar͕KFail̂Ń`FbNȂ */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    if (FAILED(hr = ShowProcessMessage(hwnd, "Decryption Failed\n"))) goto finish_func;
  } else {
    if (FAILED(hr = ShowProcessMessage(hwnd, "Decryption Completed\n"))) goto finish_func;
  }

finish_func:
  /* I */
  if (puntar) gpgez_cmd_release(&puntar);
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (pgpgmsg) string_release(&pgpgmsg);
  if (hIn != INVALID_HANDLE_VALUE) CloseHandle(hIn);
  if (hErrP != INVALID_HANDLE_VALUE) CloseHandle(hErrP);
  if (hErrC != INVALID_HANDLE_VALUE) CloseHandle(hErrC);
  if (hGpgOutP != INVALID_HANDLE_VALUE) CloseHandle(hGpgOutP);
  if (hGpgOutC != INVALID_HANDLE_VALUE) CloseHandle(hGpgOutC);
  if (hGpgErrP != INVALID_HANDLE_VALUE) CloseHandle(hGpgErrP);
  if (hGpgErrC != INVALID_HANDLE_VALUE) CloseHandle(hGpgErrC);
  if (hThread) CloseHandle(hThread);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_7zDecrypt(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_7ZDECRYPT p = (PGPGEZW_7ZDECRYPT)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_CMD punsevenzip = NULL;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_unsevenzip_create(p->p7ZipExe, p->pFileName, p->pBaseDirectory,
    p->fOverWrite, p->pPassphrase, INVALID_HANDLE_VALUE, &punsevenzip));

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_cmd_get_process_handle(punsevenzip);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* 7zip I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(punsevenzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(punsevenzip) == 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Decryption Completed\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Decryption Failed\n"));
  }

finish_func:
  /* I */
  if (punsevenzip) gpgez_cmd_release(&punsevenzip);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_UnTar(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_UNTAR p = (PGPGEZW_UNTAR)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_CMD puntar = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;
  HANDLE hErrP = INVALID_HANDLE_VALUE, hErrC = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* tarWJp\̍쐬 */
  if (FAILED(hr = gpgez_open_file_inherit_read(&hIn, p->pFileName))) goto finish_func;
  if (FAILED(hr = util_create_read_pipe(&hErrP, &hErrC))) goto finish_func;
  if (FAILED(hr = gpgez_untar_create(p->pTarExe, p->pBaseDirectory, p->fOverWrite,
    &hIn, hErrC, &puntar))) goto finish_func;
  CloseHandleClean(&hIn);
  CloseHandleClean(&hErrC);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_cmd_get_process_handle(puntar);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromTarHandle(hwnd, hErrP));
  CloseHandleClean(&hErrP);

  /* tar I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(puntar));

  /* NOTICE: bsdtar̖untar͕KFail̂Ń`FbNȂ */

finish_func:
  /* I */
  if (puntar) gpgez_cmd_release(&puntar);
  if (hIn != INVALID_HANDLE_VALUE) CloseHandle(hIn);
  if (hErrP != INVALID_HANDLE_VALUE) CloseHandle(hErrP);
  if (hErrC != INVALID_HANDLE_VALUE) CloseHandle(hErrC);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_7zGunZip(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_7ZGUNZIP p = (PGPGEZW_7ZGUNZIP)(pgl->pParam);
  HRESULT hr = S_OK;
  HANDLE hOut = INVALID_HANDLE_VALUE;
  PGPGEZ_CMD punsevenzipunzip = NULL;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* WJp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, p->pFileNameOut));
  DO_DEBUG(hr, gpgez_unsevenzipunzip_create(p->p7ZipExe, p->pFileNameIn,
    hOut, INVALID_HANDLE_VALUE, &punsevenzipunzip));
  CloseHandleClean(&hOut);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_cmd_get_process_handle(punsevenzipunzip);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* 7zip I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(punsevenzipunzip));

  /* qvZX̖߂l`FbN */
  if (gpgez_cmd_get_exitcode(punsevenzipunzip) == 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Unzip Completed\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Unzip Failed\n"));
  }

finish_func:
  /* I */
  if (punsevenzipunzip) gpgez_cmd_release(&punsevenzipunzip);
  CloseHandleClean(&hOut);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Import(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_IMPORT p = (PGPGEZW_IMPORT)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;
  HANDLE hErrR = INVALID_HANDLE_VALUE, hErrW = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* C|[gp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, p->pFileName));
  DO_DEBUG(hr, util_create_read_pipe(&hErrR, &hErrW));
  DO_DEBUG(hr, gpgez_gpg_create_import(p->pGpgExe, p->pGpgHome, p->pslSecRing, p->pslPubRing,
    FALSE, hIn, hErrW, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hErrW);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromHandle(hwnd, hErrR));
  CloseHandleClean(&hErrR);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Key Imported\n"));
  } else {
    DO_DEBUG(hr,  ShowProcessMessage(hwnd, "Key Import Process Failed\n"));
  }

finish_func:
  /* I */
  CloseHandleClean(&hIn);
  CloseHandleClean(&hErrR);
  CloseHandleClean(&hErrW);
  if (pgpg) gpgez_gpg_release(&pgpg);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Export(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_EXPORT p = (PGPGEZW_EXPORT)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hOut = INVALID_HANDLE_VALUE;
  HANDLE hErrR = INVALID_HANDLE_VALUE, hErrW = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* GNX|[gp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, p->pFileName));
  DO_DEBUG(hr, util_create_read_pipe(&hErrR, &hErrW));
  DO_DEBUG(hr, gpgez_gpg_create_export(p->pGpgExe, p->pGpgHome, p->pslSecRing, p->pslPubRing,
    FALSE, p->pFpr, p->fAscii, hOut, hErrW, &pgpg));
  CloseHandleClean(&hOut);
  CloseHandleClean(&hErrW);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromHandle(hwnd, hErrR));
  CloseHandleClean(&hErrR);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Key Exported\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Key Export Process Failed\n"));
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (hOut != INVALID_HANDLE_VALUE) CloseHandle(hOut);
  if (hErrR != INVALID_HANDLE_VALUE) CloseHandle(hErrR);
  if (hErrW != INVALID_HANDLE_VALUE) CloseHandle(hErrW);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Delete(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_DELETE p = (PGPGEZW_DELETE)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hErrR = INVALID_HANDLE_VALUE, hErrW = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* 폜\̍쐬 */
  DO_DEBUG(hr, util_create_read_pipe(&hErrR, &hErrW));
  DO_DEBUG(hr, gpgez_gpg_create_deletekey(p->pGpgExe, p->pGpgHome, p->pslSecRing, p->pslPubRing,
    FALSE, p->pFpr, hErrW, &pgpg));
  CloseHandleClean(&hErrW);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromHandle(hwnd, hErrR));
  CloseHandleClean(&hErrR);

  /* gpg I҂ */
  if (FAILED(hr = gpgez_gpg_finish(pgpg))) goto finish_func;

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    if (FAILED(hr = ShowProcessMessage(hwnd, "Key Deleted\n"))) goto finish_func;
  } else {
    if (FAILED(hr = ShowProcessMessage(hwnd, "Key Delete Process Failed\n"))) goto finish_func;
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (hErrR != INVALID_HANDLE_VALUE) CloseHandle(hErrR);
  if (hErrW != INVALID_HANDLE_VALUE) CloseHandle(hErrW);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_DetachSign(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_DETACHSIGN p = (PGPGEZW_DETACHSIGN)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE;
  HANDLE hErrR = INVALID_HANDLE_VALUE, hErrW = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, p->pFile));
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, p->pSfile));
  DO_DEBUG(hr, util_create_read_pipe(&hErrR, &hErrW));
  DO_DEBUG(hr, gpgez_gpg_create_sign(p->pGpgExe, p->pGpgHome, p->pslSecRing, p->pslPubRing,
    FALSE, p->pPassphraseSym, p->fAscii, TRUE, hIn, hOut, hErrW, &pgpg));
  CloseHandleClean(&hErrW);
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromHandle(hwnd, hErrR));
  CloseHandleClean(&hErrR);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Signature File Created\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Sign Process Failed\n"));
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (hIn != INVALID_HANDLE_VALUE) CloseHandle(hIn);
  if (hOut != INVALID_HANDLE_VALUE) CloseHandle(hOut);
  if (hErrR != INVALID_HANDLE_VALUE) CloseHandle(hErrR);
  if (hErrW != INVALID_HANDLE_VALUE) CloseHandle(hErrW);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Verify(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_VERIFY p = (PGPGEZW_VERIFY)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;
  HANDLE hErrR = INVALID_HANDLE_VALUE, hErrW = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* ؗp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, p->pFile));
  DO_DEBUG(hr, util_create_read_pipe(&hErrR, &hErrW));
  DO_DEBUG(hr, gpgez_gpg_create_verify(p->pGpgExe, p->pGpgHome, p->pslSecRing, p->pslPubRing,
    FALSE, p->pPassphraseSym, p->pSfile, hIn, hErrW, &pgpg));
  CloseHandleClean(&hErrW);
  CloseHandleClean(&hIn);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromHandle(hwnd, hErrR));
  CloseHandleClean(&hErrR);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Signature Verified\n"));
  } else {
    DO_DEBUG(hr, ShowProcessMessage(hwnd, "Verify Process Failed\n"));
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (hIn != INVALID_HANDLE_VALUE) CloseHandle(hIn);
  if (hErrR != INVALID_HANDLE_VALUE) CloseHandle(hErrR);
  if (hErrW != INVALID_HANDLE_VALUE) CloseHandle(hErrW);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD WINAPI
gpgezw_Cls_OnCreateProcess_Armor(LPVOID lpParam) {
  HWND hwnd = (HWND)lpParam;
  PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);
  PGPGEZW_ARMOR p = (PGPGEZW_ARMOR)(pgl->pParam);
  HRESULT hr = S_OK;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE;
  HANDLE hErrR = INVALID_HANDLE_VALUE, hErrW = INVALID_HANDLE_VALUE;

  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);

  /* ؗp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, p->pFileIn));
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, p->pFileOut));
  DO_DEBUG(hr, util_create_read_pipe(&hErrR, &hErrW));
  DO_DEBUG(hr, gpgez_gpg_create_armor(p->pGpgExe, p->pGpgHome, p->fEnArmor, hIn, hOut, hErrW, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOut);
  CloseHandleClean(&hErrW);

  /* LZ{^Lɂ */
  *(pgl->phProcess) = gpgez_gpg_get_process_handle(pgpg);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  /* bZ[WEChEɕ\ */
  DO_DEBUG(hr, ShowProcessMessageFromHandle(hwnd, hErrR));
  CloseHandleClean(&hErrR);

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) == 0) {
    if (p->fEnArmor) {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "Ascii Armored\n"));
    } else {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "Ascii DeArmored\n"));
    }
  } else {
    if (p->fEnArmor) {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "Ascii Armor Failed\n"));
    } else {
      DO_DEBUG(hr, ShowProcessMessage(hwnd, "Ascii DeArmor Failed\n"));
    }
  }

finish_func:
  /* I */
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (hIn != INVALID_HANDLE_VALUE) CloseHandle(hIn);
  if (hOut != INVALID_HANDLE_VALUE) CloseHandle(hOut);
  if (hErrR != INVALID_HANDLE_VALUE) CloseHandle(hErrR);
  if (hErrW != INVALID_HANDLE_VALUE) CloseHandle(hErrW);
  Edit_SetReadOnly(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static VOID
gpgezw_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    EndDialog(hwnd, (INT)ERROR_SUCCESS);
    break;
  case IDCANCEL:
    {
      PGPGEZW_LAUNCH pgl = (PGPGEZW_LAUNCH)GetWindowLong(hwnd, GWL_USERDATA);

      if ((pgl->phProcess) &&
          (*(pgl->phProcess) != INVALID_HANDLE_VALUE) &&
          (*(pgl->phProcess) != NULL)) {
        TerminateProcess(*(pgl->phProcess), -1);
      }
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDialog(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, gpgezw_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, gpgezw_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, gpgezw_Cls_OnCommand);
  case WM_CREATEPROCESS_GENKEY:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Genkey, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_ENCRYPT:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Encrypt, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_TARENCRYPT:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_TarEncrypt, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_ZIPENCRYPT:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_ZipEncrypt, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_7ZENCRYPT:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_7zEncrypt, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_TAR:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Tar, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_ENCRYPTSFX:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_EncryptSfx, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_DECRYPT:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Decrypt, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_DECRYPTTAR:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_DecryptTar, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_7ZDECRYPT:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_7zDecrypt, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_UNTAR:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_UnTar, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_7ZGUNZIP:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_7zGunZip, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_IMPORT:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Import, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_EXPORT:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Export, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_DELETE:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Delete, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_DETACHSIGN:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_DetachSign, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_VERIFY:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Verify, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  case WM_CREATEPROCESS_ARMOR:
    {
      DWORD dwThreadId;
      HANDLE hThread;

      hThread = CreateThread(NULL, 0, gpgezw_Cls_OnCreateProcess_Armor, (LPVOID)hDlg, 0, &dwThreadId);
      if (hThread) CloseHandle(hThread);
      return TRUE;
    }
    break;
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_genkey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pName, PCSTR pEMail, PCSTR pComment, PCSTR pPassphrase, PSTR *ppFpr) {
  HRESULT hr = S_OK;
  GPGEZW_GENKEY gg = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_GENKEY, (PVOID)&gg, NULL };
  int iRet;
  HANDLE hProcess = NULL;

  gg.hwnd = hwnd;
  gg.pGpgExe = pgpgezbase->pGpgExe;
  gg.pGpgHome = pgpgezbase->pConfDir;
  gg.pslSecRing = pslSecRing;
  gg.pslPubRing = pslPubRing;
  gg.pName = pName;
  gg.pEMail = pEMail;
  gg.pComment = pComment;
  gg.pPassphrase = pPassphrase;
  gg.ppFpr = ppFpr;
  gl.phProcess = &hProcess;
  gl.hr = S_OK;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_encrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFile, PCSTR pZfile, PSTRINGLIST pRecipients, PCSTR pPassphraseSym, BOOL fSign, BOOL fAscii) {
  HRESULT hr = S_OK;
  GPGEZW_ENCRYPT ge = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_ENCRYPT, (PVOID)&ge, NULL };
  int iRet;
  HANDLE hProcess = NULL;

  ge.hwnd = hwnd;
  ge.pGpgExe = pgpgezbase->pGpgExe;
  ge.pGpgHome = pgpgezbase->pConfDir;
  ge.pslSecRing = pslSecRing;
  ge.pslPubRing = pslPubRing;
  ge.pFile = pFile;
  ge.pZfile = pZfile;
  ge.pRecipients = pRecipients;
  ge.pPassphraseSym = pPassphraseSym;
  ge.fSign = fSign;
  ge.fAscii = fAscii;
  gl.phProcess = &hProcess;
  gl.hr = S_OK;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_tarencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PSTRINGLIST pRecipients, PCSTR pPassphraseSym, BOOL fSign, BOOL fAscii) {
  HRESULT hr = S_OK;
  GPGEZW_TARENCRYPT gte = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_TARENCRYPT, (PVOID)&gte, NULL };
  int iRet;
  HANDLE hProcess = NULL;

  gte.hwnd = hwnd;
  gte.pTarExe = pgpgezbase->pTarExe;
  gte.pGpgExe = pgpgezbase->pGpgExe;
  gte.pGpgHome = pgpgezbase->pConfDir;
  gte.pslSecRing = pslSecRing;
  gte.pslPubRing = pslPubRing;
  gte.pBaseDirectory = pBaseDirectory;
  gte.pFiles = pFiles;
  gte.pZfile = pZfile;
  gte.pRecipients = pRecipients;
  gte.pPassphraseSym = pPassphraseSym;
  gte.fSign = fSign;
  gte.fAscii = fAscii;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zipencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PCSTR pPassphraseSym) {
  HRESULT hr = S_OK;
  GPGEZW_ZIPENCRYPT gze = {NULL, NULL, NULL, NULL, NULL, NULL };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_ZIPENCRYPT, (PVOID)&gze, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  gze.hwnd = hwnd;
  gze.p7ZipExe = pgpgezbase->p7ZipExe;
  gze.pBaseDirectory = pBaseDirectory;
  gze.pFiles = pFiles;
  gze.pZfile = pZfile;
  gze.pPassphraseSym = pPassphraseSym;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_7zencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PCSTR pPassphraseSym, DWORD dwMode) {
  HRESULT hr = S_OK;
  GPGEZW_7ZENCRYPT g7e = {NULL, NULL, NULL, NULL, NULL, NULL, 0 };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_7ZENCRYPT, (PVOID)&g7e, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  g7e.hwnd = hwnd;
  g7e.p7ZipExe = pgpgezbase->p7ZipExe;
  g7e.pBaseDirectory = pBaseDirectory;
  g7e.pFiles = pFiles;
  g7e.pZfile = pZfile;
  g7e.pPassphraseSym = pPassphraseSym;
  g7e.dwMode = dwMode;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_tar(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, DWORD dwMode) {
  HRESULT hr = S_OK;
  GPGEZW_TAR gt = {NULL, NULL, NULL, NULL, NULL, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_TAR, (PVOID)&gt, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  gt.hwnd = hwnd;
  gt.pTarExe = pgpgezbase->pTarExe;
  gt.pBaseDirectory = pBaseDirectory;
  gt.pFiles = pFiles;
  gt.pZfile = pZfile;
  gt.dwMode= dwMode;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_encryptsfx(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PCSTR pPassphraseSym) {
  HRESULT hr = S_OK;
  GPGEZW_ENCRYPTSFX ges = {NULL, NULL, NULL, NULL, NULL, NULL };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_ENCRYPTSFX, (PVOID)&ges, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  ges.hwnd = hwnd;
  ges.p7ZipExe = pgpgezbase->p7ZipExe;
  ges.pBaseDirectory = pBaseDirectory;
  ges.pFiles = pFiles;
  ges.pZfile = pZfile;
  ges.pPassphraseSym = pPassphraseSym;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_decrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFile, PCSTR pXFile) {
  HRESULT hr = S_OK;
  GPGEZW_DECRYPT gd = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_DECRYPT, (PVOID)&gd, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  gd.hwnd = hwnd;
  gd.pGpgExe = pgpgezbase->pGpgExe;
  gd.pGpgHome = pgpgezbase->pConfDir;
  gd.pslSecRing = pslSecRing;
  gd.pslPubRing = pslPubRing;
  gd.pPassphrase = pPassphrase;
  gd.pFile = pFile;
  gd.pXFile = pXFile;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

#if 0
/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_decrypt_buff2buff(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, PSTR *ppbuff, PDWORD pdwLen, BOOL fVerify) {
  HRESULT hr = S_OK;
  BOOL fPublicEncrypted = FALSE;
  PSTRINGLIST pslVerified = NULL;
  PSTR pPassphrase = NULL;

  if (!stringlist_create(&pslVerified)) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  /* JÍΏ̈Í */
  if (FAILED(hr = gpgez_ispublicencrypted_buff(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing, pbuff, dwLen, &fPublicEncrypted))) {
    goto finish_func;
  }
  if (fPublicEncrypted) {
    if (!gpgezw_getMyFpr(pgpgezbase)) {
      hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
      goto finish_func;
    }
    if (gpgezw_fNeedPassphrase(pgpgezbase)) {
      if (!gpgezw_getPassphrseForPublicKeyEncryption(pgpgezbase, hwnd, 2)) {
        MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
        goto finish_func;
      }
    }
  } else {
    /* Ώ̈Í߂̃pXt[Y */
    CHAR buff[MAX_PATH + 1] = "";
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_REQUEST_FOR_DECRYPT_PASSPHRASE_SYM, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Input passphrase for symmetric encryption");
    }
    if (FAILED(hr = gpgezw_InputBoxEx(hwnd, m_buff, "Passphrase", buff, 0, MAX_PATH, INPUTBOX_WITHCANCEL | INPUTBOX_SECRET))) {
      goto finish_func;
    }
    if (!(pPassphrase = util_strdup(buff))) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      goto finish_func;
    }
  }

  /*  */
  if (FAILED(hr = gpgez_decrypt_buff2buff(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing, pPassphrase ? pPassphrase : pgpgezbase->pPassphrase, pbuff, dwLen, ppbuff, pdwLen, pslVerified))) {
    goto finish_func;
  }

  /* TODO: KvɉďmF */
finish_func:
  if (pPassphrase) util_free(pPassphrase);
  if (pslVerified) stringlist_release(&pslVerified);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_decrypt_buff2buff_nkf(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, PSTR *ppbuff, PDWORD pdwLen, BOOL fVerify, PCSTR pNkfOptIn, PCSTR pNkfOptOut) {
  HRESULT hr = S_OK;
  /* TODO: */
  return hr;
}
#endif
/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_decrypttar(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFileName, PCSTR pBaseDirectory, BOOL fOverWrite) {
  HRESULT hr = S_OK;
  GPGEZW_DECRYPTTAR gdt = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_DECRYPTTAR, (PVOID)&gdt, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  gdt.hwnd = hwnd;
  gdt.pGpgExe = pgpgezbase->pGpgExe;
  gdt.pGpgHome = pgpgezbase->pConfDir;
  gdt.pTarExe = pgpgezbase->pTarExe;
  gdt.pslSecRing = pslSecRing;
  gdt.pslPubRing = pslPubRing;
  gdt.pPassphrase = pPassphrase;
  gdt.pFileName = pFileName;
  gdt.pBaseDirectory = pBaseDirectory;
  gdt.fOverWrite = fOverWrite;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_7zdecrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pPassphrase, PCSTR pFileName, PCSTR pBaseDirectory, BOOL fOverWrite) {
  HRESULT hr = S_OK;
  GPGEZW_7ZDECRYPT g7d = { NULL, NULL, NULL, NULL, NULL, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_7ZDECRYPT, (PVOID)&g7d, NULL };
  int iRet;
  HANDLE hProcess = NULL;

  g7d.hwnd = hwnd;
  g7d.p7ZipExe = pgpgezbase->p7ZipExe;
  g7d.pPassphrase = pPassphrase;
  g7d.pFileName = pFileName;
  g7d.pBaseDirectory = pBaseDirectory;
  g7d.fOverWrite = fOverWrite;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_untar(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pFileName, PCSTR pBaseDirectory, DWORD dwMode, BOOL fOverWrite) {
  HRESULT hr = S_OK;
  GPGEZW_UNTAR gut = { NULL, NULL, NULL, NULL, TAR_COMPRESS_NONE, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_UNTAR, (PVOID)&gut, NULL };
  int iRet;
  HANDLE hProcess = NULL;

  gut.hwnd = hwnd;
  gut.pTarExe = pgpgezbase->pTarExe;
  gut.pFileName = pFileName;
  gut.pBaseDirectory = pBaseDirectory;
  gut.dwMode = dwMode;
  gut.fOverWrite = fOverWrite;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_7zgunzip(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pFileNameIn, PCSTR pFileNameOut) {
  HRESULT hr = S_OK;
  GPGEZW_7ZGUNZIP g7uz = { NULL, NULL, NULL, NULL };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_7ZGUNZIP, (PVOID)&g7uz, NULL };
  int iRet;
  HANDLE hProcess = NULL;

  g7uz.hwnd = hwnd;
  g7uz.p7ZipExe = pgpgezbase->p7ZipExe;
  g7uz.pFileNameIn = pFileNameIn;
  g7uz.pFileNameOut = pFileNameOut;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_import(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefault, PCSTR pFileName) {
  HRESULT hr = S_OK;
  GPGEZW_IMPORT gi = {NULL, NULL, NULL, NULL, NULL, FALSE, NULL };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_IMPORT, (PVOID)&gi, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  gi.hwnd = hwnd;
  gi.pGpgExe = pgpgezbase->pGpgExe;
  gi.pGpgHome = pgpgezbase->pConfDir;
  gi.pslSecRing = pslSecRing;
  gi.pslPubRing = pslPubRing;
  gi.fNoDefault = fNoDefault;
  gi.pFileName = pFileName;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_export(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFpr, PCSTR pFileName, BOOL fAscii) {
  HRESULT hr = S_OK;
  GPGEZW_EXPORT ge = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_EXPORT, (PVOID)&ge, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  ge.hwnd = hwnd;
  ge.pGpgExe = pgpgezbase->pGpgExe;
  ge.pGpgHome = pgpgezbase->pConfDir;
  ge.pslSecRing = pslSecRing;
  ge.pslPubRing = pslPubRing;
  ge.pFpr = pFpr;
  ge.pFileName = pFileName;
  ge.fAscii = fAscii;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_delete(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefault, PCSTR pFpr) {
  HRESULT hr = S_OK;
  GPGEZW_DELETE gd = {NULL, NULL, NULL, NULL, NULL, FALSE, NULL };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_DELETE, (PVOID)&gd, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  gd.hwnd = hwnd;
  gd.pGpgExe = pgpgezbase->pGpgExe;
  gd.pGpgHome = pgpgezbase->pConfDir;
  gd.pslSecRing = pslSecRing;
  gd.pslPubRing = pslPubRing;
  gd.fNoDefault = fNoDefault;
  gd.pFpr = pFpr;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_detach_sign(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFile, PCSTR pSfile, PCSTR pPassphraseSym, BOOL fAscii) {
  HRESULT hr = S_OK;
  GPGEZW_DETACHSIGN gd = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_DETACHSIGN, (PVOID)&gd, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  gd.hwnd = hwnd;
  gd.pGpgExe = pgpgezbase->pGpgExe;
  gd.pGpgHome = pgpgezbase->pConfDir;
  gd.pslSecRing = pslSecRing;
  gd.pslPubRing = pslPubRing;
  gd.pFile = pFile;
  gd.pSfile = pSfile;
  gd.pPassphraseSym = pPassphraseSym;
  gd.fAscii = fAscii;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_verify(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFile, PCSTR pSfile, PCSTR pPassphraseSym)
{
  HRESULT hr = S_OK;
  GPGEZW_VERIFY gv = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_VERIFY, (PVOID)&gv, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  gv.hwnd = hwnd;
  gv.pGpgExe = pgpgezbase->pGpgExe;
  gv.pGpgHome = pgpgezbase->pConfDir;
  gv.pslSecRing = pslSecRing;
  gv.pslPubRing = pslPubRing;
  gv.pFile = pFile;
  gv.pSfile = pSfile;
  gv.pPassphraseSym = pPassphraseSym;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_armor(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, PCSTR pszFileNameOut, BOOL fEnArmor)
{
  HRESULT hr = S_OK;
  GPGEZW_ARMOR ga = {NULL, NULL, NULL, NULL, NULL, FALSE };
  GPGEZW_LAUNCH gl = { WM_CREATEPROCESS_ARMOR, (PVOID)&ga, NULL };
  int iRet;
  HANDLE hProcess = NULL;
  
  ga.hwnd = hwnd;
  ga.pGpgExe = pgpgezbase->pGpgExe;
  ga.pGpgHome = pgpgezbase->pConfDir;
  ga.pFileIn = pszFileName;
  ga.pFileOut = pszFileNameOut;
  ga.fEnArmor = fEnArmor;
  gl.phProcess = &hProcess;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDialog, (LPARAM)&gl);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = gl.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn HRESULT gpgezw_listupFilesFromDrop(HDROP hDrop, PTARGET_FILELIST *pptfl)
 *
 * hbvꂽt@CATARGET_FILELIST\̂Ɋi[֐
 *
 */
/*
  FIX ME [gtH_łꍇ̏
  {IɃ[gtH_̓hbvȂ
 */
__declspec(dllexport) HRESULT
gpgezw_listupFilesFromDrop(HDROP hDrop, PTARGET_FILELIST *pptfl) {
  UINT i, iCnt;
  CHAR buff[MAX_PATH + 1];
  UINT baselen;
  HRESULT hr = S_OK;

  /* hbvꂽt@C𒲂ׂ */
  iCnt = DragQueryFile(hDrop, -1, NULL, 0);
  if (iCnt == 0) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  /* efBNg𒲂ׂ */
  baselen = DragQueryFile(hDrop, 0, NULL, 0);
  DragQueryFile(hDrop, 0, buff, baselen + 1);
  *(PathFindFileName(buff) - 1) = '\0'; /* fBNg\\͊܂߂Ȃ */
  baselen = util_strlen(buff, -1);
  target_filelist_create(pptfl, buff);

  /* t@CXg쐬 */
  for (i = 0; i < iCnt; i++) {
    DragQueryFile(hDrop, i, buff, MAX_PATH);
    if (util_strncmp(target_filelist_get_basedir(*pptfl), buff, baselen)) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = target_filelist_add_node(*pptfl, buff + baselen + 1))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }
finish_func:
  if (FAILED(hr)) target_filelist_release(pptfl);
  return hr;
}

/*------------------------------------------------------------*/
/**
 * \fn HRESULT gpgezw_prepareKeypair(HWND hwnd)
 *
 * hwndeEChEƂăvZX\Ȃ献yA𐶐
 *
 */
DllExport HRESULT
gpgezw_prepareKeypair(PGPGEZBASE pgpgezbase, HWND hwnd) {
  HRESULT hr = S_OK;
  WIN32_FIND_DATA wfd;
  HANDLE hFind;

  /* ݒtH_ꍇ͍쐬 */
  hFind = FindFirstFile(pgpgezbase->pConfDir, &wfd);
  if (hFind == INVALID_HANDLE_VALUE) {
    if (!CreateDirectory(pgpgezbase->pConfDir, NULL)) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  } else {
    FindClose(hFind);
    /* ݒtH_ƖOŁAtH_ȊỐit@CHjꍇ̓G[ */
    if (!(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  hr = gpgezw_generateGpgKey(pgpgezbase, hwnd, NULL, NULL);
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
HRESULT
decorate_fpr(PSTR pDst, INT max, PCSTR pFpr, PCSTR pMyFpr) {
  PCSTR p1 = pFpr;
  PSTR p2 = pDst;
  UINT iCnt = 0;
  HRESULT hr = S_OK;

  if (max < ((util_strlen(pFpr, -1) + 3) / 4) * 5  + 1 + 1) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  if ((pMyFpr) && (!util_strcmp(pFpr, pMyFpr))) {
    *pDst = '*';
  } else {
    *pDst = ' ';
  }

  p2 = pDst + 1;
  while (*p1) {
    if (iCnt % 4 == 0) {
      *p2 = ' ';
      ++p2;
    }
    *p2 = *p1;
    ++p1; ++p2; ++iCnt;
  }
  *p2 = '\0';
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
HRESULT
undecorate_fpr(PSTR pDst, INT max, PCSTR pFpr) {
  PCSTR p1 = pFpr;
  PSTR p2 = pDst;
  HRESULT hr = S_OK;

  if (max < (util_strlen(pFpr, -1) - 2) * 4 / 5 + 1) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  p1 = pFpr + 2;
  while (*p1) {
    while (*p1 == ' ') ++p1;
    *p2 = *p1;
    ++p1; ++p2;
  }
  *p2 = '\0';
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport PCSTR
gpgezw_getMyFpr(PGPGEZBASE pgpgezbase) {
  return pgpgezbase->pMyFpr;
}

/*------------------------------------------------------------*/
DllExport PCSTR
gpgezw_getConfDir(PGPGEZBASE pgpgezbase) {
  return pgpgezbase->pConfDir;
}

/*------------------------------------------------------------*/
DllExport BOOL
gpgezw_fNeedPassphrase(PGPGEZBASE pgpgezbase) {
  return pgpgezbase->fNeedPassphrase;
}

/*------------------------------------------------------------*/
DllExport PCSTR
gpgezw_getPubKeyDir(PGPGEZBASE pgpgezbase) {
  return pgpgezbase->pPubKeyDir;
}

/*------------------------------------------------------------*/
DllExport BOOL
gpgezw_getPassphrseForPublicKeyEncryption(PGPGEZBASE pgpgezbase, HWND hwnd, DWORD dwRetry) {
  BOOL fRet = FALSE;
  CHAR buff[261] = "";
  CHAR m_buff[MAX_PATH + 1] = "";
  PGPGEZ_GPG pgpg = NULL;
  HRESULT hr = S_OK;

  if (!gpgezw_fNeedPassphrase(pgpgezbase)) {
    fRet = TRUE;
    goto finish_func;
  }

  if (LoadString(g_hInstance, IDS_MESSAGE_NEEDPASSPHRASE, m_buff, MAX_PATH) == 0) {
    util_strcpy(m_buff, MAX_PATH, "Input Passphrase for Using Secret Key");
  }
  while (TRUE) {
    if (FAILED(hr = gpgezw_InputBoxEx(hwnd, m_buff, "Passphrase", buff, 0, 260, INPUTBOX_WITHCANCEL | INPUTBOX_SECRET))) {
      goto finish_func;
    }
    if (pgpgezbase->pPassphrase) util_free(pgpgezbase->pPassphrase);
    if (!(pgpgezbase->pPassphrase = util_strdup(buff))) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      goto finish_func;
    }

    /* SignăpXt[YmF */
    DO_DEBUG(hr, gpgez_gpg_create_sign(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, NULL, NULL,
      FALSE, pgpgezbase->pPassphrase, FALSE, TRUE, INVALID_HANDLE_VALUE, INVALID_HANDLE_VALUE, INVALID_HANDLE_VALUE, &pgpg));
    DO_DEBUG(hr, gpgez_gpg_finish(pgpg));
    if (gpgez_gpg_get_exitcode(pgpg) == 0) {
      fRet = TRUE;
      pgpgezbase->fNeedPassphrase = FALSE;
      gpgez_gpg_release(&pgpg);
      break;
    } else {
      fRet = FALSE;
      util_free(pgpgezbase->pPassphrase);
      pgpgezbase->pPassphrase = NULL;
    }
    gpgez_gpg_release(&pgpg);

    if (dwRetry == 0) break;
    --dwRetry;
  }
finish_func:
  if (pgpg) gpgez_gpg_release(&pgpg);
  return fRet;
}

/*------------------------------------------------------------*/
/**
 * \fn static BOOL checkkeyprotected(PGPGEZBASE pgpgezbase);
 *
 * JÍgp邽߂̔閧pXt[Yŕی삳Ă邩ǂ̊mF
 * ی삳ĂTRUEAłȂFALSEԂ
 *
 */
static BOOL
checkkeyprotected(PGPGEZBASE pgpgezbase) {
  HRESULT hr = S_OK;
  PCMDEZCONTEXT pgctx = NULL;
  BOOL fProtected = FALSE;
  PREADFILEBUFF pfilebuff = NULL;
  HANDLE hStdOutP = INVALID_HANDLE_VALUE, hStdOutC = INVALID_HANDLE_VALUE;
  CHAR buff[MAX_LINE + 1];
  CHAR szKey[MAX_PATH + 1];
  BOOL fFirstBlock = TRUE;

  DO_DEBUG(hr, cmdez_initialize(&pgctx, pgpgezbase->pGpgExe, NULL));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--homedir", pgpgezbase->pConfDir, TRUE));
  DO_DEBUG(hr, cmdez_add_option(pgctx, "--list-packets", NULL, FALSE));
  DO_DEBUG(hr, util_strcpy(szKey, MAX_PATH, pgpgezbase->pConfDir));
  DO_DEBUG(hr, util_strcat(szKey, MAX_PATH, "\\secring.gpg"));
  DO_DEBUG(hr, cmdez_add_operand(pgctx, szKey, TRUE));
  DO_DEBUG(hr, util_create_read_pipe(&hStdOutP, &hStdOutC));
  DO_DEBUG(hr, cmdez_set_handle(pgctx, HANDLE_TYPE_STDOUT, hStdOutC));
  DO_DEBUG(hr, cmdez_execute(pgctx));
  CloseHandle(hStdOutC);
  hStdOutC = INVALID_HANDLE_VALUE;
  DO_DEBUG(hr, CreateReadFileBuff(&pfilebuff, hStdOutP));
  while (!fEofFileLine(pfilebuff)) {
    DO_DEBUG(hr, ReadFileLine(pfilebuff, buff, MAX_LINE));
    if (fFirstBlock) {
      if ((*buff == ':') && util_strncmp(":secret", buff, 7)) {
        fFirstBlock = FALSE;
      }
    }
    if (fFirstBlock && (!util_strncmp(buff, "\tencrypted ", 11))) {
      fProtected = TRUE;
    }
  }
  DO_DEBUG(hr, DetatchReadFileLine(pfilebuff));
  CloseHandle(hStdOutP);
  hStdOutP = INVALID_HANDLE_VALUE;
  cmdez_finish(pgctx);
  cmdez_release(&pgctx);
  pgctx = NULL;

finish_func:
  if (pfilebuff) ReleaseReadFileBuff(&pfilebuff);
  if (hStdOutP != INVALID_HANDLE_VALUE) CloseHandle(hStdOutP);
  if (hStdOutC != INVALID_HANDLE_VALUE) CloseHandle(hStdOutC);
  if (pgctx) cmdez_release(&pgctx);
  return fProtected;
}

/*------------------------------------------------------------*/
HRESULT
selectGpgFileKey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pubkeyring, PSTRINGLIST recipients)
{
  /* t@C̑I */
  OPENFILENAME ofn;
  CHAR szPubkeyFileName[MAX_PATH + 1] = "";
  PSTRINGLIST psl = NULL;
  PSTRINGLIST_NODE pnode;
  HRESULT hr = S_OK;

  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = "gpg key(*.gpg)\0*.gpg\0\0";
  ofn.lpstrFile = szPubkeyFileName;
  ofn.lpstrTitle = "Select Public Key File";
  ofn.lpstrInitialDir = pgpgezbase->pPubKeyDir;
  ofn.nMaxFile = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_NOCHANGEDIR;
  if (!GetOpenFileName(&ofn)) {
    hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
    goto finish_func;
  }
  if (FAILED(hr = stringlist_add_node(pubkeyring, szPubkeyFileName))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (recipients) {
    /* J擾 */
    if (FAILED(hr =  stringlist_create(&psl))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = gpgez_listkeys(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, NULL, pubkeyring, TRUE, FALSE, psl))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    for (pnode = stringlist_get_head(psl); pnode; pnode = stringlist_get_next(pnode)) {
      PCSTR p0 = stringlist_get_string(pnode);
      PCSTR p1 = util_strchr(p0, ':');
      if (p1) {
        CHAR buff[MAX_PATH + 1] = "";
        if (FAILED(hr = util_strncpy(buff, MAX_PATH, p0, p1 - p0))) {
          DEBUGERROR(0, hr);
          stringlist_release(&psl);
          goto finish_func;
        }
        buff[p1 - p0] = '\0';
        if (FAILED(hr = stringlist_add_node(recipients, buff))) {
          DEBUGERROR(0, hr);
          stringlist_release(&psl);
          goto finish_func;
        }
      }
    }
  }
finish_func:
  if (psl) stringlist_release(&psl);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_initialize(PGPGEZBASE *ppgpgezbase, PCSTR pGpgExe, PCSTR pConfDir, PCSTR pPubKeyDir) {
  HRESULT hr = S_OK;
  HANDLE hFind;
  WIN32_FIND_DATA wfd;

  *ppgpgezbase = NULL;

  if (!pGpgExe) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  if (!pConfDir) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }

  *ppgpgezbase = (PGPGEZBASE)util_malloc(sizeof(GPGEZBASE));
  if (!*ppgpgezbase) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  ZeroMemory(*ppgpgezbase, sizeof(GPGEZBASE));
  (*ppgpgezbase)->pGpgExe = NULL;
  (*ppgpgezbase)->pTarExe = NULL;
  (*ppgpgezbase)->p7ZipExe = NULL;
  (*ppgpgezbase)->pNkfExe = NULL;
  (*ppgpgezbase)->pConfDir = NULL;
  (*ppgpgezbase)->pPubKeyDir = NULL;
  (*ppgpgezbase)->fPublic = FALSE;
  (*ppgpgezbase)->fNeedPassphrase = FALSE;
  (*ppgpgezbase)->pMyFpr = NULL;
  (*ppgpgezbase)->pPassphrase = NULL;

  if (pGpgExe && !((*ppgpgezbase)->pGpgExe = util_strdup(pGpgExe))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  if (pConfDir && !((*ppgpgezbase)->pConfDir = util_strdup(pConfDir))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  if (pPubKeyDir && !((*ppgpgezbase)->pPubKeyDir = util_strdup(pPubKeyDir))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  /* Jg邩ǂ̊mF */
  (*ppgpgezbase)->fPublic = FALSE;
  (*ppgpgezbase)->fNeedPassphrase = FALSE;

  /* ݒtH_ȂΌJÍ͎gps */
  hFind = FindFirstFile((*ppgpgezbase)->pConfDir, &wfd);
  if (hFind != INVALID_HANDLE_VALUE) {
    FindClose(hFind);

    /* ݒtH_ƓÕt@C݂ꍇAɏ󋵂[Ȃ̂ŃG[\ */
    if (!(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    if (FAILED(hr = gpgez_loadmyfpr((*ppgpgezbase)->pGpgExe, (*ppgpgezbase)->pConfDir, NULL, NULL, &((*ppgpgezbase)->pMyFpr)))) {
      goto finish_func;
    }
    if ((*ppgpgezbase)->pMyFpr) {
      (*ppgpgezbase)->fPublic = TRUE;
      if (checkkeyprotected(*ppgpgezbase)) {
        (*ppgpgezbase)->fNeedPassphrase = TRUE;
/*
        if (!gpgezw_getPassphrseForPublicKeyEncryption(hwnd, 2)) {
          MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
          return FALSE;
        }
*/
      }
    }
  }
finish_func:
  if (FAILED(hr) && *ppgpgezbase) gpgezw_release(ppgpgezbase);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_initialize_tar(PGPGEZBASE pgpgezbase, PCSTR pTarExe) {
  HRESULT hr = S_OK;

  if (pTarExe && !(pgpgezbase->pTarExe = util_strdup(pTarExe))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_initialize_7Zip(PGPGEZBASE pgpgezbase, PCSTR p7ZipExe) {
  HRESULT hr = S_OK;

  if (p7ZipExe && !(pgpgezbase->p7ZipExe = util_strdup(p7ZipExe))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_initialize_nkf(PGPGEZBASE pgpgezbase, PCSTR pNkfExe) {
  HRESULT hr = S_OK;

  if (pNkfExe && !(pgpgezbase->pNkfExe = util_strdup(pNkfExe))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_release(PGPGEZBASE *ppgpgezbase) {

  if (!*ppgpgezbase) goto finish_func;

  if ((*ppgpgezbase)->pGpgExe) util_free((*ppgpgezbase)->pGpgExe);
  if ((*ppgpgezbase)->pTarExe) util_free((*ppgpgezbase)->pTarExe);
  if ((*ppgpgezbase)->p7ZipExe) util_free((*ppgpgezbase)->p7ZipExe);
  if ((*ppgpgezbase)->pNkfExe) util_free((*ppgpgezbase)->pNkfExe);
  if ((*ppgpgezbase)->pConfDir) util_free((*ppgpgezbase)->pConfDir);
  if ((*ppgpgezbase)->pPubKeyDir) util_free((*ppgpgezbase)->pPubKeyDir);
  if ((*ppgpgezbase)->pMyFpr) util_free((*ppgpgezbase)->pMyFpr);
  if ((*ppgpgezbase)->pPassphrase) util_free((*ppgpgezbase)->pPassphrase);
  util_free(*ppgpgezbase);
  *ppgpgezbase = NULL;
finish_func:
  return S_OK;
}

/*------------------------------------------------------------*/
HRESULT
create_encrypt_filefilter(PENCFILETYPE pet, PSTR *ppfilter)
{
  HRESULT hr = S_OK;
  PSTRING pstrfilter = NULL;
  PSTR p = NULL;

  *ppfilter = NULL;

  if (FAILED(hr = string_create(&pstrfilter))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  while (pet->filetype != FILETYPE_NONE) {
    if (pet->filedesc1) {
      if (FAILED(hr = string_append(pstrfilter, pet->filedesc1))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    } else {
      if (FAILED(hr = string_append(pstrfilter, pet->filedesc))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
    if (FAILED(hr = string_append(pstrfilter, "\n*."))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = string_append(pstrfilter, pet->fileext))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = string_append(pstrfilter, "\n"))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    ++pet;
  }

  if (!(*ppfilter = string_dup_str(pstrfilter))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  p = *ppfilter;
  while (*p) {
    if (*p == '\n') {
      *p = '\0';
    }
    ++p;
  }
finish_func:
  if (pstrfilter) string_release(&pstrfilter);
  return hr;
}

/*------------------------------------------------------------*/
UINT CALLBACK
OFNHookProcSaveFile(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch(msg){
    case WM_INITDIALOG:
      SetWindowLong(hdlg, GWL_USERDATA, lParam);
      return TRUE;
    case WM_NOTIFY:
      {
        LPOFNOTIFY pofn = (LPOFNOTIFY)lParam;
        if (pofn->hdr.code == CDN_TYPECHANGE) {
          PENCFILETYPE penc = (PENCFILETYPE)(pofn->lpOFN->lCustData);
          int i = 0;
          UINT iIndex = 1;

          while (penc->filetype != FILETYPE_NONE) {
            if (pofn->lpOFN->nFilterIndex == iIndex) {
              break;
            }
            ++iIndex;
            ++penc;
          }
          SendMessage(hdlg, CDM_SETDEFEXT, 0, (LPARAM)(penc->fileext));
          return TRUE;
        }
      }
     return FALSE;
  }
  return FALSE;
}

/*------------------------------------------------------------*/
/**
 * \fn static HRESULT gpgezw_zgpencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PTARGET_FILELIST ptfl, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslRecipients, BOOL fExtraKey, BOOL fSign, PENCFILETYPE pet);
 *
 *
 */
__declspec(dllexport) HRESULT
gpgezw_zgpencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PTARGET_FILELIST ptfl, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslRecipients, BOOL fExtraKey, BOOL fSign, PENCFILETYPE pet) {
  HRESULT hr = S_OK;
  PSTRINGLIST recipients = NULL;
  PSTRINGLIST pslKeyRing = NULL;
  OPENFILENAME ofn;
  CHAR szFileName[MAX_PATH + 1] = "";
  PSTR pfilter = NULL;
  PENCFILETYPE penc1 = pet;

  DO_DEBUG(hr, create_encrypt_filefilter(pet, &pfilter));
  DO_DEBUG(hr, stringlist_create(&pslKeyRing));
  if (pslPubRing) {
    DO_DEBUG(hr, stringlist_concat(pslKeyRing, pslPubRing));
  }

  DO_DEBUG(hr, util_strcpy(szFileName, MAX_PATH, stringlist_get_string(stringlist_get_head(target_filelist_get_files(ptfl)))));
  {
    LPSTR p = PathFindExtension(szFileName);
    if (*p) *p = '\0';
  }

  /* ot@Cw */
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = pfilter;
  ofn.lpstrDefExt = "gpg";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.lCustData = (DWORD)pet;
  ofn.lpfnHook = OFNHookProcSaveFile;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_ENABLEHOOK;
  if (!GetSaveFileName(&ofn)) {
    DWORD dwErr = CommDlgExtendedError();
    /* G[`FbN */
    goto finish_func;
  }

  {
    UINT i;

    for (i = 1; i < ofn.nFilterIndex; ++i) {
      ++penc1;
    }
  }

  /* Recipients w */
  DO_DEBUG(hr, stringlist_create(&recipients));
  if (penc1->fpub) {
    if (fExtraKey) {
      if (FAILED(hr = selectGpgFileKey(pgpgezbase, hwnd, pslKeyRing, recipients))) {
        if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
          hr = S_OK;
        } else {
          DEBUGERROR(0, hr);
        }
        goto finish_func;
      }
    } else {
      if (gpgezw_fNeedPassphrase(pgpgezbase)) {
        if (!gpgezw_getPassphrseForPublicKeyEncryption(pgpgezbase, hwnd, 2)) {
          MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
          goto finish_func;
        }
      }

      /* pslRecipients ̎w肪΂gBȂ΃[U[Ɏw肳 */
      if (pslRecipients) {
        if (FAILED(hr = stringlist_concat(recipients, pslRecipients))) {
          DEBUGERROR(0, hr);
          goto finish_func;
        }
      } else {
        if (FAILED(hr = gpgezw_selectGpgKey(pgpgezbase, hwnd, pslSecRing, pslKeyRing, recipients, pet))) {
          if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
            hr = S_OK;
          } else {
            DEBUGERROR(0, hr);
          }
          goto finish_func;
        }
      }
    }
  } else if (penc1->passlen != -1) {
    CHAR buff[MAX_PATH + 1] = "";
    BOOL fSfx = FALSE;
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_REQUEST_FOR_PASSPHRASE_SYM, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Input passphrase for symmetric encryption");
    }
    if (FAILED(hr = gpgezw_InputBoxEx(hwnd, m_buff, "Passphrase", buff, penc1->passlen, MAX_PATH, INPUTBOX_CONFIRM | INPUTBOX_SECRET))) {
      if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
        hr = S_OK;
      } else {
        DEBUGERROR(0, hr);
      }
      goto finish_func;
    }
    DO_DEBUG(hr, stringlist_add_node(recipients, buff));
  }

  /* o͐̃t@CAÍ̃tH_Ɋ܂܂ĂȂ`FbN */
  {
    BOOL b = FALSE;
    DO_DEBUG(hr, util_is_file_included_in_a_folder(szFileName, ptfl, &b));
    if (b) {
      CHAR m_buff[MAX_PATH + 1];

      if (LoadString(g_hInstance, IDS_ERROR_INVALIDENCRYPTEDFILEPOSITION, m_buff, MAX_PATH) == 0) {
        util_strcpy(m_buff, MAX_PATH, "Invalid Encrypted File Position");
      }
      MessageBox(hwnd, m_buff, "Error", MB_OK);
      hr = S_OK; /* Iy[VG[Ȃ̂ */
      goto finish_func;
    }
  }

  switch(penc1->filetype) {
  case FILETYPE_GPGENCPUBLIC:
    /* fall thru */
  case FILETYPE_GPGENCPUBLIC_ASCII:
    {
      BOOL fsf = FALSE;
      CHAR szFileName1[MAX_PATH + 2] = "";

      if (target_filelist_get_number_of_files(ptfl) == 1) {
        HANDLE hFind;
        WIN32_FIND_DATA wfd;

        DO_DEBUG(hr, util_strcpy(szFileName1, MAX_PATH + 1, target_filelist_get_basedir(ptfl)));
        DO_DEBUG(hr, util_strcat(szFileName1, MAX_PATH + 1, "\\"));
        DO_DEBUG(hr, util_strcat(szFileName1, MAX_PATH + 1, stringlist_get_string(stringlist_get_head(target_filelist_get_files(ptfl)))));
        hFind = FindFirstFile(szFileName1, &wfd);
        if (hFind == INVALID_HANDLE_VALUE) {
          DEBUGERROR(0, hr);
          goto finish_func;
        }
        FindClose(hFind);
        if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
          fsf = FALSE;
        } else {
          fsf = TRUE;
        }
      }
      if (fsf) {
        DO_DEBUG(hr, gpgezw_encrypt(pgpgezbase, hwnd, pslSecRing, pslKeyRing, szFileName1, szFileName, recipients, pgpgezbase->pPassphrase, fSign, penc1->fascii));
      } else {
        DO_DEBUG(hr, gpgezw_tarencrypt(pgpgezbase, hwnd, pslSecRing, pslKeyRing, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, recipients, pgpgezbase->pPassphrase, fSign, penc1->fascii));
      }
    }
    break;
  case FILETYPE_GPGENCSYMMETRIC:
    /* fall thru */
  case FILETYPE_GPGENCSYMMETRIC_ASCII:
    {
      BOOL fsf = FALSE;
      CHAR szFileName1[MAX_PATH + 2] = "";

      if (target_filelist_get_number_of_files(ptfl) == 1) {
        HANDLE hFind;
        WIN32_FIND_DATA wfd;

        DO_DEBUG(hr, util_strcpy(szFileName1, MAX_PATH + 1, target_filelist_get_basedir(ptfl)));
        DO_DEBUG(hr, util_strcat(szFileName1, MAX_PATH + 1, "\\"));
        DO_DEBUG(hr, util_strcat(szFileName1, MAX_PATH + 1, stringlist_get_string(stringlist_get_head(target_filelist_get_files(ptfl)))));
        hFind = FindFirstFile(szFileName1, &wfd);
        if (hFind == INVALID_HANDLE_VALUE) {
          DEBUGERROR(0, hr);
          goto finish_func;
        }
        FindClose(hFind);
        if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
          fsf = FALSE;
        } else {
          fsf = TRUE;
        }
      }
      
      if (fsf) {
        DO_DEBUG(hr, gpgezw_encrypt(pgpgezbase, hwnd, pslSecRing, pslKeyRing, szFileName1, szFileName, NULL, stringlist_get_string(stringlist_get_head(recipients)), FALSE, penc1->fascii));
      } else {
        DO_DEBUG(hr, gpgezw_tarencrypt(pgpgezbase, hwnd, pslSecRing, pslKeyRing, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, NULL, stringlist_get_string(stringlist_get_head(recipients)), FALSE, penc1->fascii));
      }
    }
    break;
  case FILETYPE_ZIPENC:
    {
      DO_DEBUG(hr, gpgezw_7zencrypt(pgpgezbase, hwnd, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, stringlist_get_string(stringlist_get_head(recipients)), SEVENZIP_ZIPENC));
    }
    break;
  case FILETYPE_ZIPENCAES:
    {
      DO_DEBUG(hr, gpgezw_7zencrypt(pgpgezbase, hwnd, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, stringlist_get_string(stringlist_get_head(recipients)), SEVENZIP_ZIPENCAES));
    }
    break;
  case FILETYPE_7ZENC:
    {
      DO_DEBUG(hr, gpgezw_7zencrypt(pgpgezbase, hwnd, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, stringlist_get_string(stringlist_get_head(recipients)), SEVENZIP_7ZENC));
    }
    break;
  case FILETYPE_ZIPPLAIN:
    {
      DO_DEBUG(hr, gpgezw_7zencrypt(pgpgezbase, hwnd, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, NULL, SEVENZIP_ZIPPLAIN));
    }
    break;
  case FILETYPE_7ZPLAIN:
    {
      DO_DEBUG(hr, gpgezw_7zencrypt(pgpgezbase, hwnd, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, NULL, SEVENZIP_7ZPLAIN));
    }
    break;
  case FILETYPE_TARGZ:
    /* fall thru */
  case FILETYPE_TARBZ2:
    /* fall thru */
  case FILETYPE_TAR:
    {
      DWORD dwMode = TAR_COMPRESS_NONE;
      if (penc1->filetype == FILETYPE_TARGZ) {
        dwMode = TAR_COMPRESS_GZIP;
      } else if (penc1->filetype == FILETYPE_TARBZ2) {
        dwMode = TAR_COMPRESS_BZIP2;
      } else {
        dwMode = TAR_COMPRESS_NONE;
      }
      DO_DEBUG(hr, gpgezw_tar(pgpgezbase, hwnd, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, dwMode));
    }
    break;
  case FILETYPE_SFX:
    {
      DO_DEBUG(hr, gpgezw_encryptsfx(pgpgezbase, hwnd, target_filelist_get_basedir(ptfl), target_filelist_get_files(ptfl), szFileName, stringlist_get_string(stringlist_get_head(recipients))));
    }
    break;
  default:
    break;
  }
finish_func:
  if (pslKeyRing) stringlist_release(&pslKeyRing);
  if (recipients) stringlist_release(&recipients);
  if (pfilter) util_free(pfilter);

  return hr;
}

#if 0
/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_encrypt_buff2buff(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, PSTR *ppbuff, PDWORD pdwLen, BOOL fSign, BOOL fAscii) {
  HRESULT hr = S_OK;
  BOOL fPublic;
  CHAR szFileName[MAX_PATH + 1] = "";
  PSTRINGLIST recipients = NULL;

  /* Recipients w */
  DO_DEBUG(hr, stringlist_create(&recipients));
  if (fSign) {
    if (gpgezw_fNeedPassphrase(pgpgezbase)) {
      if (!gpgezw_getPassphrseForPublicKeyEncryption(pgpgezbase, hwnd, 2)) {
        MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
        goto finish_func;
      }
    }
  }
  if (FAILED(hr = gpgezw_selectGpgEncType(pgpgezbase, hwnd, pslSecRing, pslPubRing, recipients, &fPublic))) {
    if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
      hr = S_OK;
    } else {
      DEBUGERROR(0, hr);
    }
    goto finish_func;
  }

  if (fPublic) {
    hr = gpgez_encrypt_buff2buff(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing, pbuff, dwLen, ppbuff, pdwLen, recipients, pgpgezbase->pPassphrase, TRUE, fAscii);
  } else {
    hr = gpgez_encrypt_buff2buff(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing, pbuff, dwLen, ppbuff, pdwLen, NULL, stringlist_get_string(stringlist_get_head(recipients)), FALSE, fAscii);
  }
  if (FAILED(hr)) {
    MessageBox(hwnd, "Encryption Failed.", "Message", MB_OK);
  }
finish_func:
  if (recipients) stringlist_release(&recipients);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_encrypt_buff2file(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, BOOL fSign, BOOL fAscii) {
  HRESULT hr = S_OK;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_decrypt_file2buff(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTR *ppbuff, PDWORD pdwLen, BOOL fVerify) {
  HRESULT hr = S_OK;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_decrypt_file2buff_nkf(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTR *ppbuff, PDWORD pdwLen, BOOL fVerify, PCSTR pNkfIn, PCSTR pNkfOut) {
  HRESULT hr = S_OK;
  return hr;
}
#endif

/*------------------------------------------------------------*/
/*
 * ꂽt@CõtH_w肷
 *
 */
/*------------------------------------------------------------*/
static WNDPROC procOld;
#define IDC_OVERWRITE 1001

typedef struct struct_extract_folder_param {
  CHAR szInitFolder[MAX_PATH + 1];
  HWND hOverwrite;
  BOOL fOverwrite;
} EXTRACT_FOLDER_PARAM, *PEXTRACT_FOLDER_PARAM;

/*------------------------------------------------------------*/
static int CALLBACK
procSelectExtractFolderUser(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {

  if ((uMsg == WM_COMMAND) && (LOWORD(wParam) == IDC_OVERWRITE)) {
    PEXTRACT_FOLDER_PARAM pparam;

    pparam = (PEXTRACT_FOLDER_PARAM)GetWindowLong((HWND)lParam, GWL_USERDATA);
    if (Button_GetCheck(pparam->hOverwrite)) {
      pparam->fOverwrite = FALSE;
    } else {
      pparam->fOverwrite = TRUE;
    }
    Button_SetCheck(pparam->hOverwrite, pparam->fOverwrite);
  } else if (uMsg == WM_DESTROY) {
    SetWindowLong(hwnd, GWL_WNDPROC, (LONG)procOld);
  }
  return(CallWindowProc(procOld, hwnd, uMsg, wParam, lParam ) );
}

/*------------------------------------------------------------*/
static int CALLBACK
procSelectExtractFolder(HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData) {
  PEXTRACT_FOLDER_PARAM pparam = (PEXTRACT_FOLDER_PARAM)lpData;

  if(uMsg == BFFM_INITIALIZED){
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_CONFIRM_OVERWRITEFILES, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Overwrite Files");
    }
    pparam->hOverwrite = CreateWindow("BUTTON", m_buff, WS_CHILD | WS_VISIBLE | BS_CHECKBOX, 24, 2, 120, 40, hwnd, (HMENU)IDC_OVERWRITE, g_hInstance, NULL);
    pparam->fOverwrite = TRUE;
    Button_SetCheck(pparam->hOverwrite, pparam->fOverwrite);
    SendMessage(pparam->hOverwrite, WM_SETFONT, (WPARAM)GetStockObject(DEFAULT_GUI_FONT), FALSE);
    procOld = (WNDPROC)SetWindowLong(hwnd, GWL_WNDPROC, (LONG)procSelectExtractFolderUser);
    SetWindowLong(pparam->hOverwrite, GWL_USERDATA, (LONG)pparam);

    SendMessage(hwnd, BFFM_SETSELECTION, (WPARAM)TRUE, (LPARAM)(pparam->szInitFolder));
  }
  return 0;
}


/*------------------------------------------------------------*/
static int
decide_extension(PCSTR pszFileName, PDECFILETYPE pdt) {
  PDECFILETYPE pdt1 = pdt;
  int iRet = EXTENSION_ERROR;
  DWORD dwLenFileName = util_strlen(pszFileName, -1);
  
  while (pdt1->iext != EXTENSION_ERROR) {
    DWORD dwLen = util_strlen(pdt1->sext, -1);
    if ((dwLenFileName > dwLen) &&
        (*(pszFileName + dwLenFileName - dwLen -1) == '.') &&
        (!util_stricmp(pszFileName + dwLenFileName - dwLen, pdt1->sext))) {
      iRet = pdt1->iext;
      break;
    }
    ++pdt1;
  }
  return iRet;
}

/*------------------------------------------------------------*/
static BOOL
selectfiletype_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  int i = 0;
  HWND hcf = GetDlgItem(hwnd, IDC_COMBO_FILETYPE);
  PDECFILETYPE pdt1 = (PDECFILETYPE)lParam;

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)pdt1);

  while (pdt1->iext != EXTENSION_ERROR) {
    long l = 0;
    l = SendMessage(hcf, CB_ADDSTRING, 0, (LPARAM)(pdt1->sext));
    ++pdt1;
  }
  SendMessage(hcf, CB_SETCURSEL, 0, 0L);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
selectfiletype_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)EXTENSION_ERROR);
  return;
}

/*------------------------------------------------------------*/
static VOID
selectfiletype_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      INT i = 0;
      PDECFILETYPE pdt = (PDECFILETYPE)GetWindowLong(hwnd, GWL_USERDATA);
      i = SendDlgItemMessage(hwnd, IDC_COMBO_FILETYPE, CB_GETCURSEL, 0, 0L);
      while (i > 0) {
        ++pdt;
        --i;
      }
      EndDialog(hwnd, pdt->iext);
    }
    break;
  case IDCANCEL:
    {
      EndDialog(hwnd, (INT)EXTENSION_ERROR);
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgSelectFileType(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, selectfiletype_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, selectfiletype_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, selectfiletype_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
static DWORD
force_select_filetype(HWND hwnd, PDECFILETYPE pdt) {
  return DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_SELECT_INPUTFILETYPE), hwnd, procDlgSelectFileType, (LPARAM)pdt);
}

/*------------------------------------------------------------*/
static HRESULT
gpgezw_confirm_sign(HWND hwnd, PSTRINGLIST pslVerifiedName, PSTRINGLIST pslVerifiedFpr, PCSTR pMyFpr) {
  HRESULT hr = S_OK;
  PSTRING pstr = NULL;

  /* dqmFȂꍇ̏ */
  if (!stringlist_get_head(pslVerifiedName)) {
    CHAR m_buff[MAX_PATH + 1];

    if (FAILED(hr = string_create(&pstr))) {
      goto finish_func;
    }
    if (LoadString(g_hInstance, IDS_ALERT_SIGN_NOT_VERIFIED, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Signatures Not Verified. \n\n");
    }
    if (FAILED(hr = string_append(pstr, m_buff))) {
      goto finish_func;
    }
    if (LoadString(g_hInstance, IDS_ASKIF_CONTINUE, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Continue ?");
    }
    if (FAILED(hr = string_append(pstr, m_buff))) {
      goto finish_func;
    }
    if (MessageBox(hwnd, string_c_str(pstr), "Alert", MB_YESNO) != IDYES) {
      hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
      goto finish_func;
    }
    string_release(&pstr);
  } else {
    BOOL fMyFpr = FALSE;
    PSTRINGLIST_NODE pnode = NULL;

    for (pnode = stringlist_get_head(pslVerifiedFpr); pnode; pnode = stringlist_get_next(pnode)) {
      if (!util_strcmp(stringlist_get_string(pnode), pMyFpr)) {
        fMyFpr = TRUE;
        break;
      }
    }
    /* ȊȌ̏ꍇ͊mF */
    if (!fMyFpr) {
      PSTRINGLIST_NODE pnode = NULL;
      CHAR m_buff[MAX_PATH + 1];

      if (FAILED(hr = string_create(&pstr))) {
        goto finish_func;
      }
      if (LoadString(g_hInstance, IDS_CONFIRM_SIGN_VERIFIED, m_buff, MAX_PATH) == 0) {
        util_strcpy(m_buff, MAX_PATH, "Signatures Verified. \n\n");
      }
      if (FAILED(hr = string_append(pstr, m_buff))) {
        goto finish_func;
      }
      for (pnode = stringlist_get_head(pslVerifiedName); pnode; pnode = stringlist_get_next(pnode)) {
        if (FAILED(hr = string_append(pstr, stringlist_get_string(pnode)))) {
          goto finish_func;
        }
        if (FAILED(hr = string_append(pstr, "\n"))) {
          goto finish_func;
        }
      }
      if (FAILED(hr = string_append(pstr, "\n"))) {
        goto finish_func;
      }
      if (LoadString(g_hInstance, IDS_ASKIF_CONTINUE, m_buff, MAX_PATH) == 0) {
        util_strcpy(m_buff, MAX_PATH, "Continue ?");
      }
      if (FAILED(hr = string_append(pstr, m_buff))) {
        goto finish_func;
      }
      if (MessageBox(hwnd, string_c_str(pstr), "Confirmation", MB_YESNO) == IDNO) {
        hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
        goto finish_func;
      }
      string_release(&pstr);
    }
  }
finish_func:
  string_release(&pstr);
  return hr;
}

/*------------------------------------------------------------*/
static HRESULT
gpgezw_gpg_prepare_decrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslKeyRing, BOOL fVerify, BOOL fExtraKey, PCSTR pszFileName, PSTR *ppPassphrase, PSTR szXFile, DWORD dwMax, PDWORD pdwFileSize, PBOOL pfExtract) {
  HRESULT hr = S_OK;
  BOOL fPublicEncrypted = FALSE;
  PSTRINGLIST pslVerifiedName = NULL;
  PSTRINGLIST pslVerifiedFpr = NULL;

  if (FAILED(hr = gpgez_ispublicencrypted(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, NULL, NULL, pszFileName, &fPublicEncrypted))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (fPublicEncrypted) {
    if (!gpgezw_getMyFpr(pgpgezbase)) {
      hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
      goto finish_func;
    }
    if (gpgezw_fNeedPassphrase(pgpgezbase)) {
      if (!gpgezw_getPassphrseForPublicKeyEncryption(pgpgezbase, hwnd, 2)) {
        MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
        goto finish_func;
      }
    }
    if (pgpgezbase->pPassphrase) {
      if (!(*ppPassphrase = util_strdup(pgpgezbase->pPassphrase))) {
        hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
        goto finish_func;
      }
    }
    if (fExtraKey) {
      if (FAILED(hr = selectGpgFileKey(pgpgezbase, hwnd, pslKeyRing, NULL))) {
        if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
          hr = S_OK;
        } else {
          DEBUGERROR(0, hr);
        }
        goto finish_func;
      }
    }
    if (FAILED(hr = stringlist_create(&pslVerifiedName))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = stringlist_create(&pslVerifiedFpr))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = gpgez_isdecryptverified(pgpgezbase->pGpgExe, pgpgezbase->pTarExe, pgpgezbase->pConfDir, pslSecRing, pslKeyRing, pgpgezbase->pPassphrase, pszFileName, pslVerifiedName, pslVerifiedFpr, szXFile, dwMax, pdwFileSize, pfExtract))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }

    if (fVerify) {
      if (FAILED(hr = gpgezw_confirm_sign(hwnd, pslVerifiedName, pslVerifiedFpr, gpgezw_getMyFpr(pgpgezbase)))) {
        goto finish_func;
      }
    }
    stringlist_release(&pslVerifiedName);
    stringlist_release(&pslVerifiedFpr);
  } else {
    /* Ώ̈Í߂̃pXt[Y */
    CHAR buff[MAX_PATH + 1] = "";
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_REQUEST_FOR_DECRYPT_PASSPHRASE_SYM, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Input passphrase for symmetric encryption");
    }
    if (FAILED(hr = gpgezw_InputBoxEx(hwnd, m_buff, "Passphrase", buff, 0, MAX_PATH, INPUTBOX_WITHCANCEL | INPUTBOX_SECRET))) {
      goto finish_func;
    }
    if (!(*ppPassphrase = util_strdup(buff))) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      goto finish_func;
    }
    if (FAILED(hr = gpgez_isdecryptverified(pgpgezbase->pGpgExe, pgpgezbase->pTarExe, pgpgezbase->pConfDir, pslSecRing, pslKeyRing, *ppPassphrase, pszFileName, NULL, NULL, szXFile, MAX_PATH, pdwFileSize, pfExtract))) {
      if (HRESULT_CODE(hr) == ERROR_DECRYPTION_FAILED) {
        CHAR m_buff[MAX_PATH + 1];
        if (LoadString(g_hInstance, IDS_GPG_BADPASSPHRASE, m_buff, MAX_PATH) == 0) {
          util_strcpy(m_buff, MAX_PATH, "Incorrect passphrase.");
        }
        MessageBox(hwnd, m_buff, "Decryption Failed", MB_OK);
      }
      goto finish_func;
    }
  }
finish_func:
  stringlist_release(&pslVerifiedName);
  stringlist_release(&pslVerifiedFpr);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpdecrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, LPCSTR pDir, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fExtraKey, BOOL fVerify, PDECFILETYPE pdt) {
  HRESULT hr = S_OK;
  CHAR szDir[MAX_PATH + 1] = "";
  CHAR szXFile[MAX_PATH + 1] = "";
  LPITEMIDLIST lpid;
  BROWSEINFO bi;
  LPMALLOC pMalloc = NULL;
  BOOL fOverwrite = TRUE;
  PSTR pPassphrase = NULL;
  BOOL fExtract = TRUE;
  INT iExt = EXTENSION_ERROR;
  PSTRINGLIST pslKeyRing = NULL;
  DWORD dwSize = 0;

  if (FAILED(hr = stringlist_create(&pslKeyRing))) {
    DEBUGERROR(0, hr); 
    goto finish_func;
  }
  if (pslPubRing) {
    if (FAILED(hr = stringlist_concat(pslKeyRing, pslPubRing))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  /*
    gq gpg zip 7z tgz tar.gz tar.bz2 gz bz2 tar
  */
  iExt = decide_extension(pszFileName, pdt);
  if (iExt == EXTENSION_ERROR) {
    /* WJ[U[ɑI */
    iExt = force_select_filetype(hwnd, pdt);
    if (iExt == EXTENSION_ERROR) {
      goto finish_func;
    }
  }

  if (iExt == EXTENSION_GPG) {
    if (FAILED(hr = gpgezw_gpg_prepare_decrypt(pgpgezbase, hwnd, pslSecRing, pslKeyRing, fVerify, fExtraKey, pszFileName, &pPassphrase, szXFile, MAX_PATH, &dwSize, &fExtract))) {
      goto finish_func;
    }

    if (!fExtract) {
      /* TODO: vr[\ */
      PCSTR pext = NULL, p = szXFile;

      for (p = szXFile; *p; ++p) {
        if (*p == '.') pext = p;
      }
      if (!util_strcmp(szXFile, "_CONSOLE")) {
        /* for your eyes only */
        hr = gpgezw_decrypt_fyeo(pgpgezbase, hwnd, pslSecRing, pslKeyRing, pPassphrase, pszFileName);
        goto finish_func;
      } else if ((szXFile[0] == '\0') || (!util_stricmp(pext, ".txt"))) {
        /* eLXgt@C */
      } else if ((!util_stricmp(pext, ".bmp")) ||
                 (!util_stricmp(pext, ".gif")) ||
                 (!util_stricmp(pext, ".jpeg")) ||
                 (!util_stricmp(pext, ".jpg"))) {
        /* 摜t@C */
      }
    }
  } else if ((iExt == EXTENSION_ZIP) || (iExt == EXTENSION_7Z)) {
    BOOL fEncrypted = FALSE;

    if (FAILED(hr = gpgez_is7zencrypted(pgpgezbase->p7ZipExe, pszFileName, &fEncrypted))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (fEncrypted) {
      CHAR buff[MAX_PATH + 1] = "";
      CHAR m_buff[MAX_PATH + 1];

      if (LoadString(g_hInstance, IDS_REQUEST_FOR_DECRYPT_PASSPHRASE_SYM, m_buff, MAX_PATH) == 0) {
        util_strcpy(m_buff, MAX_PATH, "Input passphrase for symmetric encryption");
      }
      if (FAILED(hr = gpgezw_InputBoxEx(hwnd, m_buff, "Passphrase", buff, 0, MAX_PATH, INPUTBOX_WITHCANCEL | INPUTBOX_SECRET))) {
        goto finish_func;
      }
      if (!(pPassphrase = util_strdup(buff))) {
        hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
        goto finish_func;
      }
    }
    fExtract = TRUE;
  } else if ((iExt == EXTENSION_LZH) || (iExt == EXTENSION_CAB)) {
    fExtract = TRUE;
  } else if ((iExt == EXTENSION_TGZ) || (iExt == EXTENSION_TARGZ) || (iExt = EXTENSION_TARBZ2)) {
    /* untar邩ǂ[U[ɑI */

    int iRet = 0;
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_REQUEST_IF_UNTAR, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "not save one file but untar after extension ?");
    }
    iRet = MessageBox(hwnd, m_buff, "Confirmation", MB_YESNOCANCEL);
    if (iRet == IDYES) {
      fExtract = TRUE;
    } else if (iRet == IDNO) {
      fExtract = FALSE;
    } else {
      goto finish_func;
    }
  } else if (iExt == EXTENSION_TAR) {
    fExtract = TRUE;
  }

  if (fExtract) {
    /* otH_w */
    if (!pDir) {
      EXTRACT_FOLDER_PARAM param;

      param.fOverwrite = TRUE;
      param.hOverwrite = NULL;
      param.szInitFolder[0] = '\0';
      if (FAILED(hr = util_strcpy(param.szInitFolder, MAX_PATH, pszFileName))) {
        goto finish_func;
      }
      *(PathFindFileName(param.szInitFolder)) = '\0';

      if (FAILED(hr = SHGetMalloc(&pMalloc))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
      ZeroMemory(&bi, sizeof(BROWSEINFO));
      bi.hwndOwner = hwnd;
      bi.lpfn = procSelectExtractFolder;
      bi.lpszTitle = "";
      bi.ulFlags = BIF_RETURNONLYFSDIRS | BIF_NEWDIALOGSTYLE;
      bi.lParam = (LPARAM)&param;
      if (!(lpid = SHBrowseForFolder(&bi))) {
        hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
        goto finish_func;
      }
      SHGetPathFromIDList(lpid, szDir);
      /*
      if (szDir[STRLEN(szDir) - 1] != '\\')
        strcat(szDir, "\\");
       */
      pMalloc->lpVtbl->Free(pMalloc, lpid);
      pMalloc->lpVtbl->Release(pMalloc);
      fOverwrite = param.fOverwrite;
    } else {
      if (FAILED(hr = util_strcpy(szDir, MAX_PATH, pDir))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
      fOverwrite = TRUE;
    }

    switch(iExt) {
    case EXTENSION_GPG:
      {
        hr = gpgezw_decrypttar(pgpgezbase, hwnd, pslSecRing, pslKeyRing, pPassphrase, pszFileName, szDir, fOverwrite);
        if (FAILED(hr)) {
          goto finish_func;
        }
      }
      break;
    case EXTENSION_ZIP:
      /* fall thru */
    case EXTENSION_7Z:
      /* fall thru */
    case EXTENSION_LZH:
      /* fall thru */
    case EXTENSION_CAB:
      {
        if (FAILED(hr = gpgezw_7zdecrypt(pgpgezbase, hwnd, pPassphrase, pszFileName, szDir, fOverwrite))) {
          goto finish_func;
        }
      }
      break;
    case EXTENSION_TGZ:
      /* fall thru */
    case EXTENSION_TARGZ:
      {
        if (FAILED(hr = gpgezw_untar(pgpgezbase, hwnd, pszFileName, szDir, TAR_COMPRESS_GZIP, fOverwrite))) {
          goto finish_func;
        }
      }
      break;
    case EXTENSION_TBZ2:
      /* fall thru */
    case EXTENSION_TARBZ2:
      {
        if (FAILED(hr = gpgezw_untar(pgpgezbase, hwnd, pszFileName, szDir, TAR_COMPRESS_BZIP2, fOverwrite))) {
          goto finish_func;
        }
      }
      break;
    case EXTENSION_TAR:
      {
        if (FAILED(hr = gpgezw_untar(pgpgezbase, hwnd, pszFileName, szDir, TAR_COMPRESS_NONE, fOverwrite))) {
          goto finish_func;
        }
      }
      break;
    default:
      break;
    }
  }
  /* Extract Ɏs1t@C𕜍ꍇ̂ŁAU if ubN𕪂ĂAӖ͂ȂȂ */
  if (!fExtract) {
    OPENFILENAME ofn;
    CHAR szFileName[MAX_PATH + 1] = "";

    if (pDir) {
      if (FAILED(hr = util_strcat(szFileName, MAX_PATH, pDir))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
      if (*(szFileName + util_strlen(pDir, -1) - 1) != '\\') {
        if (FAILED(hr = util_strcat(szFileName, MAX_PATH, "\\"))) {
          DEBUGERROR(0, hr);
          goto finish_func;
        }
      }
    }
    if (FAILED(hr = util_strcat(szFileName, MAX_PATH, szXFile))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    ZeroMemory(&ofn, sizeof(OPENFILENAME));
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = hwnd;
    ofn.lpstrFilter = "All files(*.*)\0*.*\0\0";
    ofn.lpstrFile = szFileName;
    ofn.lpstrFileTitle = NULL;
    ofn.nMaxFile = MAX_PATH;
    ofn.nMaxFileTitle = MAX_PATH;
    ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT;
    ofn.lpstrDefExt = NULL;
    if (!GetSaveFileName(&ofn)) {
      /* G[`FbN */
      goto finish_func;
    }
    if (!util_stricmp(szFileName, pszFileName)) {
      CHAR m_buff[MAX_PATH + 1];

      if (LoadString(g_hInstance, IDS_ERROR_OVERWRITE_SOURCEFILE, m_buff, MAX_PATH) == 0) {
        util_strcpy(m_buff, MAX_PATH, "Cannot overwrite source file.");
      }
      MessageBox(hwnd, m_buff, "ERROR", MB_OK | MB_ICONERROR);
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
    switch(iExt) {
    case EXTENSION_GPG:
      {
        if (FAILED(hr = gpgezw_decrypt(pgpgezbase, hwnd, pslSecRing, pslKeyRing, pPassphrase, pszFileName, szFileName))) {
          goto finish_func;
        }
      }
      break;
    case EXTENSION_TGZ:
      /* fall thru */
    case EXTENSION_TARGZ:
      /* fall thru */
    case EXTENSION_GZ:
      /* fall thru */
    case EXTENSION_TARBZ2:
      /* fall thru */
    case EXTENSION_BZ2:
      {
        if (FAILED(hr = gpgezw_7zgunzip(pgpgezbase, hwnd, pszFileName, szFileName))) {
          goto finish_func;
        }
      }
      break;
    default:
      break;
    }
  }
  if (FAILED(hr)) goto finish_func;
finish_func:
  if (pslKeyRing) stringlist_release(&pslKeyRing);
  if (pPassphrase) util_free(pPassphrase);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpdecryptonly(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, LPCSTR pDir, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fVerify) {
  HRESULT hr = S_OK;
  CHAR szXFile[MAX_PATH + 1] = "";
  PSTR pPassphrase = NULL;
  OPENFILENAME ofn;
  CHAR szFileName[MAX_PATH + 1] = "";
  DWORD dwSize = 0;
  BOOL fExtract = FALSE;
  PSTRINGLIST pslKeyRing = NULL;

  if (FAILED(hr = stringlist_create(&pslKeyRing))) {
    goto finish_func;
  }
  if (pslPubRing) {
    if (FAILED(hr = stringlist_concat(pslKeyRing, pslPubRing))) {
      goto finish_func;
    }
  }

  /*
   *  gq gpg ̂
   */

  /* Ít@C擾 */
  if (FAILED(hr = gpgezw_gpg_prepare_decrypt(pgpgezbase, hwnd, pslSecRing, pslKeyRing, fVerify, FALSE, pszFileName, &pPassphrase, szXFile, MAX_PATH, &dwSize, &fExtract))) {
    goto finish_func;
  }

  /* for your eyes only */
  if (!util_strcmp(szXFile, "_CONSOLE")) {
    /* for your eyes only */
    hr = gpgezw_decrypt_fyeo(pgpgezbase, hwnd, pslSecRing, pslKeyRing, pPassphrase, pszFileName);
    goto finish_func;
  }

  /* o̓t@Cw */
  if (pDir) {
    if (FAILED(hr = util_strcpy(szFileName, MAX_PATH, pDir))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (*(szFileName + util_strlen(pDir, -1) - 1) != '\\') {
      if (FAILED(hr = util_strcat(szFileName, MAX_PATH, "\\"))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
  }
  if (FAILED(hr = util_strcat(szFileName, MAX_PATH, szXFile))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = "All files(*.*)\0*.*\0\0";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT;
  ofn.lpstrDefExt = NULL;
  if (!GetSaveFileName(&ofn)) {
    /* G[`FbN */
    goto finish_func;
  }
  if (!util_stricmp(szFileName, pszFileName)) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_OVERWRITE_SOURCEFILE, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Cannot overwrite source file.");
    }
    MessageBox(hwnd, m_buff, "ERROR", MB_OK | MB_ICONERROR);
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  if (FAILED(hr = gpgezw_decrypt(pgpgezbase, hwnd, NULL, NULL, pPassphrase, pszFileName, szFileName))) {
    goto finish_func;
  }

finish_func:
  if (pslKeyRing) stringlist_release(&pslKeyRing);
  if (pPassphrase) util_free(pPassphrase);
  return hr;
}

/*------------------------------------------------------------*/
#if 0
static UINT CALLBACK
OFNHookProcSaveFileSign(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch(msg){
    case WM_INITDIALOG:
      SetWindowLong(hdlg, GWL_USERDATA, lParam);
      return TRUE;
    case WM_NOTIFY:
      {
        LPOFNOTIFY pofn = (LPOFNOTIFY)lParam;
        if (pofn->hdr.code == CDN_TYPECHANGE) {
          PCSTR p;

          switch (pofn->lpOFN->nFilterIndex) {
          case 1: p = "sig"; break;
          case 2: p = "txt"; break;
          default: p = "sig"; break;
          }
          SendMessage(hdlg, CDM_SETDEFEXT, 0, (LPARAM)p);
          return TRUE;
        }
      }
     return FALSE;
  }
  return FALSE;
}
#endif

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpsign(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, DWORD dwEncType, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PENCFILETYPE pet) {
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR szFileNameSig[MAX_PATH + 1] = "";
  BOOL fAscii = FALSE;
  PSTR pfilter = NULL;
  PENCFILETYPE penc1 = pet;

  DO_DEBUG(hr, create_encrypt_filefilter(pet, &pfilter));
  if (FAILED(hr = util_strcpy(szFileNameSig, MAX_PATH, pszFileName))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (util_strlen(szFileNameSig, -1) < MAX_PATH - 4) {
    util_strcat(szFileNameSig, MAX_PATH, ".sig");
  } else {
    szFileNameSig[0] = '\0';
  }

  /* ot@Cw */
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = pfilter;
  ofn.lpstrDefExt = "sig";
  ofn.lpstrFile = szFileNameSig;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.lCustData = (DWORD)pet;
  ofn.lpfnHook = OFNHookProcSaveFile;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_ENABLEHOOK;
  if (!GetSaveFileName(&ofn)) {
    DWORD dwErr = CommDlgExtendedError();
    /* G[`FbN */
    goto finish_func;
  }

  {
    UINT i;

    for (i = 1; i < ofn.nFilterIndex; ++i) {
      ++penc1;
    }
  }

  hr = gpgezw_detach_sign(pgpgezbase, hwnd, pslSecRing, pslPubRing, pszFileName, szFileNameSig, pgpgezbase->pPassphrase, penc1->fascii);
  if (hr == E_FAIL) {
    hr = S_OK;
  }
  if (FAILED(hr)) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
finish_func:
  if (pfilter) util_free(pfilter);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpverify(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fExtraKey) {
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR szFileNameSign[MAX_PATH + 1] = "";
  CHAR szPubKeyFileName[MAX_PATH + 1] = "";
  PSTRINGLIST pslKeyRing = NULL;

  if (FAILED(hr = util_strcpy(szFileNameSign, MAX_PATH, pszFileName))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (util_strlen(szFileNameSign, -1) < MAX_PATH - 4) {
    util_strcat(szFileNameSign, MAX_PATH, ".sig");
  } else {
    szFileNameSign[0] = '\0';
  }

  /* t@Cw */
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = "sig(*.sig)\0*.sig\0ascii armored(*.txt;*.asc)\0*.txt;*.asc\0All Files(*.*)\0*.*\0\0";
  ofn.lpstrDefExt = "sig";
  ofn.lpstrFile = szFileNameSign;
  ofn.lpstrFileTitle = NULL;
  ofn.lpstrTitle = "Select sign file";
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.lpfnHook = NULL;
  ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_EXPLORER;
  if (!GetOpenFileName(&ofn)) {
    DWORD dwErr = CommDlgExtendedError();
    /* G[`FbN */
    goto finish_func;
  }

  if (fExtraKey) {
    ZeroMemory(&ofn, sizeof(OPENFILENAME));
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = hwnd;
    ofn.lpstrFilter = "gpg(*.gpg)\0*.gpg\0\0";
    ofn.lpstrDefExt = "gpg";
    ofn.lpstrFile = szPubKeyFileName;
    ofn.lpstrInitialDir = pgpgezbase->pPubKeyDir;
    ofn.lpstrTitle = "Select public key for verification";
    ofn.nMaxFile = MAX_PATH;
    ofn.nMaxFileTitle = MAX_PATH;
    ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_EXPLORER;
    if (!GetOpenFileName(&ofn)) {
      DWORD dwErr = CommDlgExtendedError();
      /* G[`FbN */
      goto finish_func;
    }
    if (FAILED(hr = stringlist_create(&pslKeyRing))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = stringlist_add_node(pslKeyRing, szPubKeyFileName))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (pslPubRing) {
      if (FAILED(hr = stringlist_concat(pslKeyRing, pslPubRing))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
  }
  hr = gpgezw_verify(pgpgezbase, hwnd, pslSecRing, pslKeyRing ? pslKeyRing : pslPubRing, pszFileName, szFileNameSign, pgpgezbase->pPassphrase);
  if (hr == E_FAIL) {
    hr = S_OK;
  }
  if (FAILED(hr)) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
finish_func:
  if (pslKeyRing) {
    stringlist_release(&pslKeyRing);
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgparmor(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, BOOL fEnArmor) {
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR szFileNameOut[MAX_PATH + 1] = "";

  /* ot@Cw */
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  if (fEnArmor) {
    ofn.lpstrTitle = "Select an ascii-armored file";
    ofn.lpstrFilter = "Ascii Armored(*.gpg;*.asc;*.txt)\0*.gpg;*.asc;*.txt\0All files(*.*)\0*.*\0\0";
    ofn.lpstrDefExt = "asc";
  } else {
    ofn.lpstrTitle = "Select a binary gpg file";
    ofn.lpstrFilter = "gpg(*.gpg)\0*.gpg\0All files(*.*)\0*.*\0\0";
    ofn.lpstrDefExt = "gpg";
  }
  ofn.lpstrFile = szFileNameOut;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT;
  if (!GetSaveFileName(&ofn)) {
    DWORD dwErr = CommDlgExtendedError();
    /* G[`FbN */
    goto finish_func;
  }

  gpgezw_armor(pgpgezbase, hwnd, pszFileName, szFileNameOut, fEnArmor);
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_listkeys(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefaultKeyRing, BOOL fSec, PSTRINGLIST pKeys) {
  return gpgez_listkeys(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing, fNoDefaultKeyRing, fSec, pKeys);
}

/*------------------------------------------------------------*/
