/**
 * gpgezw_clipboard.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/gpgezw_clipboard.c"

#include <windows.h>
#include <commctrl.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <windowsx.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

typedef struct struct_selectcharset {
  INT icharset;
  PCSTR charset;
  PCSTR nkfcs;
} SELECTCHARSET, *PSELECTCHARSET;

enum {
  CHARSET_UTF8 = 0,
  CHARSET_UTF16,
  CHARSET_SJIS,
  CHARSET_EUC_JP,
  CHARSET_JIS,
  CHARSET_DEFAULT
};

SELECTCHARSET g_selectcharset[] = {
  { CHARSET_UTF8, "UTF8", "UTF-8" },
  { CHARSET_UTF16, "UTF16", "UTF-16" },
  { CHARSET_SJIS, "SJIS", "Shift_JIS" },
  { CHARSET_EUC_JP, "EUC-JP", "EUC-JP" },
  { CHARSET_JIS, "JIS", "ISO-2022-JP" },
  { CHARSET_DEFAULT, NULL, "UTF-8" }
};

/*------------------------------------------------------------*/
typedef struct struct_buffhandle {
  PHANDLE h;
  PVOID buff;
  DWORD len;
} BUFFHANDLE, *PBUFFHANDLE;

/*------------------------------------------------------------*/

#define DO_DEBUG(hr, fn) if (FAILED((hr) = (fn))) { DEBUGERROR(0, (hr)); goto finish_func;}
#define DEBUGWIN32ERROR() { DWORD dwErr = GetLastError(); hr = HRESULT_FROM_WIN32(dwErr); DEBUGERROR(0, hr); goto finish_func; }

/*------------------------------------------------------------*/
static BOOL
selectcharset_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  int i = 0;
  HWND hcf = GetDlgItem(hwnd, IDC_COMBO_CHARSET);
  PSELECTCHARSET psc1 = (PSELECTCHARSET)lParam;

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)psc1);

  while (psc1->charset) {
    long l = 0;
    l = SendMessage(hcf, CB_ADDSTRING, 0, (LPARAM)(psc1->charset));
    ++psc1;
  }
  SendMessage(hcf, CB_SETCURSEL, 0, 0L);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
selectcharset_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)CHARSET_DEFAULT);
  return;
}

/*------------------------------------------------------------*/
static VOID
selectcharset_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      INT i = 0;
      PSELECTCHARSET psc = (PSELECTCHARSET)GetWindowLong(hwnd, GWL_USERDATA);
      i = SendDlgItemMessage(hwnd, IDC_COMBO_CHARSET, CB_GETCURSEL, 0, 0L);
      while (i > 0) {
        ++psc;
        --i;
      }
      EndDialog(hwnd, psc->icharset);
    }
    break;
  case IDCANCEL:
    {
      EndDialog(hwnd, (INT)CHARSET_DEFAULT);
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgSelectCharSet(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, selectcharset_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, selectcharset_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, selectcharset_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
static DWORD
force_select_charset(HWND hwnd, PSELECTCHARSET pcs) {
  return DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_SELECT_CHARSET), hwnd, procDlgSelectCharSet, (LPARAM)pcs);
}

/*------------------------------------------------------------*/
static DWORD WINAPI
procWriteBuff(LPVOID lpv) {
  DWORD dwErr = ERROR_SUCCESS;
  PBUFFHANDLE pbh = (PBUFFHANDLE)lpv;
  DWORD dwWritten;

  if (!WriteFile(pbh->h, pbh->buff, pbh->len, &dwWritten, NULL)) {
    dwErr = GetLastError();
  }
  CloseHandle(pbh->h);
  return dwErr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpencrypt_textbuff_to_buff(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                               HWND hwnd, PVOID pbuff, DWORD dwLen, PCSTR pCharSet, BOOL fSign, BOOL fAscii,
                               PSTR *ppbuff, PDWORD pdwLen)
{
  HRESULT hr = S_OK;
  BOOL fPublic = FALSE;
  PSTRINGLIST recipients = NULL;
  PGPGEZ_CMD pnkf = NULL;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hNkfOutP = INVALID_HANDLE_VALUE, hNkfOutC = INVALID_HANDLE_VALUE;
  HANDLE hInP = INVALID_HANDLE_VALUE, hInC = INVALID_HANDLE_VALUE;
  HANDLE hOutP = INVALID_HANDLE_VALUE, hOutC = INVALID_HANDLE_VALUE;
  HANDLE hThreadWrite = NULL;
  BUFFHANDLE bh;
  DWORD dwThreadId = 0;
  PBINBUFF pbinbuff = NULL;

  /* ߂l  */
  *ppbuff = NULL;
  *pdwLen = 0;

  /* Recipients w */
  DO_DEBUG(hr, stringlist_create(&recipients));

  /* 閧gꍇ̓pXt[YmF */
  if (fSign) {
    if (gpgezw_fNeedPassphrase(pgpgezbase)) {
      if (!gpgezw_getPassphrseForPublicKeyEncryption(pgpgezbase, hwnd, 2)) {
        MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
        goto finish_func;
      }
    }
  }

  /* Í`I */
  if (FAILED(hr = gpgezw_selectGpgEncType(pgpgezbase, hwnd, pslSecRing, pslPubRing, recipients, &fPublic))) {
    if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
      hr = S_OK;
    } else {
      DEBUGERROR(0, hr);
    }
    goto finish_func;
  }

  /* nkfp\̍쐬 */
  DO_DEBUG(hr, util_create_write_pipe(&hInP, &hInC));
  DO_DEBUG(hr, util_create_read_pipe(&hNkfOutP, &hNkfOutC));
  DO_DEBUG(hr, gpgez_nkf_create(pgpgezbase->pNkfExe, pCharSet, "UTF-8",
    hInC, hNkfOutC, INVALID_HANDLE_VALUE, &pnkf));
  CloseHandleClean(&hNkfOutC);
  CloseHandleClean(&hInC);

  /* nhtւihNkfOutPpł悤ɂj */
  {
    HANDLE h = hNkfOutP;

    if (!DuplicateHandle(GetCurrentProcess(), h, GetCurrentProcess(), &hNkfOutP, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      goto finish_func;
    }
    CloseHandle(h);
  }

  /* gpgp\̍쐬 */
  DO_DEBUG(hr, binbuff_create(&pbinbuff));
  DO_DEBUG(hr, util_create_read_pipe(&hOutP, &hOutC));
  if (fPublic) {
    DO_DEBUG(hr, gpgez_gpg_create_encrypt(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing,
      FALSE, recipients, pgpgezbase->pPassphrase, fSign, fAscii, NULL,
      hNkfOutP, hOutC, INVALID_HANDLE_VALUE, &pgpg));
  } else {
    DO_DEBUG(hr, gpgez_gpg_create_encrypt(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing,
      FALSE, NULL, stringlist_get_string(stringlist_get_head(recipients)), FALSE, fAscii, NULL,
      hNkfOutP, hOutC, INVALID_HANDLE_VALUE, &pgpg));
  }
  CloseHandleClean(&hNkfOutP);
  CloseHandleClean(&hOutC);

  /* f[^o */
  bh.h = hInP;
  bh.buff = pbuff;
  bh.len = dwLen;
  hThreadWrite = CreateThread(NULL, 0, procWriteBuff, (LPVOID)&bh, 0, &dwThreadId);
  if (!hThreadWrite) {
    DWORD dwErr = GetLastError();
    CloseHandleClean(&hInP);
    hr = HRESULT_FROM_WIN32(dwErr);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  hInP = INVALID_HANDLE_VALUE;

  /* qvZX̕Wo͓ǂݎ */
  {
    CHAR buff1[MAX_LINE + 1];
    DWORD dwRead;

    while (ReadFile(hOutP, buff1, MAX_LINE, &dwRead, NULL)) {
      DO_DEBUG(hr, binbuff_append(pbinbuff, buff1, dwRead));
    }
    CloseHandleClean(&hOutP);
  }

  WaitForSingleObject(hThreadWrite, INFINITE);
  CloseHandle(hThreadWrite);
  hThreadWrite = NULL;
  /* gpgI҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
    goto finish_func;
  }

  /* ߂lm */
  *pdwLen = binbuff_len(pbinbuff);
  *ppbuff = binbuff_detach(pbinbuff);
  binbuff_release(&pbinbuff);

finish_func:
  /* I */
  if (recipients) stringlist_release(&recipients);
  if (hThreadWrite) CloseHandle(hThreadWrite);
  if (pbinbuff) binbuff_release(&pbinbuff);
  if (pnkf) gpgez_cmd_release(&pnkf);
  if (pgpg) gpgez_gpg_release(&pgpg);
  CloseHandleClean(&hNkfOutC);
  CloseHandleClean(&hNkfOutP);
  CloseHandleClean(&hInC);
  CloseHandleClean(&hInP);
  CloseHandleClean(&hOutC);
  CloseHandleClean(&hOutP);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpencrypt_textbuff_to_clipboard(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                               HWND hwnd, PVOID pbuff, DWORD dwLen, PCSTR pCharSet, BOOL fSign)
{
  HRESULT hr = S_OK;
  PSTR buff_enc = NULL;
  DWORD buff_enc_len = 0;
  HANDLE hGlobal = NULL;
  PSTR pMem = NULL;
  BOOL fClipboardOpen = FALSE;

  DO_DEBUG(hr, gpgezw_zgpencrypt_textbuff_to_buff(pgpgezbase, pslSecRing, pslPubRing, hwnd,
    pbuff, dwLen, pCharSet, fSign, TRUE, &buff_enc, &buff_enc_len));

  hGlobal = GlobalAlloc(GHND, buff_enc_len + 1);
  if (!hGlobal) {
    DEBUGWIN32ERROR();
  }
  pMem = GlobalLock(hGlobal);
  if (!pMem) {
    DEBUGWIN32ERROR();
  }
  CopyMemory(pMem, buff_enc, buff_enc_len);
  *(pMem + buff_enc_len) = '\0';
  GlobalUnlock(hGlobal);
  if (!OpenClipboard(hwnd)) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = TRUE;
  if (!EmptyClipboard()) {
    DEBUGWIN32ERROR();
  }
  if (!SetClipboardData(CF_TEXT, hGlobal)) {
    DEBUGWIN32ERROR();
  }
  hGlobal = NULL;
  if (!CloseClipboard()) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = FALSE;

finish_func:
  /* I */
  if (fClipboardOpen) CloseClipboard();
  if (hGlobal) GlobalFree(hGlobal);
  if (buff_enc) util_free(buff_enc);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpencrypt_textbuff_to_file(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                                   HWND hwnd, PCSTR pbuff, DWORD dwLen, PCSTR pCharSet, 
                                   BOOL fSign, PENCFILETYPE pet)
{
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR szFileName[MAX_PATH + 1] = "";
  PSTR pfilter = NULL;
  PENCFILETYPE penc1 = pet;
  PSTR buff_enc = NULL;
  DWORD buff_enc_len = 0;
  DWORD dwWritten;
  HANDLE hFile = INVALID_HANDLE_VALUE;

  DO_DEBUG(hr, create_encrypt_filefilter(pet, &pfilter));
  /* ot@Cw */
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.hInstance = g_hInstance;
  ofn.lpstrFilter = pfilter;
  ofn.lpstrDefExt = "gpg";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.lCustData = (DWORD)pet;
  ofn.lpfnHook = OFNHookProcSaveFile;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_ENABLEHOOK;
  if (!GetSaveFileName(&ofn)) {
    DWORD dwErr = CommDlgExtendedError();
    /* G[`FbN */
    goto finish_func;
  }

  {
    UINT i;

    for (i = 1; i < ofn.nFilterIndex; ++i) {
      ++penc1;
    }
  }

  /* f[^obt@ňÍ */
  DO_DEBUG(hr, gpgezw_zgpencrypt_textbuff_to_buff(pgpgezbase, pslSecRing, pslPubRing, hwnd,
    (PVOID)pbuff, dwLen, pCharSet, fSign, penc1->fascii, &buff_enc, &buff_enc_len));

  /* Íf[^o */
  hFile = CreateFile(szFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFile == INVALID_HANDLE_VALUE) {
    DEBUGWIN32ERROR();
  }
  if (!WriteFile(hFile, buff_enc, buff_enc_len, &dwWritten, NULL)) {
    DEBUGWIN32ERROR();
  }
  CloseHandleClean(&hFile);

finish_func:
  if (buff_enc) util_free(buff_enc);
  CloseHandleClean(&hFile);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpencrypt_clipboard_text(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                                 HWND hwnd, BOOL fSign)
{
  HRESULT hr = S_OK;
  PSTR buff = NULL, buff_enc = NULL;
  DWORD buff_len = 0, buff_enc_len;
  HANDLE hGlobalIn = NULL;
  PSTR pMemIn = NULL;
  HANDLE hGlobalOut = NULL;
  PSTR pMemOut = NULL;
  BOOL fClipboardOpen = FALSE;

  /* buff ɃNbv{[h̒go */
  if (!OpenClipboard(hwnd)) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = TRUE;

  hGlobalIn = GetClipboardData(CF_TEXT);
  if (!hGlobalIn) {
    DEBUGWIN32ERROR();
  }
  pMemIn = GlobalLock(hGlobalIn);
  if (!pMemIn) {
    DEBUGWIN32ERROR();
  }
  buff = util_strdup(pMemIn);
  if (!buff) {
    DEBUGWIN32ERROR();
  }
  GlobalUnlock(hGlobalIn);
  pMemIn = NULL;
  hGlobalIn = NULL; /* ̎_ŁAhGlobalInWindowŝ̂ɂȂ */
  CloseClipboard();
  fClipboardOpen = FALSE;

  /* buff ÍāAbuff_enc ɕێ */
  buff_len = util_strlen(buff, -1);
  DO_DEBUG(hr, gpgezw_zgpencrypt_textbuff_to_buff(pgpgezbase, pslSecRing, pslPubRing, hwnd,
    buff, buff_len, "CP932", fSign, TRUE, &buff_enc, &buff_enc_len));

  /* buff_enc̓eNbv{[hɃRs[ */
  hGlobalOut = GlobalAlloc(GHND, buff_enc_len + 1);
  if (!hGlobalOut) {
    DEBUGWIN32ERROR();
  }
  pMemOut = GlobalLock(hGlobalOut);
  if (!pMemOut) {
    DEBUGWIN32ERROR();
  }
  *(buff_enc + buff_enc_len) = '\0';
  CopyMemory(pMemOut, buff_enc, buff_enc_len + 1);
  GlobalUnlock(hGlobalOut);
  pMemOut = NULL;
  if (!OpenClipboard(hwnd)) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = TRUE;
  if (!EmptyClipboard()) {
    DEBUGWIN32ERROR();
  }
  if (!SetClipboardData(CF_TEXT, hGlobalOut)) {
    DEBUGWIN32ERROR();
  }
  hGlobalOut = NULL;
  if (!CloseClipboard()) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = FALSE;

finish_func:
  /* I */
  if (fClipboardOpen) CloseClipboard();
  if (pMemIn) GlobalUnlock(hGlobalIn);
  if (buff) util_free(buff);
  if (buff_enc) util_free(buff_enc);
  if (pMemOut) GlobalUnlock(hGlobalOut);
  if (hGlobalOut) GlobalFree(hGlobalOut);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpdecrypt_textbuff_from_buff(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                                     PVOID pbuff, DWORD dwLen, HWND hwnd, BOOL fVerify,
                                     PCSTR pCharSet, PSTR *ppbuff, PDWORD pdwLen)
{
  HRESULT hr = S_OK;
  PSELECTCHARSET pcs;
  INT dwcset = CHARSET_DEFAULT;
  BOOL fClipboardOpen = FALSE;
  BOOL fPublic = FALSE;
  HANDLE hGlobal = NULL;
  PSTR pbuff_cb = NULL;
  DWORD len_pbuff_cb = 0;
  PSTR pPassphrase = NULL;
  PSTRINGLIST pslVerifiedName = NULL;
  PSTRINGLIST pslVerifiedFpr = NULL;
  PGPGEZ_CMD pnkf = NULL;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hGpgOutP = INVALID_HANDLE_VALUE, hGpgOutC = INVALID_HANDLE_VALUE;
  HANDLE hInP = INVALID_HANDLE_VALUE, hInC = INVALID_HANDLE_VALUE;
  HANDLE hOutP = INVALID_HANDLE_VALUE, hOutC = INVALID_HANDLE_VALUE;
  HANDLE hThreadWrite = NULL;
  BUFFHANDLE bh;
  DWORD dwThreadId = 0;
  PBINBUFF pbinbuff = NULL;

  *pdwLen = 0;
  *ppbuff = NULL;

  /* R[hI */
  dwcset = force_select_charset(hwnd, g_selectcharset);
  pcs = g_selectcharset;
  while (pcs->icharset != dwcset) ++pcs;

  /* JÍΏۈÍׂ */
  DO_DEBUG(hr, gpgez_ispublicencrypted_buff(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing,
    pbuff, dwLen, &fPublic));

  if (fPublic) {
    if (!gpgezw_getMyFpr(pgpgezbase)) {
      hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
      goto finish_func;
    }
    if (gpgezw_fNeedPassphrase(pgpgezbase)) {
      if (!gpgezw_getPassphrseForPublicKeyEncryption(pgpgezbase, hwnd, 2)) {
        MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
        goto finish_func;
      }
    }
    if (pgpgezbase->pPassphrase) {
      if (!(pPassphrase = util_strdup(pgpgezbase->pPassphrase))) {
        hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
  } else {
    /* Ώ̈Í߂̃pXt[Y */
    CHAR buff[MAX_PATH + 1] = "";
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_REQUEST_FOR_DECRYPT_PASSPHRASE_SYM, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Input passphrase for symmetric encryption");
    }
    if (FAILED(hr = gpgezw_InputBoxEx(hwnd, m_buff, "Passphrase", buff, 0, MAX_PATH, INPUTBOX_WITHCANCEL | INPUTBOX_SECRET))) {
      goto finish_func;
    }
    if (!(pPassphrase = util_strdup(buff))) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      goto finish_func;
    }
  }

  /* gpgp\̍쐬 */
  DO_DEBUG(hr, util_create_write_pipe(&hInP, &hInC));
  DO_DEBUG(hr, util_create_read_pipe(&hGpgOutP, &hGpgOutC));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(pgpgezbase->pGpgExe, pgpgezbase->pConfDir, pslSecRing, pslPubRing,
    FALSE, !fVerify, pPassphrase, hInC, hGpgOutC, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hInC);
  CloseHandleClean(&hGpgOutC);
  bh.h = hInP;
  bh.buff = pbuff;
  bh.len = dwLen;
  hThreadWrite = CreateThread(NULL, 0, procWriteBuff, (LPVOID)&bh, 0, &dwThreadId);
  if (!hThreadWrite) {
    DWORD dwErr = GetLastError();
    CloseHandleClean(&hInP);
    hr = HRESULT_FROM_WIN32(dwErr);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  hInP = INVALID_HANDLE_VALUE;

  /* nhtւihGpgOutPpł悤ɂj */
  {
    HANDLE h = hGpgOutP;

    if (!DuplicateHandle(GetCurrentProcess(), h, GetCurrentProcess(), &hGpgOutP, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      goto finish_func;
    }
    CloseHandle(h);
  }

  /* nkfp\̍쐬 */
  DO_DEBUG(hr, binbuff_create(&pbinbuff));
  DO_DEBUG(hr, util_create_read_pipe(&hOutP, &hOutC));
  DO_DEBUG(hr, gpgez_nkf_create(pgpgezbase->pNkfExe, pCharSet, pcs->nkfcs, hGpgOutP, hOutC, INVALID_HANDLE_VALUE, &pnkf));
  CloseHandleClean(&hGpgOutP);
  CloseHandleClean(&hOutC);

  /* qvZX̕Wo͓ǂݎ */
  {
    CHAR buff1[MAX_LINE + 1];
    DWORD dwRead;

    while (ReadFile(hOutP, buff1, MAX_LINE, &dwRead, NULL)) {
      DO_DEBUG(hr, binbuff_append(pbinbuff, buff1, dwRead));
    }
    CloseHandleClean(&hOutP);
  }

  WaitForSingleObject(hThreadWrite, INFINITE);
  CloseHandle(hThreadWrite);
  hThreadWrite = NULL;
  /* gpgI҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* nkfI҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(pnkf));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    goto finish_func;
  }

  /* ߂lZbg */
  *pdwLen = binbuff_len(pbinbuff);
  *ppbuff = binbuff_detach(pbinbuff);

finish_func:
  if (pslVerifiedName) stringlist_release(&pslVerifiedName);
  if (pslVerifiedFpr) stringlist_release(&pslVerifiedFpr);
  if (hThreadWrite) CloseHandle(hThreadWrite);
  if (pbinbuff) binbuff_release(&pbinbuff);
  CloseHandleClean(&hGpgOutC);
  CloseHandleClean(&hGpgOutP);
  CloseHandleClean(&hInC);
  CloseHandleClean(&hInP);
  CloseHandleClean(&hOutC);
  CloseHandleClean(&hOutP);
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (pnkf) gpgez_cmd_release(&pnkf);
  util_free(pPassphrase);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpdecrypt_textbuff_from_clipboard(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                                 HWND hwnd, BOOL fVerify, PCSTR pCharSet, PSTR *ppbuff, PDWORD pdwLen)
{
  HRESULT hr = S_OK;
  INT dwcset = CHARSET_DEFAULT;
  BOOL fClipboardOpen = FALSE;
  BOOL fPublic = FALSE;
  HANDLE hGlobal = NULL;
  PSTR pMem = NULL;
  DWORD dwMemLen = 0;

  /* ߂l */
  *ppbuff = NULL;
  *pdwLen = 0;

  /* Nbv{[h̃f[^I */
  if (!OpenClipboard(hwnd)) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = TRUE;

  hGlobal = GetClipboardData(CF_TEXT);
  if (!hGlobal) {
    DEBUGWIN32ERROR();
  }
  pMem = GlobalLock(hGlobal);
  if (!pMem) {
    DEBUGWIN32ERROR();
  }
  hGlobal = NULL; /* ̎_ŁAhGlobalWindowŝ̂ɂȂ */
  dwMemLen = util_strlen(pMem, -1);

  DO_DEBUG(hr, gpgezw_zgpdecrypt_textbuff_from_buff(pgpgezbase, pslSecRing, pslPubRing,
    pMem, dwMemLen, hwnd, fVerify, pCharSet, ppbuff, pdwLen));

finish_func:
  if (fClipboardOpen) CloseClipboard();
  if (pMem) GlobalUnlock(hGlobal);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpdecrypt_textbuff_from_file(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                                     HWND hwnd, BOOL fVerify, PCSTR pCharSet, PSTR *ppbuff, PDWORD pdwLen)
{
  HRESULT hr = S_OK;
  OPENFILENAME ofn;
  CHAR szFileName[MAX_PATH + 1] = "";
  HANDLE hFile = INVALID_HANDLE_VALUE;
  PSTR buff = NULL;
  DWORD dwLen = 0, dwRead;

  /* ߂l */
  *ppbuff = NULL;
  *pdwLen = 0;

  /* t@CI */
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.hInstance = g_hInstance;
  ofn.lpstrFilter = "gpg Files(*.gpg)\0*.gpg\0All Files(*.*)\0*.*\0\0";
  ofn.lpstrFile =szFileName;
  ofn.nMaxFile = MAX_PATH;
  ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_EXPLORER;
  if (!GetOpenFileName(&ofn)) {
    /* TODO: G[ */
    goto finish_func;
  }

  /* t@C̓ebuffɓǂݎ */
  hFile = CreateFile(szFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFile == INVALID_HANDLE_VALUE) {
    DEBUGWIN32ERROR();
  }
  dwLen = GetFileSize(hFile, NULL);
  buff = util_malloc(dwLen + 1);
  if (!buff) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!ReadFile(hFile, buff, dwLen, &dwRead, NULL)) {
    DEBUGWIN32ERROR();
  }
  CloseHandleClean(&hFile);

  /* buffÍA*ppbuffɕێ */
  DO_DEBUG(hr, gpgezw_zgpdecrypt_textbuff_from_buff(pgpgezbase, pslSecRing, pslPubRing, buff, dwLen,
    hwnd, fVerify, pCharSet, ppbuff, pdwLen));

finish_func:
  if (buff) util_free(buff);
  CloseHandleClean(&hFile);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_zgpdecrypt_clipboard_text(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing,
                               HWND hwnd, PCSTR pCharSet, BOOL fVerify)
{
  HRESULT hr = S_OK;
  PSTR buff = NULL, buff_enc = NULL;
  DWORD buff_len = 0, buff_enc_len;
  HANDLE hGlobalIn = NULL;
  PSTR pMemIn = NULL;
  HANDLE hGlobalOut = NULL;
  PSTR pMemOut = NULL;
  BOOL fClipboardOpen = FALSE;

  /* Nbv{[h̃f[^buffɃRs[ */
  if (!OpenClipboard(hwnd)) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = TRUE;

  hGlobalIn = GetClipboardData(CF_TEXT);
  if (!hGlobalIn) {
    DEBUGWIN32ERROR();
  }
  pMemIn = GlobalLock(hGlobalIn);
  if (!pMemIn) {
    DEBUGWIN32ERROR();
  }
  buff = util_strdup(pMemIn);
  if (!buff) {
    DEBUGWIN32ERROR();
  }
  GlobalUnlock(hGlobalIn);
  pMemIn = NULL;
  hGlobalIn = NULL; /* ̎_ŁAhGlobalWindowŝ̂ɂȂ */
  CloseClipboard();

  /* buffÍAbuff_encɕێ */
  buff_len = util_strlen(buff, -1);
  DO_DEBUG(hr, gpgezw_zgpdecrypt_textbuff_from_buff(pgpgezbase, pslSecRing, pslPubRing, 
    buff, buff_len, hwnd, fVerify, pCharSet, &buff_enc, &buff_enc_len));

  hGlobalOut = GlobalAlloc(GHND, buff_enc_len + 1);
  if (!hGlobalOut) {
    DEBUGWIN32ERROR();
  }
  pMemOut = GlobalLock(hGlobalOut);
  if (!pMemOut) {
    DEBUGWIN32ERROR();
  }
  *(buff_enc + buff_enc_len) = '\0';
  CopyMemory(pMemOut, buff_enc, buff_enc_len + 1);
  GlobalUnlock(hGlobalOut);
  pMemOut = NULL;
  if (!OpenClipboard(hwnd)) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = TRUE;
  if (!EmptyClipboard()) {
    DEBUGWIN32ERROR();
  }
  if (!SetClipboardData(CF_TEXT, hGlobalOut)) {
    DEBUGWIN32ERROR();
  }
  hGlobalOut = NULL;
  if (!CloseClipboard()) {
    DEBUGWIN32ERROR();
  }
  fClipboardOpen = FALSE;

finish_func:
  /* I */
  if (fClipboardOpen) CloseClipboard();
  if (pMemIn) GlobalUnlock(hGlobalIn);
  if (buff) util_free(buff);
  if (buff_enc) util_free(buff_enc);
  if (pMemOut) GlobalUnlock(hGlobalOut);
  if (hGlobalOut) GlobalFree(hGlobalOut);
  return hr;
}
/*------------------------------------------------------------*/
