/**
 * gpgezw_fyeo.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/gpgezw_fyeo.c"

#include <windows.h>
#include <commctrl.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <windowsx.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

/*------------------------------------------------------------*/
typedef struct struct_encrypt_fyeo {
  PGPGEZBASE pgpgezbase;
  HWND hwnd;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PSTRINGLIST pslRecipients;
  BOOL fExtraKey;
  BOOL fSign;
  PENCFILETYPE pet;
  HRESULT hr;
} GPGEZW_ENCRYPT_FYEO, *PGPGEZW_ENCRYPT_FYEO;

/*------------------------------------------------------------*/
typedef struct struct_gpgezw_decrypt_fyeo {
  HWND hwnd;
  PCSTR pGpgExe;
  PCSTR pGpgHome;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PCSTR pPassphrase;
  PCSTR pFile;
} GPGEZW_DECRYPTFYEOREAD, *PGPGEZW_DECRYPTFYEOREAD;

/*------------------------------------------------------------*/
#define DO_DEBUG(hr, fn) if (FAILED((hr) = (fn))) { DEBUGERROR(0, (hr)); goto finish_func;}

/*------------------------------------------------------------*/
static BOOL
encryptfyeo_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  SetWindowLong(hwnd, GWL_USERDATA, (LONG)lParam);
  SendMessage(GetDlgItem(hwnd, IDC_EDIT1), EM_LIMITTEXT, 0, 0L);
  SetFocus(GetDlgItem(hwnd, IDC_EDIT1));
  return FALSE;
}

/*------------------------------------------------------------*/
static VOID
encryptfyeo_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)ERROR_CANCELLED);
  return;
}

/*------------------------------------------------------------*/
static HRESULT
encryptfyeo_Cls_OnCommand_IDOK(PGPGEZW_ENCRYPT_FYEO pge, HWND hEdit)
{
  HRESULT hr = S_OK;
  HWND hwnd = pge->hwnd;
  OPENFILENAME ofn;
  CHAR szFileName[MAX_PATH + 1] = "";
  PSTRINGLIST recipients = NULL;
  PSTRINGLIST pslKeyRing = NULL;
  PSTR pfilter = NULL;
  PENCFILETYPE penc1 = pge->pet;
  PGPGEZ_GPG pgpg = NULL;
  PGPGEZ_CMD pnkf = NULL;
  HANDLE hNkfOutP = INVALID_HANDLE_VALUE, hNkfOutC = INVALID_HANDLE_VALUE;
  HANDLE hInP = INVALID_HANDLE_VALUE, hInC = INVALID_HANDLE_VALUE;
  HANDLE hOut = INVALID_HANDLE_VALUE;
  DWORD dwLen = 0;
  PSTR pbuff = NULL;
  DWORD dwWritten;

  if (FAILED(hr = stringlist_create(&pslKeyRing))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (pge->pslPubRing) {
    if (FAILED(hr = stringlist_concat(pslKeyRing, pge->pslPubRing))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }

  DO_DEBUG(hr, create_encrypt_filefilter(pge->pet, &pfilter));
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = pfilter;
  ofn.lpstrDefExt = "gpg";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.lCustData = (DWORD)(pge->pet);
  ofn.lpfnHook = OFNHookProcSaveFile;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_ENABLEHOOK;
  if (!GetSaveFileName(&ofn)) {
    DWORD dwErr = CommDlgExtendedError();
    /* G[`FbN */
    goto finish_func;
  }

  {
    UINT i;

    for (i = 1; i < ofn.nFilterIndex; ++i) {
      ++penc1;
    }
  }

  /* Recipients w */
  DO_DEBUG(hr, stringlist_create(&recipients));
  if (penc1->fpub) {
    if (pge->fSign) {
      if (gpgezw_fNeedPassphrase(pge->pgpgezbase)) {
        if (!gpgezw_getPassphrseForPublicKeyEncryption(pge->pgpgezbase, hwnd, 2)) {
          MessageBox(hwnd, "Incorrect Passphrase", "Error", MB_OK);
          goto finish_func;
        }
      }
    }
    if (pge->pslRecipients) {
      if (FAILED(hr = stringlist_concat(recipients, pge->pslRecipients))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    } else {
      if (pge->fExtraKey) {
        if (FAILED(hr = selectGpgFileKey(pge->pgpgezbase, hwnd, pslKeyRing, recipients))) {
          if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
            hr = S_OK;
          } else {
            DEBUGERROR(0, hr);
          }
          goto finish_func;
        }
      } else {
        if (FAILED(hr = gpgezw_selectGpgKey(pge->pgpgezbase, hwnd, pge->pslSecRing, pge->pslPubRing, recipients, pge->pet))) {
          if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
            hr = S_OK;
          } else {
            DEBUGERROR(0, hr);
          }
          goto finish_func;
        }
      }
    }
  } else {
    CHAR buff[MAX_PATH + 1] = "";
    BOOL fSfx = FALSE;
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_REQUEST_FOR_PASSPHRASE_SYM, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Input passphrase for symmetric encryption");
    }
    if (FAILED(hr = gpgezw_InputBoxEx(hwnd, m_buff, "Passphrase", buff, penc1->passlen, MAX_PATH, INPUTBOX_CONFIRM | INPUTBOX_SECRET))) {
      if (HRESULT_CODE(hr) == ERROR_CANCELLED) {
        hr = S_OK;
      } else {
        DEBUGERROR(0, hr);
      }
      goto finish_func;
    }
    DO_DEBUG(hr, stringlist_add_node(recipients, buff));
  }

  /* nkfp\̍쐬 */
  DO_DEBUG(hr, util_create_write_pipe(&hInP, &hInC));
  DO_DEBUG(hr, util_create_read_pipe(&hNkfOutP, &hNkfOutC));
  DO_DEBUG(hr, gpgez_nkf_create(pge->pgpgezbase->pNkfExe, "CP932", "UTF-8", hInC, hNkfOutC, INVALID_HANDLE_VALUE, &pnkf));
  CloseHandleClean(&hNkfOutC);
  CloseHandleClean(&hInC);

  /* nhtւihNkfOutPpł悤ɂj */
  {
    HANDLE h = hNkfOutP;

    if (!DuplicateHandle(GetCurrentProcess(), h, GetCurrentProcess(), &hNkfOutP, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      goto finish_func;
    }
    CloseHandle(h);
  }

  /* Íp\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_write(&hOut, szFileName));
  switch(penc1->filetype) {
  case FILETYPE_GPGENCPUBLIC:
    /* fall thru */
  case FILETYPE_GPGENCPUBLIC_ASCII:
    DO_DEBUG(hr, gpgez_gpg_create_encrypt(pge->pgpgezbase->pGpgExe, pge->pgpgezbase->pConfDir,
      pge->pslSecRing, pslKeyRing, FALSE, recipients, pge->pgpgezbase->pPassphrase,
      TRUE, penc1->fascii, "_CONSOLE", hNkfOutP, hOut, INVALID_HANDLE_VALUE, &pgpg));
    break;
  case FILETYPE_GPGENCSYMMETRIC:
    /* fall thru */
  case FILETYPE_GPGENCSYMMETRIC_ASCII:
    DO_DEBUG(hr, gpgez_gpg_create_encrypt(pge->pgpgezbase->pGpgExe, pge->pgpgezbase->pConfDir,
      pge->pslSecRing, pslKeyRing, FALSE, NULL, stringlist_get_string(stringlist_get_head(recipients)),
      FALSE, penc1->fascii, "_CONSOLE", hNkfOutP, hOut, INVALID_HANDLE_VALUE, &pgpg));
    break;
  }
  CloseHandleClean(&hNkfOutP);
  CloseHandleClean(&hOut);

  /* obt@o */
  dwLen = Edit_GetTextLength(hEdit);
  if (FAILED(pbuff = util_malloc(dwLen + 2))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  Edit_GetText(hEdit, pbuff, dwLen + 1);
  WriteFile(hInP, pbuff, dwLen, &dwWritten, NULL);
  CloseHandleClean(&hInP);
  util_free(pbuff);
  pbuff = NULL;

  /* nkf I҂ */
  DO_DEBUG(hr, gpgez_cmd_finish(pnkf));

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_ENCRYPTION_FAILED);
  }
  if (SUCCEEDED(hr)) {
    MessageBox(hwnd, "Encryption Completed.", "Message", MB_OK);
  }
finish_func:
  CloseHandleClean(&hNkfOutC);
  CloseHandleClean(&hNkfOutP);
  CloseHandleClean(&hInC);
  CloseHandleClean(&hInP);
  CloseHandleClean(&hOut);
  if (pnkf) gpgez_cmd_release(&pnkf);
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (pbuff) util_free(pbuff);
  if (recipients) stringlist_release(&recipients);
  if (pslKeyRing) stringlist_release(&pslKeyRing);
  if (pfilter) util_free(pfilter);
  return hr;
}

/*------------------------------------------------------------*/
static VOID
encryptfyeo_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      HRESULT hr = S_OK;
      HWND hEdit1 = GetDlgItem(hwnd, IDC_EDIT1);
      PGPGEZW_ENCRYPT_FYEO pge = (PGPGEZW_ENCRYPT_FYEO)GetWindowLong(hwnd, GWL_USERDATA);

      hr = encryptfyeo_Cls_OnCommand_IDOK(pge, hEdit1);
      if (FAILED(hr)) {
        /* G[\ */
      }
      pge->hr = hr;
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDCANCEL:
    {
      EndDialog(hwnd, (INT)ERROR_CANCELLED);
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgEncryptFyeo(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, encryptfyeo_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, encryptfyeo_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, encryptfyeo_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
__declspec(dllexport) HRESULT
gpgezw_encrypt_fyeo(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslRecipients, BOOL fExtraKey, BOOL fSign, PENCFILETYPE pet) {
  HRESULT hr = S_OK;
  GPGEZW_ENCRYPT_FYEO ge = { NULL, NULL, NULL, NULL, NULL, FALSE, FALSE, NULL, S_OK };
  int iRet;
  HANDLE hProcess = NULL;

  ge.hwnd = hwnd;
  ge.pgpgezbase = pgpgezbase;
  ge.pslSecRing = pslSecRing;
  ge.pslPubRing = pslPubRing;
  ge.pslRecipients = pslRecipients;
  ge.fExtraKey = fExtraKey;
  ge.fSign = fSign;
  ge.pet = pet;
  ge.hr = S_OK;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDlgEncryptFyeo, (LPARAM)&ge);
  if (iRet != ERROR_SUCCESS) {
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
  } else {
    hr = ge.hr;
  }
  return hr;
}

/*------------------------------------------------------------*/
static BOOL
decryptfyeo_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  PGPGEZW_DECRYPTFYEOREAD pgdf = (PGPGEZW_DECRYPTFYEOREAD)lParam;
  PGPGEZ_GPG pgpg = NULL;
  HANDLE hIn = INVALID_HANDLE_VALUE;
  HANDLE hOutP = INVALID_HANDLE_VALUE, hOutC = INVALID_HANDLE_VALUE;
  PBINBUFF pbinbuff = NULL;
  PSTR pbuff = NULL;
  DWORD dwLen = 0;
  PSTR psjis = NULL, psjis2 = NULL;
  HRESULT hr = S_OK;

  /*  */
  DO_DEBUG(hr, binbuff_create(&pbinbuff));

  /* p\̍쐬 */
  DO_DEBUG(hr, gpgez_open_file_inherit_read(&hIn, pgdf->pFile));
  DO_DEBUG(hr, util_create_read_pipe(&hOutP, &hOutC));
  DO_DEBUG(hr, gpgez_gpg_create_decrypt(pgdf->pGpgExe, pgdf->pGpgHome, pgdf->pslSecRing, pgdf->pslPubRing,
    FALSE, TRUE, pgdf->pPassphrase, hIn, hOutC, INVALID_HANDLE_VALUE, &pgpg));
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOutC);

  /* qvZX̕Wo͓ǂݎ */
  {
    CHAR buff1[MAX_LINE + 1];
    DWORD dwRead;

    while (ReadFile(hOutP, buff1, MAX_LINE, &dwRead, NULL)) {
      DO_DEBUG(hr, binbuff_append(pbinbuff, buff1, dwRead));
    }
    CloseHandleClean(&hOutP);
  }

  /* gpg I҂ */
  DO_DEBUG(hr, gpgez_gpg_finish(pgpg));

  /* qvZX̖߂l`FbN */
  if (gpgez_gpg_get_exitcode(pgpg) != 0) {
    hr = HRESULT_FROM_WIN32(ERROR_DECRYPTION_FAILED);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  gpgez_gpg_release(&pgpg);

  /* obt@tւ */
  dwLen = binbuff_len(pbinbuff);
  binbuff_append(pbinbuff, "\0", 1);
  pbuff = binbuff_detach(pbinbuff);
  binbuff_release(&pbinbuff);

  /* R[hϊ */
  if (FAILED(hr = util_utf8tosjis(pbuff, &psjis))) {
    goto finish_func;
  }
  util_free(pbuff);
  pbuff = NULL;

  /* sR[hϊ */
  if (FAILED(hr = util_lftocrlf(psjis, &psjis2))) {
    goto finish_func;
  }
  util_free(psjis);
  psjis = NULL;

  /* \ */
  SendMessage(GetDlgItem(hwnd, IDC_EDIT1), EM_LIMITTEXT, 0, 0L);
  SetWindowText(GetDlgItem(hwnd, IDC_EDIT1), psjis2);
  util_free(psjis2);
  psjis2 = NULL;

finish_func:
  CloseHandleClean(&hIn);
  CloseHandleClean(&hOutP);
  CloseHandleClean(&hOutC);
  if (pgpg) gpgez_gpg_release(&pgpg);
  if (pbinbuff) binbuff_release(&pbinbuff);
  if (psjis2) util_free(psjis2);
  if (psjis) util_free(psjis);
  if (pbuff) util_free(pbuff);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
decryptfyeo_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)ERROR_SUCCESS);
  return;
}

/*------------------------------------------------------------*/
static VOID
decryptfyeo_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgDecryptFyeo(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, decryptfyeo_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, decryptfyeo_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, decryptfyeo_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_decrypt_fyeo(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFileName) {
  GPGEZW_DECRYPTFYEOREAD gdf = { NULL, NULL, NULL, NULL, NULL, NULL, NULL };
  int iRet;
  HRESULT hr = S_OK;

  gdf.hwnd = hwnd;
  gdf.pGpgExe = pgpgezbase->pGpgExe;
  gdf.pGpgHome = pgpgezbase->pConfDir;
  gdf.pslSecRing = pslSecRing;
  gdf.pslPubRing = pslPubRing;
  gdf.pPassphrase = pPassphrase;
  gdf.pFile = pFileName;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_FYEOREAD), hwnd, procDlgDecryptFyeo, (LPARAM)&gdf);
  if (iRet == ERROR_SUCCESS) {
    hr = S_OK;
  } else {
    hr = HRESULT_FROM_WIN32(iRet);
  }
  return hr;
}

/*------------------------------------------------------------*/

