/**
 * inputbox_ex.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/inputbox_ex.c"

#include <windows.h>
#include <windowsx.h>

#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

typedef struct struct_inputbox_param {
  LPCSTR pPrompt;
  LPCSTR pTitle;
  PSTR pBuff;
  UINT uMin;
  UINT uMax;
  DWORD dwStyle;
  DWORD dwErr;
} INPUTBOX_PARAM, *PINPUTBOX_PARAM;

#define MAXINPUT 256

extern HINSTANCE g_hInstance;

/*------------------------------------------------------------*/
static BOOL
inputbox_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  PINPUTBOX_PARAM pib;

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)lParam);
  pib = (PINPUTBOX_PARAM)lParam;

  SetWindowText(hwnd, pib->pTitle);
  Static_SetText(GetDlgItem(hwnd, IDC_STATIC_LABEL1), pib->pPrompt);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_INPUT), pib->uMax);
  if (pib->dwStyle & INPUTBOX_CONFIRM) {
    Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_INPUT2), pib->uMax);
  }
  if (pib->dwStyle & INPUTBOX_SECRET) {
    Edit_SetPasswordChar(GetDlgItem(hwnd, IDC_EDIT_INPUT), '*');
    if (pib->dwStyle & INPUTBOX_CONFIRM) {
      Edit_SetPasswordChar(GetDlgItem(hwnd, IDC_EDIT_INPUT2), '*');
    }
  } else {
    Edit_SetPasswordChar(GetDlgItem(hwnd, IDC_EDIT_INPUT), 0);
    if (pib->dwStyle & INPUTBOX_CONFIRM) {
      Edit_SetPasswordChar(GetDlgItem(hwnd, IDC_EDIT_INPUT2), 0);
    }
  }
  if (!(pib->dwStyle & INPUTBOX_WITHCANCEL)) {
    ShowWindow(GetDlgItem(hwnd, IDCANCEL), SW_HIDE);
  }
  Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_INPUT), pib->pBuff);

  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
inputbox_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)ERROR_CANCELLED);
  return;
}

/*------------------------------------------------------------*/
static VOID
inputbox_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      CHAR buff[MAXINPUT + 1];
      PINPUTBOX_PARAM pib;

      pib = (PINPUTBOX_PARAM)GetWindowLong(hwnd, GWL_USERDATA);
      GetWindowText(GetDlgItem(hwnd, IDC_EDIT_INPUT), pib->pBuff, pib->uMax);
      if (pib->dwStyle & INPUTBOX_CONFIRM) {
        GetWindowText(GetDlgItem(hwnd, IDC_EDIT_INPUT2), buff, pib->uMax);
        if (util_strcmp(pib->pBuff, buff)) {
          CHAR m_buff[MAX_PATH + 1] = "";
          
          if (LoadString(g_hInstance, IDS_ERROR_INPUTSTRINGNOMATCH, m_buff, MAX_PATH) == 0) {
            util_strcpy(m_buff, MAX_PATH, "Input Missmatch");
          }
          MessageBox(hwnd, m_buff, "Error", MB_OK | MB_ICONEXCLAMATION);
          break;
        }
        if (util_strlen(pib->pBuff, pib->uMax) < (INT)pib->uMin) {
          CHAR m_buff[MAX_PATH + 1] = "";

          if (LoadString(g_hInstance, IDS_ERROR_INPUTSTRINGTOOSHORT, m_buff, MAX_PATH) == 0) {
            util_strcpy(m_buff, MAX_PATH, "Input too short");
          }
          MessageBox(hwnd, m_buff, "Error", MB_OK | MB_ICONEXCLAMATION);
          break;
        }
      }
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDCANCEL:
    {
      EndDialog(hwnd, (INT)ERROR_CANCELLED);
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procInputBoxEx(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, inputbox_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, inputbox_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, inputbox_Cls_OnCommand);
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_InputBoxEx(HWND hwnd, LPCSTR pPrompt, LPCSTR pTitle, LPSTR pBuff, UINT uMin, UINT uMax, DWORD dwStyle) {
  INPUTBOX_PARAM ip;
  INT iRet;

  ip.pPrompt = pPrompt;
  ip.pTitle = pTitle;
  ip.pBuff = pBuff;
  ip.uMin = uMin;
  ip.uMax = (uMax > MAXINPUT) ? MAXINPUT : uMax;
  ip.dwStyle = dwStyle;
  ip.dwErr = ERROR_SUCCESS;

  iRet = DialogBoxParam(g_hInstance, (dwStyle & INPUTBOX_CONFIRM) ? MAKEINTRESOURCE(IDD_DIALOG_INPUTBOX_CONFIRM) : MAKEINTRESOURCE(IDD_DIALOG_INPUTBOX), hwnd, procInputBoxEx, (LPARAM) &ip);
  return HRESULT_FROM_WIN32((DWORD)iRet);
}

/*------------------------------------------------------------*/
