/**
 * selectGpgEncType.c
 * Copyright (C) 2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/selectEncType.c"

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

typedef struct struct_selectenctype_params {
  PGPGEZBASE pgpgezbase;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PSTRINGLIST pslSelected;
  BOOL fPublic;
} SELECTENCTYPE_PARAMS, *PSELECTENCTYPE_PARAMS;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

/*------------------------------------------------------------*/
/*
fPub܂fSecfpr:ȂszFprZbg
fPub܂fSecuid:ȂszUidZbgĂăXgr[ɒǉ
ȊOȂfPubAfSecAszFprAszUidZbg
@Apub:ȂfPubAsec:ȂfSecZbg
*/
static HRESULT
selectenctype_show(HWND hwnd) {
  HRESULT hr = S_OK;
  PSTRINGLIST pslkey = NULL;
  PSTRINGLIST_NODE pnode;
  PSELECTENCTYPE_PARAMS pset_params = NULL;
  HWND hwndList;
  LV_ITEM item;
  INT iLine = 0;

  if (FAILED(hr = stringlist_create(&pslkey))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  pset_params = (PSELECTENCTYPE_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
  hwndList = GetDlgItem(hwnd, IDC_LIST_KEYS);
  if (!ListView_DeleteAllItems(hwndList)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  if (FAILED(hr = gpgez_listkeys(pset_params->pgpgezbase->pGpgExe, pset_params->pgpgezbase->pConfDir, pset_params->pslSecRing, pset_params->pslPubRing, TRUE, FALSE, pslkey))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  for (pnode = stringlist_get_head(pslkey); pnode; pnode = stringlist_get_next(pnode)) {
    /* ListView ɒǉ */
    CHAR buff[MAX_LINE + 1];
    PSTR p = NULL;

    if (FAILED(hr = util_strcpy(buff, MAX_LINE, stringlist_get_string(pnode)))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    p = buff + (util_strchr(buff, ':') - buff);
    if (p) *p = '\0';
    item.mask = LVIF_TEXT;
    if (p) {
      item.pszText = p + 1;
    } else {
      item.pszText = "";
    }
    item.iItem = iLine;
    item.iSubItem = 0;
    ListView_InsertItem(hwndList, &item);

    {
      CHAR buff2[MAX_LINE + 3] = "   ";

      if (FAILED(hr = decorate_fpr(buff2, MAX_LINE + 3, buff, gpgezw_getMyFpr(pset_params->pgpgezbase)))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }

      item.pszText = buff2;
      item.iItem = iLine;
      item.iSubItem = 1;
      ListView_SetItem(hwndList, &item);
    }

    ++iLine;
  }

finish_func:
  if (pslkey) stringlist_release(&pslkey);
  return hr;
}

/*------------------------------------------------------------*/
static BOOL
selectenctype_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  LV_COLUMN lvcol;

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)lParam);
  EnableWindow(GetDlgItem(hwnd, IDC_LIST_KEYS), TRUE);
/*
  EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_KEYMANAGER), TRUE);
*/
  Button_SetCheck(GetDlgItem(hwnd, IDC_ENCPUB), TRUE);
  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_INPUT), FALSE);
  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_INPUT2), FALSE);

  lvcol.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
  lvcol.fmt = LVCFMT_LEFT;
  lvcol.cx = 200;
  lvcol.pszText = "Name";
  lvcol.iSubItem = 0;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 0, &lvcol);

  lvcol.cx = 300;
  lvcol.pszText = "Fingerprint";
  lvcol.iSubItem = 1;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 1, &lvcol);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
selectenctype_Cls_OnShowWindow(HWND hwnd, BOOL fShow, UINT status) {
  selectenctype_show(hwnd);
  return;
}

/*------------------------------------------------------------*/
static VOID
selectenctype_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)ERROR_CANCELLED);
  return;
}

/*------------------------------------------------------------*/
static VOID
selectenctype_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      PSELECTENCTYPE_PARAMS pset;
      HRESULT hr = S_OK;

      pset = (PSELECTENCTYPE_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);

      if (Button_GetCheck(GetDlgItem(hwnd, IDC_ENCPUB))) {
        INT iNext = -1;
        HWND hwndKeyList;
        LVITEM item;
        CHAR szLine[MAX_LINE + 1];

        item.mask = LVIF_TEXT;
        item.pszText = szLine;
        item.cchTextMax = MAX_LINE;
        hwndKeyList = GetDlgItem(hwnd, IDC_LIST_KEYS);
        while (TRUE) {
          CHAR buff2[MAX_LINE + 3];

          iNext = ListView_GetNextItem(hwndKeyList, iNext, LVNI_ALL | LVNI_SELECTED);
          if (iNext == -1) break;
          item.iItem = iNext;
          item.iSubItem = 1;
          ListView_GetItem(hwndKeyList, &item);
          if (FAILED(hr = undecorate_fpr(buff2, MAX_LINE + 3, item.pszText))) {
            DEBUGERROR(0, hr);
            EndDialog(hwnd, (INT)HRESULT_CODE(hr));
          }
          if (FAILED(hr = stringlist_add_node(pset->pslSelected, buff2))){
            DEBUGERROR(0, hr);
            EndDialog(hwnd, (INT)HRESULT_CODE(hr));
            break;
          }
        }
        if (stringlist_get_len(pset->pslSelected) == 0) {
          CHAR m_buff[MAX_PATH + 1];

          if (LoadString(g_hInstance, IDS_ERROR_NORECIPIENTSSELECTED, m_buff, MAX_PATH) == 0) {
            util_strcpy(m_buff, MAX_PATH, "No Recipients Selected.\nAborting");
          }
          if (IDNO == MessageBox(hwnd, m_buff, "Confirmation", MB_YESNO | MB_ICONQUESTION)) {
            break;
          }
        }
        pset->fPublic = TRUE;
        EndDialog(hwnd, (INT)ERROR_SUCCESS);
      } else {
        CHAR buff1[MAX_PATH + 1] = "", buff2[MAX_PATH + 1] = "";
        
        Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_INPUT), buff1, MAX_PATH);
        Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_INPUT2), buff2, MAX_PATH);
        if (util_strcmp(buff1, buff2)) {
          MessageBox(hwnd, "Input Miss Match", "Error", MB_OK);
          break;
        } else {
          if (FAILED(hr = stringlist_add_node(pset->pslSelected, buff2))) {
            DEBUGERROR(0, hr);
            EndDialog(hwnd, (INT)HRESULT_CODE(hr));
            break;
          }
          pset->fPublic = FALSE;
          EndDialog(hwnd, (INT)ERROR_SUCCESS);
        }
      }
    }
    break;
  case IDCANCEL:
    {
      EndDialog(hwnd, (INT)ERROR_CANCELLED);
    }
    break;
  case IDC_ENCPUB:
    {
      EnableWindow(GetDlgItem(hwnd, IDC_LIST_KEYS), TRUE);
    /*
      EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_KEYMANAGER), TRUE);
    */
      EnableWindow(GetDlgItem(hwnd, IDC_EDIT_INPUT), FALSE);
      EnableWindow(GetDlgItem(hwnd, IDC_EDIT_INPUT2), FALSE);
    }
    break;
  case IDC_ENCSYM:
    {
      EnableWindow(GetDlgItem(hwnd, IDC_LIST_KEYS), FALSE);
    /*
      EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_KEYMANAGER), FALSE);
    */
      EnableWindow(GetDlgItem(hwnd, IDC_EDIT_INPUT), TRUE);
      EnableWindow(GetDlgItem(hwnd, IDC_EDIT_INPUT2), TRUE);
    }
    break;
/*
  case IDC_BUTTON_KEYMANAGER:
    {
      DWORD hr = S_OK;
      PSELECTKEY_PARAMS psk_params;

      psk_params = (PSELECTKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
      hr = gpgezw_editGpgKey(psk_params->pgpgezbase, hwnd, psk_params->pslSecRing, psk_params->pslPubRing, FALSE, psk_params->dwKeyType);
      if (SUCCEEDED(hr)) {
        selectkey_show(hwnd);
      }
    }
    break;
*/
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgSelectEncType(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, selectenctype_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_SHOWWINDOW, selectenctype_Cls_OnShowWindow);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, selectenctype_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, selectenctype_Cls_OnCommand);
  case WM_NOTIFY:
    if (wParam == IDC_LIST_KEYS) {
      LPNMHDR lpnmh = (LPNMHDR)lParam;
      if (lpnmh->code == NM_DBLCLK) {
        SendMessage(hDlg, WM_COMMAND, (WPARAM)IDOK, (LPARAM)NULL);
        return TRUE;
      }
    }
    return FALSE;
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_selectGpgEncType(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslSelected, PBOOL pfPublic) {
  HRESULT hr = S_OK;
  INT iRet = 0;
  SELECTENCTYPE_PARAMS set_params;

  if (!pslSelected) {
    return HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
  }
  if (gpgezw_getMyFpr(pgpgezbase)) {
    set_params.pgpgezbase = pgpgezbase;
    set_params.pslSecRing = pslSecRing;
    set_params.pslPubRing = pslPubRing;
    set_params.pslSelected = pslSelected;
    set_params.fPublic = FALSE;
    iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_SELENCTYPE), hwnd, procDlgSelectEncType, (LPARAM)&set_params);
    hr = HRESULT_FROM_WIN32((DWORD)iRet);
    *pfPublic = set_params.fPublic;
  } else {
    CHAR buff[MAX_PATH + 1] = "";
    CHAR m_buff[MAX_PATH + 1] = "";

    if (LoadString(g_hInstance, IDS_REQUEST_FOR_PASSPHRASE_SYM, m_buff, MAX_PATH) == 0) {
      util_strcpy(m_buff, MAX_PATH, "Input passphrase for symmetric encryption");
    }
    hr = gpgezw_InputBoxEx(hwnd, m_buff, "Passphrase(more than 8 chars)", buff, 8, MAX_PATH, INPUTBOX_CONFIRM | INPUTBOX_SECRET);
    *pfPublic = FALSE;
  }
  return hr;
}

/*------------------------------------------------------------*/

