/**
 * selectKey.c
 * Copyright (C) 2006-2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "gpgezw/selectKey.c"

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include "../include/util.h"
#include "../include/gpgez.h"
#include "../include/gpgezw.h"
#include "gpgezw_util.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

typedef struct selectkey_params {
  PGPGEZBASE pgpgezbase;
  PSTRINGLIST pslSecRing;
  PSTRINGLIST pslPubRing;
  PSTRINGLIST pslSelected;
  PENCFILETYPE pet;
} SELECTKEY_PARAMS, *PSELECTKEY_PARAMS;

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024

/*------------------------------------------------------------*/
/*
fPub܂fSecfpr:ȂszFprZbg
fPub܂fSecuid:ȂszUidZbgĂăXgr[ɒǉ
ȊOȂfPubAfSecAszFprAszUidZbg
@Apub:ȂfPubAsec:ȂfSecZbg
*/
static HRESULT
selectkey_show(HWND hwnd) {
  HRESULT hr = S_OK;
  PSTRINGLIST pslkey = NULL;
  PSTRINGLIST_NODE pnode;
  PSELECTKEY_PARAMS psk_params = NULL;
  HWND hwndList;
  LV_ITEM item;
  INT iLine = 0;

  if (FAILED(hr = stringlist_create(&pslkey))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  psk_params = (PSELECTKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
  hwndList = GetDlgItem(hwnd, IDC_LIST_KEYS);
  if (!ListView_DeleteAllItems(hwndList)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  if (FAILED(hr = gpgez_listkeys(psk_params->pgpgezbase->pGpgExe, psk_params->pgpgezbase->pConfDir, psk_params->pslSecRing, psk_params->pslPubRing, TRUE, FALSE, pslkey))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  for (pnode = stringlist_get_head(pslkey); pnode; pnode = stringlist_get_next(pnode)) {
    /* ListView ɒǉ */
    CHAR buff[MAX_LINE + 1];
    PSTR p = NULL;

    if (FAILED(hr = util_strcpy(buff, MAX_LINE, stringlist_get_string(pnode)))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    p = buff + (util_strchr(buff, ':') - buff);
    if (p) *p = '\0';
    item.mask = LVIF_TEXT;
    if (p) {
      item.pszText = p + 1;
    } else {
      item.pszText = "";
    }
    item.iItem = iLine;
    item.iSubItem = 0;
    ListView_InsertItem(hwndList, &item);

    {
      CHAR buff2[MAX_LINE + 3] = "   ";

      if (FAILED(hr = decorate_fpr(buff2, MAX_LINE + 3, buff, gpgezw_getMyFpr(psk_params->pgpgezbase)))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }

      item.pszText = buff2;
      item.iItem = iLine;
      item.iSubItem = 1;
      ListView_SetItem(hwndList, &item);
    }

    ++iLine;
  }

finish_func:
  if (pslkey) stringlist_release(&pslkey);
  return hr;
}

/*------------------------------------------------------------*/
static BOOL
selectkey_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  LV_COLUMN lvcol;

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)lParam);
  EnableWindow(GetDlgItem(hwnd, IDC_LIST_KEYS), TRUE);
  EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_KEYMANAGER), TRUE);

  lvcol.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
  lvcol.fmt = LVCFMT_LEFT;
  lvcol.cx = 200;
  lvcol.pszText = "Name";
  lvcol.iSubItem = 0;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 0, &lvcol);

  lvcol.cx = 300;
  lvcol.pszText = "Fingerprint";
  lvcol.iSubItem = 1;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 1, &lvcol);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
selectkey_Cls_OnShowWindow(HWND hwnd, BOOL fShow, UINT status) {
  selectkey_show(hwnd);
  return;
}

/*------------------------------------------------------------*/
static VOID
selectkey_Cls_OnClose(HWND hwnd) {
  PSELECTKEY_PARAMS psk_params;

  psk_params = (PSELECTKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
  EndDialog(hwnd, (INT)ERROR_CANCELLED);
  return;
}

/*------------------------------------------------------------*/
static VOID
selectkey_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      PSELECTKEY_PARAMS pjsk;
      INT iNext = -1;
      HWND hwndKeyList;
      LVITEM item;
      CHAR szLine[MAX_LINE + 1];
      DWORD hr = S_OK;

      pjsk = (PSELECTKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);

      item.mask = LVIF_TEXT;
      item.pszText = szLine;
      item.cchTextMax = MAX_LINE;
      hwndKeyList = GetDlgItem(hwnd, IDC_LIST_KEYS);
      while (1) {
        CHAR buff2[MAX_LINE + 3];

        iNext = ListView_GetNextItem(hwndKeyList, iNext, LVNI_ALL | LVNI_SELECTED);
        if (iNext == -1) break;
        item.iItem = iNext;
        item.iSubItem = 1;
        ListView_GetItem(hwndKeyList, &item);
        if (FAILED(hr = undecorate_fpr(buff2, MAX_LINE + 3, item.pszText))) {
          DEBUGERROR(0, hr);
          EndDialog(hwnd, (INT)HRESULT_CODE(hr));
        }
        if (FAILED(hr = stringlist_add_node(pjsk->pslSelected, buff2))){
          DEBUGERROR(0, hr);
          EndDialog(hwnd, (INT)HRESULT_CODE(hr));
          break;
        }
      }
      if (stringlist_get_len(pjsk->pslSelected) == 0) {
        CHAR m_buff[MAX_PATH + 1];

        if (LoadString(g_hInstance, IDS_ERROR_NORECIPIENTSSELECTED, m_buff, MAX_PATH) == 0) {
          util_strcpy(m_buff, MAX_PATH, "No Recipients Selected.\nAborting");
        }
        if (IDNO == MessageBox(hwnd, m_buff, "Confirmation", MB_YESNO | MB_ICONQUESTION)) {
          break;
        }
      }
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDCANCEL:
    {
      PSELECTKEY_PARAMS psk_params;

      psk_params = (PSELECTKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
      EndDialog(hwnd, (INT)ERROR_CANCELLED);
    }
    break;
  case IDC_BUTTON_KEYMANAGER:
    {
      DWORD hr = S_OK;
      PSELECTKEY_PARAMS psk_params;

      psk_params = (PSELECTKEY_PARAMS)GetWindowLong(hwnd, GWL_USERDATA);
      hr = gpgezw_editGpgKey(psk_params->pgpgezbase, hwnd, psk_params->pslSecRing, psk_params->pslPubRing, FALSE, psk_params->pet);
      if (SUCCEEDED(hr)) {
        selectkey_show(hwnd);
      }
    }
    break;
  }
  return;
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgSelectKey(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_DLG_MSG(hDlg, WM_INITDIALOG, selectkey_Cls_OnInitDialog);
    HANDLE_DLG_MSG(hDlg, WM_SHOWWINDOW, selectkey_Cls_OnShowWindow);
    HANDLE_DLG_MSG(hDlg, WM_CLOSE, selectkey_Cls_OnClose);
    HANDLE_DLG_MSG(hDlg, WM_COMMAND, selectkey_Cls_OnCommand);
  case WM_NOTIFY:
    if (wParam == IDC_LIST_KEYS) {
      LPNMHDR lpnmh = (LPNMHDR)lParam;
      if (lpnmh->code == NM_DBLCLK) {
        SendMessage(hDlg, WM_COMMAND, (WPARAM)IDOK, (LPARAM)NULL);
        return TRUE;
      }
    }
    return FALSE;
  }
  return FALSE;
}

/*------------------------------------------------------------*/
DllExport HRESULT
gpgezw_selectGpgKey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslSelected, PENCFILETYPE pet) {
  INT iRet = 0;
  SELECTKEY_PARAMS sk_params;

  sk_params.pgpgezbase = pgpgezbase; 
  sk_params.pslSecRing = pslSecRing;
  sk_params.pslPubRing = pslPubRing;
  sk_params.pslSelected = pslSelected;
  sk_params.pet = pet;
  iRet = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_SELKEY), hwnd, procDlgSelectKey, (LPARAM)&sk_params);
  return HRESULT_FROM_WIN32((DWORD)iRet);
}

/*------------------------------------------------------------*/

