#ifndef GPGEZW_H
#define GPGEZW_H

#define DllExport __declspec(dllexport)

#include "../include/util_stringlist.h"
#include "../include/util_target_filelist.h"

#ifdef __cplusplus
extern "C" {
#endif

enum {
  FILETYPE_NONE = 0,
  FILETYPE_GPGENCPUBLIC,
  FILETYPE_GPGENCPUBLIC_ASCII,
  FILETYPE_GPGENCSYMMETRIC,
  FILETYPE_GPGENCSYMMETRIC_ASCII,
  FILETYPE_ZIPENC,
  FILETYPE_ZIPENCAES,
  FILETYPE_7ZENC,
  FILETYPE_ZIPPLAIN,
  FILETYPE_7ZPLAIN,
  FILETYPE_TARGZ,
  FILETYPE_TARBZ2,
  FILETYPE_TAR,
  FILETYPE_SFX,
  FILETYPE_KEYBIN,
  FILETYPE_KEYASC,
  FILETYPE_SIGBIN,
  FILETYPE_SIGASC
};

enum {
  EXTENSION_GPG = 0,
  EXTENSION_ZIP,
  EXTENSION_7Z,
  EXTENSION_LZH,
  EXTENSION_CAB,
  EXTENSION_TGZ,
  EXTENSION_TARGZ,
  EXTENSION_TBZ2,
  EXTENSION_TARBZ2,
  EXTENSION_TAR,
  EXTENSION_GZ,
  EXTENSION_BZ2,
  EXTENSION_ERROR
};

typedef struct struct_encfiletype {
  DWORD filetype;
  PCSTR filedesc;
  DWORD dwStr;
  PSTR filedesc1;
  PCSTR fileext;
  BOOL fpub;
  BOOL fascii;
  DWORD passlen;
} ENCFILETYPE, *PENCFILETYPE;

typedef struct struct_decfiletype {
  INT iext;
  PCSTR sext;
} DECFILETYPE, *PDECFILETYPE;

typedef struct struct_gpgezbase *PGPGEZBASE;

/*------------------------------------------------------------*/
DllExport VOID gpgezw_free(PVOID p);
/*------------------------------------------------------------*/

DllExport HRESULT gpgezw_InputBoxEx(HWND hwnd, LPCSTR pPrompt, LPCSTR pTitle, LPSTR pBuff, UINT uMin, UINT uMax, DWORD dwStyle);
DllExport HRESULT gpgezw_listupFilesFromDrop(HDROP hDrop, PTARGET_FILELIST *pptfl);


/*------------------------------------------------------------*/
DllExport HRESULT gpgezw_initialize(PGPGEZBASE *ppgpgezbase, PCSTR pGpgExe, PCSTR pConfDir, PCSTR pPubKeyDir);
DllExport HRESULT gpgezw_initialize_tar(PGPGEZBASE pgpgezbase, PCSTR pTarExe);
DllExport HRESULT gpgezw_initialize_7Zip(PGPGEZBASE pgpgezbase, PCSTR p7ZipExe);
DllExport HRESULT gpgezw_initialize_nkf(PGPGEZBASE pgpgezbase, PCSTR pNkfExe);
DllExport HRESULT gpgezw_release(PGPGEZBASE *ppgpgezbase);
DllExport PCSTR gpgezw_getMyFpr(PGPGEZBASE pgpgezbase);
DllExport PCSTR gpgezw_getConfDir(PGPGEZBASE pgpgezbase);
DllExport BOOL gpgezw_fNeedPassphrase(PGPGEZBASE pgpgezbase);
DllExport PCSTR gpgezw_getPubKeyDir(PGPGEZBASE pgpgezbas);

/*------------------------------------------------------------*/
DllExport HRESULT gpgezw_selectGpgKey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslSelected, PENCFILETYPE pet);
DllExport HRESULT gpgezw_editGpgKey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefault, PENCFILETYPE pet);
DllExport HRESULT gpgezw_editMyGpgKey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefault, PENCFILETYPE pet);
DllExport HRESULT gpgezw_selectGpgEncType(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslSelected, PBOOL pfPublic);
DllExport HRESULT gpgezw_generateGpgKey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing);
DllExport HRESULT gpgezw_changeUserId(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pUid, PCSTR pFpr, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing);
DllExport HRESULT gpgezw_changeGpgPassphrase(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pSelfFpr);
DllExport BOOL gpgezw_getPassphrseForPublicKeyEncryption(PGPGEZBASE pgpgezbase, HWND hwnd, DWORD dwRetry);
DllExport HRESULT gpgezw_prepareKeypair(PGPGEZBASE pgpgezbase, HWND hwnd);

/*------------------------------------------------------------*/
DllExport HRESULT gpgezw_zgpencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PTARGET_FILELIST ptfl, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslRecipients, BOOL fExtraKey, BOOL fSign, PENCFILETYPE pet);
DllExport HRESULT gpgezw_encrypt_fyeo(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTRINGLIST pslRecipients, BOOL fExtraKey, BOOL fSign, PENCFILETYPE pet);
#if 0
DllExport HRESULT gpgezw_encrypt_buff2buff(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, PSTR *ppbuff, PDWORD pdwLen, BOOL fSign, BOOL fAscii);
DllExport HRESULT gpgezw_encrypt_buff2file(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, BOOL fSign, BOOL fAscii);
#endif
DllExport HRESULT gpgezw_zgpdecrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, LPCSTR pDir, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fExtraKey, BOOL fVerify, PDECFILETYPE pdt);
DllExport HRESULT gpgezw_decrypt_fyeo(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFileName);
#if 0
DllExport HRESULT gpgezw_decrypt_buff2buff(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, PSTR *ppbuff, PDWORD pdwLen, BOOL fVerify);
DllExport HRESULT gpgezw_decrypt_file2buff(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTR *ppbuff, PDWORD pdwLen, BOOL fVerify);
DllExport HRESULT gpgezw_decrypt_buff2buff_nkf(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pbuff, DWORD dwLen, PSTR *ppbuff, PDWORD pdwLen, BOOL fVerify, PCSTR pNkOptIn, PCSTR pNkfOptOut);
DllExport HRESULT gpgezw_decrypt_file2buff_nkf(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PSTR *ppbuff, PDWORD pdwLen, BOOL fVerify, PCSTR pNkfOptIn, PCSTR pNkfOptOut);
#endif
DllExport HRESULT gpgezw_zgpdecryptonly(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, LPCSTR pDir, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fVerify);
DllExport HRESULT gpgezw_zgpsign(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, DWORD dwEncType, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PENCFILETYPE pet);
DllExport HRESULT gpgezw_zgpverify(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fExtraKey);
DllExport HRESULT gpgezw_zgparmor(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, BOOL fEnArmor);

DllExport HRESULT gpgezw_zgpencrypt_textbuff_to_buff(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, HWND hwnd, PVOID pbuff, DWORD dwLen, PCSTR pCharSet, BOOL fSign, BOOL fAscii, PSTR *ppbuff, PDWORD pdwLen);
DllExport HRESULT gpgezw_zgpencrypt_textbuff_to_clipboard(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, HWND hwnd, PVOID pbuff, DWORD dwLen, PCSTR pCharSet, BOOL fSign);
DllExport HRESULT gpgezw_zgpencrypt_textbuff_to_file(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, HWND hwnd, PVOID pbuff, DWORD dwLen, PCSTR pCharSet, BOOL fSign, PENCFILETYPE pet);
DllExport HRESULT gpgezw_zgpencrypt_clipboard_text(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, HWND hwnd, BOOL fSign);
DllExport HRESULT gpgezw_zgpdecrypt_textbuff_from_buff(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PVOID pbuff, DWORD dwLen, HWND hwnd, BOOL fVerify, PCSTR pCharSet, PSTR *ppbuff, PDWORD pdwLen);
DllExport HRESULT gpgezw_zgpdecrypt_textbuff_from_clipboard(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, HWND hwnd, BOOL fVerify, PCSTR pCharSet, PVOID *ppbuff, PDWORD pdwLen);
DllExport HRESULT gpgezw_zgpdecrypt_textbuff_from_file(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, HWND hwnd, BOOL fVerify, PCSTR pCharSet, PVOID *ppbuff, PDWORD pdwLen);
DllExport HRESULT gpgezw_zgpdecrypt_clipboard_text(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, HWND hwnd, PCSTR pCharSet, BOOL fVerify);

/*------------------------------------------------------------*/
DllExport HRESULT gpgezw_genkey(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pName, PCSTR pEMail, PCSTR pComment, PCSTR pPassphrase, PSTR *ppFpr);
DllExport HRESULT gpgezw_encrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFile, PCSTR pZfile, PSTRINGLIST pRecipients, PCSTR pPassphraseSym, BOOL fSign, BOOL fAscii);
DllExport HRESULT gpgezw_tarencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PSTRINGLIST pRecipients, PCSTR pPassphraseSym, BOOL gSign, BOOL fAscii);
DllExport HRESULT gpgezw_zipencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PCSTR pPassphraseSym);
DllExport HRESULT gpgezw_7zencrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PCSTR pPassphraseSym, DWORD dwMode);
DllExport HRESULT gpgezw_tar(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, DWORD dwMode);
DllExport HRESULT gpgezw_encryptsfx(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pBaseDirectory, PSTRINGLIST pFiles, PCSTR pZfile, PCSTR pPassphraseSym);
DllExport HRESULT gpgezw_decrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFile, PCSTR pXFile);
DllExport HRESULT gpgezw_decrypttar(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pPassphrase, PCSTR pFileName, PCSTR pBaseDirectory, BOOL fOverWrite);
DllExport HRESULT gpgezw_7zdecrypt(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pPassphrase, PCSTR pFileName, PCSTR pBaseDirectory, BOOL fOverWrite);
DllExport HRESULT gpgezw_untar(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pFileName, PCSTR pBaseDirectory, DWORD dwMode, BOOL fOverwrite);
DllExport HRESULT gpgezw_7zgunzip(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pFileNameIn, PCSTR pFileNameOut);
DllExport HRESULT gpgezw_import(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefault, PCSTR pFileName);
DllExport HRESULT gpgezw_export(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFpr, PCSTR pFileName, BOOL fAscii);
DllExport HRESULT gpgezw_delete(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefault, PCSTR pFpr);
DllExport HRESULT gpgezw_detach_sign(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFile, PCSTR pSfile, PCSTR pPassphraseSym, BOOL fAscii);
DllExport HRESULT gpgezw_verify(PGPGEZBASE pgpgezbase, HWND hwnd, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, PCSTR pFile, PCSTR pSfile, PCSTR pPassphraseSym);
DllExport HRESULT gpgezw_armor(PGPGEZBASE pgpgezbase, HWND hwnd, PCSTR pszFileName, PCSTR szFileNameOut, BOOL fEnArmor);

/* create_stringlist(pKeys) ĂĂяoƁApKeyslist-key̌ʂԂ */
DllExport HRESULT gpgezw_listkeys(PGPGEZBASE pgpgezbase, PSTRINGLIST pslSecRing, PSTRINGLIST pslPubRing, BOOL fNoDefaultKeyRing, BOOL fSec, PSTRINGLIST pKeys);
/*------------------------------------------------------------*/

#ifdef __cplusplus
}
#endif

#endif
