/* cmdez.c
 * Copyright (C) 2007 Kazuyoshi Kakihara
 *
 * This file is part of KageHinata.
 *
 * KageHinata is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define THISFILENAME "util/util_cmdez.c"

#include <windows.h>
#include <crtdbg.h>
#include "../include/util.h"

typedef struct struct_cmd_option_node {
  struct struct_cmd_option_node *next;
  PSTR name;
  PSTR value;
  PSTR opr;
} CMDOPTION_NODE;

typedef struct struct_cmd_option {
  PCMDOPTION_NODE phead;
} CMDOPTION;

typedef struct struct_cmdez_context {
  PSTR cmdexe;
  PSTR currentdir;
  CMDOPTION cmdoption;
  PSTRINGLIST cmdoperand;
  HANDLE hStdIn;
  HANDLE hStdOut;
  HANDLE hStdErr;
  HANDLE hProcess;
  DWORD exitcode;
} CMDEZCONTEXT;

/*------------------------------------------------------------*/
static HRESULT
cmdez_initialize_cmdoption(PCMDOPTION poption)
{
  HRESULT hr = S_OK;
  poption->phead = NULL;
  return hr;
}

/*------------------------------------------------------------*/
static HRESULT
cmdez_add_cmdoption(PCMDOPTION poption, LPCSTR option, LPCSTR value, PCSTR opr, BOOL fDq)
{
  HRESULT hr = S_OK;
  PCMDOPTION_NODE pnode = NULL, pnode1 = NULL;

  _ASSERT(poption);

  pnode = (PCMDOPTION_NODE)util_malloc(sizeof(CMDOPTION_NODE));
  if (!pnode) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  pnode->next = NULL;
  pnode->name = NULL;
  pnode->opr = NULL;
  pnode->value = NULL;
  if (NULL == (pnode->name = util_strdup(option))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (value) {
    if (fDq) {
      if (NULL == (pnode->value = enclose_dq(value))) {
        hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    } else {
      if (NULL == (pnode->value = util_strdup(value))) {
        hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
    if (opr) {
      if (NULL == (pnode->opr = util_strdup(opr))) {
        hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    } else {
      if (NULL == (pnode->opr = util_strdup(""))) {
        hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
  }

  pnode1 = poption->phead;
  if (!pnode1) {
    poption->phead = pnode;
  } else {
    while (pnode1->next) pnode1 = pnode1->next;
    pnode1->next = pnode;
  }
finish_func:
  if (FAILED(hr)) {
    if (pnode) {
      if (pnode->name) util_free(pnode->name);
      if (pnode->value) util_free(pnode->value);
      util_free(pnode);
    }
  }
  return hr;
}

/*------------------------------------------------------------*/
static HRESULT
cmdez_release_cmdoption(PCMDOPTION poption)
{
  HRESULT hr = S_OK;
  PCMDOPTION_NODE pnode = NULL, pnode1;

  _ASSERT(poption);
  _ASSERT(poption->phead);

  pnode = poption->phead;
  while (pnode) {
    pnode1 = pnode->next;
    if (pnode->name) util_free(pnode->name);
    if (pnode->value) util_free(pnode->value);
    util_free(pnode);
    pnode = pnode1;
  }
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
cmdez_initialize(PCMDEZCONTEXT *ppctx, PCSTR cmdexe, PCSTR currentdir)
{
  HRESULT hr = S_OK;

  _ASSERT(ppctx);
  _ASSERT(!*ppctx);

  if (*ppctx) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  *ppctx = (PCMDEZCONTEXT)util_malloc(sizeof(CMDEZCONTEXT));
  if (!*ppctx) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  (*ppctx)->cmdexe = NULL;
  (*ppctx)->currentdir = NULL;
  if (FAILED(hr = stringlist_create(&((*ppctx)->cmdoperand)))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  (*ppctx)->hStdIn = INVALID_HANDLE_VALUE;
  (*ppctx)->hStdOut = INVALID_HANDLE_VALUE;
  (*ppctx)->hStdErr = INVALID_HANDLE_VALUE;
  (*ppctx)->hProcess = INVALID_HANDLE_VALUE;
  (*ppctx)->exitcode = 0;
  if (NULL == ((*ppctx)->cmdexe = enclose_dq(cmdexe))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (currentdir) {
    if (NULL == ((*ppctx)->currentdir = util_strdup(currentdir) /*enclose_dq(currentdir)*/)) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }
  if (FAILED(hr = cmdez_initialize_cmdoption(&((*ppctx)->cmdoption)))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
finish_func:
  if (FAILED(hr)) cmdez_release(ppctx);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
cmdez_release(PCMDEZCONTEXT *ppctx)
{
  HRESULT hr = S_OK;

  if (!*ppctx) goto finish_func;
  if ((*ppctx)->cmdexe) util_free((*ppctx)->cmdexe);
  if ((*ppctx)->currentdir) util_free((*ppctx)->currentdir);
  cmdez_release_cmdoption(&((*ppctx)->cmdoption));
  if ((*ppctx)->cmdoperand) stringlist_release(&((*ppctx)->cmdoperand));
/*
  if ((*ppctx)->hStdIn != INVALID_HANDLE_VALUE) CloseHandle((*ppctx)->hStdIn);
  if ((*ppctx)->hStdOut != INVALID_HANDLE_VALUE) CloseHandle((*ppctx)->hStdOut);
  if ((*ppctx)->hStdErr != INVALID_HANDLE_VALUE) CloseHandle((*ppctx)->hStdErr);
*/
  util_free(*ppctx);
  *ppctx = NULL;
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
cmdez_add_option(PCMDEZCONTEXT pctx, LPCSTR option, LPCSTR value, BOOL fDq)
{
  HRESULT hr = S_OK;

  _ASSERT(pctx);

  if (!pctx) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (FAILED(hr = cmdez_add_cmdoption(&(pctx->cmdoption), option, value, "=", fDq))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
cmdez_add_option2(PCMDEZCONTEXT pctx, LPCSTR option, LPCSTR value, LPCSTR opr, BOOL fDq)
{
  HRESULT hr = S_OK;

  _ASSERT(pctx);

  if (!pctx) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (FAILED(hr = cmdez_add_cmdoption(&(pctx->cmdoption), option, value, opr, fDq))) {
    DEBUGERROR(0, hr);
    goto finish_func;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
cmdez_add_operand(PCMDEZCONTEXT pctx, PCSTR value, BOOL fDq)
{
  HRESULT hr = S_OK;

  _ASSERT(pctx);

  if (!pctx) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (fDq) {
    PSTR ptmp = NULL;
    
    ptmp = enclose_dq(value);
    if (ptmp == NULL) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = stringlist_add_node(pctx->cmdoperand, ptmp))) {
      DEBUGERROR(0, hr);
      util_free(ptmp);
      goto finish_func;
    }
    util_free(ptmp);
  } else {
    if (FAILED(hr = stringlist_add_node(pctx->cmdoperand, value))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
cmdez_set_handle(PCMDEZCONTEXT pctx, INT type, HANDLE h)
{
  HRESULT hr = S_OK;

  _ASSERT(pctx);

  if (!pctx) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  switch (type) {
  case HANDLE_TYPE_STDIN:
    if (pctx->hStdIn != INVALID_HANDLE_VALUE) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    pctx->hStdIn = h;
    break;
  case HANDLE_TYPE_STDOUT:
    if (pctx->hStdOut != INVALID_HANDLE_VALUE) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    pctx->hStdOut = h;
    break;
  case HANDLE_TYPE_STDERR:
    if (pctx->hStdErr != INVALID_HANDLE_VALUE) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    pctx->hStdErr = h;
    break;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
cmdez_execute(PCMDEZCONTEXT pctx)
{
  HRESULT hr = S_OK;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  SECURITY_ATTRIBUTES sa;
  STARTUPINFO si;
  LPSTR pCommandLine = NULL;
  PROCESS_INFORMATION pi;

  _ASSERT(pctx);

  ZeroMemory(&sa, sizeof(sa));
  sa.nLength = sizeof(sa);
  sa.bInheritHandle = TRUE;
  sa.lpSecurityDescriptor = NULL;

  hNull = CreateFile("nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hNull == INVALID_HANDLE_VALUE) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  /* stdin, stout, stderr Zbg */
  ZeroMemory(&si, sizeof(STARTUPINFO));
  si.cb = sizeof(STARTUPINFO);
  si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
  si.wShowWindow = SW_HIDE;
  si.hStdInput = (pctx->hStdIn == INVALID_HANDLE_VALUE) ? hNull : pctx->hStdIn;
  si.hStdOutput = (pctx->hStdOut == INVALID_HANDLE_VALUE) ? hNull : pctx->hStdOut;
  si.hStdError = (pctx->hStdErr == INVALID_HANDLE_VALUE) ? hNull : pctx->hStdErr;

  /* R}hC񐶐 */
  {
    PSTRING pstr = NULL;
    PCMDOPTION_NODE pnode;
    if (FAILED(hr = string_create(&pstr))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = string_append(pstr, pctx->cmdexe))) {
      string_release(&pstr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    for (pnode = pctx->cmdoption.phead; pnode; pnode = pnode->next) {
      if (FAILED(hr = string_append(pstr, " "))) {
        string_release(&pstr);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
      if (FAILED(hr = string_append(pstr, pnode->name))) {
        string_release(&pstr);
        DEBUGERROR(0, hr);
        goto finish_func;
      }
      if (pnode->value) {
        if (FAILED(hr = string_append(pstr, pnode->opr))) {
          string_release(&pstr);
          DEBUGERROR(0, hr);
          goto finish_func;
        }
        if (FAILED(hr = string_append(pstr, pnode->value))) {
          string_release(&pstr);
          DEBUGERROR(0, hr);
          goto finish_func;
        }
      }
    }
    if (stringlist_get_len(pctx->cmdoperand) > 0) {
      PSTRINGLIST_NODE pnode;
      for (pnode = stringlist_get_head(pctx->cmdoperand); pnode; pnode = stringlist_get_next(pnode)) {
        if (FAILED(hr = string_append(pstr, " "))) {
          string_release(&pstr);
          DEBUGERROR(0, hr);
          goto finish_func;
        }
        if (FAILED(hr = string_append(pstr, stringlist_get_string(pnode)))) {
          string_release(&pstr);
          DEBUGERROR(0, hr);
          goto finish_func;
        }
      }
    }
    if (NULL == (pCommandLine = util_strdup(string_c_str(pstr)))) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      string_release(&pstr);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    string_release(&pstr);
  }

  /* gpg.exe N */
  if (!CreateProcess(NULL, pCommandLine, NULL, NULL, TRUE, 0, NULL, pctx->currentdir, &si, &pi)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }

  /* svȃnh */
  CloseHandle(pi.hThread);
  pctx->hProcess = pi.hProcess;

finish_func:
/*
  if (pctx->hStdIn != INVALID_HANDLE_VALUE) {
    CloseHandle(pctx->hStdIn);
    pctx->hStdIn = INVALID_HANDLE_VALUE;
  }
  if (pctx->hStdOut != INVALID_HANDLE_VALUE) {
    CloseHandle(pctx->hStdOut);
    pctx->hStdOut = INVALID_HANDLE_VALUE;
  }
  if (pctx->hStdErr != INVALID_HANDLE_VALUE) {
    CloseHandle(pctx->hStdErr);
    pctx->hStdErr = INVALID_HANDLE_VALUE;
  }
*/
  if (hNull != INVALID_HANDLE_VALUE) CloseHandle(hNull);
  return hr;
}
/*------------------------------------------------------------*/
DllExport PHANDLE
cmdez_get_pprocess(PCMDEZCONTEXT pctx)
{
  return &(pctx->hProcess);
}

/*------------------------------------------------------------*/
DllExport HANDLE
cmdez_get_process(PCMDEZCONTEXT pctx)
{
  return pctx->hProcess;
}
/*------------------------------------------------------------*/
DllExport DWORD
cmdez_get_exitcode(PCMDEZCONTEXT pctx)
{
  return pctx->exitcode;
}

/*------------------------------------------------------------*/
DllExport HRESULT
cmdez_finish(PCMDEZCONTEXT pctx)
{
  HRESULT hr = S_OK;

  _ASSERT(pctx);
  if (WAIT_FAILED == WaitForSingleObject(pctx->hProcess, INFINITE)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!GetExitCodeProcess(pctx->hProcess, &(pctx->exitcode))) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  CloseHandle(pctx->hProcess);
  pctx->hProcess = INVALID_HANDLE_VALUE;
finish_func:
  return hr;
}
/*------------------------------------------------------------*/

