#define THISFILENAME "util/util_readfilebuff.c"

#include <windows.h>
#include "../include/util.h"

#define FILEBUFFLINE_MAX_BUFF 4096
typedef struct struct_readfilebuff {
  HANDLE hRead;
  BYTE buff[FILEBUFFLINE_MAX_BUFF + 1];
  DWORD len;
  DWORD pos;
  BOOL eof;
} READFILEBUFF;

/*------------------------------------------------------------*/
DllExport HRESULT
CreateReadFileBuff(PREADFILEBUFF *ppfile, HANDLE h) {
  HRESULT hr = S_OK;

  *ppfile = util_malloc(sizeof(READFILEBUFF));
  if (!*ppfile) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  (*ppfile)->hRead = h;
  (*ppfile)->eof = FALSE;
  (*ppfile)->len = 0;
  (*ppfile)->pos = 0;
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
ReleaseReadFileBuff(PREADFILEBUFF *ppfile) {
  if (*ppfile) {
    util_free(*ppfile);
    *ppfile = NULL;
  }
  return S_OK;
}

/*------------------------------------------------------------*/
DllExport HRESULT
AttachReadFileLine(PREADFILEBUFF pfile, HANDLE hRead) {
  HRESULT hr = S_OK;
  pfile->hRead = hRead;
  pfile->eof = FALSE;
  pfile->len = 0;
  pfile->pos = 0;
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
ReadFileLine(PREADFILEBUFF pfile, PSTR buff, DWORD len) {
  HRESULT hr = S_OK;
  DWORD cnt = 0;
  BOOL fYenR = FALSE;

  if (fEofFileLine(pfile)) {
    hr = HRESULT_FROM_WIN32(ERROR_HANDLE_EOF);
    goto finish_func;
  }
  while (cnt < len - 1) {
    if (pfile->pos >= pfile->len) {
      if (pfile->eof) {
        break;
      }
      if (!ReadFile(pfile->hRead, pfile->buff, FILEBUFFLINE_MAX_BUFF, &(pfile->len), NULL)) {
        DWORD dwErr = GetLastError();
        if (dwErr == ERROR_BROKEN_PIPE) {
          pfile->eof = TRUE;
        } else {
          hr = HRESULT_FROM_WIN32(dwErr);
          goto finish_func;
        }
      }
      if (pfile->len == 0) {
        pfile->eof = TRUE;
        break;
      }
      pfile->pos = 0;
    }
    if (*(pfile->buff + pfile->pos) == '\r') {
      fYenR = TRUE;
      ++(pfile->pos);
      continue;
    } else if (*(pfile->buff + pfile->pos) == '\n') {
      ++(pfile->pos);
      break;
    } else if (fYenR) {
      break;
    } else {
      *(buff + cnt) = *(pfile->buff + pfile->pos);
    }
    ++cnt;
    ++(pfile->pos);
  }
  *(buff + cnt) = '\0';
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
testFileBuffEmpty(PREADFILEBUFF pfile, PBOOL pfEmpty) {
  HRESULT hr = S_OK;

  *pfEmpty = FALSE;
  if (fEofFileLine(pfile)) {
    *pfEmpty = TRUE;
    goto finish_func;
  }
  if (pfile->len > 0) {
    *pfEmpty = FALSE;
    goto finish_func;
  }
  {
    DWORD dw;
    CHAR c;
    if (!PeekNamedPipe(pfile->hRead, &c, 1, &dw, NULL, NULL)) {
      DWORD dwErr = GetLastError();
      if (dwErr == ERROR_BROKEN_PIPE) {
        pfile->eof = TRUE;
      } else {
        hr = HRESULT_FROM_WIN32(dwErr);
        goto finish_func;
      }
    }
    if (dw == 0) {
      *pfEmpty = TRUE;
    } else {
      *pfEmpty = FALSE;
    }
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport BOOL
fEofFileLine(PREADFILEBUFF pfile) {
  BOOL fEof = FALSE;
  if ((pfile->eof) && (pfile->pos >= pfile->len)) {
    fEof = TRUE;
  }
  return fEof;
}

/*------------------------------------------------------------*/
DllExport HRESULT
DetatchReadFileLine(PREADFILEBUFF pfile) {
  HRESULT hr = S_OK;
  if (pfile->hRead == INVALID_HANDLE_VALUE) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    goto finish_func;
  }
  if (!fEofFileLine(pfile)) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    goto finish_func;
  }
  pfile->hRead = INVALID_HANDLE_VALUE;
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
