#define THISFILENAME "util/util_string.c"

#include <windows.h>
#include "../include/util.h"
 
typedef struct struct_string {
  PSTR buff;
  DWORD max;
  DWORD len;
} STRING;

#ifdef STRING_BUFF_LEN_BASE
#undef STRING_BUFF_LEN_BASE
#endif
#define STRING_BUFF_LEN_BASE 1024

/*------------------------------------------------------------*/
DllExport HRESULT
string_create(PSTRING *ppstr) {
  HRESULT hr = S_OK;

  if (*ppstr) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  *ppstr = (PSTRING)util_malloc(sizeof(STRING));
  if (!*ppstr) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  (*ppstr)->buff = NULL;
  (*ppstr)->max = 0;
  (*ppstr)->len = 0;
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
string_append(PSTRING pstr, PCSTR pstr1) {
  HRESULT hr = S_OK;
  DWORD nlen = 0;

  nlen = util_strlen(pstr1, -1);
  if (nlen == -1) goto finish_func;

  /* ǉ̂Ƀobt@Ȃꍇ̓[mۂȂ */
  if (pstr->len + nlen >= pstr->max) {
    DWORD max = 0;
    PSTR buff2;
    
    max = (pstr->max > STRING_BUFF_LEN_BASE) ? pstr->max : STRING_BUFF_LEN_BASE;
    while (max < pstr->len + nlen) max *= 2;
    buff2 = (PSTR)util_malloc(max + 1);
    if (!buff2) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      goto finish_func;
    }
    if (pstr->buff) {
      if (FAILED(hr = util_strcpy(buff2, max, pstr->buff))) {
        util_free(buff2);
        goto finish_func;
      }
      util_free(pstr->buff);
    }
    pstr->buff = buff2;
    pstr->max = max;
  }

  /* ǉ */
  if (FAILED(hr = util_strcpy(pstr->buff + pstr->len, pstr->max - pstr->len, pstr1))) {
    goto finish_func;
  }
  pstr->len += nlen;

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
string_append_c_str(PSTRING pstr, PCSTR pstr1, DWORD nbegin, DWORD nlen) {
  HRESULT hr = S_OK;

  if (nlen == -1) {
    nlen = util_strlen(pstr1 + nbegin, -1);
    if (nlen == -1) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
  }
  /* ǉ̂Ƀobt@Ȃꍇ̓[mۂȂ */
  if (pstr->len + nlen >= pstr->max) {
    DWORD max = 0;
    PSTR buff2;
    
    max = (pstr->max > STRING_BUFF_LEN_BASE) ? pstr->max : STRING_BUFF_LEN_BASE;
    while (max < pstr->len + nlen) max *= 2;
    buff2 = (PSTR)util_malloc(max + 1);
    if (!buff2) {
      hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
      goto finish_func;
    }
    if (FAILED(hr = util_strcpy(buff2, max, pstr->buff))) {
      util_free(pstr);
      goto finish_func;
    }
    if (pstr->buff) util_free(pstr->buff);
    pstr->buff = buff2;
    pstr->max = max;
  }

  /* ǉ */
  if (FAILED(hr = util_strncpy(pstr->buff + pstr->len, pstr->max - pstr->len, pstr1 + nbegin, nlen))) {
    goto finish_func;
  }
  pstr->len += nlen;
  *(pstr->buff + pstr->len) = '\0';

finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport DWORD
string_len(PSTRING pstr) {
  return pstr->len;
}

/*------------------------------------------------------------*/
DllExport PCSTR
string_c_str(PSTRING pstr) {
  return pstr->buff;
}

/*------------------------------------------------------------*/
DllExport PSTR
string_dup_str(PSTRING pstr) {
  PSTR buff2 = NULL;
  HRESULT hr = S_OK;

  buff2 = (PSTR)util_malloc(pstr->len + 1);
  if (!buff2) goto finish_func;
  if (FAILED(hr = util_strcpy(buff2, pstr->len + 1, pstr->buff))) {
    util_free(buff2);
    buff2 = NULL;
    goto finish_func;
  }
finish_func:
  return buff2;
}

/*------------------------------------------------------------*/
DllExport PSTR
string_detach(PSTRING pstr) {
  PSTR p = pstr->buff;
  pstr->buff = NULL;
  pstr->len = 0;
  pstr->max = 0;
  return p;
}

/*------------------------------------------------------------*/
DllExport VOID
string_free_str(PSTR pstr1) {
  if (pstr1) util_free(pstr1);
  return;
}

/*------------------------------------------------------------*/
DllExport HRESULT
string_release(PSTRING *ppstr) {
  if (*ppstr) {
    if ((*ppstr)->buff) util_free((*ppstr)->buff);
    util_free(*ppstr);
  }
  *ppstr = NULL;
  return S_OK;
}

/*------------------------------------------------------------*/
