#define THISFILENAME "util/util_stringlist.c"

#include <windows.h>
#include <strsafe.h>
#include "../include/util.h"

typedef struct struct_stringlist_node {
  PSTRINGLIST_NODE next;
  PVOID buff;
} STRINGLIST_NODE;

typedef struct struct_stringlist {
  PSTRINGLIST_NODE phead;
  int length;
  PSTR *ppstrlist;
} STRINGLIST;

/*------------------------------------------------------------*/
DllExport HRESULT
stringlist_create(PSTRINGLIST *pplist)
{
  HRESULT hr = S_OK;

  if (*pplist) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    goto finish_func;
  }
  *pplist = (PSTRINGLIST)util_malloc(sizeof(STRINGLIST));
  if (!*pplist) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }
  (*pplist)->phead = NULL;
  (*pplist)->length = 0;
  (*pplist)->ppstrlist = NULL;
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
stringlist_release(PSTRINGLIST *pplist)
{
  HRESULT hr = S_OK;
  PSTRINGLIST_NODE pnode, pnode1;

  if (!*pplist) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_FUNCTION);
    goto finish_func;
  }
  if ((*pplist)->ppstrlist) util_free((*pplist)->ppstrlist);

  pnode = (*pplist)->phead;
  while (pnode) {
    pnode1 = pnode->next;
    if (pnode->buff) util_free(pnode->buff);
    util_free(pnode);
    pnode = pnode1;
  }
  util_free(*pplist);
  *pplist = NULL;
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
stringlist_add_node(PSTRINGLIST plist, PCSTR pstr)
{
  HRESULT hr = S_OK;
  PSTRINGLIST_NODE pnode_new = NULL, pnode;
  PSTR buff = NULL;
  int len = lstrlen(pstr);

  if (!plist) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    return hr;
  }
  /* m */
  pnode_new = (PSTRINGLIST_NODE)util_malloc(sizeof(STRINGLIST_NODE));
  if (!pnode_new) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    return hr;
  }
  buff = (PSTR)util_malloc(len + 1);
  if (!buff) {
    util_free(pnode_new);
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    return hr;
  }
  if (FAILED(hr = StringCbCopy(buff, len + 1, pstr))) {
    util_free(buff);
    util_free(pnode_new);
    return hr;
  }
  pnode_new->buff = buff;
  pnode_new->next = NULL;
  
  /* Xg֒ǉ */
  if (!plist->phead) {
    plist->phead = pnode_new;
  } else {
    pnode = plist->phead;
    while (pnode->next) {
      pnode = pnode->next;
    }
    pnode->next = pnode_new;
  }
  ++(plist->length);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
stringlist_add_node_len(PSTRINGLIST plist, PCSTR pstr, int len)
{
  HRESULT hr = S_OK;
  PSTRINGLIST_NODE pnode_new = NULL, pnode;
  PSTR buff = NULL;

  if (!plist) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    return hr;
  }
  /* m */
  pnode_new = (PSTRINGLIST_NODE)util_malloc(sizeof(STRINGLIST_NODE));
  if (!pnode_new) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    return hr;
  }
  buff = (PSTR)util_malloc(len + 1);
  if (!buff) {
    util_free(pnode_new);
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    return hr;
  }
  if (FAILED(hr = StringCbCopyN(buff, len + 1, pstr, len))) {
    util_free(buff);
    util_free(pnode_new);
    return hr;
  }
  pnode_new->buff = buff;
  pnode_new->next = NULL;
  
  /* Xg֒ǉ */
  if (!plist->phead) {
    plist->phead = pnode_new;
  } else {
    pnode = plist->phead;
    while (pnode->next) {
      pnode = pnode->next;
    }
    pnode->next = pnode_new;
  }
  ++(plist->length);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
stringlist_concat(PSTRINGLIST plist, PSTRINGLIST plist2add)
{
  HRESULT hr = S_OK;
  PSTRINGLIST_NODE pnode;

  if (!plist) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    return hr;
  }
  if (!plist2add) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    return hr;
  }
  for (pnode = stringlist_get_head(plist2add); pnode; pnode = stringlist_get_next(pnode)) {
    hr = stringlist_add_node(plist, stringlist_get_string(pnode));
    if (FAILED(hr)) {
      return hr;
    }
  }
  return  hr;
}

/*------------------------------------------------------------*/
DllExport PCSTR *stringlist_get_table(PSTRINGLIST plist)
{
  PSTR *ppstrlist_itr;
  PSTRINGLIST_NODE pnode;

  if (plist->ppstrlist) util_free(plist->ppstrlist);

  plist->ppstrlist = (PSTR *)util_malloc(plist->length + 1);
  if (!(plist->ppstrlist)) goto finish_func;

  ppstrlist_itr = plist->ppstrlist;
  for (pnode = plist->phead; pnode; pnode = pnode->next) {
    *ppstrlist_itr = pnode->buff;
    ++ppstrlist_itr;
  }
  *ppstrlist_itr = NULL;
finish_func:
  return plist->ppstrlist;
}

/*------------------------------------------------------------*/
DllExport PSTRINGLIST_NODE
stringlist_get_head(PSTRINGLIST plist)
{
  return plist->phead;
}

/*------------------------------------------------------------*/
DllExport PSTRINGLIST_NODE
stringlist_get_next(PSTRINGLIST_NODE pnode)
{
  return pnode->next;
}

/*------------------------------------------------------------*/
DllExport PCSTR
stringlist_get_string(PSTRINGLIST_NODE pnode)
{
  return pnode->buff;
}

/*------------------------------------------------------------*/
DllExport int
stringlist_get_len(PSTRINGLIST plist)
{
  return plist->length;
}

/*------------------------------------------------------------*/
