#define THISFILENAME "util/util.c"

#include <windows.h>
#include <strsafe.h>
#include <crtdbg.h>
#include "../include/util.h"

#ifdef MAX_LINE
#undef MAX_LINE
#endif
#define MAX_LINE 1024
static BOOL debug_launched = FALSE;

CHAR debug_log[MAX_PATH + 1] = "";
HANDLE hdebug_log = INVALID_HANDLE_VALUE;

/*------------------------------------------------------------*/
DllExport VOID
InitDebugError(PSTR pLogFileName)
{
  HANDLE h = INVALID_HANDLE_VALUE;
  _CrtSetReportMode(_CRT_ERROR, _CRTDBG_MODE_FILE);
  _CrtSetReportFile(_CRT_ERROR, _CRTDBG_FILE_STDOUT);
  debug_launched = FALSE;
  util_strcpy(debug_log, MAX_PATH, pLogFileName);
}

/*------------------------------------------------------------*/
DllExport VOID
ResetDebugError(VOID)
{
  debug_launched = FALSE;
  if (hdebug_log != INVALID_HANDLE_VALUE) {
    CloseHandle(hdebug_log);
    hdebug_log = INVALID_HANDLE_VALUE;
  }
}

/*------------------------------------------------------------*/
DllExport VOID
DebugError(INT level, HRESULT hr, PCSTR filename, INT line)
{
  PVOID pMsgBuf;
  CHAR szMsg[MAX_LINE * 2 + 1];

  if (!debug_launched) {
    debug_launched = TRUE;
    if (level < DEBUGLEVEL) {
      if (HRESULT_FACILITY(hr) == FACILITY_WIN32) {
        FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                      NULL, HRESULT_CODE(hr), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (PTSTR) &pMsgBuf, 0, NULL);
        StringCbPrintf(szMsg, MAX_LINE * 2, "%s\n%s:%d", pMsgBuf, filename, line);
        MessageBox(NULL, szMsg, "Error", MB_OK | MB_ICONINFORMATION);
        LocalFree(pMsgBuf);
      } else {
        StringCbPrintf(szMsg, MAX_LINE * 2, "Unknown Error at\n%s:%d", filename, line);
        MessageBox(NULL, szMsg, "Error", MB_OK | MB_ICONINFORMATION);
      }
    }
    if (debug_log[0]) {
      if (hdebug_log == INVALID_HANDLE_VALUE) {
        hdebug_log = CreateFile(debug_log, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
      }
      if (hdebug_log != INVALID_HANDLE_VALUE) {
        DWORD dwWritten;
        WriteFile(hdebug_log, szMsg, util_strlen(szMsg, -1), &dwWritten, NULL);
        WriteFile(hdebug_log, "\n", 1, &dwWritten, NULL);
      }
    }
  } else {
    _RPT2(_CRT_ERROR, "Error Trace  %s:%d", filename, line);
    if (hdebug_log != INVALID_HANDLE_VALUE) {
      DWORD dwWritten;
      StringCbPrintf(szMsg, MAX_LINE * 2, "Error Trace %s:%d", filename, line);
      WriteFile(hdebug_log, szMsg, util_strlen(szMsg, -1), &dwWritten, NULL);
      WriteFile(hdebug_log, "\n", 1, &dwWritten, NULL);
    }
  }
}

/*------------------------------------------------------------*/
DllExport HRESULT
CloseHandleClean(PHANDLE ph)
{
  HRESULT hr = S_OK;

  if (*ph != INVALID_HANDLE_VALUE) {
    if (!CloseHandle(*ph)) {
      DWORD dwErr = GetLastError();
      hr = HRESULT_FROM_WIN32(dwErr);
      goto finish_func;
    }
    *ph = INVALID_HANDLE_VALUE;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_create_write_pipe(PHANDLE phWrite, PHANDLE phRead)
{
  HRESULT hr = S_OK;
  SECURITY_ATTRIBUTES sa;
  HANDLE h = INVALID_HANDLE_VALUE;
  HANDLE hCurrent = GetCurrentProcess();

  ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;

  if (!CreatePipe(phRead, &h, &sa, 0)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    goto finish_func;
  }
  if (!DuplicateHandle(hCurrent, h, hCurrent, phWrite, 0, FALSE, DUPLICATE_SAME_ACCESS)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    CloseHandle(h);
    h = INVALID_HANDLE_VALUE;
    goto finish_func;
  }
  CloseHandle(h);
  h = INVALID_HANDLE_VALUE;
finish_func:
  if (h != INVALID_HANDLE_VALUE) CloseHandle(h);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_create_read_pipe(PHANDLE phRead, PHANDLE phWrite)
{
  HRESULT hr = S_OK;
  SECURITY_ATTRIBUTES sa;
  HANDLE h = INVALID_HANDLE_VALUE;
  HANDLE hCurrent = GetCurrentProcess();

  ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;

  if (!CreatePipe(&h, phWrite, &sa, 0)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    goto finish_func;
  }
  if (!DuplicateHandle(hCurrent, h, hCurrent, phRead, 0, FALSE, DUPLICATE_SAME_ACCESS)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    CloseHandle(h);
    h = INVALID_HANDLE_VALUE;
    goto finish_func;
  }
  CloseHandle(h);
  h = INVALID_HANDLE_VALUE;
finish_func:
  if (h != INVALID_HANDLE_VALUE) CloseHandle(h);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
WriteFileLine1(HANDLE hFile, PCSTR s1)
{
  HRESULT hr = S_OK;
  DWORD dwWritten;

  if (!WriteFile(hFile, s1, util_strlen(s1, -1), &dwWritten, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WriteFile(hFile, "\n", util_strlen("\n", -1), &dwWritten, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/



/*------------------------------------------------------------*/
DllExport PVOID
util_malloc(DWORD dwBytes)
{
  PVOID memblock = NULL;
  memblock = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwBytes);
  if (!memblock) {
    SetLastError(ERROR_OUTOFMEMORY);
  }
  return memblock;
}

/*------------------------------------------------------------*/
DllExport VOID
util_free(PVOID memblock)
{
  HeapFree(GetProcessHeap(), 0, memblock);
}

/*------------------------------------------------------------*/
DllExport INT
util_atoi(PCSTR str)
{
  INT i = 0;
  PCSTR p = str;

  SetLastError(ERROR_SUCCESS);
  if (*p == '-') ++p;
  while ((*p) && (*p >= '0') && (*p <= '9')) {
    if (i > 0x7fffffff / 2 / 10) {
      i = 0;
      SetLastError(ERROR_ARITHMETIC_OVERFLOW);
      break;
    }
    i *= 10;
    i += (*p - '0');
  }
  if (*str == '-') i *= (-1);
  return i;
}

/*------------------------------------------------------------*/
DllExport PCSTR
util_strchr(PCSTR s, WORD c)
{
  PCSTR p = s;
  while (*p) {
    if (*p == c) break;
    ++p;
  }
  if (!*p) p = NULL;
  return p;
}

/*------------------------------------------------------------*/
DllExport PSTR
util_strdup(PCSTR s)
{
  INT ilen = 0;
  PSTR s2 = NULL;
  
  ilen = util_strlen(s, UTIL_MAX_CHAR_BUFF);
  if (ilen == -1) goto finish_func;
  s2 = util_malloc(ilen + 1);
  if (!s2) goto finish_func;
  util_strcpy(s2, ilen + 1, s);
finish_func:
  return s2;
}

/*------------------------------------------------------------*/
DllExport PSTR
util_strndup(PCSTR s, INT len)
{
  PSTR s2 = NULL;

  s2 = util_malloc(len + 2);
  if (!s2) goto finish_func;
  util_strncpy(s2, len + 1, s, len);
  *(s2 + len) = '\0';
finish_func:
  return s2;
}

/*------------------------------------------------------------*/
DllExport INT
util_strlen(PCSTR s, INT max)
{
  int ilen = 0;
  PCSTR p = s;
  while (*p) {
    ++ilen;
    ++p;
    if (ilen == max) {
      ilen = -1;
      break;
    }
  }
  return ilen;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_strcpy(PSTR dst, INT max, PCSTR src)
{
  HRESULT hr = S_OK;
  hr = StringCbCopy(dst, max, src);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_strncpy(PSTR dst, INT max, PCSTR src, INT len)
{
  HRESULT hr = S_OK;
  hr = StringCbCopyN(dst, max, src, len);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_strcat(PSTR dst, INT max, PCSTR src)
{
  HRESULT hr = S_OK;
  hr = StringCbCat(dst, max, src);
  return hr;
}

/*------------------------------------------------------------*/
DllExport INT
util_strcmp(PCSTR s1, PCSTR s2)
{
  INT cmp;

  cmp = CompareString(LOCALE_USER_DEFAULT, 0, s1, -1, s2, -1);
  if (cmp == 0) {
  } else if (cmp == CSTR_LESS_THAN) {
    return -1;
  } else if (cmp == CSTR_EQUAL) {
    return 0;
  } else if (cmp == CSTR_GREATER_THAN) {
    return 1;
  }
  return 0;
}

/*------------------------------------------------------------*/
DllExport INT
util_stricmp(PCSTR s1, PCSTR s2)
{
  INT cmp;

  cmp = CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE, s1, -1, s2, -1);
  if (cmp == 0) {
  } else if (cmp == CSTR_LESS_THAN) {
    return -1;
  } else if (cmp == CSTR_EQUAL) {
    return 0;
  } else if (cmp == CSTR_GREATER_THAN) {
    return 1;
  }
  return 0;
}

/*------------------------------------------------------------*/
DllExport INT
util_strncmp(PCSTR s1, PCSTR s2, INT len)
{
  INT cmp;

  cmp = CompareString(LOCALE_USER_DEFAULT, 0, s1, len, s2, len);
  if (cmp == 0) {
  } else if (cmp == CSTR_LESS_THAN) {
    return -1;
  } else if (cmp == CSTR_EQUAL) {
    return 0;
  } else if (cmp == CSTR_GREATER_THAN) {
    return 1;
  }
  return 0;
}

/*------------------------------------------------------------*/
DllExport INT
util_strnicmp(PCSTR s1, PCSTR s2, INT len)
{
  INT cmp;

  cmp = CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE, s1, len, s2, len);
  if (cmp == 0) {
  } else if (cmp == CSTR_LESS_THAN) {
    return -1;
  } else if (cmp == CSTR_EQUAL) {
    return 0;
  } else if (cmp == CSTR_GREATER_THAN) {
    return 1;
  }
  return 0;
}

/*------------------------------------------------------------*/
DllExport INT
util_strncmpl(PCSTR s1, PCSTR s2)
{
  return util_strncmp(s1, s2, util_strlen(s1, -1));
}

/*------------------------------------------------------------*/
DllExport PSTR
enclose_dq(PCSTR s)
{
  INT ilen = 0;
  PSTR s2 = NULL;
  
  ilen = util_strlen(s, UTIL_MAX_CHAR_BUFF - 2);
  if (ilen == -1) goto finish_func;
  s2 = util_malloc(ilen + 2 + 1);
  if (!s2) goto finish_func;
  *s2 = '"';
  util_strcpy(s2 + 1, ilen + 1, s);
  *(s2 + ilen + 1) = '"';
  *(s2 + ilen + 2) = '\0';
finish_func:
  return s2;
}

/*------------------------------------------------------------*/
DllExport  PSTR
push_dh(PCSTR s)
{
  INT ilen = 0;
  PSTR s2 = NULL;
  
  ilen = util_strlen(s, UTIL_MAX_CHAR_BUFF - 2);
  if (ilen == -1) goto finish_func;
  s2 = util_malloc(ilen + 2 + 1);
  if (!s2) goto finish_func;
  *s2 = '-';
  *(s2 + 1) = '-';
  util_strcpy(s2 + 2, ilen + 1, s);
finish_func:
  return s2;
}

/*------------------------------------------------------------*/
DllExport PSTR
util_itoa(INT i, PSTR buff, INT max)
{
  int i2 = 0;
  PSTR p = buff;
  INT cnt = 0;

  SetLastError(ERROR_SUCCESS);

  /* i ̌𐔂 */
  if (i == 0) cnt = 1;
  else {
    for (i2 = i; i2; i2 /= 10) ++cnt;
    if (max < cnt - 1) {
      *buff = '\0';
       SetLastError(ERROR_ARITHMETIC_OVERFLOW);
      goto finish_func;
    }
  }

  i2 = i;
  buff[cnt] = '\0';
  while (cnt > 0) {
    --cnt;
    buff[cnt] = '0' + i2 % 10;
    i2 /= 10;
  }
finish_func:
  return buff;
}

/*------------------------------------------------------------*/
DllExport HRESULT
sv_get(PCSTR src, CHAR sep, DWORD pos, PSTR dst, DWORD len)
{
  HRESULT hr = S_OK;
  PCSTR p1 = src, p2;
  DWORD cnt = 0;

  while (cnt < pos) {
    if (!*p1) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    } else if (*p1 == sep) {
      ++cnt;
    }
    ++p1;
  }
  p2 = p1;
  while (*p2) {
    if (*p2 == sep) {
      break;
    }
    ++p2;
  }
  if (len > (DWORD)(p2 - p1)) {
    hr = util_strncpy(dst, len, p1, p2 - p1);
  } else {
    hr = util_strncpy(dst, len, p1, len);
    if (FAILED(hr)) goto finish_func;
    hr = HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER);
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport PCSTR
sv_pos(PCSTR src, CHAR sep, DWORD pos)
{
  PCSTR p = src;
  DWORD cnt = 0;

  while (cnt < pos) {
    if (!*p) {
      break;
    } else if (*p == sep) {
      ++cnt;
    }
    ++p;
  }
  return p;
}

/*------------------------------------------------------------*/
DllExport BOOL
sv_cmp(PCSTR src, CHAR sep, DWORD pos, PCSTR pcmp)
{
  BOOL bRet = FALSE;
  PCSTR p1 = src;
  DWORD cnt = 0, len = 0;

  while (cnt < pos) {
    if (!*p1) {
      bRet = FALSE;
      goto finish_func;
    } else if (*p1 == sep) {
      ++cnt;
    }
    ++p1;
  }

  len = util_strlen(pcmp, -1);
  if (util_strncmp(p1, pcmp, len)) {
    bRet = FALSE;
    goto finish_func;
  }

  p1 += len;
  if ((*p1) && (*p1 != sep)) {
    bRet = FALSE;
    goto finish_func;
  }
  bRet = TRUE;

finish_func:
  return bRet;
}

/*------------------------------------------------------------*/
static HRESULT
util_createfolderif(PCSTR folder)
{
  HANDLE hFind = INVALID_HANDLE_VALUE;
  WIN32_FIND_DATA wfd;
  HRESULT hr = S_OK;

  hFind = FindFirstFile(folder, &wfd);
  if (hFind != INVALID_HANDLE_VALUE) {
    FindClose(hFind);
    if (!(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
  } else {
    if (!CreateDirectory(folder, NULL)) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      goto finish_func;
    }
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_createfolder(PCSTR folder)
{
  CHAR szFolderName[MAX_PATH + 1];
  PSTR p = szFolderName;
  HRESULT hr = S_OK;

  if (util_strlen(folder, -1) > MAX_PATH) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  if (FAILED(hr = util_strcpy(szFolderName, MAX_PATH, folder))) {
    goto finish_func;
  }

  p = szFolderName;
  while (*p) {
    if (IsDBCSLeadByte(*p)) {
      ++p;
      if (!*p) {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        goto finish_func;
      }
    } else {
      if ((*p == '\\') || (*p == '/')) {
        CHAR c = *p;

        *p = '\0';
        if (FAILED(hr = util_createfolderif(szFolderName))) {
          goto finish_func;
        }
        *p = c;
      }
    }
    ++p;
  }
  if (FAILED(hr = util_createfolderif(szFolderName))) {
    goto finish_func;
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport PSTR
util_pathfindextension(PSTR pFileName)
{
  PSTR p = pFileName, px = pFileName;
  if (!*p) goto finish_func;
  ++p;
  while (*p) {
    if (*p == '.') {
      px = p;
    } else {
      if (IsDBCSLeadByte(*p)) {
        ++p;
        if (!*p) {
          px = p;
          goto finish_func;
        }
      }
    }
    ++p;
  }
  if (px == pFileName) px = p; /* .Ȃꍇ */
finish_func:
  return px;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_sjistoutf8(PCSTR pSjis, PSTR *ppUtf8)
{
  INT iLen;
  PWSTR pUtf16 = NULL;
  HRESULT hr = S_OK;

  if (!(iLen = MultiByteToWideChar(CP_ACP, 0, pSjis, -1, NULL, 0))) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!(pUtf16 = util_malloc((iLen + 2) * sizeof(PWSTR)))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!MultiByteToWideChar(CP_ACP, 0, pSjis, -1, pUtf16, iLen + 1)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!(iLen = WideCharToMultiByte(CP_UTF8, 0, pUtf16, -1, NULL, 0, NULL, NULL))) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!(*ppUtf8 = util_malloc((iLen + 2) * sizeof(PSTR)))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WideCharToMultiByte(CP_UTF8, 0, pUtf16, -1, *ppUtf8, iLen + 1, NULL, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  util_free(pUtf16);
  pUtf16 = NULL;
finish_func:
  if (pUtf16) util_free(pUtf16);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_utf8tosjis(PCSTR pUtf8, PSTR *ppSjis)
{
  INT iLen;
  PWSTR pUtf16 = NULL;
  HRESULT hr = S_OK;

  if (!(iLen = MultiByteToWideChar(CP_UTF8, 0, pUtf8, -1, NULL, 0))) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!(pUtf16 = util_malloc((iLen + 2) * sizeof(PWSTR)))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!MultiByteToWideChar(CP_UTF8, 0, pUtf8, -1, pUtf16, iLen + 1)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!(iLen = WideCharToMultiByte(CP_ACP, 0, pUtf16, -1, NULL, 0, NULL, NULL))) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!(*ppSjis = util_malloc((iLen + 2) * sizeof(PSTR)))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  if (!WideCharToMultiByte(CP_ACP, 0, pUtf16, -1, *ppSjis, iLen + 1, NULL, NULL)) {
    hr = HRESULT_FROM_WIN32(GetLastError());
    DEBUGERROR(0, hr);
    goto finish_func;
  }
  util_free(pUtf16);
  pUtf16 = NULL;
finish_func:
  if (pUtf16) util_free(pUtf16);
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_lftocrlf(PCSTR pSrc, PSTR *ppDst)
{
  HRESULT hr = S_OK;
  DWORD dwCntlf = 0;
  DWORD dwLen = 0;
  PCSTR ps;
  PSTR pd;
  BOOL fcr = FALSE;

  for (ps = pSrc; *ps; ++ps) {
    if (*ps == '\r') {
      fcr = TRUE;
    } else if (*ps == '\n') {
      if (!fcr) ++dwCntlf;
      fcr = FALSE;
    } else {
      fcr = FALSE;
    }
    ++dwLen;
  }
  if (!(*ppDst = (PSTR)util_malloc(dwLen + dwCntlf + 1))) {
    hr = HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
    goto finish_func;
  }

  pd = *ppDst;
  fcr = FALSE;
  for (ps = pSrc; *ps; ++ps) {
    if (*ps == '\r') {
      fcr = TRUE;
    } else if (*ps == '\n') {
      if (!fcr) {
        *pd = '\r';
        ++pd;
      }
      fcr = FALSE;
    } else {
      fcr = FALSE;
    }
    *pd = *ps;
    ++pd;
  }
  *pd = '\0';
finish_func:
  return hr;
}

/*------------------------------------------------------------*/
DllExport HRESULT
util_is_file_included_in_a_folder(PCSTR pFileName, PTARGET_FILELIST ptfl, PBOOL pbRet)
{
  CHAR szFileEx[MAX_PATH + 1] = "";
  HRESULT hr = S_OK;
  PSTRINGLIST_NODE pnode = stringlist_get_head(target_filelist_get_files(ptfl));

  *pbRet = FALSE;
  while (pnode) {
    DWORD dwAttr = -1;

    util_strcpy(szFileEx, MAX_PATH, target_filelist_get_basedir(ptfl));
    if (*(szFileEx + util_strlen(szFileEx, -1) - 1) != '\\') {
      if (FAILED(hr = util_strcat(szFileEx, MAX_PATH, "\\"))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
    }
    if (util_strlen(szFileEx, -1) + util_strlen(stringlist_get_string(pnode), -1) >= MAX_PATH) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (FAILED(hr = util_strcat(szFileEx, MAX_PATH, stringlist_get_string(pnode)))) {
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    dwAttr = GetFileAttributes(szFileEx);
    if (dwAttr == -1) {
      hr = HRESULT_FROM_WIN32(GetLastError());
      DEBUGERROR(0, hr);
      goto finish_func;
    }
    if (dwAttr & FILE_ATTRIBUTE_DIRECTORY) {
      if (FAILED(hr = util_strcat(szFileEx, MAX_PATH, "\\"))) {
        DEBUGERROR(0, hr);
        goto finish_func;
      }
      if (!util_strncmp(pFileName, szFileEx, util_strlen(szFileEx, -1))) {
        *pbRet = TRUE;
        goto finish_func;
      }
    }
    pnode = stringlist_get_next(pnode);
  }
finish_func:
  return hr;
}

/*------------------------------------------------------------*/

